#include "project.h"

int          N_Fonts = 6;

XFontStruct *Xvs_font[]   = {NULL, NULL, NULL, NULL, NULL, NULL};
int          Xvs_font_h[] = {0,    0,    0,    0,    0,    0};

/* Use raster/window coordinates for annotation, since world coordinates
 * aren't static and are susceptible to degeneracy, loss of precision,
 * etc. */

/* Raster position, previous raster position and change in raster position ... */
raster_pos       Raster_pos;
raster_pos       Raster_pos_prev = {0,0};
raster_pos       DRaster_pos;

/* Index of last used font size ... */
int              Last_isize = -1;

void dump_ANNOTE_CTL(char *string,ANNOTE_CTL the) {
   fprintf(stderr,"dump_ANNOTE_CTL: %s.\n",string);
   fprintf(stderr,"color: %d   min_fh_index: %d  max_xfrac: %g.\n",
           the.col,the.min_fh_index,the.max_xfrac);
   fprintf(stderr,"xlc: %g  ylc: %g  position code: %d.\n",
           the.xlc,the.ylc,the.pos_code);
   fprintf(stderr,"x: %g  y: %g  isize: %d  s: '%s'\n",
           the.x,the.y,the.isize,the.s ? the.s : "NULL");
}

void auto_annotate_0(WT wt,char *string,ANNOTE_CTL the) {
}
void auto_annotate_1(WT wt,char *string,ANNOTE_CTL the) {
}
fmfonthandle fm_annotate_auto_select(int wt,char *string,int min_index,
                                      double max_xfrac) { 
   fmfonthandle the = (fmfonthandle) 0;
   return the;
}

void dump_glRasterPos2d(char *s) {
   double rpos[4];
   glGetDoublev(GL_CURRENT_RASTER_POSITION,rpos);  
   fprintf(stderr,"dump_glRasterPos2d: %s: x=%g y=%g z=%g w=%g\n",
      s,rpos[0],rpos[1],rpos[2],rpos[3]);
}

void mglRasterPos2d(double x,double y) {
   int ltrace = Ltrace;
   static int ncall = 0;
   double rpos[4];

   Raster_pos_prev = Raster_pos;

   glRasterPos2d(x,y);
   glGetDoublev(GL_CURRENT_RASTER_POSITION,rpos);  
   Raster_pos.x = rpos[0];
   Raster_pos.y = rpos[1];
   if( ltrace ) {
      fprintf(stderr,"mglRasterPos2d: Raster_pos=(%g,%g)\n",Raster_pos.x,Raster_pos.y);
   }
}
 
void dglRasterPos2d(void) {
   int ltrace = Ltrace;
   double rpos[4];

   Raster_pos_prev = Raster_pos;

   glGetDoublev(GL_CURRENT_RASTER_POSITION,rpos);  
   Raster_pos.x = rpos[0];
   Raster_pos.y = rpos[1];

   DRaster_pos.x = Raster_pos.x - Raster_pos_prev.x;
   DRaster_pos.y = Raster_pos.y - Raster_pos_prev.y;

   Raster_pos_prev = Raster_pos;
   if( ltrace ) {
      fprintf(stderr,"dglRasterPos2d: Raster_pos=(%g,%g)\n",Raster_pos.x,Raster_pos.y);
   }
}

int xvs_set_font(char *family, int isize) {
   static char R[] = "xvs_set_font";
   int ltrace = Ltrace;
   static int  first = 1;
   int         nsize = 6;
   static int  sizes[] = {240, 180, 140, 120, 100, 80};
   static int  dl_base[] = {0, 0, 0, 0, 0, 0};
   static int  dl_base_next = 100000;
   static int  nglyph = 256, offset = 0;
   char        B[BUFLEN];
   int         i;

   char       *font_family = (char *) NULL;
   static char Default_font_family[] = "helvetica";
   
   if( family ) {
      fprintf(stderr,"xvs_set_font: parameter family must be NULL\n");
      exit(1);
   }  

   if( first ) {
/*    Load the display lists, need to be sharing GLXContexts for this to work. ... */
      if( ltrace ) 
         fprintf(stderr,"xvs_set_font: Initializing %d display lists\n",nsize);
/*    Query XVS_FONT_FAMILY, perversely use MS-Hostile Helvetica by default ... */
      if( (font_family = getenv("XVS_FONT_FAMILY")) ) {
         if( ltrace ) 
            fprintf(stderr,"%s: XVS_FONT_FAMILY=%s\n",R,font_family);
      }  else {
         font_family = (char *) Default_font_family;
         if( ltrace ) 
            fprintf(stderr,"%s: Default_font_family=%s\n",R,font_family);
      }
      for( i = 0; i < nsize; i++ ) {
/*       First try to load 100 dpi fonts ... */
         sprintf(B,"*%s-bold-r-normal--*-%d-100-100-*",font_family,sizes[i]);
         if( !(Xvs_font[i]   = XLoadQueryFont(fl_display,B)) ) {
/*          ... and if that fails, get any available resolution ... */
            sprintf(B,"*%s-bold-r-normal--*-%d-*-*-*",font_family,sizes[i]);
            if( !(Xvs_font[i]   = XLoadQueryFont(fl_display,B)) ) {
               fprintf(stderr,"xvs_set_font:  Warning!! Unable to load font.  Name='%s' Size=%d\n",B,sizes[i]);
            }
         }
         if( Xvs_font[i] ) {
            Xvs_font_h[i] =  Xvs_font[i]->max_bounds.ascent + 
                             Xvs_font[i]->max_bounds.descent;
            if( ltrace ) {
               fprintf(stderr,"%s: Font %2d  name='%s' height=%d\n",
                  R,i,B,Xvs_font_h[i]);
            }
            glXUseXFont(Xvs_font[i]->fid,offset,nglyph,dl_base_next);
            dl_base[i] = dl_base_next;
            dl_base_next += nglyph;
         }
      }
      first = 0;
   }

/* If isize parameter < 0, return the number of fonts which have been defined ... */
   if( isize < 0 ) return nsize;

   if( isize < nsize ) {
      if( ltrace ) fprintf(stderr,"xvs_set_font: Returning %d\n",dl_base[isize]);
      return dl_base[isize];
   } else {
      return 0;
   }
}

/* Plots a 24-pt test message ... */

void gl_test_string(WT wt, char *string) {
   char *p = string;
   int    ltrace = Ltrace;
   double x, y;
   int    i, ip;
   double wt_frac;

#define CO  wds[wt].pf.pxyf->dqc->dq
#define ICO wds[wt].winf.wiq
   if( !string || !strlen(string) ) return;

   glColor3d(1.0,1.0,0.0);
   x = CO.x1;
   y = CO.y1;
   glListBase(xvs_set_font(NULL,0));
   mglRasterPos2d(x,y);
   glCallLists(strlen(string),GL_BYTE,(GLbyte *) string);
   dglRasterPos2d();

   wt_frac = DRaster_pos.x / (ICO.x2 - ICO.x1);
   if( ltrace ) fprintf(stderr,
      "gl_test_string: Length of '%s' rendered in 24 pt type is %g, %g of domain\n",
      string,DRaster_pos.x,wt_frac);

   wt_frac = xvs_wtfracstringwidth(wt,0,string);
   if( ltrace ) fprintf(stderr,"gl_test_string: xvs_wtfracstringwidth() returns %g\n",
      wt_frac);
#undef CO
#undef ICO

}

void display_XFonts(void) {
   static char R[] = "display_XFonts";
   static char B[BUFLEN];
   static char Pattern[] = 
      "*helvetica-bold-r-normal--*-*-100-100-*-iso8859-1";
   int    ltrace = Ltrace;

   char **fontnames = (char **) NULL;
   XFontStruct *fontinfo = (XFontStruct *)  NULL;
   int maxfont = 10000, nfont;

   int i;

   if( ltrace ) fprintf(stderr,"%s: In routine.\n",R);
   if( !Xvsf_info ) {
      fprintf(stderr,"%s: FATAL: Xvsf_info form has not been created.\n",R);
      exit(1);
   }
   if( fl_form_is_visible(Xvsf_info->info) ) return;

   fl_show_form(Xvsf_info->info,FL_PLACE_GEOMETRY,FL_FULLBORDER,"info");
   fl_redraw_form(Xvsf_info->info);
   fontnames = XListFontsWithInfo(fl_display,Pattern,maxfont,&nfont,&fontinfo);
   sprintf(B,"%s: %d font names returned",R,nfont);
   fprintf(stderr,"%s\n",B);
   fl_clear_browser(Xvsf_info->browser_info);
   for( i = 0; i < nfont; i++ ) {
      fprintf(stderr,"%s: Font %d='%s'\n",R,i,fontnames[i]);
      fl_addto_browser(Xvsf_info->browser_info,(const char *) fontnames[i]);
   }
   fl_set_object_label(Xvsf_info->text_info,B);
   XFreeFontInfo(fontnames,fontinfo,nfont);
}

/* Replaces 'equate_screen_and_world_coordinates' ... */
void xvs_use_raster_coords(WT wt, int code) {
   int ltrace = Ltrace;
#define ICO wds[wt].winf.wiq
   double x1, x2, y1, y2;
   switch( code ) {
   case ON:
      glMatrixMode(GL_MODELVIEW); 
      glLoadIdentity();
      x1 = 0;
      y1 = 0;
      x2 = ICO.x2 - ICO.x1;
      y2 = ICO.y2 - ICO.y1;
      if( ltrace ) {
         fprintf(stderr,
            "xvs_use_raster_coords: Calling gluOrtho2D(%g,%g,%g,%g)\n",
            x1,x2,y1,y2);
      }                                 
      gluOrtho2D(x1,x2,y1,y2);
      break;
   case OFF:
      window_ortho2(wt);
      break;
   default:
      fprintf(stderr,"xvs_use_raster_coords:  Unrecognized code=%d\n",code);
      exit(1);
   }
#undef  ICO
}


raster_pos xvs_wtdisplaystringoffset(WT wt,int isize,char *string,int col,
                               double xoffset,double yoffset,
                               double x,double y,int code,
                               int opcode) {
#define ICO wds[wt].winf.wiq
   raster_pos   the = {-1,-1};
   double       w,   h;
   int          ltrace = OFF;
   rgb_color    color;

   static int first = 1;

   if( !is_active_wt(wt) ) return the;

   if( opcode != ANNOTATION_COMPUTE_SIZES &&  opcode != ANNOTATION_DISPLAY ) {
      fprintf(stderr,"xvs_wtdisplaystringoffset: FATAL: Invalid opcode=%d\n",
         opcode);
      exit(1);
   }

   xvs_use_raster_coords(wt,ON);

   if( opcode == ANNOTATION_COMPUTE_SIZES ) {
      w = ICO.x2 - ICO.x1;
      h = ICO.y2 - ICO.y1;
      switch( code ) {
      case DS_LL:
         the.x = xoffset * w;
         the.y = yoffset * h;
         break;
      case DS_UL:
         the.x = xoffset * w;
         the.y = h * (1.0 - yoffset);
         break;
      case DS_UR:
         the.x = w * (1.0 - xoffset) - xvs_getstrwidth(wt,isize,string);
         the.y = h * (1.0 - yoffset);
         break;
      case DS_LR:
         the.x = w * (1.0 - xoffset) - xvs_getstrwidth(wt,isize,string);
         the.y = yoffset * (ICO.y2 - ICO.y1);
         break;
      default:
         fprintf(stderr,"xvs_wtdisplaystringoffset: Unknown code: %d.\n",code);
         return the;
      }
   } else {
      if( ltrace ) {
         fprintf(stderr,"xvs_wtdisplaystringoffset: wt=%d  Geometry %d %d %d %d\n",wt,
            ICO.x1,ICO.x2,ICO.y1,ICO.y2);
         fprintf(stderr,"xvs_wtdisplaystringoffset: w=%g h=%g\n",w,h);
         fprintf(stderr,"xvs_wtdisplaystringoffset: Displaying '%s' at (%g,%g) isize=%d\n", 
            string,x,y,isize);
      }
      color = CT(col);
      glColor3d(color.r,color.g,color.b);
      mglRasterPos2d(x,y);
      if( ltrace ) dump_glRasterPos2d("xvs_wtdisplaystringoffset");
      glListBase(xvs_set_font(NULL,isize));
      glCallLists(strlen(string),GL_BYTE,(GLbyte *) string);

   }

   xvs_use_raster_coords(wt,OFF);
   set_must_remap(wt);

Return:
   return the;
#undef  ICO
}

/* Computes length of string in raster coordinates.  Does so by drawing 
 * string using various-sized fonts, and hence should be followed by 
 * raster clear.
 */
double xvs_getstrwidth(WT wt,int isize, char *string) {
   int         ltrace = Ltrace;
   int         direction, f_ascent, f_descent;
   XCharStruct xcs;
   GC          gc;
   double      rval;

#define ICO wds[wt].winf.wiq

   xvs_use_raster_coords(wt,ON);
   glColor3d(0.0,0.0,0.0);
   mglRasterPos2d(0.0,0.0);
   glListBase(xvs_set_font(NULL,isize));
   glCallLists(strlen(string),GL_BYTE,(GLbyte *) string);
   dglRasterPos2d();
   rval = DRaster_pos.x;
   if( ltrace ) {
      fprintf(stderr,
         "xvs_getstrwidth: isize=%d string=%s width=%g\n",isize,string,rval);
   }
   xvs_use_raster_coords(wt,OFF);
   return rval;
#undef  ICO
}

double xvs_wtfracstringwidth(WT wt, int isize, char *string) {
#define ICO wds[wt].winf.wiq
   if( (ICO.x2 - ICO.x1) ) {
      return xvs_getstrwidth(wt,isize,string) / ((double) (ICO.x2 - ICO.x1));
   } else {
      return 0;
   }
#undef  ICO 
}

void xvs_auto_annotate_0(WT wt,char *string,ANNOTE_CTL *pthe,int opcode,
                         int isize)  {
   int               i,  nsize;
   double            fr_sw, x, y;   
   int               ltrace = Ltrace;
   raster_pos        pos;

   if( ltrace ) {
      fprintf(stderr,
         "xvs_auto_annotate_0: wt=%d string='%s' opcode=%d isize=%d\n",
         wt,string,opcode,isize);
      dump_ANNOTE_CTL("xvs_auto_annotate_0",*pthe);
   }
   if( opcode != ANNOTATION_COMPUTE_SIZES &&  opcode != ANNOTATION_DISPLAY ) {
      fprintf(stderr,"xvs_auto_annotate_0: FATAL: Invalid opcode=%d\n",opcode);
      exit(1);
   }

   if( !pthe->on ) return;

   switch( opcode ) {
   case ANNOTATION_COMPUTE_SIZES:
      if( isize < 0 ) {
/*       Determine an appropriate font size, and register draw position and 
         string in ANNOTE_CTL structure  ... */
         nsize = xvs_set_font(NULL,-1);
         for( i = pthe->min_fh_index; i < nsize; i++ ) {
            fr_sw = xvs_wtfracstringwidth(wt,i,string);
            if( ltrace ) {
               fprintf(stderr,"xvs_auto_annotate_0: <%s> fr_sw[%d]: %g.\n",
                       string,i,fr_sw);
            }
            if( fr_sw <= pthe->max_xfrac  ) goto DoneSize;
         }
/*       Default case, use smallest size ... */
         i = nsize - 1;
      } else {
/*       Use size specified by routine parameter ... */
         i = isize;
      }
DoneSize:
      Last_isize = i;
      if( Ltrace ) fprintf(stderr,"xvs_auto_annotate_0: Last_isize=%d\n",Last_isize);
      pos = xvs_wtdisplaystringoffset(wt,i,string,
               pthe->col,pthe->xlc,pthe->ylc,x,y,pthe->pos_code,
               ANNOTATION_COMPUTE_SIZES);
      pthe->x = pos.x;
      pthe->y = pos.y;
      pthe->isize = i;
      if( pthe->s ) free(pthe->s);
      pthe->s = strdup(string);
      break;
   case ANNOTATION_DISPLAY:
      (void) xvs_wtdisplaystringoffset(wt,pthe->isize,pthe->s,
               pthe->col,pthe->xlc,pthe->ylc,pthe->x,pthe->y,
               pthe->pos_code,ANNOTATION_DISPLAY);
      break;
   }
   return;
}

#define   IQ       wds[wt].winf.wiq
#define   DQ       (wds[wt].pf.pxyf->dqc)->dq
#define   P        wds[wt].pf.pxyf->xycurr
#define   X        P->data->x
#define   Y        P->data->y
#define   PTAG     P->tdata
#define   TAG      P->tdata->tag
#define   N        P->data->n
#define   NEVENTS  wds[wt].pf.pxyf->nevents 

void xvs_auto_annotate_mouse_pos(WT wt,ANNOTE_CTL *pthe,int opcode,
                                 int isize)  {
   static char       R[] = "xvs_auto_annotate_mouse_pos";
   static char       FMT[BUFLEN], B1[BUFLEN], B2[BUFLEN];
   int               i,  nsize;
   double            fr_sw, x, y; 
   int               ltrace = Ltrace;
   raster_pos        pos;
   static FL_Coord   x_mouse,   y_mouse;
   static double     x_mouse_w, y_mouse_w;
   static int        inear, w, h;
   int               strw;
   static int        dx = 5;
   double            ystretch = 1.0;

   if( wt != S_wt || !is_mouse_in_wt(wt) || !xvs_has_focus() ) return;

   if( ltrace ) {
      fprintf(stderr,"%s: wt=%d opcode=%d isize=%d\n",R,wt,opcode,isize);
      dump_ANNOTE_CTL("xvs_auto_annotate_mouse_pos",*pthe);
   }
   if( opcode != ANNOTATION_COMPUTE_SIZES &&  opcode != ANNOTATION_DISPLAY ) {
      fprintf(stderr,"xvs_auto_annotate_mouse_pos: FATAL: Invalid opcode=%d\n",opcode);
      exit(1);
   }

   if( !pthe->on ) return;

   switch( opcode ) {
   case ANNOTATION_COMPUTE_SIZES:

/*    Get mouse position and index/coords of point closest to mouse */

      w = IQ.x2 - IQ.x1;
      h = IQ.y2 - IQ.y1;
      xymouse_wt(wt,&x_mouse,&y_mouse);
#ifndef XFORMS_OPEN
      y_mouse = wds[wt].winf.wiq.y2 - y_mouse;
      x_mouse = x_mouse - wds[wt].winf.wiq.x1;
#else
      y_mouse = h - y_mouse + 1;
#endif
      xymouse_w_wt(wt,&x_mouse_w,&y_mouse_w);
      inear = ixdv2nearest(X,Y,N,x_mouse_w,y_mouse_w,get_aspect_wt(wt)); 
      Highlight_nearest = inear;
      sprintf(FMT,"(%s , %s)",pthe->fmt,pthe->fmt);
      sprintf(B1,FMT,x_mouse_w,y_mouse_w);
      sprintf(FMT,"(%s , %s) [%%d]",pthe->fmt,pthe->fmt);
      sprintf(B2,FMT,X[inear],Y[inear],inear+1); 
      if( ltrace ) {
         fprintf(stderr,"%s: mouse=(%d,%d)\n",R,x_mouse,y_mouse);
         fprintf(stderr,"%s: (xlc,ylc)=(%lf,%lf)\n",R,pthe->xlc,pthe->ylc);
         fprintf(stderr,"%s: x=%d..%d  y=%d..%d\n",R,IQ.x1,IQ.x2,IQ.y1,IQ.y2);
         fprintf(stderr,"%s: %s\n",R,B1);
         fprintf(stderr,"%s: %s\n",R,B2);
      }
      if( isize < 0 ) {
/*       Determine an appropriate font size, and register draw position and 
         string in ANNOTE_CTL structure  ... */
         nsize = xvs_set_font(NULL,-1);
         for( i = pthe->min_fh_index; i < nsize; i++ ) {
            fr_sw = xvs_wtfracstringwidth(wt,i,B2);
            if( ltrace ) {
               fprintf(stderr,"xvs_auto_annotate_mouse_pos: <%s> fr_sw[%d]: %g.\n",
                       B2,i,fr_sw);
            }
            if( fr_sw <= pthe->max_xfrac  ) goto DoneSize;
         }
/*       Default case, use smallest size ... */
         i = nsize - 1;
      } else {
/*       Use size specified by routine parameter ... */
         i = isize;
      }
DoneSize:
      pthe->isize = i;
      strw =  xvs_getstrwidth(wt,pthe->isize,B2) + dx;
      if( x_mouse + strw > w  &&   x_mouse - strw > 0 ) {
            x_mouse -= strw;
      }
      if( y_mouse + (ystretch + 1.0) * Xvs_font_h[pthe->isize] > h ) {
         y_mouse -= (ystretch + 1.0) * Xvs_font_h[pthe->isize];
      }
/*
      Last_isize = i;
      if( Ltrace ) fprintf(stderr,"xvs_auto_annotate_0: Last_isize=%d\n",Last_isize);
      pos = xvs_wtdisplaystringoffset(wt,i,string,
               pthe->col,pthe->xlc,pthe->ylc,x,y,pthe->pos_code,
               ANNOTATION_COMPUTE_SIZES);
      pthe->x = pos.x;
      pthe->y = pos.y;
      if( pthe->s ) free(pthe->s);
      pthe->s = strdup(string);
*/
      break;
   case ANNOTATION_DISPLAY:
      (void) xvs_wtdisplaystringoffset(wt,pthe->isize,B1,
               pthe->col,pthe->xlc,pthe->ylc,
               x_mouse,y_mouse + ystretch * Xvs_font_h[pthe->isize],
               pthe->pos_code,ANNOTATION_DISPLAY);
      (void) xvs_wtdisplaystringoffset(wt,pthe->isize,B2,
               pthe->col,pthe->xlc,pthe->ylc,
               x_mouse,y_mouse,
               pthe->pos_code,ANNOTATION_DISPLAY);
      break;
   }
   return;

}
#undef    IQ
#undef    DQ
#undef    P
#undef    X
#undef    Y
#undef    PTAG
#undef    TAG
#undef    N
#undef    NEVENTS
