#include "project.h"

/* Applicators, etc. */
/* Modified to consolidate applicator info. in single table entry  ... */

APPLY3_TYPE0      Apply3_table[] = { 

   { NULL, NULL,  NULL,  NULL,  NULL,  NULL,  NULL },

   { pxydv_1minus2_vs_2minus3,
     "[1] - [2] vs [2] - [3]",
     "first window",
     "second window",
     "third window",
     wts_are_merge_compatible, 
     "[1] - [2] vs [2] - [3]" },

   { pxydv_v_of_a_lapse_and_u,
     "v(a,lapse,u)",
     "a",
     "lapse",
     "u",
     wts_are_fully_compatible, 
     "v" },

   { pxydv_V_and_T_of_V_X_and_Y,
     "V & T(V,X,Y)",
     "V",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "[V & T]" },

   { pxydv_xyofpppia,
     "[X,Y](pp,pi,a)",
     "pp",
     "pi",
     "a",
     wts_are_fully_compatible, 
     "xy" },

   { pxydv_rdXdtau,
     "r dX/dtau",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dX/dtau" },

   { pxydv_rdYdtau,
     "r dY/dtau",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dY/dtau" },

   { pxydv_kappa_of_ph_pp_zeta,
     "k(ph,pp,zeta)",
     "ph",
     "pp",
     "zeta",
     wts_are_fully_compatible, 
     "k" }

                                   };

int               N_Apply3_table =
                     sizeof(Apply3_table) / sizeof(APPLY3_TYPE0);
 
#define           SAVE             PXYDVBIND    lp = p;
#define           FOR              for( p = lp; p != NULL; p = p->next, p1 = p1->next, p2 = p2->next, p3 = p3->next )
#define           RETURN           return lp;

#define           X                p->data->x
#define           Y                p->data->y
#define           N                p->data->n
#define           T                p->time
#define           TAG              p->tag

#define           LXY1             wds[wt1].pf.pxyf->lxy
#define           NEVENTS1         wds[wt1].pf.pxyf->nevents
#define           HEAD1            wds[wt1].pf.pxyf->xyhead
#define           X1               p1->data->x
#define           Y1               p1->data->y
#define           N1               p1->data->n
#define           T1               p1->time
#define           TAG1             p1->tag

#define           LXY2             wds[wt2].pf.pxyf->lxy
#define           NEVENTS2         wds[wt2].pf.pxyf->nevents
#define           HEAD2            wds[wt2].pf.pxyf->xyhead
#define           X2               p2->data->x
#define           Y2               p2->data->y
#define           N2               p2->data->n
#define           T2               p2->time
#define           TAG2             p2->tag

#define           LXY3             wds[wt3].pf.pxyf->lxy
#define           NEVENTS3         wds[wt3].pf.pxyf->nevents
#define           HEAD3            wds[wt3].pf.pxyf->xyhead
#define           X3               p3->data->x
#define           Y3               p3->data->y
#define           N3               p3->data->n
#define           T3               p3->time
#define           TAG3             p3->tag

PXYDVBIND pxydv_1minus2_vs_2minus3(PXYDVBIND p, PXYDVBIND p1,
                                   PXYDVBIND p2,PXYDVBIND p3) {
   PXYDVBIND    lp = p, tp,  ltp, tp2,  tp3;

   tp  = clone_XYDVLIST(p1);
   ltp = tp;
   tp2 = clone_XYDVLIST(p2);
   tp3 = clone_XYDVLIST(p3);

   tp2 = pxydv_interpolate_x1_to_x2(tp2,p1);
   tp3 = pxydv_interpolate_x1_to_x2(tp3,p1);

   for( ; tp ; tp = tp->next, tp2 = tp2->next, tp3 = tp3->next ) {
      dvvs(tp2->data->y,tp3->data->y,tp->data->x,tp->data->n);
      dvvs(tp->data->y,tp2->data->y,tp->data->y,tp->data->n);
   }

   if( tp2 ) free_XYDVLIST(tp2);
   if( tp3 ) free_XYDVLIST(tp3);
   if( lp  ) free_XYDVLIST(lp);

   return ltp;
}

PXYDVBIND pxydv_v_of_a_lapse_and_u(PXYDVBIND p,PXYDVBIND p1,
                                   PXYDVBIND p2,PXYDVBIND p3) {
   SAVE;
   FOR {
      dvvdxx_(Y2,Y1,Y,&N);
      dvvm_  (Y,Y3,Y,&N);
   }
   RETURN;
}

PXYDVBIND pxydv_xyofpppia(p,p1,p2,p3) 
   PXYDVBIND    p,  p1,  p2,  p3;
{

/* Y1 --> pp;    Y2 --> pi;    Y3 --> a;    X1, X2, X3 --> r; */

   double            Rt2pi = 2.506628274631000;
   SAVE;
   FOR {
      dvvd_(Y1,Y3,X,&N);
      dvvsm_(X,X1,X,&Rt2pi,&N);
      dvvd_(Y2,Y3,Y,&N);
      dvvsm_(Y,X1,Y,&Rt2pi,&N);
   }
   RETURN;
}

/* Y1 --> a;    Y2 --> X;    Y3 --> Y;                 */
/* X1, X2, X3 --> r;                                   */

#define           AJ     Y1[j]
#define           XJ     Y2[j]
#define           YJ     Y3[j]
#define           DYDRJ  DYDR[j]
#define           DXDRJ  DXDR[j]
#define           RJ     X1[j]
#define           XY_ASSIGN(y,n)   X[j + n * N1] = X1[j]; Y[j + n * N1] = y

PXYDVBIND pxydv_rdXdtau(p,p1,p2,p3)
   PXYDVBIND      p,  p1,  p2,  p3;
{

   DVEC           DYDR;
   double         t0,    t1,    t2,    t3;
   int            j;

   SAVE;
   FOR {
      DYDR = make_DVEC(N);
      (void) dvd1nu_(Y3,X3,DYDR,&N3);
      for( j = 0; j < N1; j++ ) {
         t0 = RJ * DYDRJ / AJ;
         t1 = 0.5 * (AJ * AJ - 3.0) * YJ / AJ;
         t2 = AJ * YJ * (YJ * YJ - XJ * XJ);
         t3 = t0 + t1 + t2;
         XY_ASSIGN(t3,0); 
      }
      free_DVEC(DYDR);
   }
   RETURN;
}

PXYDVBIND pxydv_rdYdtau(p,p1,p2,p3)
   PXYDVBIND      p,  p1,  p2,  p3;
{
   DVEC           DXDR;
   double         t0,    t1,    t2,    t3;
   int            j;

   SAVE;
   FOR {
      DXDR = make_DVEC(N);
      (void) dvd1nu_(Y2,X2,DXDR,&N2);
      for( j = 0; j < N1; j++ ) {
         t0 = (RJ * DXDRJ + XJ) / AJ;
         t1 = 0.5 * (AJ * AJ - 1.0) * XJ / AJ;
         t2 = AJ * XJ * (XJ * XJ - YJ * YJ);
         t3 = t0 + t1 + t2;
         XY_ASSIGN(t3,0); 
      }
      free_DVEC(DXDR);
   }
   RETURN;

}


#undef            AJ    
#undef            XJ     
#undef            YJ    
#undef            DYDRJ
#undef            DXDRJ 
#undef            RJ   
#undef            XY_ASSIGN

PXYDVBIND pxydv_kappa_of_ph_pp_zeta(p,p1,p2,p3)
   PXYDVBIND      p, p1, p2, p3;
{
   DVEC           KAPPA;
   double         xi = 1.0;

   SAVE;
   GET_DOUBLE(xi,xi,"xi");
   if( xi < 0.0 ) xi = 1.0 / -xi;
   FOR {
      KAPPA = make_DVEC(N);
#ifndef __ 
      (void) dvkappa_of_ph_pp_zeta_(Y1,Y2,Y3,X,KAPPA,&xi,&N);
#else
      (void) dvkappa_of_ph_pp_zeta__(Y1,Y2,Y3,X,KAPPA,&xi,&N);
#endif
      (void) dvcopy_(KAPPA,Y,&N);
      free_DVEC(KAPPA);
   }
   RETURN;
}

#define    VV      Y1
#define    XX      Y2
#define    YY      Y3

PXYDVBIND  pxydv_V_and_T_of_V_X_and_Y(
             PXYDVBIND p,PXYDVBIND p1,PXYDVBIND p2,PXYDVBIND p3) {
   double  c_pi = M_PI;

   SAVE
   FOR {
      p = expand_PXYDVBIND_nfold(p,2);
      dvcopy_(X1,X,&N1);
      dvcopy_(X1,X+N1,&N1);
      dvdofs_(XX,YY,Y,&N1);
      dvvm_(X1,VV,Y+N1,&N1);
      dvvsm_(Y+N1,X1,Y+N1,&c_pi,&N1);
      dvfapl_(Y,Y,D_ABS,&N);
   }
   set_marker_multi_mode(wt_from_xyhead(lp));
   RETURN;
}

#undef     VV    
#undef     XX     
#undef     YY      



#undef            SAVE             
#undef            FOR             
#undef            RETURN              

#undef            X              
#undef            Y              
#undef            N              
#undef            T              
#undef            TAG              

#undef            LXY1             
#undef            NEVENTS1         
#undef            HEAD1            
#undef            X1               
#undef            Y1               
#undef            N1               
#undef            T1               
#undef            TAG1             

#undef            LXY2             
#undef            NEVENTS2         
#undef            HEAD2            
#undef            X2               
#undef            Y2               
#undef            N2               
#undef            T2               
#undef            TAG2             

#undef            LXY3             
#undef            NEVENTS3         
#undef            HEAD3            
#undef            X3               
#undef            Y3               
#undef            N3               
#undef            T3               
#undef            TAG3             
