#include "project.h"

/* Applicators, etc. */
/* Modified to consolidate applicator info. in single table entry  ... */

APPLY4_TYPE0      Apply4_table[] = { 

   { NULL, NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL },

   { pxydv_rdXdt,
     "r dX/dt",
     "lapse",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dX/dt" },

   { pxydv_rdXdt_terms,
     "r dX/dt & terms",
     "lapse",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dX/dt & terms" },

   { pxydv_rdXdt_linear_terms,
     "r dX/dt linear terms",
     "lapse",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dX/dt linear terms" },

   { pxydv_rdXdt_nonlinear_terms,
     "r dX/dt nonlinear terms",
     "lapse",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dX/dt nonlinear terms" },

   { pxydv_rdYdt,
     "r dY/dt",
     "lapse",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dY/dt" },

   { pxydv_rdYdt_terms,
     "r dY/dt & terms",
     "lapse",
     "a",
     "X",
     "Y",
     wts_are_fully_compatible, 
     "r dY/dt & terms" },

   { pxydv_hyperbolic_expansion,
     "expansion",
     "Gamma[th,th,r]",
     "A",
     "K[th,th]",
     "Br",
     wts_are_fully_compatible, 
     "expansion" }

                                   };

int               N_Apply4_table =
                     sizeof(Apply4_table) / sizeof(APPLY4_TYPE0);
 
#define           SAVE             PXYDVBIND    lp = p;
#define           FOR              for( p = lp; p != NULL; p = p->next, p1 = p1->next, p2 = p2->next, p3 = p3->next, p4 = p4->next )
#define           RETURN           return lp;

#define           X                p->data->x
#define           Y                p->data->y
#define           N                p->data->n
#define           T                p->time
#define           TAG              p->tag

#define           LXY1             wds[wt1].pf.pxyf->lxy
#define           NEVENTS1         wds[wt1].pf.pxyf->nevents
#define           HEAD1            wds[wt1].pf.pxyf->xyhead
#define           X1               p1->data->x
#define           Y1               p1->data->y
#define           N1               p1->data->n
#define           T1               p1->time
#define           TAG1             p1->tag

#define           LXY2             wds[wt2].pf.pxyf->lxy
#define           NEVENTS2         wds[wt2].pf.pxyf->nevents
#define           HEAD2            wds[wt2].pf.pxyf->xyhead
#define           X2               p2->data->x
#define           Y2               p2->data->y
#define           N2               p2->data->n
#define           T2               p2->time
#define           TAG2             p2->tag

#define           LXY3             wds[wt3].pf.pxyf->lxy
#define           NEVENTS3         wds[wt3].pf.pxyf->nevents
#define           HEAD3            wds[wt3].pf.pxyf->xyhead
#define           X3               p3->data->x
#define           Y3               p3->data->y
#define           N3               p3->data->n
#define           T3               p3->time
#define           TAG3             p3->tag

#define           LXY4             wds[wt4].pf.pxyf->lxy
#define           NEVENTS4         wds[wt4].pf.pxyf->nevents
#define           HEAD4            wds[wt4].pf.pxyf->xyhead
#define           X4               p4->data->x
#define           Y4               p4->data->y
#define           N4               p4->data->n
#define           T4               p4->time
#define           TAG4             p4->tag

/* Y1 --> lapse  Y2 --> a;    Y3 --> X;    Y4 --> Y        */
/* X1, X2, X3, X4 --> r;                                   */

#define           LJ     Y1[j]
#define           AJ     Y2[j]
#define           XJ     Y3[j]
#define           YJ     Y4[j]
#define           DYDRJ  DYDR[j]
#define           DXDRJ  DXDR[j]
#define           RJ     X1[j]
#define           XY_ASSIGN(y,n)   X[j + n * N1] = X1[j]; Y[j + n * N1] = y

PXYDVBIND pxydv_rdXdt(p,p1,p2,p3,p4) 
   PXYDVBIND      p,  p1,  p2,  p3,  p4;
{

   DVEC           DYDR;
   double         t0,    t1,    t2,    t3;
   int            j;

   SAVE;
   FOR {
      DYDR = make_DVEC(N);
      (void) dvd1nu_(Y4,X4,DYDR,&N4);
      for( j = 0; j < N1; j++ ) {
         t0 = LJ * RJ * DYDRJ / AJ;
         t1 = 0.5 * LJ * (AJ * AJ - 3.0) * YJ / AJ;
         t2 = AJ * LJ * YJ * (YJ * YJ - XJ * XJ);
         t3 = t0 + t1 + t2;
         XY_ASSIGN(t3,0); 
      }
      free_DVEC(DYDR);
   }
   RETURN;

}

PXYDVBIND pxydv_rdXdt_terms(p,p1,p2,p3,p4) 
   PXYDVBIND      p,  p1,  p2,  p3,  p4;
{
   DVEC           DYDR;
   double         t0,    t1,    t2,    t3;
   int            j;

   SAVE;
   FOR {
      DYDR = make_DVEC(N);
      (void) dvd1nu_(Y4,X4,DYDR,&N4);
      p = expand_PXYDVBIND_nfold(p,4);
      for( j = 0; j < N1; j++ ) {
         t0 = LJ * RJ * DYDRJ  / AJ;
         t1 = 0.5 * LJ * (AJ * AJ - 3.0) * YJ / AJ;
         t2 = AJ * LJ * YJ * (YJ * YJ - XJ * XJ);
         t3 = t0 + t1 + t2;
         XY_ASSIGN(t0,0); XY_ASSIGN(t1,1); XY_ASSIGN(t2,2); XY_ASSIGN(t3,3);
      }
      free_DVEC(DYDR);
   }
   set_marker_multi_mode(wt_from_xyhead(lp));
   RETURN;

}

PXYDVBIND pxydv_rdXdt_linear_terms(p,p1,p2,p3,p4) 
   PXYDVBIND      p,  p1,  p2,  p3,  p4;
{
   DVEC           DYDR;
   double         t0,    t1,    t3;
   int            j;

   SAVE;
   FOR {
      DYDR = make_DVEC(N);
      (void) dvd1nu_(Y4,X4,DYDR,&N4);
      for( j = 0; j < N1; j++ ) {
         t0 = LJ * RJ * DYDRJ  / AJ;
         t1 = 0.5 * LJ * (AJ * AJ - 3.0) * YJ / AJ;
         t3 = t0 + t1;
         XY_ASSIGN(t3,0); 
      }
      free_DVEC(DYDR);
   }
   RETURN;

}

PXYDVBIND pxydv_rdXdt_nonlinear_terms(p,p1,p2,p3,p4) 
   PXYDVBIND      p,  p1,  p2,  p3,  p4;
{
   double         t2;
   int            j;

   SAVE;
   FOR {
      for( j = 0; j < N1; j++ ) {
         t2 = AJ * LJ * YJ * (YJ * YJ - XJ * XJ);
         XY_ASSIGN(t2,0); 
      }
   }
   RETURN;

}

PXYDVBIND pxydv_rdYdt(p,p1,p2,p3,p4) 
   PXYDVBIND      p,  p1,  p2,  p3,  p4;
{
   DVEC           DXDR;
   double         t0,    t1,    t2,    t3;
   int            j;

   SAVE;
   FOR {
      DXDR = make_DVEC(N);
      (void) dvd1nu_(Y3,X3,DXDR,&N3);
      for( j = 0; j < N1; j++ ) {
         t0 = LJ * (RJ * DXDRJ + XJ) / AJ;
         t1 = 0.5 * LJ * (AJ * AJ - 1.0) * XJ / AJ;
         t2 = AJ * LJ * XJ * (XJ * XJ - YJ * YJ);
         t3 = t0 + t1 + t2;
         XY_ASSIGN(t3,0); 
      }
      free_DVEC(DXDR);
   }
   RETURN;

}

PXYDVBIND pxydv_rdYdt_terms(p,p1,p2,p3,p4) 
   PXYDVBIND      p,  p1,  p2,  p3,  p4;
{
   DVEC           DXDR;
   double         t0,    t1,    t2,    t3;
   int            j;

   SAVE;
   FOR {
      DXDR = make_DVEC(N);
      (void) dvd1nu_(Y3,X3,DXDR,&N3);
      p = expand_PXYDVBIND_nfold(p,4);
      for( j = 0; j < N1; j++ ) {
         t0 = LJ * (RJ * DXDRJ + XJ) / AJ;
         t1 = 0.5 * LJ * (AJ * AJ - 1.0) * XJ / AJ;
         t2 = AJ * LJ * XJ * (XJ * XJ - YJ * YJ);
         t3 = t0 + t1 + t2;
         XY_ASSIGN(t0,0); XY_ASSIGN(t1,1); XY_ASSIGN(t2,2); XY_ASSIGN(t3,3);
      }
      free_DVEC(DXDR);
   }
   set_marker_multi_mode(wt_from_xyhead(lp));
   RETURN;

}

PXYDVBIND pxydv_hyperbolic_expansion(PXYDVBIND p,
          PXYDVBIND p1,PXYDVBIND p2,PXYDVBIND p3,PXYDVBIND p4) {
   DVEC   VTMP;
   double f77_two = 2.0;
   SAVE;
   FOR {
      VTMP = make_DVEC(N);
      dvvdxx_(Y1,Y2,Y,&N);
      dvvm_(Y3,Y4,VTMP,&N);
      dvvs_(Y,VTMP,Y,&N);
      dvsm_(Y,&f77_two,Y,&N);
      free_DVEC(VTMP);
   }
   RETURN;
}


#undef            LJ     
#undef            AJ    
#undef            XJ   
#undef            YJ   
#undef            DYDRJ 
#undef            DXDRJ 
#undef            RJ   
#undef            XY_ASSIGN

#undef            SAVE             
#undef            FOR             
#undef            RETURN              

#undef            X              
#undef            Y              
#undef            N              
#undef            T              
#undef            TAG              

#undef            LXY1             
#undef            NEVENTS1         
#undef            HEAD1            
#undef            X1               
#undef            Y1               
#undef            N1               
#undef            T1               
#undef            TAG1             

#undef            LXY2             
#undef            NEVENTS2         
#undef            HEAD2            
#undef            X2               
#undef            Y2               
#undef            N2               
#undef            T2               
#undef            TAG2             

#undef            LXY3             
#undef            NEVENTS3         
#undef            HEAD3            
#undef            X3               
#undef            Y3               
#undef            N3               
#undef            T3               
#undef            TAG3             

#undef            LXY4             
#undef            NEVENTS4         
#undef            HEAD4            
#undef            X4               
#undef            Y4               
#undef            N4               
#undef            T4               
#undef            TAG4             
