/*
  asctoxvs

  Sends a 1D ASCII file in the format specified in the usage message
  to xvs.

  Copyright (c) 2009 by Matthew W. Choptuik
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <v.h>
#include <bbhutil.h>
#include <sdf_priv.h>
#include <sdf.h>
#include <math.h>
#include <cliser.h>
#include "libxvs.h"
#include "sdf_util.h"

#define IVLEN  1024
#define BUFLEN 1024
#define NTMAX  1024*1024

char *P;

void die(char *mess) {
   fprintf(stdout,"%s: %s\n",P,mess);
   exit(1);
}

void usage() {
   fprintf(stdout,"Synopsis:\n");
   fprintf(stdout,"  Sends an ASCII file to the xvs visualization server\n\n");
   fprintf(stdout,"Usage:\n");
   fprintf(stdout,"  %s file [xvsname]\n\n",P);
   fprintf(stdout,"  file    -- file to be transmitted\n");
   fprintf(stdout,"  xvsname -- xvs name for data [defaults to file]\n\n");
   fprintf(stdout,"File Format:\n\n");
   fprintf(stdout,"<time1>\n");
   fprintf(stdout,"<nx1>\n");
   fprintf(stdout,"<x_1> <y_1>\n");
   fprintf(stdout,"<x_2> <y_2>\n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"<x_nx1> <y_nx1>\n");
   fprintf(stdout,"<time2>\n");
   fprintf(stdout,"<nx2>\n");
   fprintf(stdout,"<x_1> <y_1>\n");
   fprintf(stdout,"<x_2> <y_2>\n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"<x_nx2> <y_nx2>\n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"    .     \n");
   fprintf(stdout,"    .     \n");
   exit(0);
}

int main(int argc, char **argv) {
   int     ltrace =  0;
   int     opt, argerr = 0, 
           addcoords = 0, rename = 0,  xremap = 0, defsiv = 0;
   char    siv[BUFLEN], oname[BUFLEN], buf[BUFLEN], wd[BUFLEN];
   char   *fname, *xfname = (char *) NULL;
   char   *xvsname = (char *) NULL;
   int     iv[IVLEN];
   int     ifile, nfile, nt, it, pass, noerr, vsrc;

   FILE   *stream = (FILE *) NULL;
   FILE   *file = (FILE *) NULL;
   int     i, ixy, nx;
 
   int     version, rank, dsize, csize;
   double  time; 

   int    *shape  = (int *) NULL;
   double *bbox   = (double *) NULL, *coords = (double *) NULL,
          *data   = (double *) NULL;
   char   *pname  = (char *) NULL, *cnames = (char *) NULL, 
          *tag    = (char *) NULL;

   int     versionx, rankx, dsizex, csizex, ntlevsx, xonestep, cd = 0;
   double  timex; 

   int    *shapex = (int *) NULL;
   double *bboxx  = (double *) NULL, *coordsx = (double *) NULL,
          *datax  = (double *) NULL, *tvecx   = (double *) NULL;
   char   *pnamex = (char *) NULL, *cnamesx = (char *) NULL, 
          *tagx   = (char *) NULL;

   int     itout, ntout, ndataout;
   int    *start  = (int *) NULL;
   double *xout   = (double *) NULL, *yout = (double *) NULL, 
          *tout   = (double *) NULL;
   int     maxnt  = 100000;
   int     maxxy  = 10000000;

   gft_sdf_file_data *gp, *gpx; 

   P = argv[0];
   while( (opt = getopt(argc,argv,"cdi:n:x:")) != EOF) {
      switch( opt ){
      case 'c' :
         addcoords=1;
         break;
      case 'i' :
         sprintf(siv,"output:=%s",optarg);
         defsiv = 1;
         break;
      case 'n' :
         rename=1;
         strcpy(oname,optarg);
         break;
      case 'x' :
         xremap=1;
         xfname = optarg;
         break;
      case 'd' :
         cd=1;
         break;
      default :
         argerr=1;
         break;
      }
   }

   if( ! (2 <= argc  &&  argc <= 3 ) ) usage();

   fname = argv[1];
   xvsname = (argc > 2) ? argv[2] : argv[1];

/* Disable "interactive" xvs mode ... */
#ifdef HAVE_UNSETENV
   unsetenv("XVSI");
#endif

   if( ltrace ) {
      fprintf(stderr,"fname: <%s>\n",fname);
      fprintf(stderr,"xvsname: <%s>\n",xvsname);
   }
/*
   nfile = argc - optind;
   if( argerr || nfile < 1 ) usage();
   if( rename && (nfile > 1) ) 
      die("-n argument only makes sense for a single file.");

   if( !defsiv ) sprintf(siv,"output:=1-*");
   if( !sget_ivec_param(siv,"output",iv,IVLEN) ) 
      die("Inusfficient memory to allocate index vector.");
   if( ltrace ) {
      fprintf(stderr,"siv: %s\n",siv ? siv : "NULL");
   }
   fixup_ivec(1,NTMAX,0,iv);

   gft_set_multi();

   if( xremap ) {
      if( gft_extract_sdf(xfname,1,&pnamex,&ntlevsx,&tvecx,&timex,
             &rankx,&shapex,&csizex,&cnamesx,&coordsx,&datax) ) {
         xonestep = (ntlevsx == 1);
         if( !xonestep ) {
            gft_extract_sdf(xfname,-1,&pnamex,&ntlevsx,&tvecx,&timex,
               &rankx,&shapex,&csizex,&cnamesx,&coordsx,&datax);
         }
         gft_close_all();
      } else {
         sprintf(buf,"Error opening x-remap file '%s'",xfname); die(buf);
      }
   }


*/
/* Process each file ... */
#define ABORT noerr = 0; \
        fprintf(stderr,"%s: Processing of '%s' aborted.\n",P,fname)
#define FREENULL(p) if( p ) { free(p); p = NULL; }

   if( ! (file = fopen(fname, "r")) ) {
      fprintf(stderr,"%s: Could not open '%s' for read.\n", P, fname);
      exit(1);
   }
   if( !(start = make_IVEC(maxnt + 1)) ) die("Error allocating 'start'");
   if( !(tout = make_DVEC(maxnt)) ) die("Error allocating 'tout'");
   if( !(xout = make_DVEC(maxxy)) ) die("Error allocating 'xout'");
   if( !(yout = make_DVEC(maxxy)) ) die("Error allocating 'yout'");
   ntout = 0;
   start[ntout] = 1;
   ixy = 0;
   while( ntout < maxnt ) {
      if( fscanf(file, "%lf", tout + ntout) != 1) break;
      if( (fscanf(file, "%d", &nx) != 1 ) ) {
         fprintf(stderr,"%s: Unexpected EOF reading 'nt'\n", P);
         break;
      }
      for( i = 0; i < nx; i++, ixy++ ) {
         if( (fscanf(file, "%lf %lf", xout + ixy, yout + ixy) != 2) ) {
            fprintf(stderr,"%s: Unexpected EOF reading 'x,y'\n", P);
            break;
         }
      }
      ntout++;
      start[ntout] = start[ntout-1] + nx;
      if( ltrace ) fprintf(stderr,"%s: start[%d]=%d t=%g nx=%d\n", P, 
         ntout, start[ntout], tout[ntout-1], nx);
   }
   if( ntout ) {
      if( vsmxynt(xvsname,start,tout,ntout,xout,yout) < 1 ) exit(1);
   } else {
      exit(1);
   }
/*
         if( !strstr(oname," ") && (stream = xvs_stream(XVS_OPEN)) ) {
            sprintf(buf,"go <%s> 1",oname);
            (void) xvs_control_send_message(stream,buf,"'go' failed.",P);
            if( !xvs_recv_status(stream,buf) ) fprintf(stderr,"%s: %s\n",P,buf);
            (void) xvs_stream(XVS_CLOSE);
         }

//       Change working directory of xvs window to
         1) Invocation directory by default
         2) Last_data_dir if -d option supplied ... //

         if( cd && (stream = xvs_stream(XVS_OPEN)) ) {
            sprintf(buf,"cd_ldd <%s>",oname);
            if( ltrace ) fprintf(stderr,"%s: Sending '%s'\n",P,buf);
            (void) xvs_control_send_message(stream,buf,"'cd_ldd' failed.",P);
            if( !xvs_recv_status(stream,buf) ) fprintf(stderr,"%s: %s\n",P,buf);
            (void) xvs_stream(XVS_CLOSE);
         } else if ( stream = xvs_stream(XVS_OPEN) ) {
            getcwd(wd,BUFLEN);
            sprintf(buf,"cd <%s> %s",oname,wd);
            if( ltrace ) fprintf(stderr,"%s: Sending '%s'\n",P,buf);
            (void) xvs_control_send_message(stream,buf,"'cd' failed.",P);
            if( !xvs_recv_status(stream,buf) ) fprintf(stderr,"%s: %s\n",P,buf);
            (void) xvs_stream(XVS_CLOSE);
         }
*/
   FREENULL(xout);
   FREENULL(yout);
   FREENULL(tout);
   FREENULL(start);
   exit(0);
}
