#include "project.h"

#include <sys/poll.h>

FILE *C[MAX_CONNECTION];
int   C_bulkinput[MAX_CONNECTION];

/*=========================================================================
Parses fetch request and returns single index of dataset as SDF. 
=========================================================================*/
PSDFDS  xvs_fetch_SDFDS(char *req,int opcode,int *prc) {
   static   char R[] = "xvs_fetch_SDFDS";
   int      ltrace = Ltrace;
   PXYF     pp;

   PSDFDS   the = (PSDFDS) NULL;
   int      req_index;
   int      wt;

   if( sscanf(req,"%s%[ <]%[^>]%[ >]%d",B,B,B1,B,&req_index) == 5 ) {
      if( ltrace ) 
         fprintf(stderr,"%s: Fetch name='%s' index=%d\n",R,B1,req_index);
      if( (wt = wt_from_name(B1)) >= 0 ) {
         *prc = XVS_FETCH_OK;
      } else {
         *prc = XVS_FETCH_NO_SUCH_NAME;
         sprintf(req,"%s",B1);
      }
   } else {
      *prc = XVS_FETCH_INVALID_REQUEST;
   }
   if( *prc != XVS_FETCH_OK ) return the;

   pp = wds[wt].pf.pxyf;
   if( ltrace ) 
      fprintf(stderr,"%s: wt=%d has %d data sets\n",R,wt,pp->lxy);
   if( req_index < 1 || req_index > pp->lxy ) {
      *prc = XVS_FETCH_NO_SUCH_INDEX;
      return the;
   }
   the = xvs_wt_index_to_SDFDS(wt,req_index);
   return the;
}

/*=========================================================================
Converts requested dataset index to SDF.
=========================================================================*/
PSDFDS xvs_wt_index_to_SDFDS(WT wt, int req_index) {
   int        ltrace = Ltrace;
   PSDFDS     the = (PSDFDS) NULL;
   PXYDVBIND  xy = (PXYDVBIND) NULL;
   int        n;
   double    *x, *y;

   unmake_xy_ring(wt);
   xy = nth_PXYDVBIND(wds[wt].pf.pxyf->xyhead,req_index);
   if( ltrace ) {
      fprintf(stderr,"xvs_wt_index_to_SDFDS: req_index=%d\n",req_index);
      fprintf(stderr,"xvs_wt_index_to_SDFDS: index=%d\n",xy->index);
      fprintf(stderr,"xvs_wt_index_to_SDFDS: time=%g\n",xy->time);
   }
   if( xy ) {
      the = new_SDFDS();
      n = xy->data->n;
      x = xy->data->x;
      y = xy->data->y;
      the->time   = xy->time;
      the->rank   = 1;
      the->dsize  = n;
      the->csize  = n;
      the->pname  = strdup(wds[wt].name);
      the->cnames = strdup("x");
      the->tag = strdup("");
      the->shape = (int *) malloc(sizeof(int));
      the->shape[0] = n;
      the->bbox = (double *) malloc(2 * sizeof(double));
      the->bbox[0] = x[0];
      the->bbox[1] = x[n-1];
      the->coords = Dvcopy(x,n);
      the->data = Dvcopy(y,n);
   }
   restore_ring_state(wt);
   return the;
}



/*=========================================================================
Callback for handling data streams into server, this version maintains
connection database registers callbacks for separate connections.
=========================================================================*/
FL_IO_CALLBACK cb_xvs_socket_1(int ss,FD_main *vsform) {
   static char          R[] = "cb_xvs_socket_1";
   int                  ltrace = Ltrace;
   int                  ltrace_full = OFF;
   struct sockaddr_in   cli;
   int                  lencli = sizeof(struct sockaddr_in);
   int                  cs; 

   FILE                *stream;
   PSDFDS               the = (PSDFDS) NULL, mess = (PSDFDS) NULL;
   int                  nsds = 0, rcmain = 1, rc = 1;
   char                *message;

   char                 opcode[1024], req_name[1024];
   int                  req_index = 1;
   WT                   wt;

   static int           first = 1, i;

   if( ltrace ) fprintf(stderr,"%s: Data on ss=%d, first=%d\n",R,ss,first);

   if( first ) {
      for( i = 0; i < MAX_CONNECTION; i++ ) {
         C[i] = (FILE *) NULL;
         C_bulkinput[i] = OFF;
      }
      first = 0;
   }

#ifdef __BITS_SOCKET_H
   cs = accept(ss,(struct sockaddr *) &cli, (socklen_t *) &lencli); 
#else
   cs = accept(ss,(struct sockaddr *) &cli,               &lencli); 
#endif
   if( ltrace ) fprintf(stderr,"%s: Connection on %d\n",R,cs);
   if( cs >= MAX_CONNECTION ) {
      fprintf(stderr,"%s: FATAL: Out of connections MAX_CONNECTION=%d.\n",
         R,MAX_CONNECTION);
      exit(1);
   }

   assert( !C[cs] );
   if( !(C[cs] = fdopen(cs,"r+")) ) {
      fprintf(stderr,"%s: fdopen(%d) failed\n",R,cs);
      close(cs);
      return NULL;
   } 
   C_bulkinput[cs] = OFF;
   if( ltrace ) fprintf(stderr,"%s: Opened stream on cs=%d\n",R,cs);

   return NULL;   
}

/*=========================================================================
Secondary callback routine, called from polling routines, splits off
handling of messages to cb_xvs_socket_handle_message.
=========================================================================*/
int cb_xvs_socket_1a(int ss,FD_main *vsform) {
   static char          R[] = "cb_xvs_socket_1a";
   int                  ltrace = Ltrace;
   int                  ltrace_full = OFF;

   FILE                *stream;
   PSDFDS               the = (PSDFDS) NULL, mess = (PSDFDS) NULL;
   int                  nsds = 0, rcmain = 1, rc = 1;
   int                  wt;

   stream = C[ss];
   assert( stream );
      if( ltrace ) fprintf(stderr,"%s: Calling new_read_SDFDS nsds=%d\n",
         R,nsds);
      the = new_read_SDFDS(stream,&rcmain); 
      if( ltrace ) fprintf(stderr,"%s: new_read_SDFDS returned rcmain=%d\n",
         R,rcmain);
      if( rcmain ) {
         nsds++;
         if( ltrace ) fprintf(stderr,"%s: Read SDFDS the->rank=%d\n",
            R,the->rank);

         /* Messages/op-codes signaled via rank==-1 ... */

         if( the->rank == -1 ) {

         /* Process message, previously lower-cased vis strl2c(), now don't  ... */

            cb_xvs_socket_handle_message(ss,stream,the->pname);
         } else {

            if( ltrace_full) dump_SDFDS(stderr,the,"cb_xvs_socket_1a: data",1);

         /* Process data set but ignore date with non-finite coords.; replace non-finite 
          * data values with default. */

            if( isdvfinite(the->coords,the->csize) ) {
               if( !isdvfinite(the->data,the->dsize) ) {
                  fprintf(stderr,"%s: Warning! One or more non-finite data values replaced with %g\n",
                     R,Nonfinite_val);
                  dvmkfinite(the->data,the->dsize,Nonfinite_val);
               }
               xvs_insert_SDFDS(the);
            } else {
               fprintf(stderr,"%s: Ignoring data set with non-finite coordinates\n",R);
               dvdump(the->coords,the->csize,"the->coords");
            }
         }
      } else {
         if( ltrace ) {
            fprintf(stderr,"%s: Could not read SDFDS.\n",R);
         }
      }
      if( the ) the = free_SDFDS(the);

   return rcmain;
}

/*=========================================================================
Handles non-data transmissions (messages) from clients ...
=========================================================================*/
int cb_xvs_socket_handle_message(int ss,FILE *stream,char *message) {
   static    char R[] = "cb_xvs_socket_handle_message";
   PSDFDS    sdfds, mess;
   PWTBIND   p;
   char     *name = (char *) NULL;
   int       wt, index, rc;
   int       ltrace = Ltrace;

   if( ltrace ) fprintf(stderr,"%s: In routine ss=%d message='%s'\n",R,ss,message);

   if( !strcmp(message,"stop") ) {

   /* Shutdown ... */

      fprintf(stderr,
         "cb_xvs_socket: Received 'stop'. Shutting down server.\n");
      exit(0);

   } else if( !strncmp(message,"killall",7) ) {
      xvs_close_all_wt(NO);

   } else if( !strncmp(message,"kill",4) ) {
      sscanf(message,"kill%[ <]%[^>]%[ >]",B1,B,B1);
      name = strdup(B);
      if( name ) {
         wt = wt_from_name(name);
         if( wt >= 0 ) {
            xvs_close_wt(wt,NO);
            xvs_send_status(stream,1,"");
         } else {
            sprintf(B1,"No window named '%s'.",name);
            xvs_send_status(stream,0,B1);
         }
         free(name);
      } else {
         sprintf(B1,"Could not parse 'kill' command='%s'",message);
         xvs_send_status(stream,0,B1);
      }

   } else if( !strncmp(message,"raise",5) ) {
      fl_raise_form(Xvsf->main);
      if( fl_form_is_visible(Xvsf_control->control) ) fl_raise_form(Xvsf_control->control);

   } else if( !strncmp(message,"set bulkinput",13) ) {
      C_bulkinput[ss] = ON;
      if( ltrace ) { 
         fprintf(stderr,"%s: C_bulkinput[%d]=%d\n",
            R,ss,C_bulkinput[ss]);
      }

   } else if( !strncmp(message,"unset bulkinput",15) ) {
      C_bulkinput[ss] = OFF;
      if( ltrace ) { 
         fprintf(stderr,"%s: C_bulkinput[%d]=%d\n",
            R,ss,C_bulkinput[ss]);
      }

   } else if( !strncmp(message,"nwin",4) ) {
      sprintf(B,"%d",n_windows());
      xvs_send_buf(stream,B);

   } else if( !strncmp(message,"list",4) ) {
      for( p = tail_wtal(wtal); p != NULL; p = p->prev) {
         xvs_send_status(stream,1,wds[p->wt].name);
      }
      xvs_send_status(stream,0,"");

   } else if( !strncmp(message,"saveall",7) ) {
      if( ltrace ) {
         fprintf(stderr,"%s: Received 'saveall'.  Full message='%s'",R,message);
      }
      if( n_windows() ) {
         xvs_send_status(stream,1,"");
         xvs_recv_buf(stream,B1);
         for( p = tail_wtal(wtal); p != NULL; p = p->prev) {
            if( xvs_save_wt(p->wt,B1) ) {
               sprintf(B,"%s/%s",B1,Sdf_name);
               xvs_send_status(stream,1,B);
            } else {
               sprintf(B,"Save to %s/%s aborted.",B1,Sdf_name);
               xvs_send_status(stream,1,B);
            }
         }
         xvs_send_status(stream,0,"");
      } else {
         xvs_send_status(stream,0,"");
      }

   } else if( !strncmp(message,"save",4) ) {
      if( ltrace ) {
         fprintf(stderr,"%s: Received 'save'.  Full message='%s'\n",R,message);
      }
      if( sscanf(message,"%s%[ <]%[^>]%[ >]",B1,B1,B,B1) == 4 ) {
         name = strdup(B);
         wt = wt_from_name(name);
         if( ltrace ) fprintf(stderr,"%s: wt_from_name(%s)=%d\n",R,name,wt);
      } else {
         fprintf(stderr,"%s: Fatal Could not parse name from save command='%s'\n",
            R,message);
         exit(1);
      }

      if( wt >= 0 ) {
         xvs_send_status(stream,1,"");
         xvs_recv_buf(stream,B1);
         if( xvs_save_wt(wt,B1) ) {
            sprintf(B,"%s/%s",B1,Sdf_name);
            xvs_send_status(stream,1,B);
         } else {
            xvs_send_status(stream,0,"");
         }
      } else {
         sprintf(B,"No window with name='%s'.",name);
         xvs_send_status(stream,0,B);
      }
      if( name ) free(name);


   } else if( !strncmp(message,"get",3) ) {
      if( ltrace ) fprintf(stderr, 
                  "%s: Received 'get'.  Full message='%s'\n",R,message);
      sdfds  = xvs_fetch_SDFDS(message,0,&rc);
      if( ltrace ) fprintf(stderr,"xvs_fetch returns %d\n",rc);
      switch( rc ) {
      case XVS_FETCH_OK:
         xvs_send_status(stream,1,"");
         write_SDFDS(stream,sdfds);
         rewind(stream);
         break;
      case XVS_FETCH_NO_SUCH_NAME:
         sscanf(message,"get %s",B1);
         sprintf(B,"No window named '%s'.",B1);
         xvs_send_status(stream,0,B);
         break;
      case XVS_FETCH_INVALID_REQUEST:
         xvs_send_status(stream,0,"Invalid 'get' request.");
         break;
      case XVS_FETCH_NO_SUCH_INDEX:
         sscanf(message,"get %s %s",B,B1);
         sprintf(B2,"No index %s for window named '%s'.",B1,B);
         xvs_send_status(stream,0,B2);
         break;
      default:
         fprintf(stderr,"%s: Fatal: Unrecognized xvs_fetch_SDFDS() code=%d\n",R,rc);
         exit(1);
         break;
      }
      if( sdfds ) free_SDFDS(sdfds);
   } else if( !strncmp(message,"go",2) ) {
      if( ltrace ) fprintf(stderr,"%s: Received 'go'.  Full message='%s'\n",R,message); 
      if( sscanf(message,"%s%[ <]%[^>]%[ >]%d",B1,B1,B,B1,&index) == 5 ) {
         if( (wt = wt_from_name(B)) >= 0 ) {
            if( 1 <= index && index <= n_XYDV(wt) ) {
               set_xycurr_nth(wt,index);
               set_must_redraw(wt);
               xvs_send_status(stream,1,"");
            } else {
               sprintf(B2,"No index %d for window named '%s'.",index,B);
               xvs_send_status(stream,0,B2);
            }
         } else {
            sprintf(B2,"No window named '%s'.",B);
            xvs_send_status(stream,0,B2);
         }
      } else {
         sprintf(B2,"Parse of get command '%s' failed",message);
         xvs_send_status(stream,0,B2);
      }
   } else if( !strncmp(message,"cd_ldd",6) ) {
      if( ltrace ) {
         fprintf(stderr,"%s: Received 'cd_ldd'.  Full message='%s'\n",R,message);
         fprintf(stderr,
            "%s: Last_data_dir=%s\n",R,Last_data_dir ? Last_data_dir : "NULL");
      }
      if( sscanf(message,"%s%[ ]<%[^>]%[ >]",B2,B2,B1,B2) == 4 ) {
         if( ltrace ) {
            fprintf(stderr,"%s: wt_from_name(%s)=%d\n",R,B1,wt_from_name(B1));
         }
         if( Last_data_dir && is_active_wt(wt_from_name(B1)) ) {
            xvs_cd(wt_from_name(B1),Last_data_dir);
         }
      }
      xvs_send_status(stream,1,"");
   } else if( !strncmp(message,"cd",1) ) {
      if( ltrace ) {
         fprintf(stderr,"%s: Received 'cd'.  Full message='%s'\n",R,message);
         fprintf(stderr,"%s: Last_data_wt=%d\n",R,Last_data_wt);
      }
      if( sscanf(message,"%s%[ <]%[^>]%[ >]%s",B2,B2,B1,B2,B) == 5 ) {
         if( is_active_wt(wt_from_name(B1)) ) {
            if( ltrace ) {
               fprintf(stderr,"Executing xvs_cd(%d,%s)\n",wt_from_name(B1),B);
            }
            xvs_cd(wt_from_name(B1),B);
         }
      } else {
         fprintf(stderr,"%s: sscanf() failed.\n",R);
      }
      xvs_send_status(stream,1,"");
   } else { fprintf(stderr,"%s: Unimplemented message '%s'.\n",
         R,message ? message : "NULL");
   }
   return 1;
}

struct pollfd *make_poll_array(int *pn) {
   struct pollfd *the = (struct pollfd *) NULL;
   int            i,   j;
   int            ltrace = Ltrace;
   for( i = 0, *pn = 0; i < MAX_CONNECTION; i++ ) {
      if( C[i] ) (*pn)++;
   }
   if( ltrace ) {
      fprintf(stderr,"make_poll_array: %d connections open\n",*pn);
   }
   if( *pn ) {
      the = (struct pollfd *) malloc((*pn) * sizeof(struct pollfd));
   }
   for( i = 0, j = 0; i < MAX_CONNECTION; i++ ) {
      if( C[i] ) {
         the[j].fd = i;
         the[j].events = POLLIN;
         if( ltrace ) {
            fprintf(stderr,"make_poll_array: fd=%d\n",i);
         }
         j++;
      }
   }
   return the;
}


int Alarm = OFF;

void handle_alarm(int val) {
   Alarm = ON;
}

void xvs_poll_connections(void) {
   struct pollfd *poll_array = NULL; 
   int            nconn, nret, i, ss;
   int            timeout = 0;
   int            ltrace = Ltrace;
   int            j;

   struct itimerval timer;

   poll_array = make_poll_array(&nconn);
   if( nconn ) {

      timer.it_value.tv_sec  = 0;
      timer.it_value.tv_usec = 10000;
      setitimer(ITIMER_REAL,&timer,0);

      while( nret = poll(poll_array,nconn,timeout) ) {
         if( ltrace ) {
            fprintf(stderr,"xvs_poll_connections: poll returns %d\n",nret);
         }
         for( i = 0; i < nret; i++ ) {
            if( ltrace ) {
               fprintf(stderr,"xvs_poll_connections: poll_array[%d].revents=%x\n",
                  i,poll_array[i].revents);
            }
            ss = poll_array[i].fd;
            if( C_bulkinput[ss] ) {
               xvs_poll_connections_bulkinput(ss);
            } else {
               switch( poll_array[i].revents ) {
               case 0X1:
               case 0X10:
                  if( !cb_xvs_socket_1a(ss,NULL) ) {
                     fclose(C[ss]);
                     shutdown(ss,2);
                     close(ss);
                     C[ss] = NULL; 
                  }
                  break;
               default:
                  break;
               }
            }
         }
         if( Alarm ) {
            Alarm = OFF;
            goto Return;
         }
         if( poll_array ) free(poll_array);
         poll_array = make_poll_array(&nconn);
      }
   }
Return:
   if( ltrace ) fprintf(stderr,"xvs_poll_connections: >>>>> Exiting\n");
   if( poll_array ) free(poll_array);
}

void xvs_poll_connections_bulkinput(int ss) {
   static char    R[] = "xvs_poll_connections_bulkinput";
   int            ltrace = Ltrace;
   struct pollfd  poll_array[1];
   int            nconn = 1, nret;
   int            timeout = 0;

   if( ltrace ) fprintf(stderr,"%s: Enter ss=%d\n",R,ss);
   while( 1 ) {
      if( !cb_xvs_socket_1a(ss,NULL) ) {
         fclose(C[ss]);
         shutdown(ss,2);
         close(ss);
         C[ss] = NULL; 
         if( ltrace ) fprintf(stderr,"%s: Exit ss=%d\n",R,ss);
         return;
      }
   }
}

/*=========================================================================
Original callback for handling data streams into server.

Note: On IRIX (and possibly other systems), fread's and fwrite's on same 
stream need to be separated by file-pointer positioning call such as 
rewind.
=========================================================================*/
FL_IO_CALLBACK cb_xvs_socket(int ss,FD_main *vsform) {
   static char          R[] = "cb_xvs_socket";
   int                  ltrace = Ltrace;
   int                  ltrace_full = OFF;
   struct sockaddr_in   cli;
   int                  lencli = sizeof(struct sockaddr_in);
   int                  cs; 

   FILE                *stream;
   PSDFDS               the = (PSDFDS) NULL, mess = (PSDFDS) NULL;
   int                  nsds = 0, rcmain = 1, rc = 1;

   char                 opcode[1024], req_name[1024];
   int                  req_index = 1;
   WT                   wt;

   if( ltrace ) fprintf(stderr,"cb_xvs_socket: Data on vs_ss %d\n",ss);

#ifdef __BITS_SOCKET_H
   cs = accept(ss,(struct sockaddr *) &cli, (socklen_t *) &lencli); 
#else
   cs = accept(ss,(struct sockaddr *) &cli,               &lencli); 
#endif
   if( ltrace ) fprintf(stderr,"cb_xvs_socket: Connection on %d\n",cs);

   if( !(stream = fdopen(cs,"r+")) ) {
      fprintf(stderr,"cb_xvs_socket: fdopen(%d) failed\n",cs);
      close(cs);
      return NULL;
   }

/*-------------------------------------------------------------------------
Read as much data as possible from stream.  Current implementation adds 
each data set separately.
-------------------------------------------------------------------------*/
   do {
      if( ltrace ) fprintf(stderr,"Calling new_read_SDFDS\n");
      the = new_read_SDFDS(stream,&rcmain); 
      if( ltrace ) fprintf(stderr,"new_read_SDFDS returned\n");
      if( rcmain ) {
         nsds++;
         if( ltrace ) fprintf(stderr,"cb_xvs_socket: Read SDFDS\n");
/*-------------------------------------------------------------------------
         Messages/op-codes signaled via rank==-1.
-------------------------------------------------------------------------*/
         if( the->rank == -1 ) {
            if( ltrace ) fprintf(stderr,"cb_xvs_socket: the->pname='%s'\n",
               the->pname);
            cb_xvs_socket_handle_message(ss,stream,the->pname);
         } else {
            if( ltrace_full) dump_SDFDS(stderr,the,"cb_xvs_socket: data",1);

/*-------------------------------------------------------------------------
            Process data set but ignore data with non-finite coords.;
            replace non-finite data values with default.
-------------------------------------------------------------------------*/

            if( isdvfinite(the->coords,the->csize) ) {
               if( !isdvfinite(the->data,the->dsize) ) {
                  fprintf(stderr,"%s: Warning! One or more non-finite data values replaced with %g\n",
                     R,Nonfinite_val);
                  dvmkfinite(the->data,the->dsize,Nonfinite_val);
               }
               xvs_insert_SDFDS(the);
            } else {
               fprintf(stderr,"%s: Ignoring data set with non-finite coordinates\n",R);
               dvdump(the->coords,the->csize,"the->coords");
            }
         }
      } else {
         if( nsds == 0 && ltrace ) 
            fprintf(stderr,"cb_xvs_socket: Could not read SDFDS.\n");
      }
      if( the ) the = free_SDFDS(the);
   } while( rcmain > 0 );
   
   fclose(stream);
   close(cs);
   if( ltrace ) fprintf(stderr,"cb_xvs_socket: Exiting\n");
   return NULL;   
}
