#include "project.h"

PXYDVBIND Edit_save_p = NULL;

int       Edit_sort   = OFF;
int       Edit_intord = 2;
int       Edit_x_clamp = ON;

#define   IQ       wds[wt].winf.wiq
#define   DQ       (wds[wt].pf.pxyf->dqc)->dq
#define   P        wds[wt].pf.pxyf->xycurr
#define   X        P->data->x
#define   Y        P->data->y
#define   PTAG     P->tdata
#define   TAG      P->tdata->tag
#define   N        P->data->n
#define   NEVENTS  wds[wt].pf.pxyf->nevents

void get_mouse_pos_wt(WT wt,double *px,double *py) {
   *px = xmouse_w_wt(wt);
   *py = ymouse_w_wt(wt);
}

void set_mouse_pos_wt(WT wt,double x,double y) {
   setvaluator(MOUSEX,x_dq_to_iq(x,DQ,IQ),0,XMAXSCREEN);
   setvaluator(MOUSEY,y_dq_to_iq(y,DQ,IQ),0,YMAXSCREEN);
}

void qd_EDIT(WT wt,Device code) {

   int  ltrace = OFF;

   switch( code ) {
   case RIGHTMOUSE:
      edit_save_data(P);
      if( ctrl_key_pressed() ) {
         edit_linearly_insert_point(wt);
      } else {
         edit_add_point(wt);
      }
      break;
   case LEFTMOUSE:
   case DELKEY:
      edit_save_data(P);
      edit_delete_point(wt);
      break;
   case MIDDLEMOUSE:
      edit_save_data(P);
      edit_move_point(wt);
      break;
   case UPARROWKEY:
      edit_save_data(P);
      edit_double_N(wt);
      break;
   case DOWNARROWKEY:
      edit_save_data(P);
      edit_halve_N(wt);
      break;
   case INSERTKEY:
      edit_save_data(P);
      edit_linearly_insert_point(wt);
      break;
   case UKEY:
      if( ltrace ) {
         scan_all_pxy(wt);
      }
      P = edit_recover_data(wt,P);
      if( ltrace ) {
         scan_all_pxy(wt);
      }
      set_must_remap(wt); set_must_redraw(wt);
      break;
   default:
      fprintf(stderr,"qd_EDIT: Unimplemented code: %d.\n",code);
      break;
   }
}

void edit_add_point(WT wt) {
   double   x,     y;
   int      i;
   int      prev_N = N;
   
   get_mouse_pos_wt(wt,&x,&y);
   if( Edit_sort ) {
      N = dv2insert(&X,&Y,N,x,y,SORT_UP);
   } else {
      i = ixdv2nearest(X,Y,N,x,y,get_aspect_wt(wt));
      if(        i == 0     ) {
         if( x < X[0] )   {
            N = dv2insert_nosort(&X,&Y,N,x,y,0);
         } else {
            N = dv2insert_nosort(&X,&Y,N,x,y,1);
         }
      } else if( i == (N-1) ) {
         if( x > X[N-1] ) {
            N = dv2insert_nosort(&X,&Y,N,x,y,N);
         } else {
            N = dv2insert_nosort(&X,&Y,N,x,y,N-1);
         }
      } else {
         if( fabs(X[i-1] - x) <  fabs(X[i+1] - x) ) {
            N = dv2insert_nosort(&X,&Y,N,x,y,i);
         } else {
            N = dv2insert_nosort(&X,&Y,N,x,y,i+1);
         }
      }
   }
   if( N != prev_N ) {
      ++NEVENTS;
      set_must_remap(wt); set_must_redraw(wt);
   }
}

void edit_linearly_insert_point(WT wt) {
   double   x,     y;
   int      i;
   double   aspect;
   int      prev_N = N;
   
   get_mouse_pos_wt(wt,&x,&y);
   i = ixdv2nearest(X,Y,N,x,y,aspect = get_aspect_wt(wt));
   if( ( i == 0  &&  x < X[0] ) || ( i == (N-1) && x > X[N-1] ) ) {
      scope_mess_and_bell("Can't linearly insert beyond endpoints.",1);
   } else {
      if (       i == 0     ) {
         x = 0.5 * (X[i] + X[i+1]);
         y = 0.5 * (Y[i] + Y[i+1]);
         i = 1;
      } else if( i == (N-1) ) {
         x = 0.5 * (X[i] + X[i-1]);
         y = 0.5 * (Y[i] + Y[i-1]);
      } else {
         i = ixdv2perp(X,Y,N,x,y,aspect);
         x = 0.5 * (X[i] + X[i+1]);
         y = 0.5 * (Y[i] + Y[i+1]);
         i = i + 1;
      }
      N = dv2insert_nosort(&X,&Y,N,x,y,i);
   }
   if( N != prev_N ) {
      ++NEVENTS;
      set_mouse_pos_wt(wt,x,y);
      set_must_remap(wt); set_must_redraw(wt);
   }
}

void edit_delete_point(WT wt) {
   double   x,     y,     delx,     dely;
   int      i;

   if( N <= 1 ) {
      sprintf(O_buf,"Can't delete any more points.");
      scope_mess_and_bell(O_buf,1);
   } else {
      get_mouse_pos_wt(wt,&x,&y);
      i = ixdv2nearest(X,Y,N,x,y,get_aspect_wt(wt));    
      X = dvdel_el(X,N,i);
      Y = dvdel_el(Y,N,i);
      if( PTAG ) {
         TAG = ivdel_el(TAG,N,i);
      }
      --N; --NEVENTS;
      set_must_remap(wt); set_must_redraw(wt);
   }
}

void edit_move_point(WT wt) {
   int    first = TRUE,   i;
   double x,    y;

   while( middle_mouse_down() ) {
      if( first ) {
         get_mouse_pos_wt(wt,&x,&y);
         i = ixdv2nearest(X,Y,N,x,y,get_aspect_wt(wt));
         x = X[i]; y = Y[i];
         set_mouse_pos_wt(wt,x,y);
         first = FALSE;
      } else {
         switch( get_movement() ) {
         case X_CLAMPED:
            get_mouse_pos_wt(wt,&x,Y+i);
            break;
         case Y_CLAMPED:
            get_mouse_pos_wt(wt,X+i,&y);
            break;
         case NO_CLAMP:
            get_mouse_pos_wt(wt,X+i,Y+i);
            break;
         default:
            break;
         }
         if( Edit_sort ) {
            dv2sort(X,Y,N,SORT_UP);
         }
         set_mouse_pos_wt(wt,X[i],Y[i]);
      }
      display_window_1(wt);
   }
   set_must_remap(wt); set_must_redraw(wt);
}

void edit_which_point(WT wt) {
   int    i;
   double x,    y,    nearest_x,    nearest_y;

   if( ! is_active_wt(wt) ) return;

   get_mouse_pos_wt(wt,&x,&y);
   i = ixdv2nearest(X,Y,N,x,y,get_aspect_wt(wt));
   nearest_x = X[i]; nearest_y = Y[i];
   sprintf(O_buf,"(%.3G,%.3G) -- %d (%.3G,%.3G)",
           x,y,i+1,nearest_x,nearest_y);
   fprintf(stderr,"(%.3G,%.3G) -- %d (%.3G,%.3G)\n",
           x,y,i+1,nearest_x,nearest_y);
   scope_mess(O_buf,1);
}

Boolean middle_mouse_down() {
   return getbutton(MIDDLEMOUSE);
}

void edit_display_mouse_and_nearest_data(WT wt) {
   double      x,               y,    
               prev_x,          prev_y,
               nearest_x,       nearest_y,
               first = 1;
   int         i;

   int         echo = ON;

   if( ! is_active_wt(wt) ) return;
   while( getbutton(LEFTMOUSE) ) {
      Mouse_position_display = ON;
      if( !Mp_fmt ) {
         Mp_fmt = strdup(Mp_default_fmt);
      }
      get_mouse_pos_wt(wt,&x,&y);
      if( first || (x != prev_x) || (y != prev_y) ) {
         i = ixdv2nearest(X,Y,N,x,y,get_aspect_wt(wt));
         nearest_x = X[i]; nearest_y = Y[i];
         sprintf(O_buf,"(%s, %s)",Mp_fmt,Mp_fmt);
         sprintf(Mp_strs[0],O_buf,x,y);
         if( echo ) {
            fprintf(stderr,Mp_strs[0]); fprintf(stderr,"\n");
         }
         sprintf(O_buf,"(%s, %s) [%%d]",Mp_fmt,Mp_fmt);
         sprintf(Mp_strs[1],O_buf,nearest_x,nearest_y,i + 1);
         if( echo ) {
            fprintf(stderr,Mp_strs[1]);
            fprintf(stderr,"\n\n");
         }
         display_window_1(wt);
         first = 0;
      }
      prev_x = x; prev_y = y;
   }
   Mouse_position_display = OFF;
   set_must_redraw(wt);
}

int get_movement() {

   if( alt_ctrl_keys_pressed() || !Edit_x_clamp ) {
      return NO_CLAMP;
   } else if( ctrl_key_pressed() ) {
      return Y_CLAMPED;
   } else {
      return X_CLAMPED;
   }
}

void edit_double_N(WT wt) {
   DVEC   NEWX,    NEWY;
   int    NEWN,    f77_2 = 2,  f77_int_ord;
   double f77_def_val = 0.0;
   int    save_pos;

   NEWN = 2 * (N - 1) + 1;
   if( !(NEWX = make_DVEC(NEWN)) || !(NEWY = make_DVEC(NEWN)) ) {
      printf("edit_double_N: make_DVEC(%d) failed.\n",NEWN);
   } else {
      dvprln_(X,NEWX,&f77_2,&N);
      dvav2_(NEWX,&NEWN);
      f77_int_ord = N > 4 ? Edit_intord : 2;
      if( f77_int_ord == 2 ) {
         dvprln_(Y,NEWY,&f77_2,&N);
         dvav2_(NEWY,&NEWN);
      } else {
         dvinqn_(Y,X,NEWY,NEWX,&N,&NEWN,&f77_def_val,&f77_def_val,&f77_int_ord);
      }
      free_DVEC(X);  free_DVEC(Y);
      X = NEWX; Y = NEWY; NEVENTS += (NEWN - N); N = NEWN; 
      rescale_wt_restore_display_index(wt);
      set_must_remap(wt); set_must_redraw(wt);
   }
}

void edit_halve_N(WT wt) {
   DVEC   NEWX,    NEWY;
   int    NEWN,    f77_2 = 2;
   int    save_pos;

   if( N % 2 ) {
      NEWN = (N + 1) / 2;
      if( !(NEWX = make_DVEC(NEWN)) || !(NEWY = make_DVEC(NEWN)) ) {
         printf("edit_halve_N: make_DVEC(%d) failed.\n",NEWN);
      } else {
         dvinj_(X,NEWX,&f77_2,&N);
         dvinj_(Y,NEWY,&f77_2,&N);
         free_DVEC(X);  free_DVEC(Y);
         X = NEWX; Y = NEWY; NEVENTS += (NEWN - N); N = NEWN; 
         save_pos = get_display_index(wt); 
         rescale_wt(wt);
         set_xycurr_nth(wt,save_pos);
         set_must_remap(wt); set_must_redraw(wt);
      }
   } else {
      scope_mess_and_bell("Can't halve data.",1);
   }
}

double get_aspect_wt(WT wt) {
   double    dy;
   
   dy = DQ.y2 - DQ.y1;
   return (dy == 0.0) ? 1.0 : ((DQ.x2 - DQ.x1) / dy);
}

int toggle_Edit_sort(void) {
   Edit_sort = !Edit_sort;
   scope_mess_and_bell(Edit_sort ? "Edit sort on." : "Edit sort off.",1);
   return(Edit_sort);
}

int toggle_Edit_intord(void) {
   if(        Edit_intord == 2 ) {
      Edit_intord = 4;
   } else if( Edit_intord == 4 ) {
      Edit_intord = 2;
   } else {
      fprintf(stderr,"toggle_Edit_intord: Unexpected Edit_intord: %d.\n",
              Edit_intord);
   }
   sprintf(O_buf,"Edit interpolation order: %d.",Edit_intord);
   scope_mess_and_bell(O_buf,1);
   return(Edit_intord);
}

int toggle_Edit_x_clamp(void) {
   Edit_x_clamp = !Edit_x_clamp;
   scope_mess_and_bell(Edit_x_clamp ? "Edit x-clamp on." : 
                                      "Edit x-clamp off.",1);
   return(Edit_x_clamp);
}

void edit_save_data(PXYDVBIND p) {
   free_XYDVBIND(Edit_save_p);
   Edit_save_p = copy_PXYDVBIND(p);
}

void edit_reset_save_data(void) {
   free_XYDVBIND(Edit_save_p);
   Edit_save_p = NULL;
}

PXYDVBIND edit_recover_data(WT wt,PXYDVBIND p) {
   PXYDVBIND  lp = p;
   
   if( p && Edit_save_p ) {
      lp = Edit_save_p;
      replace_PXYDVBIND_wt(wt,p,lp);
      Edit_save_p = p;
   } else {
      scope_mess_and_bell("Can't recover data.",1);
   }
   return(lp);
}

#undef    NEVENTS
#undef    N    
#undef    Y     
#undef    P      
#undef    DQ
#undef    IQ
