#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <forms.h>

#include "mfl.h"
#include "quad_util.h"
#include "plot.h"

/*-------------------------------------------------------------------------
G L O B A L   V A R I A B L E S 
-------------------------------------------------------------------------*/
LabelledXEvent AllXEvents[] = {
   {KeyPress,    "KeyPress"},
   {KeyRelease,  "KeyRelease"},
   {ButtonPress, "ButtonPress"},
   {ButtonRelease, "ButtonRelease"},
   {MotionNotify, "MotionNotify"},
   {EnterNotify, "EnterNotify"},
   {LeaveNotify, "LeaveNotify"},
   {FocusIn, "FocusIn"},
   {FocusOut, "FocusOut"},
   {Expose, "Expose"},
   {GraphicsExpose, "GraphicsExpose"},
   {NoExpose, "NoExpose"},
   {VisibilityNotify, "VisibilityNotify"},
   {CreateNotify, "CreateNotify"},
   {DestroyNotify, "DestroyNotify"},
   {UnmapNotify, "UnmapNotify"},
   {MapNotify, "MapNotify"},
   {ResizeRequest, "ResizeRequest"}
};

/* #define NAllXevents (sizeof(AllXEvents) / sizeof(LabelledXEvent)) */

color_translate CTlookup[] = {
   {BLACK,  {0.0,0.0,0.0}},
   {RED,    {1.0,0.0,0.0}},
   {GREEN,  {0.0,1.0,0.0}},
   {YELLOW, {1.0,1.0,0.0}},
   {BLUE,   {0.0,0.0,1.0}},
   {MAGENTA,{1.0,0.0,1.0}},
   {CYAN,   {0.0,1.0,1.0}},
   {WHITE,  {1.0,1.0,1.0}}
};

#define NCTlookup (sizeof(CTlookup) / sizeof(color_translate))

/*-------------------------------------------------------------------------
Translates numeric Xevent descriptor to string
-------------------------------------------------------------------------*/
char *strXEvent(int event) {
   int i;
   for( i = 0; i < NAllXevents; i++ ) {
      if( event == AllXEvents[i].event ) return(AllXEvents[i].label);
   }
   return("UnknownEvent");
}

/*-------------------------------------------------------------------------
Translates SGI colormap-indexed color to RGB.
-------------------------------------------------------------------------*/
rgb_color CT(int cindex) {
   rgb_color  the = {1.0, 1.0, 1.0};
   int        i;
   for( i = 0; i < NCTlookup; i++ ) {
      if( CTlookup[i].cindex == cindex ) {
         the = CTlookup[i].rgb;
         goto Return;
      }
   }
Return:
   return the;
}

/*-------------------------------------------------------------------------
Return internal/external display border widths 
-------------------------------------------------------------------------*/
int get_dpy_ibdr(void)  { return 5; }
int get_dpy_obdr(void)  { return 2; }

/*-------------------------------------------------------------------------
Returns 1 if any mouse button is depressed in window win
-------------------------------------------------------------------------*/
int mfl_mouse_any(Window win) {
   unsigned int keymask;
   FL_Coord     x, y;
   (void) fl_get_win_mouse(win,&x,&y,&keymask);
   return  (keymask & Button1Mask) ||  (keymask & Button2Mask) ||
           (keymask & Button3Mask);
}


/*-------------------------------------------------------------------------
Returns geometry of display computed from geoms of main form and control

Display panels are rendered within area of the width of the main form,
and extending below the maximum depth of the control widgets.
-------------------------------------------------------------------------*/
igeom mfl_get_dpy_geom(void) {
#define FM   Xvsf->main
#define FC   Xvsf->frame_ctl
#define GEOM igeom_make(get_dpy_obdr(),FC->y+FC->h+get_dpy_obdr(),\
             FM->w-2*get_dpy_obdr(),FM->h-(FC->y+FC->h+2*get_dpy_obdr()))
   igeom   the;
   the = GEOM;
   if( Ltrace ) {
      fprintf(stderr,"mfl_get_dpy_geom: Main geometry (%d,%d) %d x %d\n",
         FM->x,FM->y,FM->w,FM->h);
      fprintf(stderr,"mfl_get_dpy_geom: Frame_ctl geometry (%d,%d) %d x %d\n",
         FC->x,FC->y,FC->w,FC->h);
      igeom_dump("mfl_get_dpy_geom: Return_geom",the);
   }
   return the;
#undef  FM
#undef  FC
#undef  GEOM
}

/*-------------------------------------------------------------------------
Adjusts main form geometry so that display geometry is square. 

mfl_adjust_main_geom_up      sizes up
mfl_adjust_main_geom_down    sizes down
-------------------------------------------------------------------------*/
void mfl_adjust_main_geom_up(void) {
   igeom   dg = mfl_get_dpy_geom();
   int     del;
   del = dg.w - dg.h;
   if(        del > 0 ) {
      fl_set_form_geometry(Xvsf->main,Xvsf->main->x,Xvsf->main->y,
                                      Xvsf->main->w,Xvsf->main->h+del);
   } else if( del < 0 ) {
      fl_set_form_geometry(Xvsf->main,Xvsf->main->x,Xvsf->main->y,
                                      Xvsf->main->w-del,Xvsf->main->h);
   }
   fl_redraw_form(Xvsf->main);
}

void mfl_adjust_main_geom_down(void) {
   igeom   dg = mfl_get_dpy_geom();
   int     del;
   del = dg.w - dg.h;
   if(        del < 0 ) {
      fl_set_form_geometry(Xvsf->main,Xvsf->main->x,Xvsf->main->y,
                                      Xvsf->main->w,Xvsf->main->h+del);
   } else if( del > 0 ) {
      fl_set_form_geometry(Xvsf->main,Xvsf->main->x,Xvsf->main->y,
                                      Xvsf->main->w-del,Xvsf->main->h);
   }
   fl_redraw_form(Xvsf->main);
}

/*-------------------------------------------------------------------------
Resizes GL panels:
   obdr (I): outside border width (pixels)
   ibdr (I): inside border width (pixels)
-------------------------------------------------------------------------*/
void mfl_resize_gl_panels(int obdr, int ibdr) {
   int     nx,  ny,  x,  y,  w,  h,  i,  j,  k;
	static  int squares[MAX_PANEL_LEV];
	static  int first = 1;
	
	if( first ) {
		for( i = 0; i < MAX_PANEL_LEV; i++ ) squares[i] = (i + 1) * (i + 1);
	}
   for( i = 1; i <= MAX_PANEL; i++ ) {
		for( k = 0; k < MAX_PANEL_LEV; k++ ) {
			if( i <= squares[k] ) {
				nx = k + 1; ny = k + 1;
				goto Continue;
			} 
		}
Continue:
		for( j = 1; j <= i; j++ ) {
			compute_gl_panel_geometry(mfl_get_dpy_geom(),
				nx,ny,j,obdr,ibdr,&x,&y,&w,&h);
			fl_set_object_geometry(Xvsf->w[i-1][j-1].gl,x,y,w,h);
		}
	}
}

/*-------------------------------------------------------------------------
Computes geometry information for single Gl window in panel:
g  (I): global geometry of panel
nx (I): number of columns in panel
ny (I): number of rows in panel
ip (I): which panel (1 based)
obdr (I): outside border width (pixels)
ibdr (I): inside border width (pixels)
px, py, pw, pz (O): computed geometry 
-------------------------------------------------------------------------*/
void compute_gl_panel_geometry(igeom g, int nx, int ny, int ip, int obdr, int ibdr,
  int *px, int *py, int *pw, int *ph) {

int        fxuc = g.x, fyuc = g.y, fw = g.w, fh = g.h;
int        ix,  iy;

iy = (ip - 1) / nx;
ix = (ip - 1 ) % nx;

*pw = (fw - (2 * obdr + (nx - 1) * ibdr)) / nx;
*ph = (fh - (2 * obdr + (ny - 1) * ibdr)) / ny; 
*px = fxuc + obdr + ix * (ibdr + *pw);
*py = fyuc + obdr + iy * (ibdr + *ph);

}

/*-------------------------------------------------------------------------
Returns xvs_panel object and sets xvs_panel_array as a side-effect
-------------------------------------------------------------------------*/
FL_OBJECT *mfl_add_gl_panel(igeom g, xvs_panel_array *pxvsp, int nx, int ny, 
     int np, int obdr, int ibdr) {
   int        ltrace = Ltrace;

   FL_OBJECT *the, *gl;
   char       buffer[BUFLEN];
   int        config[] = {GLX_RGBA, GLX_DEPTH_SIZE, 16, GLX_DOUBLEBUFFER, None};
   int        fxuc = g.x, fyuc = g.y, fw = g.w, fh = g.h;
   int        pw, ph, xuc, yuc, i, j, k, ip, n;

   if( nx < 1 || ny < 1 || nx * ny < np ) {
      fprintf(stderr,"mfl_add_gl_panel: Fatal error --- bad parameters!\n");
      fprintf(stderr,"mfl_add_gl_panel:    nx = %d  ny = %d  np = %d\n",
         nx,ny,np);
      exit(1);
   }
   if( ltrace ) {
      fprintf(stderr,"mfl_add_gl_panel: Frame geometry: (%d,%d) %d x %d\n",fxuc,fyuc,fw,fh);
      fprintf(stderr,"mfl_add_gl_panel: nx=%d  ny=%d  np=%d\n",nx,ny,np);
   }
   the = fl_bgn_group();
   
   for( j = 0, n = 1; j < ny; j++ ) {
      for( i = 0; i < nx; i++, n++ ) {
         compute_gl_panel_geometry(g,nx,ny,n,obdr,ibdr,&xuc,&yuc,&pw,&ph);
         gl = fl_add_glcanvas(FL_NORMAL_CANVAS,xuc,yuc,pw,ph,""); 
         (*pxvsp)[np-1][n-1].gl = gl;
         if( ltrace ) {
            fprintf(stderr,"mfl_add_gl_panel: x=%d y=%d w=%d h=%d\n",
                  gl->x,gl->y,gl->w,gl->h);
         }
         (*pxvsp)[np-1][n-1].wt = -1;
         (*pxvsp)[np-1][n-1].font_init = 0;
         sprintf(buffer,"Panel(%d,%d,%d)",n,nx,ny);
         gl->u_cdata = strdup(buffer);
         fl_set_glcanvas_attributes(gl,config);
         for( k = 0; k < NAllXevents; k++ ) {
            fl_add_canvas_handler(gl,AllXEvents[k].event,cb_glcanvas_main,gl);
         }
         if( n == np ) goto Done;
      }
   }
Done:
   fl_end_group();
   fl_hide_object(the);

   return the;
}

void xvs_init_gl_main(void) {
   int k;
   FL_OBJECT *gl = Xvsf->gl_main;
   for( k = 0; k < NAllXevents; k++ ) {
      fl_add_canvas_handler(gl,AllXEvents[k].event,cb_glcanvas_main,gl);
   } 
}

/*-------------------------------------------------------------------------
Returns gl_panel object and sets gl_panel_array as a side-effect
-------------------------------------------------------------------------*/
FL_OBJECT *mfl_add_gl_panel_old(igeom g, gl_panel_array *pgl, int nx, int ny, 
     int np, int obdr, int ibdr) {

   FL_OBJECT *the, *gl;
   char       buffer[BUFLEN];
   int        config[] = {GLX_RGBA, GLX_DEPTH_SIZE, 16, GLX_DOUBLEBUFFER, None};
   int        fxuc = g.x, fyuc = g.y, fw = g.w, fh = g.h;
   int        pw, ph, xuc, yuc, i, j, k, ip, n;

   if( nx < 1 || ny < 1 || nx * ny < np ) {
      fprintf(stderr,"mfl_add_gl_panel: Fatal error --- bad parameters!\n");
      fprintf(stderr,"mfl_add_gl_panel:    nx = %d  ny = %d  np = %d\n",
         nx,ny,np);
      exit(1);
   }
   fprintf(stderr,"mfl_add_gl_panel: Frame geometry: (%d,%d) %d x %d\n",fxuc,fyuc,fw,fh);
   the = fl_bgn_group();
   
   for( j = 0, n = 1; j < ny; j++ ) {
      for( i = 0; i < nx; i++, n++ ) {
         compute_gl_panel_geometry(g,nx,ny,n,obdr,ibdr,&xuc,&yuc,&pw,&ph);
         gl = fl_add_glcanvas(FL_NORMAL_CANVAS,xuc,yuc,pw,ph,""); 
         (*pgl)[np-1][n-1] = gl;
         sprintf(buffer,"Panel(%d,%d,%d)",n,nx,ny);
         gl->u_cdata = strdup(buffer);
         fl_set_glcanvas_attributes(gl,config);
         for( k = 0; k < NAllXevents; k++ ) {
            fl_add_canvas_handler(gl,AllXEvents[k].event,cb_glcanvas_main,gl);
         }
         if( n == np ) goto Done;
      }
   }
Done:
   fl_end_group();
   fl_hide_object(the);

   return the;
}


/*-------------------------------------------------------------------------
Dump utility for xvs_panel_arrary
-------------------------------------------------------------------------*/
void dump_xvs_panel_array(xvs_panel_array w) {
   int ip, igl;
   for( ip = 0; ip < MAX_PANEL; ip++ ) {
      fprintf(stderr,"===========================================\n");
      fprintf(stderr,"Panel: %d\n",ip);
      fprintf(stderr,"===========================================\n");
      for( igl = 0; igl <= ip;  igl++ ) {
         fprintf(stderr,"   Canvas: %d  x=%d  y=%d  w=%d  h=%d  wt=%d\n",igl,
            w[ip][igl].gl->x,w[ip][igl].gl->y,w[ip][igl].gl->w,w[ip][igl].gl->h,
            w[ip][igl].wt);
      }
   }
}

/*-------------------------------------------------------------------------
Dump utility for gl_panel_arrary
-------------------------------------------------------------------------*/
void dump_gl_panel_array(gl_panel_array gl) {
   int ip, igl;
   for( ip = 0; ip < MAX_PANEL; ip++ ) {
      fprintf(stderr,"===========================================\n");
      fprintf(stderr,"Panel: %d\n",ip);
      fprintf(stderr,"===========================================\n");
      for( igl = 0; igl <= ip;  igl++ ) {
         fprintf(stderr,"   Canvas: %d  x=%d  y=%d  w=%d  h=%d\n",igl,
            gl[ip][igl]->x,gl[ip][igl]->y,gl[ip][igl]->w,gl[ip][igl]->h);
      }
   }
}


/*=========================================================================
Close selected windows.
=========================================================================*/
int mfl_close_selected_1(IVECN wt_list) {
   PWTBIND    p;
   
   WT         wt;
   int        i,   ltrace = OFF;

   if( wt_list.n <= 0 ) return 0;
   
   wt_list = ivecn_sortupuniq(wt_list);
   if( ltrace ) {
      fprintf(stderr,"mfl_close_selected_1: lal_dump() on exit ...\n");
      lal_dump();
      ivdump(wt_list.v,wt_list.n,"mfl_close_selected_1: Windows to be closed");
   }
   for( i = 0; i < wt_list.n; i++ ) {
      if( is_active_wt(wt_list.v[i]) ) {
         close_status_window(wt_list.v[i]);
      }
   }
   if( scope_mode == SINGLE_MODE ) {
      if( ltrace ) {
         fprintf(stderr,"mfl_close_selected_1: Going to PANEL_MODE.\n");
      }
      lal_pop();
      scope_mode = PANEL_MODE;
   } 
   for( i = 0; i < wt_list.n; i++ ) {
      if( is_active_wt(wt_list.v[i]) ) {
         if( ltrace ) {
            fprintf(stderr,"mfl_close_selected_1: Closing %d.\n",wt_list.v[i]);
         }
         free_XYF_storage(wt_list.v[i]);
         if( wds[wt_list.v[i]].winf.pwf.pxywinf ) {
            free_pxywinf(wds[wt_list.v[i]].winf.pwf.pxywinf);
         }
         winclose(wds[wt_list.v[i]].sgi_winno);
         wds[wt_list.v[i]].status = WINDOW_NOT_ACTIVE;
         free_wt(wt_list.v[i]);
      }
   }
   lal_set();
   last_S_wt = scope_wt;
   (void) vs_repanel();
   if( mouse_is_near_wt(scope_wt,0.2,3.0) ) {
      centre_mouse_in_wt(scope_wt);
   }
   if( ltrace ) {
      fprintf(stderr,"mfl_close_selected_1: lal_dump() on exit ...\n");
      lal_dump();
   }
   free_IVECN(wt_list);
   return 0;
}
