#include "project.h"

double    X_mouse,     Y_mouse;

int iclamp(int i, int imin, int imax, int idef) {
   return (imin <= i && i <= imax) ? i : idef;
}

void err1(s1,s2)
char      *s1, *s2;
{
   printf("<<< %s:: %s >>>\n",s1,s2);
   exit(0);
}

void err2(s1,s2)
char      *s1, *s2;
{
   printf("<<< %s:: %s >>>\n",s1,s2);
}

void dump_dq_stack(PDQBIND top) {
   int     i = 0;
   PDQBIND p = top;

   if( top ) {
      printf("dump_dq_stack: %d entries\n",dq_stack_depth(top));
      while( p ) {
         i++;
         printf("dump_dq_stack: dq %d\n",i);
         dump_dquad(p->dq);
         p = p->prev;
      }
   } else {
      fprintf(stderr,"dump_dq_stack: Null stack\n");
   }

}

PDQBIND copy_dq_stack(PDQBIND top) {
   if( top ) {
      return push_dq(copy_dq_stack(top->prev),top->dq);
   } else {
      return(NULL);
   }
}

PDQBIND make_pdqbind(dq,next,prev)
   DQUAD      dq;
   PDQBIND    next,    prev;
{
   PDQBIND    p;

   p = (PDQBIND) malloc(sizeof(DQBIND));
   p->dq = dq; p->next = next; p->prev = prev;
   return p;
}

PXYF make_pxyf(limits,scale_option,max_lxy,max_nevents)
   DQUAD   limits;
   int     scale_option, max_lxy, max_nevents;
{
   PXYF    pxyf;

   pxyf   = (PXYF) malloc(sizeof(XYF));
   pxyf->limits = limits;
   pxyf->limits_computed = NO_LIMITS_COMPUTED;
   pxyf->scale_option = scale_option;
   pxyf->thin_option = AUTO_THIN_OFF;
   pxyf->ring_state = RING_OFF;
   pxyf->new_data = OFF;

   pxyf->dqsz = NULL;
   pxyf->dqb = make_pdqbind(limits,NULL,NULL);
   pxyf->dqc = push_dq(NULL,limits);
   

   pxyf->xyhead = pxyf->xytail = (PXYDVBIND) NULL;

   pxyf->xymr = (PXYDVBIND) NULL;

   pxyf->xyfirst = pxyf->xycurr = pxyf->xylast = (PXYDVBIND) NULL;

   pxyf->display_index = 0;
   pxyf->lxy = 0;
   pxyf->max_lxy = max_lxy;
   pxyf->nevents = 0;
   pxyf->max_nevents = max_nevents;

   return(pxyf);
}

PXYWINF make_pxywinf()
{
   PXYWINF pxywinf;

   pxywinf = (PXYWINF) malloc(sizeof(XYWINF));
   pxywinf->back_color = DEFAULT_BACK_COLOR;
   pxywinf->border_color = Q_BORDER_COLOR;
   pxywinf->marker_type = DEFAULT_MARKER_TYPE_LO_N;
   pxywinf->marker_color = Default_marker_color;
   pxywinf->marker_size = DEFAULT_MARKER_SIZE;
   pxywinf->point_size = DEFAULT_POINT_SIZE;
   pxywinf->marker_npoly = DEFAULT_MARKER_NPOLY;
   pxywinf->marker_multi_mode = ON;
   pxywinf->marker_line_auto_break = DEFAULT_LINE_AUTO_BREAK;
   pxywinf->line_color = Default_line_color;
   pxywinf->line_width = Default_line_width;
   pxywinf->must_remap = 1;
   pxywinf->new_domain = 0;
   pxywinf->must_redraw = OFF;
   pxywinf->domain_mode = FLOATING_DOMAIN;
   pxywinf->display_mode = FREEZE_MODE;
   pxywinf->prev_display_mode = FREEZE_MODE;
   pxywinf->auto_sub_mode = AUTO_FORWARD;
   pxywinf->auto_break_mode = AUTO_BREAK_OFF;
   pxywinf->full_scale_mode = OFF;
   pxywinf->var_dquad = VAR_DQUAD_OFF;
   pxywinf->var_dquad_zoom = 1.0;
   pxywinf->heads_up = heads_up_status();
   pxywinf->heads_up_t_format = NULL;
   pxywinf->heads_up_lim = ON;

   pxywinf->range_gauge = make_default_range_gauge();

   pxywinf->accept_frequency = 1;
   pxywinf->status_window_flag = OFF;
   pxywinf->status_sgi_winno = -1;
   pxywinf->plxy = NULL; 
   pxywinf->pdisplay_index = NULL; 

   return(pxywinf);
}

void free_pxywinf(PXYWINF pxywinf)
{
   int  ltrace = Ltrace;

   if( ltrace ) {
      if( pxywinf->heads_up_t_format ) {
         fprintf(stderr,"free_pxywinf: Will free heads_up_t_format.\n");
      } else {
         fprintf(stderr,"free_pxywinf: heads_up_t_format is NULL.\n");
      }
   }
   if( pxywinf->heads_up_t_format ) {
      free(pxywinf->heads_up_t_format);
      pxywinf->heads_up_t_format = NULL;
   }
}

void window_ortho2(WT wt) {
#define P          wds[wt].pf.pxyf->dqc
#define Q          P->dq
#define CO (float) Q
   int             ltrace = 0;

   double          rpos[4];

   assert(P != NULL);
   fl_activate_glcanvas(gl_from_wt(wt)); 
   Q = replace_degenerate_DQUAD(Q);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   if( ltrace ) {
      fprintf(stderr,"window_ortho2: Calling gluOrtho2D(%g,%g,%g,%g)\n",
         CO.x1,CO.x2,CO.y1,CO.y2);
   }
   gluOrtho2D(CO.x1,CO.x2,CO.y1,CO.y2);
#undef  CO
#undef  Q 
#undef  P 
}

void  window_ortho2_old(wt)
   WT        wt;
{
#define P          wds[wt].pf.pxyf->dqc
#define Q          P->dq
#define CO (float) Q
   assert(P != NULL);
   winset(wds[wt].sgi_winno);
   Q = replace_degenerate_DQUAD(Q);
   ortho2(CO.x1,CO.x2,CO.y1,CO.y2);
#undef  CO
#undef  Q 
#undef  P 
}

void window_stats(wt)   
   WT       wt;
{
extern
   WINDOW   wds[];
   WINDOW   w;

   int      full_dump = ON;

   w = wds[wt];
   if( wt < 0 ) {
      fprintf(stderr,"window_stats: Warning ... window %d not active.\n",wt);
      return;
   }
   printf("\nStats on window tagged as %d.\n",wt);
   printf("Pnl: %d  Pnl_prev: %d\n",Pnl,Pnl_prev);
   printf("sgi winno: %d   name: '%s'   status: %d\ndouble buff. %d\n",
          w.sgi_winno, w.name, w.status,w.winf.d_buff);
   printf("working directory: %s.\n",w.wd);
#define P w.pf.pxyf 
   printf("limits_computed: %d   scale_option: %d   thin_option: %d.\n",
        P->limits_computed,   P->scale_option,   P->thin_option);
   printf("ring_state: %d  new_data: %d.\n", 
        P->ring_state,  P->new_data);
   printf("nxy: %d    max_lxy: %d   nevents: %d    max_nevents: %d.\n",
        n_XYDV(wt), P->max_lxy,   P->nevents,    P->max_nevents);
   if( full_dump ) {
      dump_PXYDVBIND("xyhead",P->xyhead);
      dump_PXYDVBIND("xytail",P->xytail);
      dump_PXYDVBIND("xymr",P->xymr);
      dump_PXYDVBIND("xyfirst",P->xyfirst);
      dump_PXYDVBIND("xycurr",P->xycurr);
      dump_PXYDVBIND("xylast",P->xylast);
   }
#define CO (float) w.pf.pxyf->limits
   printf("Virtual limits: %f %f %f %f.\n",CO.x1, CO.x2, CO.y1, CO.y2);
#undef  CO
#define CO (float) w.pf.pxyf->dqc->dq
   printf("Active  limits: %f %f %f %f.\n",CO.x1, CO.x2, CO.y1, CO.y2);
#undef  CO
#define CO (float) w.pf.pxyf->dqsz->dq
   if( w.pf.pxyf->dqsz ) {
      printf("Single zoom limits: %f %f %f %f.\n",CO.x1, CO.x2, CO.y1, CO.y2);
   } else {
      printf("Single zoom limits: <null>.\n");
   }
#undef  CO
#define ICO w.winf.wiq
   printf("Real limits: %d %d %d %d.\n",ICO.x1, ICO.x2, ICO.y1, ICO.y2);
   printf("Width: %d  Height: %d\n",ICO.x2-ICO.x1+1,ICO.y2-ICO.y1+1);
#undef  ICO
   printf("Depth of local DQ stack: %d.\n",dq_stack_depth(P->dqc));
   printf("Current DQ %s equal to base DQ.\n",
      equal_dquads(P->dqc->dq,P->dqb->dq) ? "is" : "is not");
   printf("Stack dump\n");
   dump_dq_stack(P->dqc);

   if( wt == 0 ) 
      return; /* Because rest of the junk isn't defined for TBAR window. */
#define I w.winf.pwf.pxywinf
   printf("back_color: %d   border_color: %d\n",
        I->back_color, I->border_color); 
   printf("marker_type: %d   marker_color: %d marker_size: %g\n",
        I->marker_type,   I->marker_color,   I->marker_size);
   printf("point_size: %d   marker_npoly: %d   marker_multi_mode: %d\n",
        I->point_size,   I->marker_npoly,   I->marker_multi_mode);
   printf("marker_line_auto_break: %d\n",
        I->marker_line_auto_break);
   printf("must remap: %d  must_redraw: %d   domain_mode: %d\n",
        I->must_remap,  I->must_redraw,   I->domain_mode);
   printf("new_domain: %d   display_mode: %d   prev_display_mode: %d\n",
        I->new_domain,   I->display_mode,   I->prev_display_mode);
   printf("auto_sub_mode: %d\n",    I->auto_sub_mode);
   printf("auto_break_mode: %d   lxy: %d   display_index: %d\n",
        I->auto_break_mode, *(I->plxy),*(I->pdisplay_index));
   printf("full_scale_mode: %d\n",   I->full_scale_mode);
   printf("accept_frequency: %d   status_window_flag: %d\n",
        I->accept_frequency,   I->status_window_flag);
   printf("status_sgi_winno: %d   heads_up: %d\n",
        I->status_sgi_winno,I->heads_up);
   printf("heads_up_lim: %d   var_dquad: %d  var_dquad_zoom: %lf\n",
        I->heads_up_lim,I->var_dquad,I->var_dquad_zoom);
   if( I->heads_up_t_format ) {
      printf("heads_up_t_format: %s\n",I->heads_up_t_format);
   } else {
      printf("heads_up_t_format: (NULL)\n");
   }
   dump_range_gauge(I->range_gauge);

   printf("\n");
#undef I
#undef P
}

void dump_PXYDVBIND(char *s, PXYDVBIND p) {
   int  node = 0, prev = 0, next = 0;

   if( p ) {
      node = p->data->n;
      if( p->prev ) {
         prev = p->prev->data->n;
      }
      if( p->next ) {
         next = p->next->data->n;
      }
   }
   printf("   <%s>:  Node: %d   Prev: %d   Next: %d\n",
          s, node, prev, next );
}

void dump_wt_iquad(WT wt,char *s,int ltrace) {
   if( ltrace ) {
      fprintf(stderr,"dump_wt_iquad(%s):  wt=%d  geometry=(%d,%d) %d x %d\n",
            s,wt,wds[wt].winf.wiq.x1,wds[wt].winf.wiq.y1,wds[wt].winf.wiq.x2,
            wds[wt].winf.wiq.y2);
   }
}

void window_sense(WT wt) {
   int        ltrace = OFF;
   long       x1, y1, dx, dy;
   FL_OBJECT *gl;

   assert(wt != WT_FORM);
   if( wds[wt].status != WINDOW_NOT_ACTIVE ) {
      if( gl = gl_from_wt(wt) ) {
         wds[wt].winf.wiq = make_iquad(gl->x, gl->y, gl->x+gl->w-1, gl->y+gl->h-1);
         if( ltrace ) {
            fprintf(stderr,"window_sense: wds[%d].winf.wiq=%d %d %d %d\n",wt,
               wds[wt].winf.wiq.x1,wds[wt].winf.wiq.x2,
               wds[wt].winf.wiq.y1,wds[wt].winf.wiq.y2);
         }
      } else {
         fprintf(stderr,
            "window_sense: FATAL: Could not find gl corresponding to wt=%d\n",wt);
         exit(1);
      }
   } 
}

void window_sense_old(wt)
   WT       wt;
{
   long     x1, y1, dx, dy;

   if( wds[wt].status != WINDOW_NOT_ACTIVE ) {
      winset(wds[wt].sgi_winno);
      getorigin(&x1,&y1); getsize(&dx,&dy);
      wds[wt].winf.wiq = make_iquad(x1,y1,x1+dx-1,y1+dy-1);
   } 
}


IQUAD old_iquad_window_sense(wt)
   WT       wt;
{
   long     x1, y1, dx, dy;

   if( wds[wt].status == WINDOW_ACTIVE ) {
      winset(wds[wt].sgi_winno);
      getorigin(&x1,&y1); getsize(&dx,&dy);
      return(make_iquad(x1,y1,x1+dx-1,y1+dy-1));
   } else {
      return(make_iquad(0,0,0,0));
   }
}

IQUAD iquad_window_sense(wt)
   WT       wt;
{
   if( wds[wt].status == WINDOW_ACTIVE ) {
      return iquad_from_winno(wds[wt].sgi_winno);
   } else {
      return(make_iquad(0,0,0,0));
   }
}

IQUAD iquad_from_winno(winno)
   int      winno;
{
   long     x1, y1, dx, dy;

   winset(winno);
   getorigin(&x1,&y1); getsize(&dx,&dy);
   return make_iquad(x1,y1,x1+dx-1,y1+dy-1);
}

IQUAD relative_iquad_from_winno(winno)
   int      winno;
{
   long     dx, dy;

   winset(winno);
   getsize(&dx,&dy);
   return make_iquad(0,0,dx,dy);
}

void old_dump_window_xy_data(wt) 
   WT         wt;
{
   int        max_per_dump = 10, inc;
   PXYDVBIND  p;
#define FP wds[wt].pf.pxyf
     
   printf("dump_window_xy_data: forward ...\n");
   for( p = FP->xyhead; p != NULL; p = p->next ) {
      inc = (p->data)->n / max_per_dump; if( inc == 0 ) ++inc;
      if( inc == 1) 
         printf("Index: %d    Time: %f\n",p->index,p->time);
      else
         printf("Index: %d    Time: %f (p)\n",p->index,p->time);
      dvprint_inc((p->data)->x,(p->data)->n,inc,"x"); 
      dvprint_inc((p->data)->y,(p->data)->n,inc,"y"); 
   }
   printf("dump_window_xy_data: drawrof ...\n");
   test_ring(wt);
   printf("dump_window_xy_data: reverse ...\n");
   for( p = FP->xytail; p != NULL; p = p->prev ) {
      inc = (p->data)->n / max_per_dump; if( inc == 0 ) ++inc;
      if( inc == 1)
         printf("Index: %d    Time: %f\n",p->index,p->time);
      else
         printf("Index: %d    Time: %f (p)\n",p->index,p->time);
      dvprint_inc((p->data)->x,(p->data)->n,inc,"x");
      dvprint_inc((p->data)->y,(p->data)->n,inc,"y");
   }
   printf("dump_window_xy_data: esrever ...\n");
#undef   FP
}


void dump_window_xy_data(wt) 
   WT         wt;
{
   int        max_per_dump = 10, inc;
   PXYDVBIND  p;
#define FP wds[wt].pf.pxyf
     
   unmake_xy_ring(wt);

   printf("dump_window_xy_data: begin ...\n");
   for( p = FP->xyhead; p != NULL; p = p->next ) {
      inc = (p->data)->n / max_per_dump; if( inc == 0 ) ++inc;
      if( inc == 1) 
         printf("Index: %d    Time: %f    Tag: %d    Events: %d.\n",
                 p->index,p->time,p->tag,(p->data)->n);
      else
         printf("Index: %d    Time: %f    Tag: %d    Events: %d (p).\n",
                p->index,p->time,p->tag,(p->data)->n);
      dvprint_inc((p->data)->x,(p->data)->n,inc,"x"); 
      dvprint_inc((p->data)->y,(p->data)->n,inc,"y"); 
      if( p->tdata ) {
         printf(">>> Index %d: Tag length %d.\n",p->index,p->tdata->n);
      } else {
         printf(">>> Index %d: No tag.\n",p->index);
      }
   }
   printf("dump_window_xy_data: end ...\n");

   restore_ring_state(wt);

#undef  FP
}

void scan_all_pxy(wt) 
   WT          wt;
{
#define FP wds[wt].pf.pxyf

   printf("scan_all_pxy: tag %d.\n",wt);
   unmake_xy_ring(wt);
   scan_pxydvbind(FP->xyhead,"xyhead");
   scan_pxydvbind(FP->xytail,"xytail");
   scan_pxydvbind(FP->xyfirst,"xyfirst");
   scan_pxydvbind(FP->xylast,"xylast");
   scan_pxydvbind(FP->xycurr,"xycurr");
   scan_pxydvbind(FP->xymr,"xymr");
#undef  FP
}

void scan_pxydvbind(p,s)
   PXYDVBIND      p;
   char          *s;
{
   PXYDVBIND      q;
   int            i;

   printf("\npxydvbind (forward): %s.\n",s);
   for( q = p, i = 1; q != NULL; q = q->next, i++ ) {
      printf(q->tdata ? "node %d: %d events. Tagged.\n" :
                        "node %d: %d events. No tag.\n",i,(q->data)->n);
   }
   printf("pxydvbind (reverse): %s.\n",s);
   for( q = p, i = 1; q != NULL; q = q->prev, i++ ) {
      printf(q->tdata ? "node %d: %d events. Tagged.\n" :
                        "node %d: %d events. No tag.\n",i,(q->data)->n);
   }
}

void test_ring(wt)
   WT        wt;
{
   PXYDVBIND p;
   int       i, imax = 50;
#define FP wds[wt].pf.pxyf

   make_xy_ring(wt);
   printf("test_ring: Should loop to %d unless null.\n",imax);
   for( p = FP->xyfirst, i = 0; i < imax && p != NULL; i++, p = p->next ) 
      printf("%d ",i);
   printf("\n");
   unmake_xy_ring(wt);
#undef  FP
}

/* Routine now deprecated ... */

void dump_mallinfo() 
{
/*
   struct     mallinfo   m, mallinfo(void);

   m = mallinfo();

   printf("dump_mallinfo: Begin dump.\n");
   printf("arena:%d   ordblks:%d   smblks:%d\nhblkhd:%d   hblks:%d   usmblks:%d\nfsmblks:%d   uordblks:%d   fordblks:%d   keepcost:%d\n",
          m.arena, m.ordblks, m.smblks, m.hblkhd, 
          m.hblks, m.usmblks, m.fsmblks, m.uordblks, 
          m.fordblks, m.keepcost);
   printf("dump_mallinfo: End dump.\n");
*/
}

WT old_wt_from_winno(winno)
   int      winno;
{
extern   
   PWTBIND  wtal;

     PWTBIND   p;

   if( wds[SCOPE_WT].sgi_winno == winno )
      return( SCOPE_WT );
   for( p = wtal; p != (PWTBIND) NULL; p = p->next ) {
      if( wds[p->wt].sgi_winno == winno ) {
         return( p->wt );
      }
   }
   return(-1);
}

WT wt_from_winno(winno)
   int      winno;
{
extern   
   PLALBIND  lalroot;

   PWTBIND   p;

   if( wds[SCOPE_WT].sgi_winno == winno )
      return( SCOPE_WT );
   for( p = lalroot->wtl; p != (PWTBIND) NULL; p = p->next ) {
      if( wds[p->wt].sgi_winno == winno ) {
         return( p->wt );
      }
   }
   return(-1);
}

#define       P1           wds[wt].winf.pwf.pxywinf
#define       P2           wds[wt].pf.pxyf

/* Links lxy (number of vectors stored) and display_index  of xywinf 
   structure to xyf structure. */

void link_xywinf_to_xyf(wt) 
   WT         wt;
{
   if( is_active_wt(wt) ) {
      P1->plxy = &(P2->lxy);
      P1->pdisplay_index = &(P2->display_index);
   }
}

void set_display_index(wt,display_index)
   WT        wt;
   int       display_index;
{
   if( is_active_wt(wt) ) {
      P2->display_index = display_index;
   }
}

void set_display_index_current(wt)
   WT        wt;
{
   if( is_active_wt(wt) ) {
      if( P2->xycurr ) {
         P2->display_index = P2->xycurr->index;
      }
   }
}

int get_display_index(wt)
   WT        wt;
{
   return is_active_wt(wt) ? P2->display_index : -1;
}

double display_index_to_unit_interval(WT wt) {
   if( is_active_wt(wt) ) {
      if( P2->lxy > 1 ) {
         return (get_display_index(wt) - 1.0) / (P2->lxy - 1.0);
      } else {
         return 0;
      }
   } else {
      return 0;
   }
}

#undef P1
#undef P2

PXYDVEC pnth_XYDVEC(p,n)
   PXYDVBIND     p;
   int           n;
{
   PXYDVEC       pp;

   if( n < 1 )
      return ((PXYDVEC) NULL);
   for(  ; p != NULL; p = p->next, n-- ) {
      if( n == 1 )
         return p->data;  
   }
   return ((PXYDVEC) NULL);
}

/* Set and retrieve window's must_remap values --- no error checking. */
/* As side effect, set_must_remap invokes set_must_redraw ... */

int set_must_remap(wt)
   WT       wt;
{
#define P wds[wt].winf.pwf.pxywinf 
   if( P ) {
      P->must_remap = ON;
      set_must_redraw(wt);
      return ON;
   } else {
      return OFF;
   }
}

int unset_must_remap(wt)
   WT       wt;
{
   P->must_remap = OFF;
   return OFF;
}

int get_must_remap(wt)
   WT       wt;
{
   return P ? P->must_remap : 0;
}

/* Set and retrieve window's must_redraw values --- no error checking. */

int set_must_redraw(WT wt) {
   P->must_redraw = ON;
   return ON;
}

int unset_must_redraw(wt)
   WT       wt;
{
   P->must_redraw = OFF;
   return OFF;
}

int get_must_redraw(wt)
   WT       wt;
{
   return P->must_redraw;
}

int set_domain_mode(wt,mode)
   WT       wt;
   int      mode;
{
   P->domain_mode = mode;
   return mode;
}

int get_domain_mode(wt)
   WT       wt;
{
   return P->domain_mode;
}
#define     M_TYPE P->marker_type
#define     MMM    P->marker_multi_mode

void set_marker_type(WT wt,int marker_type) {
   P->marker_type = marker_type;
}

void set_marker_size_relative(WT wt,int n_times_default) {
   P->marker_size = n_times_default * DEFAULT_MARKER_SIZE;
}

void set_marker_size(WT wt,double marker_size) {
   P->marker_size = marker_size;
}

void set_point_size(WT wt,int point_size) {
   P->point_size = point_size;
}

void set_marker_npoly(WT wt,int npoly) {
   P->marker_npoly = npoly;
}

void set_marker_multi_mode_val(WT wt,int val) {
   P->marker_multi_mode = val;
}

int set_marker_multi_mode(WT wt) {
   set_marker_multi_mode_val(wt,1);
   return 1;
}

int unset_marker_multi_mode(WT wt) {
   set_marker_multi_mode_val(wt,0);
   return 0;
}

int set_marker_multi_mode_old(WT wt) {
   MMM = MULTI_OFFSET; 
   M_TYPE = get_base_marker_type(wt) + MULTI_OFFSET;
   return MMM;
}

int set_marker_line_auto_break(WT wt) {
   return( P->marker_line_auto_break = ON );
}

int unset_marker_line_auto_break(WT wt) {
   return( P->marker_line_auto_break = OFF );
}

int unset_marker_multi_mode_old(wt)
   WT       wt;
{
   MMM = OFF;          
   M_TYPE = get_base_marker_type(wt);
   return MMM;
}

int set_marker_limits(WT wt) {
   char   *lbuffer;
Prompt:
   sprintf(O_buf,"Marker color limits (%d--%d): ",
           get_default_min_marker(),get_default_max_marker());
   lbuffer = get_string_from_window(scope_sgi_winno,fh_prompt,
                                    GREEN,BLACK,CYAN,O_buf,0.01,0.20);
   if( !lbuffer ) {
      Min_marker_color = get_default_min_marker();
      Max_marker_color = get_default_max_marker();
   } else {
      if( sscanf(lbuffer,"%d %d",&Min_marker_color,&Max_marker_color) != 2 ) 
          goto Prompt;
      if( Min_marker_color < get_abs_min_marker() || 
          Max_marker_color > get_abs_max_marker() ||
          Min_marker_color > Max_marker_color ) goto Prompt;
      free(lbuffer);
   }
   Mod_marker_color = Max_marker_color - Min_marker_color + 1;
   sprintf(O_buf,"New marker color limits %d--%d.",
           Min_marker_color,Max_marker_color);
   scope_mess(O_buf,1);

   Default_point_color = Min_marker_color + 2;
   Default_line_color =  Min_marker_color + 4;
   Default_poly_color =  Min_marker_color + 2;
   Default_marker_color = Default_point_color;

   return wt;

}

int set_marker_rainbow_mode(wt)
   WT       wt;
{
   MMM = RAINBOW_OFFSET; 
   M_TYPE = get_base_marker_type(wt) + RAINBOW_OFFSET;
   return MMM;
}

int unset_marker_rainbow_mode(wt)
   WT       wt;
{
   MMM = OFF;          
   M_TYPE = get_base_marker_type(wt);
   return MMM;
}

int get_marker_multi_mode(wt)
   WT       wt;
{
   return MMM;

#undef      MMM
#undef      M_TYPE
}

int set_border_color(wt,border_color)
   WT        wt;
   int       border_color;
{
   P->border_color = border_color;
   return wt;
}

int change_border_color(wt,border_color)
   WT        wt;
   int       border_color;
{
   if( is_active_wt(wt) ) {
      set_border_color(wt,border_color);
      redraw_window_in_panel(wt);
   }
   return wt;
}
#undef  P

/* DQ stack routines. */

PDQBIND flush_dq_stack(PDQBIND p) {
   if( p ) {
      for( ; (p = pop_dq(p)) != NULL; ){
      }
   }
   return NULL;
}

PDQBIND flush_dq_stack_no_empty(PDQBIND p) {
   if( p ) {
      p->prev = flush_dq_stack(p->prev);
      p->next = NULL;
   }
   return(p);
}

PDQBIND push_dq(p,dq)
   PDQBIND      p;
   DQUAD        dq;
{
   PDQBIND      pp;

   if( p == NULL ) {
      if( (pp = make_pdqbind(dq,NULL,NULL)) == NULL ) {
         printf("push_dq: malloc problem.\n");
      }
   } else {
      if( (pp = make_pdqbind(dq,NULL,p)) == NULL ) {
         printf("push_dq: malloc problem.\n");
      }
      p->next = pp;
   }
   return pp;
}

PDQBIND pop_dq(p)
   PDQBIND     p;
{
   PDQBIND     pp;

   if( (pp = p) != NULL ) {
      pp = p->prev;
      free(p);
      if( pp != NULL ) {
         pp->next = NULL;
      }
   }
   return pp;
}

PDQBIND pop_dq_no_empty(p)
   PDQBIND       p;
{
   PDQBIND       pp;

   if( p != NULL ) {
      if( (p->prev) != NULL ) {
         return pop_dq(p);
      } else {
         return p;
      }
   } else {
      return NULL;
   }
}

PDQBIND replace_dq(p,dq)
   PDQBIND     p;
   DQUAD       dq;
{
   if( p != NULL ) {
      p->dq = dq;
   }
   return p;
}

/* Supply with "top" of stack.  */

int dq_stack_depth(PDQBIND p) {
   if( !p ) {
      return 0;
   } else {
      return 1 + dq_stack_depth(p->prev);
   }
}

#define FP wds[wt].pf.pxyf

void dqc_from_dqb(WT wt) {
   FP->dqc = push_dq(flush_dq_stack(FP->dqc),(FP->dqb)->dq);
   set_must_remap(wt);
}

void dqc_reset(WT wt) {
   FP->dqc = push_dq(flush_dq_stack(FP->dqc),FP->limits);
}

void dqc_from_current_xrange(WT wt) {
   double    xmin = (FP->dqc)->dq.x1;
   double    xmax = (FP->dqc)->dq.x2;

   FP->dqc = flush_dq_stack(FP->dqc);
   (void) push_limits_r(wt,xmin,xmax);
   set_must_remap(wt); set_must_redraw(wt);
}

#undef  FP

DQUAD dq_wt(wt)
   WT          wt;
{
   DQUAD       dq;  

   unmake_xy_ring(wt);
   dq = replace_degenerate_DQUAD(dq_xydv_list(wds[wt].pf.pxyf->xyhead));
   restore_ring_state(wt);
   
   return dq;
}

DQUAD dq_xydv_list(p)            
   PXYDVBIND     p;
{
   double        xmin = MAXDOUBLE,    xmax = -MAXDOUBLE,
                 ymin = MAXDOUBLE,    ymax = -MAXDOUBLE;
   double        r;
   int           n;
#define          P     (p->data)

   for( ; p != NULL ; p = p->next ) {
         n = P->n;
         if( (r = dvmin(P->x,n)) < xmin ) xmin = r; 
         if( (r = dvmin(P->y,n)) < ymin ) ymin = r; 
         if( (r = dvmax(P->x,n)) > xmax ) xmax = r; 
         if( (r = dvmax(P->y,n)) > ymax ) ymax = r; 
   }

   return make_dquad(xmin,ymin,xmax,ymax);
#undef           P
}

DQUAD dq_xydv_list_r(p,xmin,xmax)            
   PXYDVBIND     p;
   double        xmin,      xmax;
{
   double        ymin = MAXDOUBLE,    ymax = -MAXDOUBLE;

   double        r;
   int           i = 0,               n;
#define          P     (p->data)

   for( ; p != NULL ; p = p->next ) {
         n = P->n;
         if( (r = dxymin_range(P->x,P->y,n,xmin,xmax)) < ymin ) ymin = r; 
         if( (r = dxymax_range(P->x,P->y,n,xmin,xmax)) > ymax ) ymax = r; 
         i++;
   }
/* printf("dq_xydv_list_r: %d vectors processed.\n",i);  */
/* printf("dq_xydv_list_r: Range: %g %g.\n",xmin,xmax);   */
/* printf("dq_xydv_list_r: Quad: ");   */
/* dump_dquad(make_dquad(xmin,ymin,xmax,ymax));  */
   return make_dquad(xmin,ymin,xmax,ymax);
#undef           P
}

DQUAD dq_xydv(p)
   PXYDVBIND     p;
{
   int           n;
#define          P     (p->data)

   if( p != NULL ) {
      n = P->n;
      return make_dquad(dvmin(P->x,n),dvmin(P->y,n),
                        dvmax(P->x,n),dvmax(P->y,n));
   } else {
      printf("dq_xydv: null pointer passed to routine.\n");
      return make_dquad(MAXDOUBLE,MAXDOUBLE,-MAXDOUBLE,-MAXDOUBLE);
   }
#undef           P
}

DQUAD dq_xydv_r(p,xmin,xmax)
   PXYDVBIND     p;
   double        xmin,                xmax;
{
   int           n;
#define          P     (p->data)

   int           ltrace = OFF;

   if( p != NULL ) {
      n = P->n;
      if( ltrace ) {
         printf("dq_xydv_r: Range: %g %g.\n",xmin,xmax);    
         printf("dq_xydv_r: Quad: ");    
         dump_dquad(make_dquad(xmin,dxymin_range(P->x,P->y,n,xmin,xmax), 
                               xmax,dxymax_range(P->x,P->y,n,xmin,xmax)));
      }
      return make_dquad(xmin,dxymin_range(P->x,P->y,n,xmin,xmax), 
                        xmax,dxymax_range(P->x,P->y,n,xmin,xmax));
   } else {
      printf("dq_xydv: null pointer passed to routine.\n");
      return make_dquad(MAXDOUBLE,MAXDOUBLE,-MAXDOUBLE,-MAXDOUBLE);
   }
#undef           P
}

#define           P        wds[wt].pf.pxyf
void  dqc_from_xycurr(wt)
   WT             wt;
{
   P->dqc = push_dq(P->dqc,dq_xydv(P->xycurr));
}

void  dqc_for_full_scale(WT wt)
{
   int    ltrace = OFF;

   if( P->dqc ) {
      if( P->xycurr ) {
         if( ltrace ) {
            fprintf(stderr,
               "dqc_for_full_scale: window_stats(%d) before push_dq.\n",wt);
            window_stats(wt);
         }
         P->dqc = pop_dq_no_empty(P->dqc);
         P->dqc = push_dq(P->dqc,
                          dq_xydv_r(P->xycurr,P->dqc->dq.x1,P->dqc->dq.x2));
         if( ltrace ) {
            fprintf(stderr,
               "dqc_for_full_scale: window_stats(%d) after  push_dq.\n",wt);
            window_stats(wt);
         }
         set_must_remap(wt);
      }
   } else {
      printf("dqc_for_full_scale:: Should not be here.\n");
   }
}

int var_dquad_display_mode(WT wt) {
#define    MODE wds[wt].winf.pwf.pxywinf->display_mode
   return  (MODE == AUTO_MODE)   ||
           (MODE == SSTEP_MODE)  ||
           (MODE == FREEZE_MODE); 
#undef     MODE
}

void  dqc_for_var_dquad_zoom(WT wt)
{
#define    CO  P->dqb->dq
#define    FP  wds[wt].winf.pwf.pxywinf

   double  s;

   if( P->dqc ) {
      if( P->xycurr ) {
         s = display_index_to_unit_interval(wt);
         P->dqc = push_dq(pop_dq_no_empty(P->dqc),
             make_dquad(CO.x1,CO.y1,
             CO.x2 * exp(s*log(FP->var_dquad_zoom)),CO.y2));
         set_must_remap(wt);
      }
   } else {
      fprintf(stderr,"dqc_for_test_var_dquad:: Should not be here.\n");
   }
#undef     FP
#undef     CO
}

/* Currently unused ... requires separate switch for full scale x ... */

void  dqc_for_full_scale_x_and_y(WT wt)
{
#define           X   P->xycurr->data->x
#define           N   P->xycurr->data->n
   if( P->dqc ) {
      P->dqc = push_dq(P->dqc,
                       dq_xydv_r(P->xycurr,dvmin_(X,&N),dvmax_(X,&N)));
      set_must_remap(wt);
   } else {
      printf("dqc_for_full_scale:: Should not be here.\n");
   }
#undef            X
#undef            N
}

void auto_border_x_wt(wt)
   WT             wt;
{
#define           FP wds[wt].pf.pxyf

   DQUAD          limits;

   double         xrange,   offset  = Auto_border_x_offset, 
                            offset0 = 2.0 * Auto_border_x_offset;
   int            ltrace = OFF;

   if( ltrace ) {
      printf("auto_border_x_wt: Window %d.\n",wt);
   }
   unmake_xy_ring(wt);
   limits = dq_xydv_list(wds[wt].pf.pxyf->xyhead);
   if( ltrace ) {
      printf("auto_border_x_wt: Current limits "); dump_dquad(limits);
   }
   xrange = limits.x2 - limits.x1;
   if( xrange == 0.0 ) {
      FP->dqc = replace_dq( FP->dqc,
                            make_dquad(limits.x1 - offset0,limits.y1,
                                       limits.x2 + offset0,limits.y2) );
   } else {
      FP->dqc = replace_dq( FP->dqc,
                            make_dquad(limits.x1 - offset * xrange,
                                       limits.y1,
                                       limits.x2 + offset * xrange,
                                       limits.y2) );
   }
   set_must_remap(wt); set_must_redraw(wt); restore_ring_state(wt);
#undef            FP
}


void auto_border_xy_wt(wt)
   WT             wt;
{
#define           FP wds[wt].pf.pxyf

   DQUAD          limits;

   double         xrange,   xoffset  = Auto_border_x_offset, 
                            xoffset0 = 2.0 * Auto_border_x_offset;
   double         yrange,   yoffset  = Auto_border_y_offset, 
                            yoffset0 = 2.0 * Auto_border_y_offset;
   int            ltrace = OFF;

   if( ltrace ) {
      printf("auto_border_xy_wt: Window %d.\n",wt);
   }
   unmake_xy_ring(wt);
   limits = dq_xydv_list(wds[wt].pf.pxyf->xyhead);
   if( ltrace ) {
      printf("auto_border_xy_wt: Current limits "); dump_dquad(limits);
   }
   xrange = limits.x2 - limits.x1;
   yrange = limits.y2 - limits.y1;
   if( xrange == 0.0 ) {
      if( yrange == 0.0 ) {
      } else {
         FP->dqc = replace_dq( FP->dqc,
                               make_dquad(limits.x1 - xoffset0,limits.y1,
                                          limits.x2 + xoffset0,limits.y2) );
      }
   } else {
      if( yrange == 0.0 ) {
      } else {
         FP->dqc = replace_dq( FP->dqc,
                               make_dquad(limits.x1 - xoffset * xrange,
                                          limits.y1 - yoffset * yrange,
                                          limits.x2 + xoffset * xrange,
                                          limits.y2 + yoffset * yrange) );
      }
   }
   set_must_remap(wt); set_must_redraw(wt); restore_ring_state(wt);
#undef            FP
}

void manual_border_wt(wt)
   WT             wt;
{
   int            ltrace = OFF;
      
   if( ltrace ) {
      printf("manual_border_wt:: Window %d.\n",wt);
   }
   sprintf(O_buf,"manual_border_wt:: Not implemented.");
   scope_mess_and_bell(O_buf,1);
}

#undef            P

void push_limits_r(wt,xmin,xmax)
   WT              wt;
   double          xmin,          xmax;
{
#define            P              wds[wt].pf.pxyf

/* printf("push_limits_r: range: %g %g.\n",xmin,xmax);  */
   if( is_active_wt(wt) ) {   
      unmake_xy_ring(wt);
      P->dqc = push_dq(P->dqc,dq_xydv_list_r(P->xyhead,xmin,xmax));
      restore_ring_state(wt);
   } else {
      printf("push_limits_r: window %d not active.\n",wt);    
   }
}
#undef             P

void safe_set_pxy_wt(wt)
   WT              wt;
{
   PXYDVBIND       lp;

   int             ltrace = OFF;

#define            P              wds[wt].pf.pxyf
   if( !P ) return;

/* Ensure that xylast, xyfirst are set properly ... */
 
   make_xy_ring(wt);
   unmake_xy_ring(wt);

   if( ltrace ) {
      scan_pxydvbind(P->xyhead,"safe_set_pxy_wt (post unmake)");
   }
   for( lp = P->xyhead; lp && lp->next; lp = lp->next );
   P->xytail  = lp;
   P->xyfirst = P->xyhead;
   P->xylast  = P->xytail;
   P->xycurr  = P->xyhead;
   P->xymr    = P->xycurr;
   restore_ring_state(wt);
}

void thin_wt_old(wt,thetac)
   WT              wt;
   double          thetac;
{
   PXYDVBIND       p;
   int             i;
   int             mess_freq = 10;

   unmake_xy_ring(wt);
   for( i = 1, p = P->xyhead; p != (PXYDVBIND) NULL; i++, p = p->next ) {
      P->xytail = p = thin_PXYDVBIND(p,thetac);
      if( (mess_freq > 0)  &&  ((i % mess_freq) == 0) ) {
         sprintf(Buffer,"Thinned frame %d of %d.",i,P->lxy);
         put_string_to_window(scope_sgi_winno,fh_prompt,
                              Buffer,RED,BLACK,0.01,0.20);
      }
   }
   if( P->xyhead && mess_freq > 0 ) {
      scope_mess("Thinning completed.",1);
   }
   safe_set_pxy_wt(wt);
   P->nevents = n_events_wt(wt);
   restore_ring_state(wt);
   set_must_remap(wt); set_must_redraw(wt);
}

void thin_wt(WT wt, double thetac) {
   PXYDVBIND       p;
   int             i;
   int             mess_freq = 10;

   if( !is_active_wt(wt) ) return;

   unmake_xy_ring(wt);
   for( i = 1, p = P->xyhead; p != (PXYDVBIND) NULL; i++, p = p->next ) {
      P->xytail = p = thin_PXYDVBIND(p,thetac);
   }
   safe_set_pxy_wt(wt);
   P->nevents = n_events_wt(wt);
   restore_ring_state(wt);
   set_must_remap(wt); set_must_redraw(wt);
}

void time_reverse_wt(WT wt) {
   int         ltrace = OFF;
   int         nnode,  inode = 0;
   PXYDVBIND   pxy,    tpxy;

   if( !is_active_wt(wt) ) return;

   unmake_xy_ring(wt);

   nnode = n_PXYDVBIND(P->xyhead);
   pxy = P->xyhead;
   while( pxy ) {
#define  NN(p) ( p ? p->data->n : 0 )
#define  SWAP(p1,p2) tpxy = p1; p1 = p2; p2 = tpxy
      inode++;
      if( ltrace ) {
         fprintf(stderr,"time_reverse_wt: Node <%d>  nprev <%d> nnext<%d>\n",
                 inode,NN(pxy->prev),NN(pxy->next));
      }
      SWAP(pxy->next,pxy->prev);
      if( ltrace ) {
         fprintf(stderr,"time_reverse_wt: Node <%d>  nprev <%d> nnext<%d>\n",
                 inode,NN(pxy->prev),NN(pxy->next));
      }
      pxy = tpxy;
   }
   if( ltrace ) {
      fprintf(stderr,"time_reverse_wt:  nnode: %d  inode: %d.\n",nnode,inode);
   }
   if( P->xyhead ) {
      SWAP(P->xyhead,P->xytail);
      P->xyfirst = P->xyhead;
      P->xylast  = P->xytail;
      P->xycurr  = P->xyhead;
      reindex_XYDVlist(P->xyhead,1);
      safe_set_pxy_wt(wt);
      P->nevents = n_events_wt(wt);
      if( ltrace ) {
         fprintf(stderr,"time_reverse_wt: nevents: %d.\n",P->nevents);
      }
   }
   restore_ring_state(wt);
   set_must_remap(wt); set_must_redraw(wt);
#undef   SWAP
#undef   NN
}

void apply_func_0_wt(wt,func)
   WT              wt;
   PXYDVBIND_FCN   func;
{
   PXYDVBIND       p;

   int             i;

   int             mess_freq = 0;
   
   unmake_xy_ring(wt);
   for( i = 1, p = P->xyhead; p != (PXYDVBIND) NULL; i++, p = p->next ) {
      P->xytail = p = (* func)(p);
      if( (mess_freq > 0)  &&  ((i % mess_freq) == 0) ) {
         sprintf(Buffer,"Processed frame %d of %d.",i,P->lxy);
         put_string_to_window(scope_sgi_winno,fh_prompt,
                              Buffer,RED,BLACK,0.01,0.20);
      }
   }
   P->limits =  dq_xydv_list(P->xyhead);

   if( P->xyhead && mess_freq > 0 ) {
      scope_mess("Processing completed.",1);
   }

   safe_set_pxy_wt(wt);
   P->nevents = n_events_wt(wt);
   dqc_reset(wt);

   set_display_index_current(wt);
   restore_ring_state(wt);
   set_must_remap(wt); set_must_redraw(wt);
}
#undef             P

WT  wt_from_name(name)  
   char      *name;
{
   WTBIND   *p;
#define      NAME wds[p->wt].name

   for( p = wtal; p != NULL; p = p->next) {
      if( !strcmp(NAME,name) ) return p->wt;
   }
   return -1;
#undef       NAME
}

#define CO wds[wt].pf.pxyf->limits

double wt_xmin(wt)
   WT         wt;
{
   return CO.x1;
}

double wt_xmax(wt)
   WT         wt;
{
   return CO.x2;
}

double wt_ymin(wt)
   WT         wt;
{
   return CO.y1;
}

double wt_ymax(wt)
   WT         wt;
{
   return CO.y2;
}

#undef  CO


int wait_for_mouse()
{
   short    val,       device;

#define     RETURN(a)  X_mouse = getvaluator(MOUSEX); Y_mouse =  getvaluator(MOUSEY); qreset(); return a

   qreset();
   while( ON ) {
      while( qtest() ) {
         device = qread(&val);
         if( val ) {
            switch (device) {
            case LEFTMOUSE:
               RETURN(-1);
            case MIDDLEMOUSE:
               RETURN(0);
            case RIGHTMOUSE:
               RETURN(1);
            default:
               break;
            }
         }
      }
   }
#undef      RETURN
}

double x_mouse_after_button()
{
   short    val;

   qreset();
   while( ON ) {
      while( qtest() ) {
         switch (qread(&val)) {
         case LEFTMOUSE:
         case MIDDLEMOUSE:
         case RIGHTMOUSE:
            if( val ) {
               qreset();
               return xmouse_w_wt(S_wt);
            }
         default:
            break;
         }
      }
   }
}

double y_mouse_after_button()
{
   short    val;

   qreset();
   while( ON ) {
      while( qtest() ) {
         switch (qread(&val)) {
         case LEFTMOUSE:
         case MIDDLEMOUSE:
         case RIGHTMOUSE:
            if( val ) {
               qreset();
               return ymouse_w_wt(S_wt);
            }
         default:
            break;
         }
      }
   }
}

double x_from_mouse_or_status_window() {
   return 0.0;
}

int clone_wt(WT wt) {
   WT       nwt;
   int      winno;
   int      ltrace = OFF;

#define     P          wds[wt].pf.pxyf
   
   if( is_active_wt(wt) ) {
      if( (nwt = new_wt_insert_before_wt(wt)) > 0 ) {
         if( ltrace ) {
            printf("dump_wtal(); rdump_wtal(); in clone_wt(%d)\n",wt);
                    dump_wtal(); rdump_wtal();
         }
         if( (winno = xvs_open_and_repanel(nwt)) > 0 ) {
            sprintf(O_buf,"%s_",wds[wt].name);
            wds_init_1(nwt,wds[nwt].sgi_winno,O_buf,
                       wds[wt].owner,wds[wt].wd);
            wds_init_2(nwt,P->limits,P->scale_option,P->max_lxy,
                       P->max_nevents);
            if( ltrace ) {
               fprintf(stderr,"clone_wt: Initialized wt=%d\n",nwt);
            }
            unmake_xy_ring(wt);
            copy_xyf(wt,nwt);
            copy_xywinf(wt,nwt);
            window_sense(nwt);
            set_xycurr_nth(nwt,P->display_index);
            set_display_index_current(nwt);
            set_must_remap(nwt); set_must_redraw(nwt);
            if( Clone_mouse_nwt ) {
               centre_mouse_in_wt(nwt);
               if( ltrace ) fprintf(stderr,
                  "clone_wt: Returned from centre_mouse_in_wt()\n");
            }
            restore_ring_state(wt);
            restore_ring_state(nwt);
            if( ltrace ) fprintf(stderr,"clone_wt: Return val=%d\n",nwt);
            return nwt;
         } else {
            return -1;
         }
      } else {
         return -1;
      }
   } else {
      return -1;
   }
#undef      P 
}

void old_copy_xyf(WT wt1,WT wt2) {

#define      P1     wds[wt1].pf.pxyf
#define      P2     wds[wt2].pf.pxyf
#define      N      (p1->data)->n
   PXYDVBIND p1,        p2,         lp2;
   int       i = 0;

   for( p1 = P1->xyhead, lp2 = NULL; p1 != NULL; p1 = p1->next, i++ ) {
      if( (p2 = (PXYDVBIND) malloc(sizeof(XYDVBIND))) ) {
         if( i == 0 ) {
            P2->xyhead = p2;
         }
         p2->index = p1->index;
         p2->time  = p1->time;
         p2->tag   = p1->tag;  
         p2->data  = make_XYDVEC(N);
         dvcopy_((p1->data)->x,(p2->data)->x,&N);
         dvcopy_((p1->data)->y,(p2->data)->y,&N);
         if( p1->tdata ) {
#define     TAG1   p1->tdata
#define     TAG2   p2->tdata
            TAG2 = make_XYTAGVEC(TAG1->n);
            TAG2->tag = IVcopy(TAG1->tag,TAG2->tag,TAG1->n);
#undef      TAG2
#undef      TAG1
         } else {
            p2->tdata = NULL;
         }
         p2->next = NULL;
         p2->prev = lp2;
         if( lp2 ) lp2->next = p2;
         lp2 = p2;
      } else {
         printf("old_copy_xyf:: malloc failure ...\n");
         exit(-1);
      }
   }
   safe_set_pxy_wt(wt2);
   P2->xymr = P2->xyhead;
   P2->dqb = make_pdqbind(P2->limits,NULL,NULL);
   dqc_from_dqb(wt2);
   P2->lxy = P1->lxy;
   P2->nevents = P1->nevents;
   P2->limits_computed = P1->limits_computed;
   P2->thin_option = P1->thin_option;
   P2->new_data = P1->new_data;
   P2->ring_state = P1->ring_state;
#undef       P1
#undef       P2
#undef       N 
}

void copy_xyf(WT wt1,WT wt2) {

#define      P1     wds[wt1].pf.pxyf
#define      P2     wds[wt2].pf.pxyf
#define      N      (p1->data)->n
   PXYDVBIND p1,        p2,         lp2;
   int       i = 0;

   for( p1 = P1->xyhead, lp2 = NULL; p1 != NULL; p1 = p1->next, i++ ) {
      if( ( p2 = copy_PXYDVBIND(p1) ) ) {
         if( i == 0 ) {
            P2->xyhead = p2;
         }
         p2->next = NULL;
         p2->prev = lp2;
         if( lp2 ) lp2->next = p2;
         lp2 = p2;
      } else {
         printf("copy_xyf:: malloc failure ...\n");
         exit(-1);
      }
   }
   safe_set_pxy_wt(wt2);
   P2->xymr = P2->xyhead;
   P2->dqb = make_pdqbind(P2->limits,NULL,NULL);
   P2->dqc = copy_dq_stack(P1->dqc);
   P2->lxy = P1->lxy;
   P2->nevents = P1->nevents;
   P2->limits_computed = P1->limits_computed;
   P2->thin_option = P1->thin_option;
   P2->new_data = P1->new_data;
   P2->ring_state = P1->ring_state;
#undef       P1
#undef       P2
#undef       N 
}

void copy_xywinf(wt1,wt2)
   WT        wt1,      wt2;
{
#define      P1     wds[wt1].winf.pwf.pxywinf
#define      P2     wds[wt2].winf.pwf.pxywinf
#define      C(a)   P2->a = P1->a
   C(marker_type); C(marker_color); C(marker_size); C(point_size);
   C(marker_npoly); C(marker_multi_mode); C(marker_line_auto_break);
   C(line_color); C(line_width);
   C(display_mode); C(prev_display_mode); C(auto_sub_mode);
   C(heads_up); C(heads_up_lim); 
   if( P1->heads_up_t_format ) {
      P2->heads_up_t_format = strdup(P1->heads_up_t_format);
   }
   C(range_gauge);
   C(var_dquad); C(var_dquad_zoom);
   C(auto_break_mode); C(accept_frequency); C(status_window_flag);
   C(full_scale_mode);
#undef       C
#undef       P1
#undef       P2
}

/*  Creates single-group window with group length equal to number of */
/*  groups in wt ...                                                 */

int create_reduced_wt(WT wt)
{
   WT         nwt;
   int        winno;
   int        ltrace = OFF;

#define       P          wds[wt].pf.pxyf
   
   if( ltrace ) {
      printf("create_reduced_wt: In routine, wt: %d.\n",wt);
   }
   if( is_active_wt(wt) ) {
      if( (nwt = new_wt_insert_before_wt(wt)) > 0 ) {
         if( ltrace ) {
            printf("dump_wtal(); rdump_wtal(); in clone_wt(%d)\n",wt);
                    dump_wtal(); rdump_wtal();
         }
         if( (winno = xvs_open_and_repanel(nwt)) > 0 ) {
            sprintf(O_buf,"%s_",wds[wt].name);
            wds_init_1(nwt,wds[nwt].sgi_winno,O_buf,
                       wds[wt].owner,wds[wt].wd);
            wds_init_2(nwt,P->limits,P->scale_option,P->max_lxy,
                       P->max_nevents);
            unmake_xy_ring(wt);
            create_reduced_xyf(wt,nwt);
            copy_xywinf(wt,nwt);
            window_sense(nwt);
            set_xycurr_nth(nwt,1);
            set_display_index_current(nwt);
            set_must_remap(nwt); set_must_redraw(nwt);
            Force_centre_mouse_in_wt = ON; centre_mouse_in_wt(nwt);
            restore_ring_state(wt);
            restore_ring_state(nwt);
            return nwt;
         } else {
            return -1;
         }
      } else {
         return -1;
      }
   } else {
      return -1;
   }
#undef      P 
}

void create_reduced_xyf(WT wt1,WT wt2)
{
#define      P1     wds[wt1].pf.pxyf
#define      P2     wds[wt2].pf.pxyf
#define      N      (p1->data)->n
   PXYDVBIND p1,        p2,         lp2;
   int       i;
   double    f77_zero = 0.0,        f77_one = 1.0;

   if( (p2 = (PXYDVBIND) malloc(sizeof(XYDVBIND))) ) {
      P2->xyhead = p2;
      p2->index = 1;
      p2->time = 0.0;
      p2->tag = 0;
      p2->data = make_XYDVEC(P1->lxy);
      if( ctrl_key_pressed() ) {
         for( p1 = P1->xyhead, i = 0; p1 != NULL; p1 = p1->next, i++ ) {
            p2->data->x[i] = p1->time;
         }
      } else {
         dvramp_((p2->data)->x,&f77_one,&f77_one,&(P1->lxy));
      }
      dvramp_((p2->data)->y,&f77_one,&f77_one,&(P1->lxy));
      p2->next = NULL;
      p2->prev = NULL;
      p2->tdata = NULL;
   } else {
      printf("create_reduced_xyf:: malloc failure ...\n");
      exit(-1);
   }
   safe_set_pxy_wt(wt2);
   P2->xymr = P2->xyhead;
   P2->limits = dq_xydv_list(P2->xyhead);
   P2->dqb = make_pdqbind(P2->limits,NULL,NULL);
   dqc_from_dqb(wt2);
   P2->lxy = 1;
   P2->nevents = P1->lxy;
   P2->limits_computed = P1->limits_computed;
   P2->thin_option = P1->thin_option;
   P2->new_data = P1->new_data;
   P2->ring_state = P1->ring_state;
#undef       P1
#undef       P2
#undef       N 
}

void centre_mouse_in_wt(WT wt) {
   int      ltrace  =  Ltrace;

#define     CO      wds[wt].winf.wiq              
   if( ltrace ) {
      fprintf(stderr,"centre_mouse_in_wt: wt=%d.\n",wt);
      fprintf(stderr,"centre_mouse_in_wt: mouse_in_ser_domain()=%d\n",
         mouse_in_ser_domain());
      fprintf(stderr,"centre_mouse_in_wt: Force_centre_mouse_in_wt=%d\n",
         Force_centre_mouse_in_wt);
   }
   if( (is_active_wt(wt) &&  mouse_in_ser_domain()) || wt == scope_wt ||
       Force_centre_mouse_in_wt ) {
      if( ltrace ) fprintf(stderr,"centre_mouse_in_wt: Calling fl_set_mouse\n");
      fl_set_mouse((Xvsf->main->x) + (CO.x1 + CO.x2) / 2,
                   (Xvsf->main->y) + (CO.y1 + CO.y2) / 2);
      if( ltrace ) fprintf(stderr,"centre_mouse_in_wt: Calling activate_and_redraw...\n");
      activate_and_redraw_window_with_mouse();
      Force_centre_mouse_in_wt = OFF;
   }
#undef      CO
}


int mouse_is_near_wt(WT wt,double fracx,double fracy) {
   static char      R[] = "mouse_is_near_wt";
   FL_Coord         x, y;
   unsigned int     keymask;
   int              ltrace = OFF;

#define     CO      wds[wt].winf.wiq              
   double           x1, y1, x2, y2, xmouse, ymouse;

   if( ltrace ) {
      fprintf(stderr,"%s: wt=%d  geometry (user coords)=[%d %d %d %d]\n",
            R,wt,CO.x1,CO.x2,CO.y1,CO.y2);
   }
   if( is_active_wt(wt) || wt == scope_wt ) {
      x1 = CO.x1 - fracx * (CO.x2 - CO.x1);
      y1 = CO.y1 - fracy * (CO.y2 - CO.y1);
      x2 = CO.x2 + fracx * (CO.x2 - CO.x1);
      y2 = CO.y2 + fracy * (CO.y2 - CO.y1);
      fl_get_mouse(&x,&y,&keymask);
      xmouse = x;
      ymouse = y;
      if( ltrace ) {
         fprintf(stderr,"mouse_is_near_wt: Quad %lf %lf %lf %lf\n",x1,x2,y1,y2);
         fprintf(stderr,"mouse_is_near_wt: Mouse %lf %lf\n",xmouse,ymouse);
      }
      return ( x1 <= xmouse && xmouse <= x2 && y1 <= ymouse && ymouse <= y2 );
   } else {
      return (0);
   }
#undef      CO

}

#define     P            wds[wt].winf.pwf.pxywinf

int heads_up_status(void) {
   return ((N_F11Key % 2) ? ON : OFF);
}

void toggle_heads_up(wt)
   WT       wt;
{
   int      ltrace = OFF; 

   wds[wt].winf.pwf.pxywinf->heads_up = !wds[wt].winf.pwf.pxywinf->heads_up;
   set_must_redraw(wt);
   if( ltrace ) {
      printf("toggle_heads_up: Toggled %d.\n",wt);
   }
}

void set_heads_up(wt)
   WT       wt;
{
   wds[wt].winf.pwf.pxywinf->heads_up = ON;
   set_must_redraw(wt);
}

void unset_heads_up(wt)
   WT       wt;
{
   wds[wt].winf.pwf.pxywinf->heads_up = OFF;
   set_must_redraw(wt);
}

void toggle_heads_up_lim(wt)
   WT       wt;
{
   int      ltrace = OFF; 

   wds[wt].winf.pwf.pxywinf->heads_up_lim = !wds[wt].winf.pwf.pxywinf->heads_up_lim;
   set_must_redraw(wt);
   if( ltrace ) {
      printf("toggle_heads_up_lim: Toggled %d.\n",wt);
   }
}

void set_heads_up_lim(wt)
   WT       wt;
{
   wds[wt].winf.pwf.pxywinf->heads_up_lim = ON;
   set_must_redraw(wt);
}

void unset_heads_up_lim(wt)
   WT       wt;
{
   wds[wt].winf.pwf.pxywinf->heads_up_lim = OFF;
   set_must_redraw(wt);
}

void scan_windows()
{
   PWTBIND      p;

   printf("scan_windows: In routine ... \n");
   for( p = wtal; p != NULL; p = p->next ) 
   {
      printf("scan_windows: Window %d: Name: %s.\n",
             p->wt,wds[p->wt].name);
   }
   printf("scan_windows: Out of routine ... \n");
}

/* Needed modification for Indigo ...*/ 

int is_next_key_ESCKEY()
{
   Device       dev;
   short        val,       ESC_val_4D = 27,   ESC_val_Indigo = 1;

   qreset();
   qdevice(KEYBD);
   while( 1 ) {
      while( qtest() ) {
         dev = qread(&val);
         if( val ) {
            unqdevice(KEYBD);
            qreset();
            return( (val == ESC_val_4D)  || (val == ESC_val_Indigo) );
         }
      }
   }
}

void set_line_color(wt,line_color)
   WT         wt;
   int        line_color;
{
   P->line_color = line_color;
   set_must_redraw(wt);
}

void set_line_width(wt,line_width)
   WT         wt;
   int        line_width;
{
   P->line_width = line_width;
   set_must_redraw(wt);
}

void set_marker_color(wt,marker_color)
   WT         wt;
   int        marker_color;
{
   P->marker_color = marker_color;
   set_must_redraw(wt);
}

/* Returns quiescent border color (different if blocking). */

int Q_border_color(wt) 
   WT         wt;
{
   if( is_active_wt(wt) ) {
      switch( wds[wt].status ) {
      case WINDOW_ACTIVE:
         return Q_BORDER_COLOR;
      case WINDOW_BLOCKING:
         return BLOCK_BORDER_COLOR;
      default:
         printf("Q_border_color: Unexpected wds[%d].status: %d\n",wt,
                wds[wt].status);
         return 0;
      }
   } else {
      return 0;
   }
}

void set_status_wt(wt,status)
   WT         wt;
   int        status;
{
#define       BC     P->border_color

   if( is_active_wt(wt) ) {
      switch( status ) {
      case WINDOW_ACTIVE:
         if( BC != A_BORDER_COLOR ) {
            BC = Q_BORDER_COLOR;
         }
         wds[wt].status = status;       
         set_must_redraw(wt);
         break;
      case WINDOW_BLOCKING:
         if( BC != A_BORDER_COLOR ) {
            BC = BLOCK_BORDER_COLOR;
         }
         wds[wt].status = status;       
         set_must_redraw(wt);
         ringbell();
         break;
      default:
         printf("set_status_wt: Unexpected status: %d\n",status);
      }
   }
#undef        BC
}

#undef        P            

/* Pre-insertion not implemented ... */

void move_window(void) {
   move_window_1(0);
}

void move_window_1(int opcode) {
   int    ltrace = Ltrace;
   WT     wt_to_move,    wt_target;
   int    preinsert;

   if( scope_mode == SINGLE_MODE ) return;

   if( n_windows() < 2 ) {
      scope_mess_and_bell("Need at least two windows ...",1);
      return;
   }

   if( n_windows() == 2 ) {
      move_wt_after_wt(wt_from_index(1),wt_from_index(2));
      centre_mouse_in_wt(S_wt);
      return;
   }

   if( opcode == 0 ) {
      Menu_deferred = MENU_DEFERRED_MOVE;
      Menu_deferred_wt = S_wt; 
      Menu_picking = MENU_PICKING_ON; 
      pick_prompts_init(2);
      pick_prompts_set("Pick window to be moved ...");
      pick_prompts_set("Pick target window ...");
      mpick_wt(MPICK_START,2);
      return;
   }
   assert(Picked_wt_list.n == 2);
   wt_to_move = Picked_wt_list.v[0];
   wt_target  = Picked_wt_list.v[1];
   if( ltrace ) {
      fprintf(stderr,"move_window_1: Calling move_wt_after_wt(%d,%d)\n",
         wt_to_move,wt_target);
   }
   if( wt_to_move != wt_target ) {
      if( move_wt_after_wt(wt_to_move,wt_target) ) {
         centre_mouse_in_wt(wt_to_move);
      }
   } else {
      centre_mouse_in_wt(wt_to_move);
   }
}

#define CO  wds[wt].pf.pxyf->dqc->dq
#define ICO wds[wt].winf.wiq

double wtstringwidth(WT wt,fmfonthandle fh,char *string) {
   return fmgetstrwidth(fh,string) * (CO.x2 - CO.x1) / (ICO.x2 - ICO.x1);
}

double wtfracstringwidth(WT wt,fmfonthandle fh,char *string) {
   return fmgetstrwidth(fh,string) / ((double) (ICO.x2 - ICO.x1));
}

void old_wtdisplaystringoffset(WT wt,fmfonthandle fh,char *string,int col,
                               double xoffset,double yoffset,int code) {
   double    x,    y;
   int       ltrace = OFF;

   switch( code ) {
   case DS_LL:
      x = CO.x1 + xoffset * (CO.x2 - CO.x1);
      y = CO.y1 + yoffset * (CO.y2 - CO.y1);
      break;
   case DS_UL:
      x = CO.x1 + xoffset * (CO.x2 - CO.x1);
      y = CO.y2 - yoffset * (CO.y2 - CO.y1);
      break;
   case DS_UR:
      x = CO.x2 - (xoffset * (CO.x2 - CO.x1) + wtstringwidth(wt,fh,string));
      y = CO.y2 - yoffset * (CO.y2 - CO.y1);
      break;
   case DS_LR:
      x = CO.x2 - (xoffset * (CO.x2 - CO.x1) + wtstringwidth(wt,fh,string));
      y = CO.y1 + yoffset * (CO.y2 - CO.y1);
      break;
   default:
      printf("wtdisplaystringoffset: Unknown code: %d.\n",code);
      return;
   }
   if( wds[wt].sgi_winno > 0 ) {
      if( ltrace ) {
         fprintf(stderr,"wtdisplaystringoffset: displaying <%s>.\n",string);
      }
      winset(wds[wt].sgi_winno);
      cmov2((Coord) x,(Coord) y);
      color(col); fmsetfont(fh); fmprstr(string);
   } 
   return;
}

/*  Uses IQUAD coordinate system to stop text ``jitter'' on zoom. */

void wtdisplaystringoffset(WT wt,fmfonthandle fh,char *string,int col,
                           double xoffset,double yoffset,int code) {
   double    x,    y;
   int       ltrace = OFF;

   if( wds[wt].sgi_winno <= 0 ) return;

   equate_screen_and_world_coordinates(wds[wt].sgi_winno);

   switch( code ) {
   case DS_LL:
      x = ICO.x1 + xoffset * (ICO.x2 - ICO.x1);
      y = ICO.y1 + yoffset * (ICO.y2 - ICO.y1);
      break;
   case DS_UL:
      x = ICO.x1 + xoffset * (ICO.x2 - ICO.x1);
      y = ICO.y2 - yoffset * (ICO.y2 - ICO.y1);
      break;
   case DS_UR:
      x = ICO.x2 - (xoffset * (ICO.x2 - ICO.x1) + fmgetstrwidth(fh,string));
      y = ICO.y2 - yoffset * (ICO.y2 - ICO.y1);
      break;
   case DS_LR:
      x = ICO.x2 - (xoffset * (ICO.x2 - ICO.x1) + fmgetstrwidth(fh,string));
      y = ICO.y1 + yoffset * (ICO.y2 - ICO.y1);
      break;
   default:
      printf("wtdisplaystringoffset: Unknown code: %d.\n",code);
      return;
   }
   if( ltrace ) {
      fprintf(stderr,"wtdisplaystringoffset: displaying <%s>.\n",string);
   }
   winset(wds[wt].sgi_winno);
   cmov2((Coord) x,(Coord) y);
   color(col); fmsetfont(fh); fmprstr(string);

   set_must_remap(wt);

   return;
}

void wtdisplaystringabsolute(WT wt,fmfonthandle fh,char *string,int col,
                             int x,int y) {
   int      ltrace = OFF;
   int      act_x,      act_y;
   int      fuzz_x = 2;
   int      fuzz_y = 3;

   act_x = imin(ICO.x2-fuzz_x,imax(ICO.x1+fuzz_x,x));
   act_y = imin(ICO.y2-fuzz_y,imax(ICO.y1+fuzz_y,y));

   if( ltrace ) {
      fprintf(stderr,"wtdisplaystringabsolute: ");
      fprintf(stderr,"requested display at (%d,%d) in (%d:%d,%d:%d)\n",
              x,y,ICO.x1,ICO.x2,ICO.y1,ICO.y2);
      fprintf(stderr,"wtdisplaystringabsolute: ");
      fprintf(stderr,"actual display at (%d,%d)\n",act_x,act_y);
   }

   if( wds[wt].sgi_winno <= 0 ) return;

   cmov2((Coord) act_x,(Coord) act_y);
   color(col); fmsetfont(fh); fmprstr(string);

   return;
}

#undef  ICO
#undef  CO

#define GAUGE  wds[wt].winf.pwf.pxywinf->range_gauge

BAR_GAUGE make_default_range_gauge(void) {
   BAR_GAUGE    the;
      
   the.state = BAR_GAUGE_STATE_OFF;
   the.mode  = BAR_GAUGE_MODE_NONE;
    
   the.border_color = -1;
   the.background_color = -1;
   the.foreground_color = -1;
   the.dquad = make_dquad(0.0,0.0,1.0,1.0);

   return(the);
}

void dump_range_gauge(BAR_GAUGE range_gauge) {
   printf("range_gauge.state: %d\n",range_gauge.state);
   printf("range_gauge.mode:  %d\n",range_gauge.mode);
   printf("range_gauge.border_color:  %d\n",range_gauge.border_color);
   printf("range_gauge.background_color:  %d\n",range_gauge.background_color);
   printf("range_gauge.foreground_color:  %d\n",range_gauge.foreground_color);
   printf("range_gauge.dquad:\n");
   dump_dquad(range_gauge.dquad);
}

#undef  GAUGE

/* Routines for implementing 'point-wise' zoom ... */

int ix_first_visible(DQUAD dq,DVEC x,int n) {
   int      i;
   double   fuzz = 1.0e-5,  xmin = dq.x1;

   for( i = 0; i < n; i++ ) {
      if( fuzz_ge(x[i],xmin,fuzz) ) {
         return( i );
      }
   }
   return( -1 );
}

int ix_last_visible(DQUAD dq,DVEC x,int n) {
   int      i;
   double   fuzz = 1.0e-5,  xmax = dq.x2;

   for( i = n - 1; i >= 0;  i-- ) {
      if( fuzz_le(x[i],xmax,fuzz) ) {
         return( i );
      }
   }
   return( -1 );
}

int n_visible(DQUAD dq,DVEC x,int n) {
   int       ix_first = ix_first_visible(dq,x,n);
   int       ix_last  = ix_last_visible(dq,x,n);

   if( ix_first < 0 || ix_last < 0 ) {
      return( 0 );
   } else {
      return(ix_last - ix_first + 1);
   }
}

int n_visible_xy(DQUAD dq,DVEC x,DVEC y,int n) {
   int   i, nvis = 0;
   for( i = 0; i < n; i++ ) {
      if( (dq.x1 <= x[i]) && (x[i] <= dq.x2) &&
          (dq.y1 <= y[i]) && (y[i] <= dq.y2) ) nvis++;
   }
   return nvis;
}


#define   P wds[wt].pf.pxyf
#define   I P->xycurr->index
#define   X P->xycurr->data->x
#define   N P->xycurr->data->n
#define   Q P->dqc->dq

void get_ix_visible(WT wt,int *ix_min,int *ix_max) {
   if( is_active_wt(wt) ) {
      *ix_min = ix_first_visible(Q,X,N);
      *ix_max = ix_last_visible(Q,X,N);
   } else {
      *ix_min = -1;
      *ix_max = -1;
   }
}
#undef    P
#undef    I
#undef    X
#undef    N
#undef    Q

#define   PP     wds[wt].winf.pwf.pxywinf
void set_full_scale_mode(WT wt) {
   if( is_active_wt(wt) ) {
      set_must_redraw(wt);
      PP->full_scale_mode = ON;
   }
}

void unset_full_scale_mode(WT wt) {
   if( is_active_wt(wt) ) {
      dqc_from_current_xrange(wt);
      PP->full_scale_mode = OFF;
   }
}

void toggle_full_scale_mode(WT wt) {
   if( is_active_wt(wt) ) {
      if( PP->full_scale_mode ) {
         unset_full_scale_mode(wt);
      } else {
         set_full_scale_mode(wt);
      }
   }
}

/* Sleeps for the specified number of (floating point) seconds ... */

void fsleep(float secs) {
   struct   itimerval the_timer;

   if( secs == 0.0 ) return;

   the_timer.it_interval.tv_sec = 0;
   the_timer.it_interval.tv_usec = 0;
   the_timer.it_value.tv_sec = (int) secs;
   the_timer.it_value.tv_usec =  1.0e6 * (secs - (int) secs);
   signal(SIGALRM,_fsleep);
   setitimer(ITIMER_REAL,&the_timer,&the_timer);
   pause();
}

/* Helper function for fsleep() ... */

void _fsleep(void) {
   return;
}
#undef    PP  
