#include "project.h"

int  Dx_state = DX_UNDEF;

/* Modified versions of non ..._multi routines which switch color when  */
/* abscissa sense changes.                                              */

void plot_dot_multi(PXYDVEC pxydv,int col,int point_size) {
   rgb_color   color;
   int         j,   n;

   double      new_x,   old_x;
   double      new_dx;
   int         i,       revs = 0;

#define        FIRST_X  (pxydv->x)[0]
#define        NEW_X    (pxydv->x)[i]

   glPointSize(point_size);
   color = CT(col);
   glColor3d(color.r,color.g,color.b);
   n = pxydv->n;
   dx_state_reset();
   old_x  = FIRST_X;
   for( i = 0; i < n; i++ ) {
      glBegin(GL_POINTS);
      new_x = NEW_X;
      new_dx = new_x - old_x;
      if( dx_state_changed(new_dx) ) {
         col = inc_ring(col,Min_marker_color,Max_marker_color);
         color = CT(col);
         glColor3d(color.r,color.g,color.b);
      }
      glVertex2d((pxydv->x)[i],(pxydv->y)[i]); 
      old_x  = new_x;
      glEnd();
   }
}

void plot_line_multi(PXYDVEC pxydv,int col,int lw) {
   rgb_color   color;

   int         j,   n;
   double      new_x,   old_x;
   double      new_dx;
   int         i,       revs = 0;

#define        FIRST_X  (pxydv->x)[0]
#define        NEW_X    (pxydv->x)[i]

   color = CT(col);
   glColor3d(color.r,color.g,color.b);
   glLineWidth((float) lw);
   glEnable(GL_LINE_SMOOTH); 
   n = pxydv->n;
   dx_state_reset();
   old_x  = FIRST_X;
   glBegin(GL_LINE_STRIP);
   for( i = 0; i < n; i++ ) {
      new_x = NEW_X;
      new_dx = new_x - old_x;
      if( dx_state_changed(new_dx) ) {
         glEnd();
         col = inc_ring(col,Min_marker_color,Max_marker_color);
         color = CT(col);
         glColor3d(color.r,color.g,color.b);
         glBegin(GL_LINE_STRIP);
      }
      glVertex2d((pxydv->x)[i],(pxydv->y)[i]);
      old_x  = new_x;
   }
   glEnd();
}


void plot_poly_multi(PXYDVEC pxydv,int npoly,
                     double r,double xsc,double ysc,int col) {
   rgb_color   color; 
   double      dx[MAX_NPOLY],  dy[MAX_NPOLY];
   double      xc,   yc;
   int         j,   n;

   double      new_x,   old_x;
   double      new_dx;
   int         i,       revs = 0;

#define        FIRST_X  (pxydv->x)[0]
#define        NEW_X    (pxydv->x)[i]

   for( i = 0;  i < npoly; i++ ) {
      dx[i] = r * xsc * Cosphi[npoly][i];
      dy[i] = r * ysc * Sinphi[npoly][i];
   }

   color = CT(col);
   glColor3d(color.r,color.g,color.b);
   dx_state_reset();
   old_x  = FIRST_X;
   n = pxydv->n;
   for( i = 0; i < n; i++ ) {
      new_x = NEW_X;
      new_dx = new_x - old_x;
      if( dx_state_changed(new_dx) ) {
         col = inc_ring(col,Min_marker_color,Max_marker_color);
         color = CT(col);
         glColor3d(color.r,color.g,color.b);
      }
      xc = new_x;
      yc = (pxydv->y)[i];
      glBegin(GL_POLYGON);
      for( j = 0; j < npoly; j++ ) {
         glVertex2d(xc+dx[j],yc+dy[j]);
      }
      glEnd(); 
      old_x  = new_x;
   }
}

void plot_line_and_poly_multi(PXYDVEC pxydv,int npoly,double r,
                              double xsc,double ysc,int lcol,int lw,int pcol) {
   rgb_color   color; 
   double      dx[MAX_NPOLY],  dy[MAX_NPOLY];
   double      v2[2],  xc,   yc;
   int         j,   n;

   double      new_x,   old_x;
   double      new_dx;

   int         i,       revs = 0;
#define        FIRST_X  (pxydv->x)[0]
#define        NEW_X    (pxydv->x)[i]

   for( i = 0;  i < npoly; i++ ) {
      dx[i] = r * xsc * Cosphi[npoly][i];
      dy[i] = r * ysc * Sinphi[npoly][i];
   }

   dx_state_reset();
   old_x  = FIRST_X;
   color = CT(pcol);
   glColor3d(color.r,color.g,color.b);
   n = pxydv->n;
   for( i = 0; i < n; i++ ) {
      new_x = NEW_X;
      new_dx = new_x - old_x;
      if( dx_state_changed(new_dx) ) {
         pcol = inc_ring(pcol,Min_marker_color,Max_marker_color);
         color = CT(pcol);
         glColor3d(color.r,color.g,color.b);
      }
      xc = new_x;
      yc = (pxydv->y)[i];
      glBegin(GL_POLYGON);
      for( j = 0; j < npoly; j++ ) {
         glVertex2d(xc+dx[j],yc+dy[j]);
      }
      glEnd(); 
      old_x  = new_x;
   }

   color = CT(lcol);
   glColor3d(color.r,color.g,color.b);
   glLineWidth((float) lw);
   glEnable(GL_LINE_SMOOTH);
   dx_state_reset();
   old_x  = FIRST_X;
   revs   = 0; 
   glBegin(GL_LINE_STRIP);
   for( i = 0; i < n; i++ ) {
      new_x = NEW_X;
      new_dx = new_x - old_x;
      if( dx_state_changed(new_dx) ) {
         glEnd();
         lcol = inc_ring(lcol,Min_marker_color,Max_marker_color);
         color = CT(lcol);
         glColor3d(color.r,color.g,color.b);
         glBegin(GL_LINE_STRIP);
      }
      glVertex2d((pxydv->x)[i],(pxydv->y)[i]);
      old_x  = new_x;
   }
   glEnd();
}
#undef    NEW_X

#define COL  (Min_marker_color + ((col + (pxytag->tag[i] - Min_marker_color)) % Mod_marker_color))
#define PCOL (Min_marker_color + ((pcol + (pxytag->tag[i] - Min_marker_color)) % Mod_marker_color))
#define LCOL (Min_marker_color + ((lcol + (pxytag->tag[i] - Min_marker_color)) % Mod_marker_color))

void plot_dot_multi_tag(PXYDVEC pxydv, PXYTAGVEC   pxytag,
                        int col,int point_size) {
   rgb_color   color;
   int         i,   n;

   glPointSize(point_size);
   glBegin(GL_POINTS);
   n = pxydv->n;
   for( i = 0; i < n; i++ ) {
      color = CT(COL);
      glColor3d(color.r,color.g,color.b);
      glVertex2d((pxydv->x)[i],(pxydv->y)[i]); 
   }
   glEnd();
}

void plot_line_multi_tag(PXYDVEC pxydv,PXYTAGVEC pxytag,int col,int lw) {
   rgb_color   color;

   int         j,   n;
   double      new_x,   old_x;
   double      new_dx;
   int         i,       revs = 0;

#define        FIRST_X  (pxydv->x)[0]
#define        NEW_X    (pxydv->x)[i]

   glLineWidth((float) lw);
   glEnable(GL_LINE_SMOOTH); 
   n = pxydv->n;
   dx_state_reset();
   old_x  = FIRST_X;
   glBegin(GL_LINE_STRIP);
   for( i = 0; i < n; i++ ) {
      new_x = NEW_X;
      new_dx = new_x - old_x;
      if( dx_state_changed(new_dx) ) {
         glEnd();
         glBegin(GL_LINE_STRIP);
      }
      color = CT(COL);
      glColor3d(color.r,color.g,color.b);
      glVertex2d((pxydv->x)[i],(pxydv->y)[i]);
      old_x  = new_x;
   }
   glEnd();
}

void plot_poly_multi_tag(PXYDVEC pxydv,PXYTAGVEC pxytag,int npoly,double r,
               double xsc,double ysc,int col) {

   rgb_color   color;
   double      dx[MAX_NPOLY],  dy[MAX_NPOLY];
   double      v2[2],          xc,              yc;
   int         i,              j,               n;

   for( i = 0;  i < npoly; i++ ) {
      dx[i] = r * xsc * Cosphi[npoly][i];
      dy[i] = r * ysc * Sinphi[npoly][i];
   }

   n = pxydv->n;
   for( i = 0; i < n; i++ ) {
      xc = (pxydv->x)[i];
      yc = (pxydv->y)[i];
      color = CT(COL);
      glColor3d(color.r,color.g,color.b);
      glBegin(GL_POLYGON);
      for( j = 0; j < npoly; j++ ) {
         glVertex2d(xc+dx[j],yc+dy[j]);
      }
      glEnd();
   }
}

void plot_line_and_poly_multi_tag(PXYDVEC pxydv,PXYTAGVEC pxytag,int npoly,double r,
                              double xsc,double ysc,int lcol,int lw,int pcol) {
   rgb_color   color; 
   double      dx[MAX_NPOLY],  dy[MAX_NPOLY];
   double      v2[2],  xc,   yc;
   int         j,   n;

   double      new_x,   old_x;
   double      new_dx;

   int         i,       revs = 0;
#define        FIRST_X  (pxydv->x)[0]
#define        NEW_X    (pxydv->x)[i]

   for( i = 0;  i < npoly; i++ ) {
      dx[i] = r * xsc * Cosphi[npoly][i];
      dy[i] = r * ysc * Sinphi[npoly][i];
   }

   n = pxydv->n;
   for( i = 0; i < n; i++ ) {
      color = CT(PCOL);
      glColor3d(color.r,color.g,color.b);
      xc = (pxydv->x)[i];
      yc = (pxydv->y)[i]; 
      glBegin(GL_POLYGON);
      for( j = 0; j < npoly; j++ ) {
         glVertex2d(xc+dx[j],yc+dy[j]);
      }
      glEnd(); 
   }

   glLineWidth((float) lw);
   glEnable(GL_LINE_SMOOTH);
   dx_state_reset();
   old_x  = FIRST_X;
   revs   = 0; 
   glBegin(GL_LINE_STRIP);
   for( i = 0; i < n; i++ ) {
      new_x = NEW_X;
      new_dx = new_x - old_x;
      if( dx_state_changed(new_dx) ) {
         glEnd();
         glBegin(GL_LINE_STRIP);
      }
      color = CT(LCOL);
      glColor3d(color.r,color.g,color.b);
      glVertex2d((pxydv->x)[i],(pxydv->y)[i]);
      old_x  = new_x;
   }
   glEnd();
}
#undef    FIRST_X
#undef    NEW_X


#undef        COL  
#undef        PCOL 
#undef        LCOL 

void dx_state_reset(void) {
   
   Dx_state = DX_UNDEF;

}

int  dx_state_changed(double new_dx) {
   int     rc;

   switch( Dx_state ) {

   case DX_UNDEF:
      if(        new_dx > 0 ) {
         Dx_state = DX_POS;
         rc = 0;
      } else if( new_dx < 0 ) {
         Dx_state = DX_NEG;
         rc = 0;
      } else {
         rc = 0;
      }
   break;

   case DX_POS:
      if(        new_dx > 0 ) {
         rc = 0;
      } else if( new_dx < 0 ) {
         Dx_state = DX_UNDEF;
         rc = 1;
      } else {
         rc = 0;
      }
   break;

   case DX_NEG:
      if(        new_dx > 0 ) {
         Dx_state = DX_UNDEF;
         rc = 1;
      } else if( new_dx < 0 ) {
         rc = 0;
      } else {
         rc = 0;
      }
   break;

   }
   return rc;

}

/* Modified versions of non ..._rainbow routines which switch color */
/* every point.                                                     */
   
void plot_dot_rainbow(pxydv,col,winno,wt)
register
   PXYDVEC     pxydv;
   int         col,   winno;
   WT          wt; 
{
   double      v2[2];
   int         i,    j,   n;

   n = pxydv->n;
   color(col);
   for( i = 0; i < n; ) {
      bgnpoint();
      for( j = 0; (j < 256) && (i < n); ++i, ++j ) {
         v2[0] = (pxydv->x)[i];
         v2[1] = (pxydv->y)[i];
         v2d(v2);
         col = inc_ring(col,Min_marker_color,Max_marker_color);
         color(col);
      }
      endpoint();
   }
}

void plot_line_rainbow(pxydv,col,lw,winno,wt)
register
   PXYDVEC     pxydv;
   int         col,   lw,   winno;
   WT          wt; 
{
   double      v2[2];
   int         hits,  i,     n;

   n = pxydv->n;
   color(col);
   linewidth(lw);
   bgnline();
   for( i = 0, hits = 0; i < n; ) {
      v2[0] = (pxydv->x)[i];
      v2[1] = (pxydv->y)[i];
      v2d(v2);
      if( hits++ % 2 ) {
         endline();
         col = inc_ring(col,Min_marker_color,Max_marker_color);
         color(col);
         bgnline();
      } else {
         i++;
      }
   }
   endline();
}

void plot_poly_rainbow(pxydv,npoly,r,xsc,ysc,col,winno,wt)
register
   PXYDVEC     pxydv;
   double      r,      xsc,     ysc;
   int         npoly,  col,   winno;
   WT          wt; 
{
   double      dx[MAX_NPOLY],  dy[MAX_NPOLY];
   double      v2[2],  xc,   yc;
   int         i,    j,   n;

   for( i = 0;  i < npoly; i++ ) {
      dx[i] = r * xsc * Cosphi[npoly][i];
      dy[i] = r * ysc * Sinphi[npoly][i];
   }

   n = pxydv->n;
   color(col);
   for( i = 0; i < n; i++ ) {
      xc = (pxydv->x)[i];
      yc = (pxydv->y)[i];
      bgnpolygon();
      for( j = 0; j < npoly; j++ ) {
         v2[0] = xc + dx[j];
         v2[1] = yc + dy[j];
         v2d(v2);
      }
      endpolygon();
      col = inc_ring(col,Min_marker_color,Max_marker_color);
      color(col);
   }
}

void plot_line_and_poly_rainbow(pxydv,npoly,r,xsc,ysc,lcol,lw,pcol,winno,wt)
register
   PXYDVEC     pxydv;
   double      r,      xsc,     ysc;
   int         npoly,  lcol,    pcol,   winno,    lw;
   WT          wt; 
{
   double      dx[MAX_NPOLY],  dy[MAX_NPOLY];
   double      v2[2],  xc,   yc;
   int         hits,   i,    j,   n;

   for( i = 0;  i < npoly; i++ ) {
      dx[i] = r * xsc * Cosphi[npoly][i];
      dy[i] = r * ysc * Sinphi[npoly][i];
   }

   n = pxydv->n;
   color(pcol);
   for( i = 0; i < n; i++ ) {
      xc = (pxydv->x)[i];
      yc = (pxydv->y)[i];
      bgnpolygon();
      for( j = 0; j < npoly; j++ ) {
         v2[0] = xc + dx[j];
         v2[1] = yc + dy[j];
         v2d(v2);
      }
      endpolygon();
      pcol = inc_ring(pcol,Min_marker_color,Max_marker_color);
      color(pcol);
   }

   color(lcol);
   linewidth(lw);
   bgnline();
   for( i = 0, hits = 0; i < n; ) {
      v2[0] = (pxydv->x)[i];
      v2[1] = (pxydv->y)[i];
      v2d(v2);
      if( hits++ % 2 ) {
         endline();
         lcol = inc_ring(lcol,Min_marker_color,Max_marker_color);
         color(lcol);
         bgnline();
      } else {
         i++;
      }
   }
   endline();
}

void plot_poly_rainbow_multi_size(pxydv,npoly,r,xsc,ysc,col,winno,wt)
register
   PXYDVEC     pxydv;
   double      r,      xsc,     ysc;
   int         npoly,  col,   winno;
   WT          wt; 
{
   double      dx[MAX_NPOLY],  dy[MAX_NPOLY];
   double      v2[2],  xc,   yc;
   int         i,    j,   n;

   double      shrink_factor = 0.75, scale;

   for( i = 0;  i < npoly; i++ ) {
      dx[i] = r * xsc * Cosphi[npoly][i];
      dy[i] = r * ysc * Sinphi[npoly][i];
   }

   n = pxydv->n;
   color(col);
   scale = 1.0;
   for( i = 0; i < n; i++ ) {
      xc = (pxydv->x)[i];
      yc = (pxydv->y)[i];
      bgnpolygon();
      for( j = 0; j < npoly; j++ ) {
         v2[0] = xc + scale * dx[j];
         v2[1] = yc + scale * dy[j];
         v2d(v2);
      }
      endpolygon();
      col = inc_ring(col,Min_marker_color,Max_marker_color);
      color(col);
      scale *= shrink_factor;
   }
}

