#include <quad_util.h>

DPAIR make_dpair(double x,double y) {
   DPAIR p;
   
   p.x = x; p.y = y;
   return p;
}

int equal_dpairs(DPAIR dp1,DPAIR dp2) {
   return( dp1.x == dp2.x  && dp1.y == dp2.y );
}

void dump_dpair(DPAIR p) {
   fprintf(stderr,"%g %g\n",p.x,p.y);
}

DQUAD make_dquad(double x1,double y1,double x2,double y2) {
   DQUAD q;

   q.x1 = x1; q.y1 = y1; q.x2 = x2; q.y2 = y2;
   return q;
}

/* ... scalar * DQUAD ... */

DQUAD dqsm(DQUAD dq,double xscale,double yscale) {
   return make_dquad(xscale * dq.x1,yscale * dq.y1,
                     xscale * dq.x2,yscale * dq.y2);
}

/* ... (scalar * DQUAD) + (scalar * DQUAD) ... */

DQUAD dqsdqsa(DQUAD dq1,DQUAD dq2,double s1,double s2) {
   return make_dquad(s1 * dq1.x1 + s2 * dq2.x1,
                     s1 * dq1.y1 + s2 * dq2.y1,
                     s1 * dq1.x2 + s2 * dq2.x2,
                     s1 * dq1.y2 + s2 * dq2.y2);
}

DQUAD replace_degenerate_DQUAD(DQUAD dq) {
   if( dq.x1 == dq.x2 ) {
      dq.x2 += 1.0;
   } 
   if( dq.y1 == dq.y2 ) {
      if(        dq.y1 == 0.0 ) {
         dq.y2 = 1.0;
      } else if( dq.y1 >  0.0 ) {
         dq.y1 = 0.0;
      } else {
         dq.y2 = 0.0; 
      }
   }
   return dq;
}

int is_dp_in_dq(DPAIR dp,DQUAD dq) {
   return dq.x1 <= dp.x && dp.x <= dq.x2 &&
          dq.y1 <= dp.y && dp.y <= dq.y2;
}


IPAIR make_ipair(int x,int y) {
   IPAIR p;
   
   p.x = x; p.y = y;
   return p;
}

IQUAD make_iquad(int x1,int y1,int x2,int y2) {
   IQUAD q;

   q.x1 = x1; q.y1 = y1; q.x2 = x2; q.y2 = y2;
   return q;
}

void dump_iquad(IQUAD q) {
   fprintf(stderr,"%d %d %d %d.\n",q.x1,q.y1,q.x2,q.y2);
}

void dump_dquad(DQUAD q) {
   fprintf(stderr,"%g %g %g %g.\n",q.x1,q.y1,q.x2,q.y2);
}

int equal_iquads(IQUAD iq1,IQUAD iq2) {
   if( (iq1.x1 == iq2.x1)  &&  (iq1.x2 == iq2.x2)  &&
       (iq1.y1 == iq2.y1)  &&  (iq1.y2 == iq2.y2) ) {
      return(1);
   } else {
      return(0);
   }
}
  
int equal_dquads(DQUAD dq1,DQUAD dq2) {
   return (dq1.x1 == dq2.x1)  &&  (dq1.x2 == dq2.x2)  &&
          (dq1.y1 == dq2.y1)  &&  (dq1.y2 == dq2.y2);
}
  
double x_iq_to_dq(int i,IQUAD iq,DQUAD dq) {
   int            di; 

   if(         i < iq.x1 ) {
      return dq.x1;
   } else if ( i > iq.x2 ) {
      return dq.x2;
   } else if ( (di = iq.x2 - iq.x1) > 0 ) {
      return dq.x1 + ((double) (i - iq.x1)) / (double) di * 
             (dq.x2 - dq.x1);
   } else {
      return dq.x1;
   }
}

double y_iq_to_dq(int i,IQUAD iq,DQUAD dq) {
   int            di; 

   if(         i < iq.y1 ) {
      return dq.y1;
   } else if ( i > iq.y2 ) {
      return dq.y2;
   } else if ( (di = iq.y2 - iq.y1) > 0 ) {
      return dq.y1 + ((double) (i - iq.y1)) / (double) di * 
             (dq.y2 - dq.y1);
   } else {
      return dq.y1;
   }
}

double x_iq_to_dp(int i,IQUAD iq,DPAIR dp) {
   int            di; 

   if(         i < iq.x1 ) {
      return dp.x;
   } else if ( i > iq.x2 ) {
      return dp.y;
   } else if ( (di = iq.x2 - iq.x1) > 0 ) {
      return dp.x + ((double) (i - iq.x1)) / (double) di * 
             (dp.y - dp.x);
   } else {
      return dp.x;
   }
}

double y_iq_to_dp(int i,IQUAD iq,DPAIR dp) {
   int            di; 

   if(         i < iq.y1 ) {
      return dp.x;
   } else if ( i > iq.y2 ) {
      return dp.y;
   } else if ( (di = iq.y2 - iq.y1) > 0 ) {
      return dp.x + ((double) (i - iq.y1)) / (double) di * 
             (dp.y - dp.x);
   } else {
      return dp.x;
   }
}

int x_dq_to_iq(double x,DQUAD dq,IQUAD iq)
{
   double      dx;

   if(         x < dq.x1 ) {
      return iq.x1;
   } else if ( x > dq.x2 ) {
      return iq.x2;
   } else if ( (dx = dq.x2 - dq.x1) > 0 ) {
      return iq.x1 +  ((x - dq.x1) / dx * (iq.x2 - iq.x1));
   } else {
      return iq.x1;
   }
}

int y_dq_to_iq(double y,DQUAD dq,IQUAD iq)
{
   double      dy;

   if(         y < dq.y1 ) {
      return iq.y1;
   } else if ( y > dq.y2 ) {
      return iq.y2;
   } else if ( (dy = dq.y2 - dq.y1) > 0 ) {
      return iq.y1 +  ((y - dq.y1) / dy * (iq.y2 - iq.y1));
   } else {
      return iq.y1;
   }
}

int x_dp_to_iq(double x,DPAIR dp,IQUAD iq)
{
   double      dx;

   if(         x < dp.x ) {
      return iq.x1;
   } else if ( x > dp.y ) {
      return iq.x2;
   } else if ( (dx = dp.y - dp.x) > 0 ) {
      return iq.x1 +  ((x - dp.x) / dx * (iq.x2 - iq.x1));
   } else {
      return iq.x1;
   }
}

int y_dp_to_iq(double y,DPAIR dp,IQUAD iq)
{
   double      dy;

   if(         y < dp.x ) {
      return iq.y1;
   } else if ( y > dp.y ) {
      return iq.y2;
   } else if ( (dy = dp.y - dp.x) > 0 ) {
      return iq.y1 +  ((y - dp.x) / dy * (iq.y2 - iq.y1));
   } else {
      return iq.y1;
   }
}

IQUAD iq_from_rel_dq(IQUAD iq_in,DQUAD dq_in,DQUAD dq_out) {
   return(
      make_iquad(
         x_dq_to_iq(dq_out.x1,dq_in,iq_in),
         y_dq_to_iq(dq_out.y1,dq_in,iq_in),
         x_dq_to_iq(dq_out.x2,dq_in,iq_in),
         y_dq_to_iq(dq_out.y2,dq_in,iq_in)
      )
   );
}

IQUAD iq_from_frac(IQUAD iq,double xmn,double xmx,double ymn,double ymx) {
   if( 0.0 <= xmn && xmn < xmx && xmx <= 1.0  &&
       0.0 <= ymn && ymn < ymx && ymx <= 1.0 ) {
      return(iq_from_rel_dq(iq,make_unit_dq(),make_dquad(xmn,ymn,xmx,ymx)));
         
   } else {
      return(iq);
   }
}

DQUAD dq_trans_rel_unit_dq(DQUAD dq_in,DQUAD dq_trans) {
   return(
      make_dquad(
         dq_in.x1 + dq_trans.x1 * (dq_in.x2 - dq_in.x1),
         dq_in.y1 + dq_trans.y1 * (dq_in.y2 - dq_in.y1),
         dq_in.x1 + dq_trans.x2 * (dq_in.x2 - dq_in.x1),
         dq_in.y1 + dq_trans.y2 * (dq_in.y2 - dq_in.y1)
      )
   );
}

int dq1_contained_in_dq2(DQUAD dq1,DQUAD dq2) {
   return(
      dq1.x1 >= dq2.x1  &&  dq1.x2 <= dq2.x2 &&
      dq1.y1 >= dq2.y1  &&  dq1.y2 <= dq2.y2 
   );
}

DQUAD iq_to_dq(IQUAD iq) {
   return(make_dquad((double) iq.x1,(double) iq.y1,
                     (double) iq.x2,(double) iq.y2));
}

IQUAD shift_iq(IQUAD iq,int dx,int dy) {
   return(make_iquad(iq.x1+dx,iq.y1+dy,iq.x2+dx,iq.y2+dy));
}

DQUAD make_unit_dq(void) {
   return(make_dquad(0.0,0.0,1.0,1.0));
}

int in_iq(IQUAD iq,int x,int y) {
   return( (iq.x1 <= x) && (x <= iq.x2) && (iq.y1 <= y) && (y <= iq.y2));
}

void igeom_dump(char *s,igeom ig) {
   fprintf(stderr,"%s=[x=%d y=%d w=%d h=%d]\n",s,ig.x,ig.y,ig.w,ig.h);
}

int igeom_eq(igeom ig1,igeom ig2) {
   return (ig1.x == ig2.x && ig1.y == ig2.y && 
           ig1.w == ig2.w && ig1.h == ig2.h); 
}

int igeom_eq_size(igeom ig1,igeom ig2) {
   return (ig1.w == ig2.w && ig1.h == ig2.h); 
}

void igeom_cpy(igeom *pig1,igeom *pig2) {
   (*pig1).x = (*pig2).x;
   (*pig1).y = (*pig2).y;
   (*pig1).w = (*pig2).w;
   (*pig1).h = (*pig2).h;
}

igeom igeom_make(int x,int y,int w,int h) {
   igeom the;
   the.x = x;
   the.y = y;
   the.w = w;
   the.h = h;
   return the;
}
