#include "project.h"

/* Modified to consolidate applicator info. in single table entry  ... */

/* Applicators producing single group ... */

REDUCE_TYPE0      Reduce_table[] =       {
 
   {NULL, NULL, NULL},

   { pxydv_l2_norm,
     "l2 norm",
     "||%s||_2" },

   { pxydv_l1_norm,
     "l1 norm",
     "||%s||_1" },

   { pxydv_linf_norm ,
     "linf norm %l",
     "||%s||_inf" },

   { pxydv_log10_l2_norm,
     "log[l2 norm]",
     "log||%s||_2" },

   { pxydv_log10_l1_norm,
     "log[l1 norm]",
     "log||%s||_1" },

   { pxydv_log10_linf_norm ,
     "log[linf norm] %l",
     "log||%s||_inf" },

   { pxydv_reduce_ls_fit_slope,
     "l.s. fit slope %l",
     "l.s. fit slope[%s]" },

   { pxydv_trap_Int,
     "trap. Int",
     "Int[%s]" },

   { pxydv_trap_abs_Int,
     "trap. Int (abs)",
     "Int[|%s|]" },

   { pxydv_expectation_value_trap_Int,
     "trap. Int / (xmax - xmin) %l",
     "<%s>" },

   { pxydv_reduce_t,
     "t",
     "t" },

   { pxydv_reduce_N,
     "N %l",
     "N" },

   { pxydv_reduce_y1_t_to_x,
     "y[1]",
     "%s[inner]" },

   { pxydv_reduce_yj_t_to_x,
     "y[j]",
     "%s[j]" },

   { pxydv_reduce_yN_t_to_x,
     "y[N] %l",
     "%s[outer]" },

   { pxydv_reduce_xN_t_to_x,
     "x[N] %l",
     "%s[outer x]" },

   { pxydv_reduce_qfitres,
     "qfit res. %l",
     "qfit res.[%s]" },

   { pxydv_reduce_ynminusy1,
     "y[n] - y[1] %l",
     "%s([n] - [1])" },

   { pxydv_reduce_average,
     "average[y]",
     "avg[%s]" },

   { pxydv_reduce_ymax,
     "max[y]",
     "max[%s]" },

   { pxydv_reduce_ymin,
     "min[y]",
     "min[%s]" },

   { pxydv_reduce_yamax,
     "absmax[y]",
     "absmax[%s]" },

   { pxydv_reduce_yamin,
     "absmin[y]",
     "absmin[%s]" },

   { pxydv_reduce_xymax,
     "x[max[y]]",
     "x[max[%s]]" },

   { pxydv_reduce_xymin,
     "x[min[y]]",
     "x[min[%s]]" },

   { pxydv_reduce_xyamax,
     "x[absmax[y]]",
     "x[absmax[%s]]" },

   { pxydv_reduce_xyamin,
     "x[absmin[y]] %l",
     "x[absmin[%s]]" },

   { pxydv_reduce_first_x0,
     "first 0-crossing",
     "0-crossing[%s]" }

                                         };
              
int               N_Reduce_table =
                     sizeof(Reduce_table) / sizeof(REDUCE_TYPE0);

#define   X       pnew->data->x
#define   Y       pnew->data->y
#define   T       pnew->time
#define   N       pnew->data->n

#define   XOLD    pold->data->x
#define   YOLD    pold->data->y
#define   TOLD    pold->time
#define   NOLD    pold->data->n  

PXYDVBIND pxydv_l2_norm(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvnrm2_(pold->data->y,&(pold->data->n));
   }
   if( i == 1 ) T = Y[0];
   return pnew;
}

PXYDVBIND pxydv_l1_norm(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvnrm1_(pold->data->y,&(pold->data->n));
   }
   if( i == 1 ) T = Y[0];
   return pnew;
}

PXYDVBIND pxydv_linf_norm(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvlinf_(pold->data->y,&(pold->data->n));
   }
   if( i == 1 ) T = Y[0];
   return pnew;
}

PXYDVBIND pxydv_log10_l2_norm(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvnrm2_(pold->data->y,&(pold->data->n));
      Y[i] = Y[i] == 0.0 ? 0.0 : log10(Y[i]);
   }
   if( i == 1 ) T = Y[0];
   return pnew;
}

PXYDVBIND pxydv_log10_l1_norm(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvnrm1_(pold->data->y,&(pold->data->n));
      Y[i] = Y[i] == 0.0 ? 0.0 : log10(Y[i]);
   }
   if( i == 1 ) T = Y[0];
   return pnew;
}

PXYDVBIND pxydv_log10_linf_norm(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvlinf_(pold->data->y,&(pold->data->n));
      Y[i] = Y[i] == 0.0 ? 0.0 : log10(Y[i]);
   }
   if( i == 1 ) T = Y[0];
   return pnew;
}

PXYDVBIND pxydv_trap_Int(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvtrap_(pold->data->x,pold->data->y,&(pold->data->n));
   }
   return pnew;
}

PXYDVBIND pxydv_expectation_value_trap_Int(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = dvtrap_(pold->data->x,pold->data->y,&(pold->data->n)) /
             ((pold->data->x)[pold->data->n -1] -
              (pold->data->x)[0]);
   }
   return pnew;
}

PXYDVBIND pxydv_trap_abs_Int(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   DVEC TY;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      if( (TY = make_DVEC(pold->data->n)) != NULL ) {
         dvfapl_(pold->data->y,TY,D_ABS,&(pold->data->n));
         X[i] = pold->time;
         Y[i] = dvtrap_(pold->data->x,TY,&(pold->data->n));
         free_DVEC(TY);
      } else {
         fprintf(stderr,"pxydv_trap_abs_Int: make_DVEC(%d) fails.\n",
                 pold->data->n);
         Y[i] = 0.0;
      }
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_t(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      Y[i] = pold->time;
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_y1(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = pold->data->y[0];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_yN(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = pold->data->y[pold->data->n - 1];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_y1_t_to_x(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = pold->data->y[0];
   }
   return pnew;
}


PXYDVBIND pxydv_reduce_yj_t_to_x(PXYDVBIND pnew, PXYDVBIND pold) {
   PXYDVBIND  psave = pold;
   int        i;
   int        j,  default_j = 0;
   double     dj;
   
Prompt:
   sprintf(O_buf,"0-based grid index (%d)",default_j);
   GET_DOUBLE(dj,(double) default_j,O_buf);
   j = dj; 
   if( j < 0 ) goto Prompt;
   for( i = 0; pold != NULL; pold = pold->next ) {
      if( j > (NOLD - 1) ) {
         scope_mess_and_bell("Index out of range for one or more groups",1);
         goto Prompt;
      }
   }
   
   for( i = 0, pold = psave; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = pold->data->y[j];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_yN_t_to_x(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = pold->data->y[pold->data->n - 1];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_xN_t_to_x(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = pold->data->x[pold->data->n - 1];
   }
   return pnew;
}


PXYDVBIND pxydv_reduce_N(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = pold->time;
      Y[i] = pold->data->n;
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_qfitres(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = dvqfitres_(YOLD,&NOLD);
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_ynminusy1(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = YOLD[NOLD-1] - YOLD[0];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_average(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = dvsum(YOLD,NOLD) / NOLD;
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_ymax(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = dvmax(YOLD,NOLD);
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_ymin(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = dvmin(YOLD,NOLD);
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_yamax(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = dvabsmax(YOLD,NOLD);
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_yamin(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = dvabsmin(YOLD,NOLD);
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_xymax(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = XOLD[ixdvmax(YOLD,NOLD)];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_xymin(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = XOLD[ixdvmin(YOLD,NOLD)];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_ls_fit_slope(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;

   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      DVEC       TV;
      double     f77_a,   f77_b;
      TV = make_DVEC(NOLD);
      clls_(XOLD,YOLD,TV,&NOLD,&f77_a,&f77_b);
      X[i] = TOLD;
      Y[i] = f77_b;
      if( TV ) free(TV);
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_xyamax(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = XOLD[ixdvabsmax(YOLD,NOLD)];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_xyamin(PXYDVBIND pnew, PXYDVBIND pold) {
   int  i;
   
   for( i = 0; pold != NULL; pold = pold->next, i++ ) {
      X[i] = TOLD;
      Y[i] = XOLD[ixdvabsmin(YOLD,NOLD)];
   }
   return pnew;
}

PXYDVBIND pxydv_reduce_first_x0(PXYDVBIND pnew, PXYDVBIND pold) {
   int         i,         j,      jx,     nnew,     maxnnew,   f77_one = 1;
   double      x0;
   PXYDVBIND   psave = pold;

   int         ltrace = OFF;

/* This routine can change the length of X, Y vectors ... */

   for( maxnnew = 0, nnew = 0, i = 0, pold = psave; pold != NULL; pold = pold->next, i++ ) {
      fdvx0_(XOLD,YOLD,&NOLD,&f77_one,&jx);
      if( jx ) nnew++;
      maxnnew++;
   }
   if( ltrace ) {
      fprintf(stderr,"pxydv_reduce_first_x0: %d groups\n",maxnnew);
      fprintf(stderr,"pxydv_reduce_first_x0: %d groups with 0-crossings detected\n",nnew);
   }
   if( nnew ) {
      if( nnew != maxnnew ) {
         free(X); X = make_DVEC(nnew);
         free(Y); Y = make_DVEC(nnew);
         N = nnew;
      } 
      for(  j = 0, i = 0, pold = psave; pold != NULL; pold = pold->next, i++ ) {
         x0 = fdvx0_(XOLD,YOLD,&NOLD,&f77_one,&jx);
         if( jx ) {
            X[j] = TOLD;
            Y[j] = x0;
            j++;
         }
      }
   } else {
      scope_mess_and_bell("No 0-crossings detected",1);
      free(X); X = make_DVEC(1);
      free(Y); Y = make_DVEC(1);
      N = 1;
      X[0] = 0.0;
      Y[0] = 0.0;
   }
   return pnew;
}


#undef    X
#undef    Y
#undef    T

#undef    XOLD
#undef    YOLD
#undef    TOLD
#undef    NOLD
