#include <sysincludes.h>
#include "constants.h"
#include "wt.h"
#include "win_types.h"
#include "vs.h"
#include "xy.h"
#include "xywinf.h" 
#include "scope.h"
#include "plot.h"
#include "misc.h"
#include "sw.h"
#include "lal.h"
#include "util.h"
#include "quad_util.h"

WT         scope_wt;
int        scope_sgi_winno;

WT         First_wt;
WT         Last_wt;

int        Scope_wid = DEFAULT_SCOPE_WID;
int        Scope_hit = DEFAULT_SCOPE_HIT;
int        scope_xlc,  scope_xuc;
int        scope_ylc,  scope_yuc;

int        panel_lev;
int        panel_xlc[MAX_PANEL_LEV];
int        panel_xuc[MAX_PANEL_LEV];
int        panel_ylc[MAX_PANEL_LEV];
int        panel_yuc[MAX_PANEL_LEV];
int        panel_wid;
int        panel_hit;

int        scope_mode;
int        current_border_color = A_BORDER_COLOR;
int        border_width = BORDER_WIDTH;

char       scope_name[] = "vs";

void compute_panel(int call) {
   int   nw, pl, i;

   if(         scope_mode == PANEL_MODE ) {
      nw = n_windows(); 
   } else if(  scope_mode == SINGLE_MODE ) {
      nw = n_windows_1(); 
   } else {
      err1("compute_panel","unexpected scope mode");
   }
   if(        nw < 2 ) {
      panel_lev = 0;
   } else if( nw < 5 ) {
      panel_lev = 1;
   } else if( nw < 10 ) {
      panel_lev = 2;
   } else if( nw < 17 ) {
      panel_lev = 3;
   } else if( nw < 26 ) {
      panel_lev = 4;
   } else {
      printf("compute_panel: Too many windows %d.\n",nw);
   }
   pl = panel_lev;
   if( call == INITIAL_CALL ) {
      set_init_scope_corners();
   } else {
      scope_xlc = iquad_window_sense(scope_wt).x1;
      scope_ylc = iquad_window_sense(scope_wt).y1 - SCOPE_HEAD_SPACE - Scope_hit;
   }
   panel_wid = (int) (0.5 + (float) (Scope_wid - pl * GRID_WH) / (pl + 1));
   panel_hit = (int) (0.5 + (float) (Scope_hit - pl * GRID_WH) / (pl + 1));

   panel_xlc[0] = scope_xlc;
   for( i = 0; i <= panel_lev; i++ ) {
      if( i > 0 ) {
         panel_xlc[i] = panel_xuc[i-1] + GRID_WH;
      }
      panel_xuc[i] = panel_xlc[i] + (panel_wid - 1);
   }
   scope_xuc = panel_xuc[panel_lev];

   panel_ylc[pl] = scope_ylc;
   for( i = pl; i >= 0; i-- ) {
      if( i < pl ) {
         panel_ylc[i] = panel_yuc[i+1] + GRID_WH;
      }
      panel_yuc[i] = panel_ylc[i] + (panel_hit - 1);
   }
   scope_yuc = panel_yuc[0];
}

void set_init_scope_corners(void) {
   long     screen_x = fl_scrw - 1;
   long     screen_y = fl_scrh - 1;
   int      ltrace = ON;

   if( ltrace ) {
      fprintf(stderr,"Screen is %d x %d pixels.\n",screen_x + 1,screen_y + 1);
   }

   if(          (screen_x == XMAXSCREEN) && (screen_y == YMAXSCREEN) ) {
      scope_xlc = BIG_XLC;
      scope_ylc = BIG_YLC;
   } else if(   (screen_x == XMAXMEDIUM) && (screen_y == YMAXMEDIUM) ) {
      scope_xlc = SMALL_XLC;
      scope_ylc = SMALL_YLC;
   } else {
      fprintf(stderr,"set_init_scope_corners: Unrecognized screen size.\n");
      fprintf(stderr,"set_init_scope_corners: Will use medium settings.\n");
      scope_xlc = SMALL_XLC;
      scope_ylc = SMALL_YLC;
   }
}

/* Scope header now opened as regular xy window, should always return 
   0 window tag.  This routine also defines pop--up menu. */

WT scope_open()
{
   double  xmin = 0.0, xmax = 1.0, ymin = 0.0, ymax = 1.0;

   WT      wt;

	static  char R[] = "scope_open";
	int     ltrace = 1;


	if( ltrace ) fprintf(stderr,"%S: In routine\n",R);
   wt = S_wt = last_S_wt = SCOPE_WT;
   scope_mode = PANEL_MODE;

   compute_panel(INITIAL_CALL);

   prefposition(scope_xlc,scope_xuc,
                scope_yuc + SCOPE_HEAD_SPACE,
                scope_yuc + SCOPE_HEAD_SPACE + SCOPE_HEAD_HIT);
   foreground(); 
   scope_sgi_winno = winopen(scope_name);    
   doublebuffer();  gconfig(); 
   queue_devices();
   winset(scope_sgi_winno); color(SCOPE_COLOR); clear(); 
   swapbuffers(); 

   wds[wt].sgi_winno = scope_sgi_winno;
   wds[wt].namelen = strlen(scope_name);
   wds[wt].name = (char *) malloc(wds[wt].namelen);
   wds[wt].wd = strdup(getenv("HOME"));
   strcpy(wds[wt].name,scope_name);
   wds[wt].status = WINDOW_ACTIVE;
   wds[wt].pf.pxyf = make_pxyf(make_dquad(xmin,ymin,xmax,ymax),
                               NO_SCALE_AS_LOADED,0,0);
   dqc_from_dqb(wt);
   wds[wt].winf.d_buff = D_BUFF_ON;
   wds[wt].winf.pwf.pxywinf = make_pxywinf();

   window_sense(wt);

   main_menu = make_main_menu();    
 
/* ... have to initialize font before call to do_csd_wt ... */

   if( V_prompt_fh_name ) {
      fh_prompt = fmscalefont(fmfindfont(V_prompt_fh_name),FH_PROMPT_DEF_SIZE);
   } else {
      fprintf(stderr,"scope_open: Warning -- <V_prompt_fh_name> not set.  ");
      fprintf(stderr,"Using  <%s>\n",VS_FONTNAME);
      fh_prompt = fmscalefont(fmfindfont(VS_FONTNAME),FH_PROMPT_DEF_SIZE);
   }

/* Initialize working directory from $HOME/sd, if it exists ... */

   if( getenv("HOME") ) {
      sprintf(Sd_name,"%s/sd",getenv("HOME"));
      do_csd_wt(wt,ON);
   }

   if( mouse_is_near_wt(wt,0.2,3.0) ) {
      centre_mouse_in_wt(wt);
   }

   return(wt);
}

int  srx(abs_x)
   int   abs_x;
{
   return(abs_x - scope_xlc);
}

int  sry(abs_y)
   int   abs_y;
{
   return(abs_y - scope_ylc);
}

/* Opens window in panel in frame. Returns SGI window tag. */
/* Backbuffer is BLACKed but not bordered, then swapped. */

int open_in_panel(name)
   char    *name;
{   
   int     sgi_winno, ix,   iy,   nw; 

/* printf("open_in_panel: before n_windows call.\n");  */
   if( scope_mode == SINGLE_MODE ) {
      (void) scope_to_PANEL_MODE_no_repanel();
   }
   nw = n_windows();
/* printf("open_in_panel: after n_windows call.\n");  */
   repanel(nw - 1);

   iy = (nw - 1) / (panel_lev + 1);
   ix = (nw - 1) % (panel_lev + 1);
   noborder();
   prefposition(panel_xlc[ix],panel_xuc[ix],     
                panel_ylc[iy],panel_yuc[iy]);
   sgi_winno = winopen(name);
   doublebuffer(); gconfig();
   color(BLACK);
   clear(); 
   swapbuffers();

   set_First_and_Last_wt();

   return(sgi_winno);
}

void repanel(int nw) {
   xvs_repanel(nw);
}

/* Repositions first nw windows in panel. */

void repanel_old(nw)
   int              nw;
{
   PWTBIND          wtp;
   WT               wt;
   int              iw,    ix,   iy;

   compute_panel(NOT_INITIAL_CALL);
/* wtp = tail_wtal(wtal);  */
   wtp = tail_wtal(lal_wtal());    
   redraw_scope();
   if( wtp != NULL ) {
      iw = 1;
      while( wtp != NULL && iw <= nw ) {
         wt = wtp->wt;
         iy = (iw - 1) / (panel_lev + 1);
         ix = (iw - 1) % (panel_lev + 1);
         winset(wds[wt].sgi_winno);
         winposition(panel_xlc[ix],panel_xuc[ix],panel_ylc[iy],panel_yuc[iy]);
         reshapeviewport();
                 
         window_sense(wt);    
         redraw_window_in_panel(wt);
         
         ++iw; wtp = wtp->prev;
      }
   }
   apply_all_wt_0((PFV_I) set_must_redraw);
   set_First_and_Last_wt();
}

/* Combines open and repanelling operations ... */

int open_and_repanel(wt_to_open,name)
   WT               wt_to_open;
   char            *name;
{
   PWTBIND          wtp;
   WT               wt;
   int              nw,  iw,    ix,   iy;
   int              sgi_winno;

   if( scope_mode == SINGLE_MODE ) {
      (void) scope_to_PANEL_MODE_no_repanel();
   }

   nw = n_windows();
   compute_panel(NOT_INITIAL_CALL);
   wtp = tail_wtal(lal_wtal());    
   redraw_scope();
   if( wtp != NULL ) {
      iw = 1;
      while( wtp != NULL && iw <= nw ) {
         wt = wtp->wt;
         iy = (iw - 1) / (panel_lev + 1);
         ix = (iw - 1) % (panel_lev + 1);
         if( wt == wt_to_open ) {
            noborder();
            prefposition(panel_xlc[ix],panel_xuc[ix],
                         panel_ylc[iy],panel_yuc[iy]);
            sgi_winno = winopen(name);
            wds[wt].sgi_winno = sgi_winno;
            doublebuffer(); 
            gconfig();
            color(BLACK);
            clear();
            swapbuffers();
         }
         winset(wds[wt].sgi_winno);
         winposition(panel_xlc[ix],panel_xuc[ix],panel_ylc[iy],panel_yuc[iy]);
         reshapeviewport();
                 
         window_sense(wt);    
         if( wt != wt_to_open ) redraw_window_in_panel(wt);

         ++iw; wtp = wtp->prev;
      }
      set_First_and_Last_wt();
      apply_all_wt_0((PFV_I) set_must_redraw);
      return wt_to_open;
   } else {
      return -1;
   }
}

void redraw_window_in_panel(WT wt) {
   if( is_active_wt(wt) ) {
      fl_raise_form(Xvsf->main);
      display_window_1(wt);    
   }
}

void redraw_window_in_panel_old(wt)
   WT     wt;
{
   winset(wds[wt].sgi_winno);
   winpop();
   if( wds[wt].pf.pxyf && wds[wt].winf.pwf.pxywinf ) {    
      display_window_1(wt);    
   }    
}


void old_border_window(wt) 
   WT     wt;
{
   float  fuzz = 0.0001, dx, dy;

   int    ltrace = OFF;

   if( ltrace ) {
      printf("border_window: wt: %d.\n",wt);
   }

   color(wds[wt].winf.pwf.pxywinf->border_color);
   linewidth(border_width);
#define CO (float) (wds[wt].pf.pxyf->dqc)->dq
   dx = fuzz * (CO.x2 - CO.x1);
   dy = fuzz * (CO.y2 - CO.y1);
   rect(CO.x1+dx,CO.y1+dy,CO.x2-dx,CO.y2-dy);
#undef  CO
}

void new_border_window(WT wt) 
{
   int    w;

   int    ltrace = OFF;

   if( ltrace ) {
      printf("border_window: wt: %d.\n",wt);
   }

   equate_screen_and_world_coordinates(wds[wt].sgi_winno);
#define IQ wds[wt].winf.wiq
   outline_IQUAD(   
      make_iquad(IQ.x1,IQ.y1,IQ.x2+1,IQ.y2),
      wds[wt].winf.pwf.pxywinf->border_color
                );
   for( w = 0; w < border_width; w++ ) {
      outline_IQUAD(   
         make_iquad(IQ.x1+w,IQ.y1+w,IQ.x2-w-1,IQ.y2-w-1),
         wds[wt].winf.pwf.pxywinf->border_color
                   );
   }
#undef  IQ 

   window_ortho2(wt);
}

void outline_IQUAD(IQUAD iq,int col) {
   double    v2[2];

   color(col);
   linewidth(1);
   bgnline();
   v2[0] = iq.x1;  v2[1] = iq.y1;   v2d(v2);
   v2[0] = iq.x1;  v2[1] = iq.y2;   v2d(v2);
   v2[0] = iq.x2;  v2[1] = iq.y2;   v2d(v2);
   v2[0] = iq.x2;  v2[1] = iq.y1;   v2d(v2);
   v2[0] = iq.x1;  v2[1] = iq.y1;   v2d(v2);
   endline();

}

void redraw_scope()
{
   winset(scope_sgi_winno); 
   winpop();
   reshapeviewport();
   color(SCOPE_COLOR); clear();
   swapbuffers();
}

int scope_to_SINGLE_MODE(WT wt) {
   static char R[] = "scope_to_SINGLE_MODE";
   int    ltrace = Ltrace;

   int    iw;

   assert(scope_mode == PANEL_MODE);
   if( is_active_wt(wt) && wds[wt].pf.pxyf->lxy > 0 ) {
      Pnl_prev = Pnl;
      Pnl = 0;
      if( Pnl_prev >= 0 ) {
         for( iw = 0; iw <= Pnl_prev; iw++ ) {
            (Xvsf->w)[Pnl_prev][iw].wt = UNDEF;
         }
         if( Pnl_prev != Pnl ) {
            if( ltrace ) fprintf(stderr,"%s: Hiding panel %d\n",R,Pnl_prev);
            xvs_hide_panel(Pnl_prev);
         }
      }
      if( Pnl_prev != Pnl ) {
         if( ltrace ) fprintf(stderr,"%s: Displaying panel %d\n",R,Pnl);
         xvs_show_panel(Pnl);
      }

      lal_push(make_WTBIND(wt,NULL,NULL));    
      scope_mode = SINGLE_MODE;
      (void) vs_repanel();
      centre_mouse_in_wt(wt);
      return 1;
   } else {
      return 0;
   }
}

int scope_to_PANEL_MODE(void) {
   static char R[] = "scope_to_SINGLE_MODE";
   int    ltrace = Ltrace;

   int    iw, nw;

   int    single_wt;

   assert(scope_mode == SINGLE_MODE);
   assert(Pnl == 0);
   single_wt = lal_wtal()->wt;
   nw = n_windows();
   if( ltrace ) {
      fprintf(stderr,"%s: n_windows()=%d\n",R,nw);
   }
   if( nw > Max_panel ) {
      fprintf(stderr,"%s: FATAL Number of windows=%d exceeds %d\n",
         R,nw,Max_panel);
   }
   Pnl_prev = 0;
   (Xvsf->w)[Pnl_prev][0].wt = UNDEF;
   Pnl = nw - 1;

   if( Pnl_prev != Pnl ) {
      if( ltrace ) fprintf(stderr,"%s: Hiding panel %d\n",R,Pnl_prev);
      xvs_hide_panel(Pnl_prev);
      if( ltrace ) fprintf(stderr,"%s: Displaying panel %d\n",R,Pnl);
      xvs_show_panel(Pnl);
   }

   lal_pop();
   scope_mode = PANEL_MODE;
   vs_repanel();
   centre_mouse_in_wt(single_wt);
   return(0);
}

int scope_to_PANEL_MODE_no_repanel() {
   int      ltrace = Ltrace;
   int      rc;

   if( scope_mode == SINGLE_MODE ) {
      lal_pop();
      scope_mode = PANEL_MODE;
   }
   return(0);
}

int mouse_in_ser_domain(void) {
   FL_Coord x, y;
   unsigned int keymask;

   (void) fl_get_mouse(&x,&y,&keymask);
   return ((Xvsf->main->x) <= x)  && (x <= (Xvsf->main->x)+(Xvsf->main->w)-1) &&
          ((Xvsf->main->y) <= y)  && (y <= (Xvsf->main->y)+(Xvsf->main->h)-1);
}

/* Sets global vbls.  First_wt and Last_wt to tags of first and last */
/* currently visible windows ...                                     */

void set_First_and_Last_wt(void) {
   PWTBIND    wtp;
   WT         wt;
   int        iw,    nw;
   int        ltrace = OFF;

   nw = n_windows_1();
   if( ltrace ) {
      fprintf(stderr,"set_First_and_Last_wt: nw: %d.\n",nw);
   }
   for( wtp = tail_wtal(lal_wtal()), iw = 1; wtp && iw <= nw; 
        ++iw, wtp = wtp->prev ) {
      wt = wtp->wt;
      if(         iw == 1 ) {
         First_wt = wt;
      } 
      if ( iw == nw ) {
         Last_wt  = wt;
      }
   }
   if( ltrace ) {
      printf("set_First_and_Last_wt: First wt: %d  Last wt: %d.\n",
             First_wt,Last_wt);
      lal_dump();
   }
}
