#include <sysincludes.h>
#include "wt.h"
#include "win_types.h"
#include "vs.h"
#include "xywinf.h"
#include "plot.h"
#include "scope.h"
#include "lal.h"
#include "crit.h"
#include "sw.h"
#include "quad_util.h"

#define    DEFAULT_STATUS_HIT   160 
#define    DEFAULT_STATUS_WID   DEFAULT_SCOPE_WID
#define    STATUS_HEAD_SPACE    2 * GRID_WH
#define    STATUS_BACK_COLOR    BLACK 
#define    STATUS_BORDER_COLOR  MAGENTA
#define    STATUS_TEXT_COLOR    GREEN 

int        Status_wid   =       DEFAULT_STATUS_WID;
int        Status_hit   =       DEFAULT_STATUS_HIT;

#define    STAT_XMIN            0
#define    STAT_XMAX            DEFAULT_STATUS_WID
#define    STAT_YMIN            0
#define    STAT_YMAX            DEFAULT_STATUS_HIT

#define    SW_NS                6

double                  sw_xc[SW_NS]  = { 0.0500,  0.6250,  0.0500,  0.6250,
                                          0.0500,  0.0500 };
double                  sw_yc[SW_NS]  = { 0.7500,  0.7500,  0.5500,  0.5500,
                                          0.3500,  0.1500 };
float                   sw_fs[SW_NS]  = { 20.0,    20.0,    20.0,    20.0,
                                          20.0,    20.0   };
/*
char                   *sw_fn[SW_NS]  = { "Helvetica-Bold",
                                          "Helvetica-Bold",
                                          "Helvetica-Bold",
                                          "Helvetica-Bold",
                                          "Helvetica-Bold",
                                          "Helvetica-Bold" 
                                        };
*/
int                     sw_col[SW_NS] = { GREEN,   GREEN,   GREEN,   GREEN, 
                                          CYAN,    CYAN };
int                     sw_ixc[SW_NS];
int                     sw_iyc[SW_NS];
fmfonthandle            sw_fh1;
fmfonthandle            sw_fh40;
fmfonthandle            sw_fh60;
fmfonthandle            sw_fh100;
fmfonthandle            sw_fh38;
fmfonthandle            sw_fh25;
fmfonthandle            sw_fh[SW_NS];

fmfonthandle            fhTB05;
fmfonthandle            fhTB10;
fmfonthandle            fhTB15;
fmfonthandle            fhTB25;
fmfonthandle            fhTB40;


#define                 PP     wds[wt].winf.pwf.pxywinf
#define                 FP     wds[wt].pf.pxyf
#define                 DP     FP->xycurr
#define                 DATA   FP->xycurr->data
#define                 X      DATA->x
#define                 Y      DATA->y
#define                 N      DATA->n
#define                 ICO    wds[wt].winf.wiq
#define                 CO     (FP->dqc)->dq
#define                 GL_CO  FP->limits
#define                 W      (float) (STAT_XMAX - STAT_XMIN)
#define                 H      (float) (STAT_YMAX - STAT_YMIN)

void sw_init()
{
   char    *sw_fontname = NULL;
   char    *sw_fontname1 = NULL;
   int      i;
   int      ltrace = OFF;

   sw_fontname  = vs_default_fontname();
   sw_fontname1 = vs_default_fontname1();
   if( ltrace ) {
      fprintf(stderr,"sw_init: sw_fontname:  <%s>\n",sw_fontname);
      fprintf(stderr,"sw_init: sw_fontname1: <%s>\n",sw_fontname1);
   }
   for( i = 0; i < SW_NS; i++ ) {
      if( (sw_fh1 = fmfindfont(sw_fontname)) == 0 ) {
         printf("sw_init: could not find font %s.\n",sw_fontname);
         exit(-1);
      }
      sw_fh[i] = fmscalefont(sw_fh1,sw_fs[i]);
      sw_ixc[i] = STAT_XMIN + (int) (sw_xc[i] * W );
      sw_iyc[i] = STAT_YMIN + (int) (sw_yc[i] * H );
/*    printf("sw_init: line %d  Corner: %d  %d.\n",i,sw_ixc[i],sw_iyc[i]);  */
   }

   sw_fh40 = fmscalefont(sw_fh1,40.0);
   sw_fh60 = fmscalefont(sw_fh1,60.0);
   sw_fh100 = fmscalefont(sw_fh1,100.0);
   sw_fh38 = fmscalefont(sw_fh1,38.0);
   sw_fh25 = fmscalefont(sw_fh1,25.0);

   fhTB05 = fmscalefont(sw_fh1,5.0);
   fhTB10 = fmscalefont(sw_fh1,10.0);
   fhTB15 = fmscalefont(sw_fh1,15.0);
   fhTB25 = fmscalefont(sw_fh1,25.0);
   fhTB40 = fmscalefont(sw_fh1,40.0);

   for( i = MIN_VIDEO_STATE; i <= MAX_VIDEO_STATE; i++ ) {
      V_Heads_up_fh[i] = 
         make_fh(V_heads_up_fh_name[i],V_Heads_up_fh_size[i]);
      V_Heads_up_lim_fh[i] = 
         make_fh(V_heads_up_lim_fh_name[i],V_Heads_up_lim_fh_size[i]);
   }
   free( sw_fontname );
   free( sw_fontname1 );
}

fmfonthandle make_fh(char *name, float size) {
   return fmscalefont(fmfindfont(name),size);
}


int get_status_window_flag(wt)
   WT            wt;
{
   int           ltrace = OFF;

   if( is_active_wt(wt) ) {
      return PP->status_window_flag;
   } else {
      if( ltrace ) {
         printf("get_status_window: window %d not active.\n",wt);
      }
      return -1;
   }
}

void set_status_window_flag(wt)
   WT            wt;
{
   if( is_active_wt(wt) ) {
      PP->status_window_flag = ON;
   } 
}

void unset_status_window_flag(wt)
   WT            wt;
{
   if( is_active_wt(wt) ) {
      PP->status_window_flag = OFF;
   }
}

int open_status_window(wt)
   WT            wt;
{
static
   char          status_name[] = "status";
   int           xlc,              ylc,
                 winno; 

   if( is_active_wt(wt) && (scope_mode == SINGLE_MODE) ) {
      xlc = scope_xlc;
      ylc = scope_ylc - (Status_hit + STATUS_HEAD_SPACE);
      noborder();
      prefposition(xlc,xlc + Status_wid,ylc,ylc + Status_hit );
      winno = winopen(status_name);
      winset(winno);
      doublebuffer(); gconfig();
   
      PP->status_sgi_winno    = winno;

      ortho2((Coord) STAT_XMIN,(Coord) Status_wid,
             (Coord) STAT_YMIN,(Coord) Status_hit);

      color(STATUS_BACK_COLOR);

      clear(); 
      border_status_window(winno);
      swapbuffers();
       
      set_must_redraw(wt);

      return winno;
   } else {
      return -1;
   }
}

void border_status_window(winno)
   int            winno;
{
   winset(winno);
   linewidth(border_width);
   color(STATUS_BORDER_COLOR);
   rect(STAT_XMIN,STAT_YMIN,STAT_XMAX,STAT_YMAX);
}

int close_status_window(wt)
   WT             wt;
{
   int            rc;

   int            ltrace = OFF;

   if( get_status_window_flag(wt) && scope_mode == SINGLE_MODE 
       && PP->status_sgi_winno > 0 ) {
      if( ltrace ) {
         printf("close_status_window: wt: %d   status_sgi_winno: %d.\n",
                wt,PP->status_sgi_winno); 
         printf("close_status_window: status_window_flag: %d.\n",
               get_status_window_flag(wt));
      }
      winclose(PP->status_sgi_winno);
      PP->status_sgi_winno = -1;
   } 
   return 0;
}

void display_status_window(wt)
   WT        wt;
{
   int       winno,        i = 0;
   char      buffer[1024];

#define      WRITELN1      cmov2i(sw_ixc[i],sw_iyc[i]); color(sw_col[i]);
#define      WRITELN       WRITELN1; fmsetfont(sw_fh[i]); fmprstr(buffer); ++i

   if( scope_mode == SINGLE_MODE  && get_status_window_flag(wt) ) {
      if( (winno = PP->status_sgi_winno) < 0 ) 
         return;
      winset(winno);
      color(STATUS_BACK_COLOR); clear();    
      border_status_window(winno);
      color(STATUS_TEXT_COLOR);

/*    sprintf(buffer,"%d:   %-20.20s", 
              wt,wds[wt].name); */
      sprintf(buffer,"%d:   %-18.18s", 
              wt,wds[wt].name); 
      WRITELN;

      sprintf(buffer,"%d  of  %d (%d)",
              FP->display_index,FP->lxy,N);
      WRITELN;
      
/*    No longer display status window time in long format ... */
      switch ( sw_Time_format ) {
      case DOUBLE_SHORT_FORMAT:
      case DOUBLE_LONG_FORMAT: 
         sprintf(buffer,"Time: %15.8e", 
                 DP->time);
         break;
      default:
         printf("Unrecognized sw_Time_format: %d.\n",sw_Time_format);
         sprintf(buffer,"Time: ?");   
      }
      WRITELN;

      sprintf(buffer,"Tag:  %d",
              DP->tag);
      WRITELN;

      sprintf(buffer,"x:   %12.5e     %12.5e",
              CO.x1,CO.x2);
      WRITELN;
   
      sprintf(buffer,"y:   %12.5e     %12.5e",
              CO.y1,CO.y2);
      WRITELN;
 
      swapbuffers();
   } 
#undef       WRITELN1
#undef       WRITELN 
}

void display_status_window_1(wt)
   WT            wt;
{
}

#define          G       PP->range_gauge

void old_display_range_gauge(WT wt) {

   DQUAD         dq_gauge,   dq_bar;

   switch( G.state ) {

   case BAR_GAUGE_STATE_DISPLAY:
      switch( G.mode ) {
      case BAR_GAUGE_MODE_X_RANGE:
         dq_gauge = dq_trans_rel_unit_dq(CO,G.dquad);
         dq_bar = make_dquad( 
            dq_gauge.x1 + (CO.x1 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y1,
            dq_gauge.x1 + (CO.x2 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y2
         );
         display_filled_dquad(dq_gauge,G.background_color);
         display_filled_dquad(dq_bar,  G.foreground_color);
         display_outlined_dquad(dq_gauge,G.border_color);
         break;
      case BAR_GAUGE_MODE_Y_RANGE:
         dq_gauge = dq_trans_rel_unit_dq(CO,G.dquad);
         dq_bar = make_dquad( 
            dq_gauge.x1,
            dq_gauge.y1 + (CO.y1 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1),
            dq_gauge.x2,
            dq_gauge.y1 + (CO.y2 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1)
         );
         display_filled_dquad(dq_gauge,G.background_color);
         display_filled_dquad(dq_bar,  G.foreground_color);
         display_outlined_dquad(dq_gauge,G.border_color);
         break;
      case BAR_GAUGE_MODE_XY_RANGE:
         dq_gauge = dq_trans_rel_unit_dq(CO,G.dquad);
         dq_bar = make_dquad( 
            dq_gauge.x1 + (CO.x1 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y1 + (CO.y1 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1),
            dq_gauge.x1 + (CO.x2 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y1 + (CO.y2 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1)
         );
         display_filled_dquad(dq_gauge,G.background_color);
         display_filled_dquad(dq_bar,  G.foreground_color);
         display_outlined_dquad(dq_gauge,G.border_color);
         break;
      default:
         break;
      }
      break;
   default:
      break;

   }
}

void display_range_gauge(WT wt) {

   DQUAD         dq_gauge,   dq_bar;

   switch( G.state ) {

   case BAR_GAUGE_STATE_DISPLAY:
      switch( G.mode ) {
      case BAR_GAUGE_MODE_X_RANGE:
         equate_screen_and_world_coordinates(wds[wt].sgi_winno);
         dq_gauge = dq_trans_rel_unit_dq(iq_to_dq(ICO),G.dquad);
         dq_bar = make_dquad( 
            dq_gauge.x1 + (CO.x1 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y1,
            dq_gauge.x1 + (CO.x2 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y2
         );
         display_filled_dquad(dq_gauge,G.background_color);
         display_filled_dquad(dq_bar,  G.foreground_color);
         display_outlined_dquad(dq_gauge,G.border_color);
         set_must_remap(wt);
         break;
      case BAR_GAUGE_MODE_Y_RANGE:
         equate_screen_and_world_coordinates(wds[wt].sgi_winno);
         dq_gauge = dq_trans_rel_unit_dq(iq_to_dq(ICO),G.dquad);
         dq_bar = make_dquad( 
            dq_gauge.x1,
            dq_gauge.y1 + (CO.y1 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1),
            dq_gauge.x2,
            dq_gauge.y1 + (CO.y2 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1)
         );
         display_filled_dquad(dq_gauge,G.background_color);
         display_filled_dquad(dq_bar,  G.foreground_color);
         display_outlined_dquad(dq_gauge,G.border_color);
         set_must_remap(wt);
         break;
      case BAR_GAUGE_MODE_XY_RANGE:
         equate_screen_and_world_coordinates(wds[wt].sgi_winno);
         dq_gauge = dq_trans_rel_unit_dq(iq_to_dq(ICO),G.dquad);
         dq_bar = make_dquad( 
            dq_gauge.x1 + (CO.x1 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y1 + (CO.y1 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1),
            dq_gauge.x1 + (CO.x2 - GL_CO.x1) / (GL_CO.x2 - GL_CO.x1) *
                          (dq_gauge.x2 - dq_gauge.x1),
            dq_gauge.y1 + (CO.y2 - GL_CO.y1) / (GL_CO.y2 - GL_CO.y1) *
                          (dq_gauge.y2 - dq_gauge.y1)
         );
         display_filled_dquad(dq_gauge,G.background_color);
         display_filled_dquad(dq_bar,  G.foreground_color);
         display_outlined_dquad(dq_gauge,G.border_color);
         set_must_remap(wt);
         break;
      default:
         break;
      }
      break;
   default:
      break;

   }
}

void display_filled_dquad(DQUAD dq,int dq_color) {
   color(dq_color);
   rectf((Coord) dq.x1, (Coord) dq.y1, (Coord) dq.x2, (Coord) dq.y2);
}

void display_outlined_dquad(DQUAD dq,int dq_color) {
   color(dq_color);
   linewidth(1);
   rect ((Coord) dq.x1, (Coord) dq.y1, (Coord) dq.x2, (Coord) dq.y2);
}

#undef           G       

#undef           PP
#undef           FP
#undef           DP
#undef           DATA
#undef           X
#undef           Y
#undef           N
#undef           CO
#undef           GL_CO
#undef           W 
#undef           H 
