#include "project.h"

int            T_border_color = DEFAULT_TITLE_BORDER_COLOR;
int            T_border_width = DEFAULT_TITLE_BORDER_WIDTH;
int            T_back_color   = DEFAULT_BACK_COLOR;

char           T_default_fontname[] = "Times-Roman";
double         T_default_fontsize = 30.0;
int            T_default_fontcolor = 3;
int            T_default_justify = JUSTIFY_CENTRE;
double         T_baselineskip = 0.3;

char          *T_Fontname[MAX_NLINES];
char          *T_Text[MAX_NLINES];
double         T_Fontsize[MAX_NLINES];
int            T_Fontcolor[MAX_NLINES];
int            T_Justify[MAX_NLINES];
fmfonthandle   T_fh[MAX_NLINES];
double         T_Fontheight[MAX_NLINES];
double         T_Fontysize[MAX_NLINES];
int            T_Iline = 0;

char *my_fgets(s,n,file)
   char      *s;
   int        n;
   FILE      *file;
{
   if( fgets(s,n,file) ) {
      int    i;
      for( i = 0; s[i] != '\n'; i++ );
      s[i] = '\0';
      return s;
   } else {
      return NULL;
   }
}

int check_title(file)
   FILE       *file;
{
   int         nlines = 0, rc;
   char        buffer[TBUFLEN];

   int         ltrace = OFF; 
   
   T_Iline = 0;
   while( (rc = ractcl(file)) > 0 ) {
      if( my_fgets(buffer,TBUFLEN,file) ) {
         T_Text[T_Iline++] = strdup(buffer);
         if( ltrace ) {
            printf("check_title: Text line: %d:\n%s\n\n",T_Iline,buffer);
         }
      } else {
         return -1;
      }
   }
   return rc < 0 ? -1 : nlines;
}

/* Read and check .title control line. */

int ractcl(file)
   FILE       *file;
{
   char        buffer[TBUFLEN],      sfontname[TBUFLEN],
               sfontsize[TBUFLEN],   sfontcolor[TBUFLEN],
               sjustify[TBUFLEN];

   int         nitems;
   int         ltrace = OFF;

   if( my_fgets(buffer,TBUFLEN,file) ) {
      if( ( nitems = sscanf(buffer,"%s %s %s %s",
                     sfontname,sfontsize,sfontcolor,sjustify) ) == 4) {
         if( ltrace ) {
            printf("Font field: %s\nSize field: %s\nColor field: %s\nJustify field: %s\n",sfontname,sfontsize,sfontcolor,sjustify);
         }
         if( is_available_font(sfontname) ) {
            T_Fontname[T_Iline] = strdup(sfontname);
         } else if( sfontname[0] == '.' || sfontname[0] == '"' ) {
            T_Fontname[T_Iline] =  T_Iline == 0 ? strdup(T_default_fontname) :
                                                  strdup(T_Fontname[T_Iline-1]);
         } else {
            printf("ractcl: Unknown font <%s>.\n",sfontname);
            return -1;
         }
         if( sscanf(sfontsize,"%lf",T_Fontsize + T_Iline) != 1 ) {
            if( sfontsize[0] == '.' || sfontsize[0] == '"' ) {
               T_Fontsize[T_Iline] =  T_Iline == 0 ? T_default_fontsize :
                                                     T_Fontsize[T_Iline-1];
            } else {
               printf("ractcl: Invalid fontsize <%s>.\n",sfontsize);
               return -1;
            }
         }
         if( sscanf(sfontcolor,"%d",T_Fontcolor + T_Iline) != 1 ) {
            if( sfontcolor[0] == '.' || sfontcolor[0] == '"' ) {
               T_Fontcolor[T_Iline] =  T_Iline == 0 ? T_default_fontcolor :
                                                     T_Fontcolor[T_Iline-1];
            } else {
               printf("ractcl: Invalid fontcolor <%d>.\n",sfontcolor);
               return -1;
            }
         }
         switch( sjustify[0] ) {
         case 'L':
         case 'l':
            T_Justify[T_Iline] = JUSTIFY_LEFT;
            break;
         case 'R':
         case 'r':
            T_Justify[T_Iline] = JUSTIFY_RIGHT;
            break;
         case 'C':
         case 'c':
            T_Justify[T_Iline] = JUSTIFY_CENTRE;
            break;
         case '.':
         case '"':
            T_Justify[T_Iline] = T_Iline == 0 ? T_default_justify : 
                                                T_Justify[T_Iline-1];
            break;
         default:
            printf("ractcl: Invalid justify code <%1s>.\n",sjustify);
            return -1;
         }
         return 1;
      } else {
         return -1; 
      }
   } else {
      return 0;
   }
}

void dump_title_info()
{
   int     iline;

   for( iline = 0; iline < T_Iline; iline++ ) {
      printf("Title line: %d\nFont: %s\nSize: %g\nColor: %4d    Justify: %4d\n",
             iline,T_Fontname[iline],T_Fontsize[iline],T_Fontcolor[iline],
             T_Justify[iline]);
      printf("%s\n\n",T_Text[iline]);
   }
}

void allocate_title_fonts()
{
   fmfonthandle    fh1;
   int             i;
   
   for( i = 0; i < T_Iline; i++ ) {
      fh1 = fmfindfont(T_Fontname[i]);
      T_fh[i] = fmscalefont(fh1,T_Fontsize[i]);
      fmfreefont(fh1);
   }
}

void free_title_fonts()
{
   int             i;

   for( i = 0; i < T_Iline; i++ ) {
      fmfreefont(T_fh[i]);
   }
}

void get_title_font_heights()
{
   int             i;
   fmfontinfo      fontinfo;

   int             ltrace = OFF;

   for( i = 0; i < T_Iline; i++ ) {
      fmgetfontinfo(T_fh[i],&fontinfo);
      T_Fontheight[i] = fontinfo.height;
      T_Fontysize[i] = fontinfo.ysize;
      if( ltrace ) {
         printf("get_title_font_heights: Font %d: Height: %g  Ysize: %g.\n",
                i,T_Fontheight[i],T_Fontysize[i]);
      }
   }
}

void compute_title_dimensions(pxsize,pysize)
   double       *pxsize, *pysize;
{
   double        xsize = 0,          ysize = 0,
                 line_length;
   int           i;
   
   for( i = 0; i < T_Iline; i++ ) {
      line_length = fmgetstrwidth(T_fh[i],T_Text[i]);
      xsize = line_length > xsize ? line_length : xsize;
      ysize += (i != T_Iline - 1 ? (1.0 + T_baselineskip) : 1.0) *
               T_Fontheight[i];
   }
   *pxsize = xsize; *pysize = ysize;
}

void do_title_2(wt)
   WT       wt;
#define     WINNO         wds[wt].sgi_winno
{
   FILE    *title_file;
   long     orgx,         orgy,
            dx,           dy;
   double   xsize,        ysize;

   int      ltrace = OFF;

   choose_file_by_ext(wds[wt].wd,".title","TITLE",Selected_file_name);
   if( strlen(Selected_file_name) ) {
      sprintf(O_buf,"%s/%s",wds[wt].wd,Selected_file_name);
      if( ltrace ) {
         printf("Selected title file name: %s\n",Selected_file_name);
         printf("Will open: %s\n",O_buf);
      }
      if( title_file = fopen(O_buf,"r") ) {
         if( check_title(title_file) == 0 ) {
            winset(WINNO);

            frontbuffer(ON); backbuffer(OFF); 
            equate_screen_and_world_coordinates(WINNO);
            getsize(&dx,&dy); getorigin(&orgx,&orgy);
            color(T_back_color); clear();
            color(T_border_color); linewidth(T_border_width);
            rect(orgx,orgy,orgx+dx-2,orgy+dy-2);

            allocate_title_fonts();
            get_title_font_heights();
            compute_title_dimensions(&xsize,&ysize);
            if( xsize < dx && ysize < dy ) {
               if( ltrace ) {
                  printf("Title dimensions: %g %g.\n",xsize,ysize);
                  printf("Window dimensions: %d %d.\n",dx,dy);
               }

               plot_title(orgx,orgy,dx,dy,xsize,ysize);

            } else {
               scope_mess_and_bell("Title too large for this window.",1);
               printf("Title dimensions: %g %g.\n",xsize,ysize);
               printf("Window dimensions: %d %d.\n",dx,dy);
               free_title_fonts();
               goto Bail_out;
            }
            free_title_fonts();

/*          swapbuffers();   */
            backbuffer(ON); frontbuffer(OFF); 

            video_set_sequence_label(Selected_file_name);
            cursoff();
            title_process_queue();
            curson();
            video_set_sequence_label(NULL);

/*          x_mouse_after_button();  */
         } else {
            sprintf(O_buf,"Syntax error in %s.",Selected_file_name);
            scope_mess_and_bell(O_buf,1);
         }
      } else {
         sprintf(O_buf,"Open for %s failed.",Selected_file_name);
         scope_mess_and_bell(O_buf,1);
      }
   }
Bail_out:
   set_must_remap(wt); set_must_redraw(wt);
}

void title_process_queue(void) 
{
   short    val;
#define     KEYcase(a,b) case a: if( val ) { b; } break

   while(1) {
      while( qtest() ) {
         switch( qread(&val) ) {
         KEYcase(MIDDLEMOUSE,return);
         KEYcase(  LEFTMOUSE,return);
         KEYcase( RIGHTMOUSE,return);
         KEYcase(F4KEY,(void) nvs_video_record_n_seconds(-1.0,0));
         KEYcase(F5KEY,(void) nvs_video_record_n_seconds(-5.0,0));
         KEYcase(INSERTKEY,(void) nvs_video_record_n_seconds(-1.0,1));
         KEYcase(PRINTSCREENKEY,
            (void) nvs_video_record_n_seconds(-1.0/FRAMES_PER_SECOND,0));
         }
      }
   }

#undef      KEYcase
}

void plot_title(orgx,orgy,dx,dy,xsize,ysize)
   int       orgx, orgy, dx, dy;
   double    xsize,      ysize;
{
   int       xcursor,    ycursor;
   int       i,          length_Text;

   int       ltrace = OFF;

   ycursor = orgy + (dy - ysize) / 2.0 + ysize - T_Fontheight[0];
   if( ltrace ) {
      printf("plot_title: In routine ... %d lines.\n",T_Iline);
   }
   for( i = 0; i < T_Iline; i++ ) {
      if( i > 0 ) {
         ycursor -= (T_baselineskip * T_Fontheight[i-1] + T_Fontheight[i]);
      }
      length_Text = fmgetstrwidth(T_fh[i],T_Text[i]);
      switch( T_Justify[i] ) {
      case JUSTIFY_LEFT:
         xcursor = orgx + LEFT_JUSTIFY_MARGIN * (float) dx;
         break;
      case JUSTIFY_CENTRE:
         xcursor = orgx + (dx - length_Text) / 2.0;
         break;
      case JUSTIFY_RIGHT:
         xcursor = orgx + dx - length_Text - RIGHT_JUSTIFY_MARGIN * (float) dx;
         break;
      }
      fmsetfont(T_fh[i]);
      if( ltrace ) { 
         printf("Plotting %s at %d %d.\n",T_Text[i],xcursor,ycursor);
      }
      cmov2i(xcursor,ycursor);
      color(T_Fontcolor[i]);
      fmprstr(T_Text[i]);
Bail_out:;
   }
}

int get_new_T_border_color(void) {
   double  dcolor;

   sprintf(O_buf,"Border color (%d)",T_border_color);
   GET_DOUBLE(dcolor,(double) T_border_color,O_buf);
   return (int) dcolor;
}

int get_new_T_back_color(void) {
   double  dcolor;

   sprintf(O_buf,"Background color (%d)",T_back_color);
   GET_DOUBLE(dcolor,(double) T_back_color,O_buf);
   return (int) dcolor;
}

int get_new_T_border_width(void) {
   double  dwidth;

   sprintf(O_buf,"Border width (%d)",T_border_width);
   GET_DOUBLE(dwidth,(double) T_border_width,O_buf);
   return (int) dwidth;
}

int get_new_Heads_up_color(void) {
   double  dcolor;

   sprintf(O_buf,"Heads up color (%d)",Heads_up_color);
   GET_DOUBLE(dcolor,(double) Heads_up_color,O_buf);
   return (int) dcolor;
}

int get_new_Heads_up_lim_color(void) {
   double  dcolor;

   sprintf(O_buf,"Heads up lim color (%d)",Heads_up_lim_color);
   GET_DOUBLE(dcolor,(double) Heads_up_lim_color,O_buf);
   return (int) dcolor;
}
