#include "video.h"

typedef enum {
	VIDEO_AUTO_SEQUENCE_OFF,
	VIDEO_AUTO_SEQUENCE_ON
} Video_enum;

jmp_buf     Video_ldr_off_jbuf;
int         Video_ldr_state = UNKNOWN;
int         Video_ldr_mode  = UNKNOWN;
int         Video_ldr_off_time = 1;

int         Video_videolab_state = UNKNOWN;
char        Video_videolab_init_cmd[] = "/usr/video/vli/bin/vlicmd init";
char       *Video_videolab_initfile = (char *) NULL;
int         Video_videolab_init_sleep = 5;

int         Video_rs232 = -1;
char        Video_rs232_pseudo_file[] = "/dev/ttyd2";
int         Video_trace = OFF;

int         Video_record_trace = OFF;
int         Video_record_trace_prev;
char       *Video_sequence_label = NULL;
char        Video_sequence_file[BUFSIZ] = ".video_trace";
char        Video_sequence_file_prev[BUFSIZ] = "";
FILE       *Video_sequence_FILE = NULL;
int         Video_sequence_count = 0;

int         Video_start_frame = VIDEO_NO_START_FRAME;
int         Video_finish_frame;

char        Video_buffer[1024];

int              Video_auto_sequence_state = VIDEO_AUTO_SEQUENCE_OFF;
char            *Video_auto_sequence_filename = NULL;
char            *Video_auto_sequence_label = NULL;
int              Video_auto_sequence_bgn = VIDEO_NO_START_FRAME;
int              Video_auto_sequence_end = VIDEO_NO_START_FRAME;
VIDEO_SEQ_LIST   Video_auto_seq_list = {NULL, 0, NULL, NULL, NULL, NULL};
int              Video_auto_sequence_trace = ON;


/* These routines assume that all ldr control is from software other than
   intial power on and load ... if this assumption is incorrect (i.e. if 
   you fiddle with the front panel) internal state may become confused */

/* Slightly modified version of vtalk() written by Kelvin Thompson. */
/* Extracted from pablo/rob's string codes.                         */

int video_open_rs232(void) {
#ifndef DARWIN
   int temp;
   char cc;
   char *filename;
   struct termio termdata,prevdata;

	/* Is port apparently already open?  If so just return ... */

	if( Video_rs232 >= 0 ) return( 1 );

   /* try to open the device file */
   Video_rs232 = open(  Video_rs232_pseudo_file, O_RDWR );
   if ( Video_rs232 < 0 )
   {
      fprintf(stderr,
         "video_open_rs232: Can't open file '%s'.\n",Video_rs232_pseudo_file);
      return(0);
   }

   /* fetch current terminal modes for tty device */
   temp = ioctl( Video_rs232, TCGETA, &termdata );
   if ( temp == -1 )
   {
      fprintf( stderr, "Can't get modes for '%s'.\n",Video_rs232_pseudo_file );
      return(0);
   }

   /* change to the correct modes */
   termdata.c_iflag = 0;  /* no special input processing */
   termdata.c_oflag = 0;  /* no special output processing */
   termdata.c_cflag = 0
       | B1200    /* 1200 baud */
   | CS8      /* 8 data bits */
   & ~CSTOPB  /* one stop bit */
   | CREAD    /* enable receiver */
   & ~PARENB  /* no parity */
   | CLOCAL;  /* local line (not dialup) */
   termdata.c_lflag = 0;       /* no processing */
   termdata.c_cc[VMIN] = 1;

   /* set the modes */
   temp = ioctl( Video_rs232, TCSETA, &termdata );

   if ( temp == -1 )
   {
      fprintf( stderr, "Can't set modes for '%s'.\n",Video_rs232_pseudo_file );
      return(0);
   }
   return(1);
#endif
}

int video_close_rs232(void) {
   if( !video_rs232_not_open("video_close_rs232") ) {
      close(Video_rs232);
      Video_rs232 = -1;
   }
   return(1);
}

static int fake = 0;
int fake_video_record_one_frame_0(void) {
	return( fake++ );
}
int fake_video_get_ldr_frame(void) {
	return(fake);
}

int video_record_one_frame(void) {

   int  rc;
   int  start_frame,   finish_frame;

	int  iseq;

   int  ltrace = ON;

   if( !video_init_ldr("video_record_one_frame",VIDEO_RECORD) ) return(0);    

	if( Video_auto_sequence_state == VIDEO_AUTO_SEQUENCE_ON ) {
		Video_auto_seq_list = 
         get_video_seq_list_from_file(Video_auto_sequence_filename);
		if( Video_auto_sequence_bgn == VIDEO_NO_START_FRAME ) {
			Video_auto_seq_list = append_seq_list(Video_auto_seq_list);
			iseq = Video_auto_seq_list.nseq - 1;
			Video_auto_sequence_bgn = video_get_ldr_frame();
			Video_auto_seq_list.bgn[iseq]   = Video_auto_sequence_bgn;
			Video_auto_seq_list.speed[iseq] = 1;
			Video_auto_seq_list.labels[iseq] = Video_auto_sequence_label ?
            strdup(Video_auto_sequence_label) : strdup(" ");
		}
		rc = video_record_one_frame_0();
		iseq = Video_auto_seq_list.nseq - 1;
		Video_auto_seq_list.end[iseq] = video_get_ldr_frame() - 1;
		put_video_seq_list_to_file(Video_auto_seq_list,
                                 Video_auto_sequence_filename);
	} else {
		if( Video_record_trace ) {
			start_frame = video_get_ldr_frame();
		}
		rc = video_record_one_frame_0();
		if( Video_record_trace ) {
			finish_frame = video_get_ldr_frame() - 1;
			video_record_trace_sequence(start_frame,finish_frame); 
			if( ltrace ) {    
				fprintf(stderr,"video_record_one_frame: Recorded frames ");
				fprintf(stderr,"<%d> ... <%d>\n",start_frame,finish_frame);
			}
		}
	}

   return(rc);
}

int video_record_one_frame_0(void) {
   char stx = 0x02;
   char etx = 0x03;
   int  rc;

   int  start_frame,   finish_frame;

   int  ltrace = Video_trace;

   if( !video_get_ldr_state("video_record_one_frame_0") ) {
      return(0);
   }

   video_send_ldr_command("GS");
   rc = video_get_ldr_response("GS");

   if( Video_trace ) {
      fprintf(stderr,"video_record_one_frame_0: wrote frame.\n");
   }
   return(rc == 0);
}

int video_record_one_frame_no_cursor(void) {
   char stx = 0x02;
   char etx = 0x03;
   int  rc;

   int  start_frame,   finish_frame;

   int  ltrace = OFF;

   if( !video_get_ldr_state("video_record_one_frame_no_cursor") ) return(0);

   if( Video_record_trace ) {
      start_frame = video_get_ldr_frame();
   }

   video_send_ldr_command("GS");
   rc = video_get_ldr_response("GS");

   if( Video_record_trace ) {
      finish_frame = video_get_ldr_frame() - 1;
      video_record_trace_sequence(start_frame,finish_frame); 
      if( ltrace ) {    
         fprintf(stderr,"video_record_one_frame: Recorded frames ");
         fprintf(stderr,"<%d> ... <%d>\n",start_frame,finish_frame);
      }
   }

   if( Video_trace ) {
      fprintf(stderr,"video_record_one_frame: wrote frame.\n");
   }

   return(rc == 0);
}

int video_record_n_seconds(double n) {
   int    start_frame,   finish_frame;
   int    n_frames, n_record = 0;

   int    ltrace = OFF;
      
   
   if( !video_get_ldr_state("video_record_n_seconds") ) return(0);

   if( Video_record_trace ) {
      start_frame = video_get_ldr_frame();
   }
   for( n_frames = n * FRAMES_PER_SECOND; n_frames > 0; n_frames-- ) {
       n_record += video_record_one_frame_0();
   }
   if( Video_record_trace ) {
      finish_frame = video_get_ldr_frame() - 1;
      video_record_trace_sequence(start_frame,finish_frame); 
      if( ltrace ) {    
         fprintf(stderr,"video_record_n_seconds: Recorded frames ");
         fprintf(stderr,"<%d> ... <%d>\n",start_frame,finish_frame);
      }
   }

   return(n_record);
}

void video_trace_on(void) {
   Video_trace = ON;
}

void video_trace_off(void) {
   Video_trace = OFF;
}

int video_get_ldr_response(char *com) {
   char       rc;
   int        i;

   if( !video_get_ldr_state("video_get_ldr_response") ) return(0);

   for( i = 0; i < strlen(com); ) {
      if( read(Video_rs232,&rc,1) == 1 ) {
         if(        rc == com[i] ) {
            i++;
         } else if( rc == 'E' ) {
            return(-video_get_ldr_error_code());
         }
      }
   }
   return(0);
}

int video_get_ldr_error_code(void) {
   char       rc;
static
   char       nums[] = "0123456789";
   int        val;
   
   if( !video_get_ldr_state("video_get_ldr_error_code") ) return(0);

   while( 1 ) {
      if( read(Video_rs232,&rc,1) == 1 ) {
         if( strchr(nums,rc) ) {
            val = 10 * (rc - '0');
            while( 1 ) {
               if( read(Video_rs232,&rc,1) == 1 ) {
                  if( strchr(nums,rc) ) {
                     val += rc - '0';
                     return(val);
                  }
               }
            }
         }
      }
   }
}

char *video_get_ldr_general_response(void) {
   char    rc;
   char   *response = (char *) NULL;
   int     curs = 0, pos = 0;
static 
   char    alphas[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
   char    nums  [] = "0123456789";

   if( !video_get_ldr_state("video_get_ldr_general_response") ) 
       return( (char *) NULL );
   while( 1 ) {
      if( read(Video_rs232,&rc,1) == 1 ) {
         if(  strchr(alphas,rc) ) {
            pos = rc - 'A';
            if( !response ) {
               response = (char *) malloc(4 * sizeof(char) );
               response[3] = '\0';
               curs = 0;
            }
            response[curs] = alphas[pos];
            curs++;
            if( curs >= 3 ) return( response );
         }
      } 
   }
}
char *old_video_get_ldr_general_response(void) {
   char    rc;
   char    locbuf[1024];
   int     curs = 0;

   if( !video_get_ldr_state("video_get_ldr_general_response") ) 
       return( (char *) NULL );
   while( 1 ) {
      if( read(Video_rs232,&rc,1) == 1 ) {
         locbuf[curs] = rc; curs++;
         if( rc == 'Z' ) {
            fprintf(stderr,"Z\n");
         }
         if( rc == '\n' ) {
            fprintf(stderr,"curs: %d\n",curs);
            locbuf[curs] = '\0';
            return( strdup(locbuf) );
         }
      } 
   }
}

int video_set_ldr_frame(int frame_number) {
   int      ltrace = OFF;

   if( !video_get_ldr_state("video_set_ldr_frame") ) return(0);

	if( video_get_ldr_frame() == frame_number ) return(1);

   sprintf(Video_buffer,"SR%6d:",frame_number);
   if( ltrace ) {
      printf("video_set_ldr_frame: Command to ldr <%s> (%d).\n",
             Video_buffer,strlen(Video_buffer));
   }
   video_send_ldr_command(Video_buffer);
   return(video_get_ldr_response("SR"));
}

int video_get_ldr_frame(void) {
   int       ltrace = OFF;
   char      number_buf[5];
   
   int       i,  frame;

   if( !video_get_ldr_state("video_get_ldr_frame") ) return(0);

   video_send_ldr_command("NO");
   if( !video_get_ldr_response("NO") ) {
      for( i = 0; i < 5; i++ ) {
         if( read(Video_rs232,number_buf+i,1) != 1 ) {
            return(-1);
         }
      }
   }
   if( ltrace ) {
      printf("video_get_ldr_frame: Frame number: %s.\n",number_buf);
   }
   if( sscanf(number_buf,"%5d",&frame) == 1 ) {
      return(frame);
   } else {
      return(-1);
   }
}

void video_send_ldr_command(char *command) {
   char     stx = 0x02;
   char     etx = 0x03;

	int      ltrace = OFF;

   write(Video_rs232,&stx,1);
	if( ltrace ) {
		fprintf(stderr,"video_send_ldr_command: Sent STX\n");
	}
   write(Video_rs232,command,strlen(command));
	if( ltrace ) {
		fprintf(stderr,"video_send_ldr_command: Sent %s\n",command);
	}
   write(Video_rs232,&etx,1);
	if( ltrace ) {
		fprintf(stderr,"video_send_ldr_command: Sent ETX\n");
	}
}

int int_in_range(int n, int lo, int hi) {
   return((lo <= n) && (n <= hi));
}

int non_zero_int_in_range(int n, int lo, int hi) {
   return (int_in_range(n,lo,hi) && n);
}

/* speed > 0 : slow forward ...                               */
/* speed < 0 : fast forward ...                               */

int video_ldr_playback_with_pause(int first_frame, int last_frame, 
                                  int speed, 
                                  int pre_pause, int post_pause)
{
   int     rc;
   int     forward;
   double  runtime;
   int     ltrace = OFF;

   if( !video_get_ldr_state("video_ldr_playback_with_pause") ) return(0);

   if( last_frame < first_frame ) {
      fprintf(stderr,
              "video_ldr_playback_with_pause: Sequence error %d -- %d.\n",
              first_frame,last_frame);
      return(-1);
   }

   forward = speed > 0;

   if( !forward ) speed = -speed;
   if( non_zero_int_in_range(speed,-10,256) ) {
      if( !(rc = video_set_ldr_frame(first_frame)) ) {
         if( pre_pause < 0 ) {
            printf("\nEnter anything to begin recording sequence: ");
            (void) getchar();
            if( forward ) {
               runtime = (((double) (last_frame - first_frame + 1))  * speed) / 
                         (double) FRAMES_PER_SECOND; 
            } else {
               runtime = (((double) (last_frame - first_frame + 1))  / speed) / 
                         (double) FRAMES_PER_SECOND; 
            }
            if( runtime > 100.0 ) {
               printf("\nSequence will end in %3.1f minutes.\n",runtime / 60.0);
            } else {
               printf("\nSequence will end in %d seconds.\n",
                      (int) (runtime + 0.5) );
            }
         } else {
            countdown("Playback sequence will begin",pre_pause);
         }
         if( speed == 1 ) {
            sprintf(Video_buffer,"PF%6d:",last_frame);
         } else {
            sprintf(Video_buffer,forward ? "LF%3d:%6d:" : "FF%3d:%6d:",
                    speed,last_frame);
         }
         if( ltrace ) {
            printf("video_ldr_playback_with_pause: Command to ldr <%s> (%d).\n",
                    Video_buffer,strlen(Video_buffer));
         }
         video_send_ldr_command(Video_buffer);
         rc = video_get_ldr_response(speed == 1 ? "PF" : 
                                     (forward ? "LF" : "FF"));
         if( post_pause < 0 ) {
            printf("\nEnter anything to end recording sequence: ");
            (void) getchar();
         } else {
            countdown("Playback sequence complete ... exiting",post_pause);
         }
         return(rc);
      } else {
         fprintf(stderr,
                 "video_ldr_playback_with_pause: set_ldr_frame returns %d.\n",
                 rc);
         return(rc);
      }
   } else {
      fprintf(stderr,
              "video_ldr_playback_with_pause: Invalid speed: %d.\n",
              speed);
      return(-1);
   }
}

int video_ldr_playback(
   int first_frame, int last_frame, int speed) 
{
   if( !video_get_ldr_state("video_ldr_playback") ) return(0);

   return(
      video_ldr_playback_with_pause(first_frame,last_frame,speed,0,0));
}

int video_ldr_playback_with_prompt(
   int first_frame, int last_frame, int speed) 
{
   if( !video_get_ldr_state("video_ldr_playback_with_prompt") ) return(0);

   return (
      video_ldr_playback_with_pause(first_frame,last_frame,speed,-1,-1));
}

void countdown(char *mess, int nsecs) {
   if( nsecs > 0 ) {
      for( ; nsecs > 0; nsecs-- ) {
         fprintf(stderr,"%s in %d seconds.\n",mess,nsecs);
         sleep(1);
      }
   }
}

int video_ldr_cancel(void) {
   if( !video_get_ldr_state("video_ldr_cancel") ) return(0);

   video_send_ldr_command("CS");
   return(video_get_ldr_response("CS"));
}

void video_set_record_trace(void) {
   if( Video_record_trace == ON ) {
      fprintf(stderr,"video_set_record_trace: Tracing already enabled.\n");
   } else {
      Video_record_trace = ON;
   }
}

void video_unset_record_trace(void) {
   Video_record_trace = OFF;
}

void video_set_sequence_label(char *label) {
   if( Video_sequence_label ) {
      free(Video_sequence_label);
      Video_sequence_label = NULL;
   }
   if( label ) {
      Video_sequence_label = strdup(label);
   } 
}

void video_record_trace_sequence(int start,int end) {
   int     ltrace = OFF; 
   char   *wd;

   if( Video_record_trace == ON ) {
      if( !Video_sequence_FILE ) {
         if( ! ( Video_sequence_FILE = fopen(Video_sequence_file,"a") ) ) {
            fprintf(stderr,"video_record_trace_sequence: Error opening ");
            fprintf(stderr,"%s for appending.\n",Video_sequence_file);
            Video_record_trace = OFF;
            return;
         }
      } 

      wd = getcwd(NULL,BUFSIZ);
      fprintf(stderr,"video_record_trace_sequence: tracing on ");
      fprintf(stderr,"%s/%s\n",wd,Video_sequence_file);
      free(wd);

      if( Video_sequence_label ) {
         fprintf(Video_sequence_FILE,"%5d   %5d    1   %s\n",
                 start,end,Video_sequence_label);
         if( ltrace ) {
            fprintf(stderr,"%5d   %5d    1    %s\n",
                    start,end,Video_sequence_label);
         }
      } else {
         fprintf(Video_sequence_FILE,"%5d   %5d    1\n",start,end);
         if( ltrace ) {
            fprintf(stderr,"%5d   %5d    1\n",start,end);
         }
      }
      fclose(Video_sequence_FILE);  Video_sequence_FILE = NULL;
   } else {
      if( ltrace ) {
         fprintf(stderr,"video_record_trace_sequence: Tracing not enabled.\n");
      }
   }
}

/* Function for maintaining and recording start and finish frames of       */
/* video sequence ... provides alternative to "automatic" labelling        */
/* afforded by various "record" routines when Video_record_trace == ON ... */

/* Call immediately before and after sequence is recorded ... */

void video_record_trace(void) {
	int   ltrace = OFF;
   if( Video_start_frame == VIDEO_NO_START_FRAME ) {
      Video_record_trace_prev = Video_record_trace;
      Video_record_trace = OFF;
      Video_start_frame = video_get_ldr_frame();
		if( ltrace ) {
			fprintf(stderr,"video_record_trace: Mode 0\n");
		}
   } else {
      Video_finish_frame = video_get_ldr_frame() - 1;
      Video_record_trace = ON;
      if( Video_finish_frame >= Video_start_frame ) {
         video_record_trace_sequence(Video_start_frame,Video_finish_frame);
      }
      Video_start_frame = VIDEO_NO_START_FRAME;
      Video_record_trace = Video_record_trace_prev;
		if( ltrace ) {
			fprintf(stderr,"video_record_trace: Mode 1\n");
		}
   }
}

int video_rs232_not_open(char *routine) {
   if( Video_rs232 < 0 ) {
      fprintf(stderr,"%s: Video port not open.\n",routine);
      return 1;
   } else {
      return 0;
   }
}

int video_on_line(void) {
   int    rc;

   if( !video_get_ldr_state("video_on_line") ) return(0);
   if( !strcmp(video_player_status(),"ZEJ") ) {
      fprintf(stderr,"%s: Please load the disk.\n","video_on_line");
      return( 0 );
   }
   video_send_ldr_command("ON0:");
   rc = video_get_ldr_response("ON");
   return( rc );
}

int video_off_line(void) {
   int    rc;

   if( !video_get_ldr_state("video_off_line") ) return(0);
   if( !strcmp(video_player_status(),"ZEJ") ) {
      fprintf(stderr,"%s: Please load the disk.\n","video_off_line");
      return( 0 );
   }
   video_send_ldr_command("OF");
   rc = video_get_ldr_response("OF");
   return( rc );
}

int video_set_record_mode_0(void) {
   int    rc;
	int    ltrace = Video_trace;

   if( !video_get_ldr_state("video_set_record_mode_0") ) return(0);
	if( strcmp(video_player_status(),"ZRM") ) {
		if( ltrace ) {
			fprintf(stderr,"Sending RM to ldr.\n");
		}
		video_send_ldr_command("RM");
		rc = video_get_ldr_response("RM");
	}
   return( rc );

}

int video_set_record_mode(void) {
   int    rc;
   rc = video_init_ldr("video_set_record_mode",1);
   return( rc );
}

int video_clear_record_mode(void) {
   int    rc;
   rc = video_init_ldr("video_clear_record_mode",0);
   return( rc );
}

int video_clear_record_mode_0(void) {
   int    rc;
   if( !video_get_ldr_state("video_clear_record_mode_0") ) return(0);
   video_send_ldr_command("RC");
   rc = video_get_ldr_response("RC");
   return( rc );

}

int video_on_line_lock(void) {
   int    rc;
   if( !video_get_ldr_state("video_on_line_lock") ) return(0);
   video_send_ldr_command("OL1:");
   rc = video_get_ldr_response("OL");
   return( rc );

}

int video_on_line_unlock(void) {
   int    rc;
   if( !video_get_ldr_state("video_on_line_unlock") ) return(0);
   video_send_ldr_command("OL0:");
   rc = video_get_ldr_response("OL");
   return( rc );

}

int video_eject(void) {
   int    rc;
   rc = video_init_ldr("video_eject",0);
   if( rc ) {
      video_send_ldr_command("EJ");
      rc = video_get_ldr_response("EJ");
   }
   return( rc );

}

int video_load(void) {
   int    rc;
   if( !video_get_ldr_state("video_load") ) return(0);
   video_send_ldr_command("LD");
   rc = video_get_ldr_response("LD");
   return( rc );

}

char *video_player_status(void) {
   char   *status = (char *) NULL;

   if( !video_get_ldr_state("video_player_status") ) return( status );
   video_send_ldr_command("PS0:");
   status = video_get_ldr_general_response();
   return( status );
}

void video_onalarm(void) {

   Video_ldr_state = OFF;
   signal(SIGALRM,SIG_IGN);
   longjmp(Video_ldr_off_jbuf,0);
}

int video_get_ldr_state(char *routine) {
   int    rc;
   int    ltrace = OFF;
   
	if( Video_ldr_state == UNKNOWN ) {
		if( !video_open_rs232() ) {
			fprintf(stderr,"%s: Can't open RS-232 port.\n",routine);
			return( OFF );
		}
		signal(SIGALRM,video_onalarm);
		if( ltrace ) {
			fprintf(stderr,
					  "video_get_ldr_state: Setting an alarm for %d seconds.\n",
					  Video_ldr_off_time);
		} 
		alarm(Video_ldr_off_time);
		setjmp(Video_ldr_off_jbuf);
		if(        Video_ldr_state == UNKNOWN ) {
			video_send_ldr_command("PS0:");
			if( read(Video_rs232,&rc,1) == 1 ) {
				signal(SIGALRM,SIG_IGN);
				Video_ldr_state = ON;
			} 
		} else if( Video_ldr_state == OFF ) {
			fprintf(stderr,
                 "%s: Laser disk recorder appears to be off.\n",routine);
		} else {
			fprintf(stderr,
					  "talarm: Unrecognized Video_ldr_state: %d.\n",Video_ldr_state);
		}
	}
   
   return( Video_ldr_state );
}

/* Main routine for setting up laser disk recorder ...               
   
   usage:   

#include   <video.h>                                        or
#include   "/usr/people/matt/include/video.h"

           video_init_ldr("<calling routine name>",VIDEO_PLAY);
           video_init_ldr("<calling routine name>",VIDEO_RECORD);

    to set up in playback/record mode respectively ...                    */

int video_init_ldr(char *routine,int mode) {
   int    rc;
   int    ltrace;
	char   hostname[BUFLEN];
   
/* Software should only be used on einstein ... */

	ltrace = Video_trace;

	gethostname(hostname,BUFLEN);
	if( strcmp(hostname,"einstein") ) {
		fprintf(stderr,"%s: Can only run this software on einstein, not %s.\n",
			     routine,hostname);
		return( 0 );
	}

/* Check to see if software has already been initialized to desired state ... */

	if( ltrace ) {
		fprintf(stderr,"video_init_ldr: mode <%d>  Video_ldr_mode <%d>\n",
              mode,Video_ldr_mode);
	}

	if( (mode == VIDEO_PLAY)    &&  (Video_ldr_mode == VIDEO_PLAY) ) 
      return( 1 );
	if( (mode == VIDEO_RECORD)  &&  (Video_ldr_mode == VIDEO_RECORD) ) 
      return( 1 );

   if( video_get_ldr_state(routine) ) {
      if( !strcmp(video_player_status(),"ZEJ") ) {
         fprintf(stderr,"%s: Please load the disk.\n",routine);
         return( 0 );
      }
      if(      mode == VIDEO_PLAY ) {
         video_on_line();
         if( ltrace ) {
            fprintf(stderr,"video_init_ldr: Called from <%s>\n",routine);
            fprintf(stderr,"video_init_ldr: State prior to record clear <%s>\n",
                    video_player_status());
         }
         if( !strcmp(video_player_status(),"ZRM") ) {
            video_clear_record_mode_0();
         } else {
            if( ltrace ) {
               fprintf(stderr,"video_init_ldr: Record mode not set.\n");
            }
         }
			Video_ldr_mode = mode;
      } else if( mode == VIDEO_RECORD ) {

/*       Initialize videolab board per application and only if RECORD 
         mode has been requested ... 

			Modified to initialize using file specified by VLIINIT environment
         variable ...

	      ... or file specified in call to video_set_videolab_initfile() ...
*/

	      if (  Video_videolab_state != ON ) {
				if( Video_videolab_initfile ) {
					sprintf(Video_buffer,"%s %s",Video_videolab_init_cmd,
                      Video_videolab_initfile);
				} else if( getenv("VLIINIT") ) {
					sprintf(Video_buffer,"%s %s",Video_videolab_init_cmd,
                       getenv("VLIINIT"));
				} else {
					sprintf(Video_buffer,"%s",Video_videolab_init_cmd);
				}
				if( ltrace ) {
					fprintf(stderr,"video_init_ldr: executing system(%s)\n",
							  Video_buffer);
				}
				system(Video_buffer);
				if( ltrace ) {	
					fprintf(stderr,"video_init_ldr: Sleeping for %d seconds\n",
							  Video_videolab_init_sleep);
				}
				sleep(Video_videolab_init_sleep);
				Video_videolab_state = ON;
			}

         video_on_line();
         if( ltrace ) {
            fprintf(stderr,"video_init_ldr: Called from <%s>\n",routine);
            fprintf(stderr,"video_init_ldr: State prior to record set <%s>\n",
                    video_player_status());
         }
         if( strcmp(video_player_status(),"ZRM") ) {
            video_set_record_mode_0();
         } else {
            if( ltrace ) {
               fprintf(stderr,"video_init_ldr: Record mode already set.\n");
            }
         }
			Video_ldr_mode = mode;
      } else {
			fprintf(stderr,"video_init_ldr: Unknown mode %d\n",mode);
      }
      return( 1 );
   } else {
      return( 0 );
   }
}
/* Initializes videolab board from specfied file ... */

int  video_set_videolab_initfile(char *initfile) {
	
   if ( Video_videolab_initfile ) free( Video_videolab_initfile );
	Video_videolab_initfile = strdup(initfile);
	return( 1 );
}

int video_standard_controls(Video_Control_Type op, double speed,int frame_num) {
	int     rc = 0;

	int     ltrace = OFF;

	if( ltrace ) {
		video_trace_on();
	} else {
		video_trace_off();
	}

	if(  Video_ldr_state == OFF ) {
		fprintf(stderr,"Laser disk recorder appears to be off.\n");
		return( rc );
	}

	if( Video_ldr_state == UNKNOWN ) {
		if( !video_init_ldr("video_standard_controls",VIDEO_PLAY) ) return( rc );
	}

	switch( op ) {

	case Video_Stop:
		if( Video_ldr_mode == VIDEO_RECORD ) {
			rc = video_clear_record_mode();
		} else {
			rc = video_still();
		}
		break;
	case Video_Search:
		if( Video_ldr_mode == VIDEO_RECORD ) {
         rc = video_clear_record_mode();
      } else {
         rc = video_still();
      }
		if( video_get_ldr_frame() != frame_num ) {
			rc =  video_set_ldr_frame(frame_num);
		} else {
			rc = 0;
		}
		break;
	case Video_Forward:
		if( !video_init_ldr("video_standard_controls",VIDEO_PLAY) ) return( rc );
		rc = video_play0( 1.0 );
		break;
	case Video_Backward:
		if( !video_init_ldr("video_standard_controls",VIDEO_PLAY) ) return( rc );
		rc = video_play0( -1.0 );
		break;
	case Video_Fast_Forward:
		if( !video_init_ldr("video_standard_controls",VIDEO_PLAY) ) return( rc );
		rc = video_play0( speed );
		break;
	case Video_Fast_Backward:
		if( !video_init_ldr("video_standard_controls",VIDEO_PLAY) ) return( rc );
		rc = video_play0( -speed );
		break;
	case Video_Single_Forward:
		if( !video_init_ldr("video_standard_controls",VIDEO_PLAY) ) return( rc );
		rc = video_step( 1 );
		break;
	case Video_Single_Backward:
		if( !video_init_ldr("video_standard_controls",VIDEO_PLAY) ) return( rc );
		rc = video_step( -1 );
		break;
	case Video_Record_Mode:
		rc = video_set_record_mode();
		break;
	case Video_Online:
		rc = video_on_line();
		break;
	case Video_Offline:
		rc = video_off_line();
		break;
	case Video_Eject:
		rc = video_eject();
		break;
	default:
		fprintf(stderr,"video_standard_controls: Unrecognized op: %d\n",op);
		rc = 0;
		break;

	}

	return( video_get_ldr_frame() );
}

int video_play0(double speed) {
	char  lbuf[256];

	if( speed == 0.0 ) return( video_still() );
	if( fabsf(speed) >= 1.0 ) {
		if(        speed ==  1 ) {
			video_send_ldr_command("PF");
			return( video_get_ldr_response("PF") );
		} else if( speed == -1 ) {
			video_send_ldr_command("PR");
			return( video_get_ldr_response("PR") );
		} else if( speed > 0  &&  speed < 11 ) {
			sprintf(lbuf,"FF%3d:",(int) speed);
			video_send_ldr_command(lbuf);
			return( video_get_ldr_response("FF") );
		} else if( speed < 0  &&  speed > -11 ) {
			sprintf(lbuf,"FR%3d:",(int) -speed);
			video_send_ldr_command(lbuf);
			return( video_get_ldr_response("FR") );
		} 
	} else {
		if( speed > 0.0 ) {
			speed = (int) (1.0 / speed + 0.5);
			if( speed >= 2  &&  speed <= 256 ) {
				sprintf(lbuf,"LF%3d:",(int) speed);
				video_send_ldr_command(lbuf);
				return( video_get_ldr_response("LF") );
			}
		} else {
			speed = (int) (1.0 / (-speed) + 0.5);
			if( speed >= 2  &&  speed <= 256 ) {
				sprintf(lbuf,"LR%3d:",(int) speed);
				video_send_ldr_command(lbuf);
				return( video_get_ldr_response("LR") );
			}
		}
	}
	return 0;
}

int video_still(void) {
	if( strcmp(video_player_status(),"ZTF") ) {
		video_send_ldr_command("TF");
		return( video_get_ldr_response("TF") );
	}
	return 0;
}

int video_step(int nsteps) {
	if( nsteps == 0 ) return( 0 );
	if( nsteps > 0 ) {
		if( strcmp(video_player_status(),"ZTF") ) {
			video_send_ldr_command("TF");
			video_get_ldr_response("TF");
		} 
		while( nsteps > 0 ) {
			video_send_ldr_command("TF");
			video_get_ldr_response("TF");
			nsteps--;
		}
	} else {
		nsteps = -nsteps;
		if( strcmp(video_player_status(),"ZTR") ) {
			video_send_ldr_command("TR");
			video_get_ldr_response("TR");
		} 
		while( nsteps > 0 ) {
			video_send_ldr_command("TR");
			video_get_ldr_response("TR");
			nsteps--;
		}
	}
	return 0;
}

/* Speed < 0 for slow playback ... */

int video_sequence_playback_forward(int bgn,int end,int speed) {
	int    rc = 0;
	int    ltrace = Video_trace;
	char   lbuf[256];

	if( bgn  && end  && (bgn < end) ) {
		if(        speed > 0 ) {
			if( speed == 1 ) {
				video_set_ldr_frame(bgn);
				sprintf(lbuf,"PF%6d:",end);
				video_send_ldr_command(lbuf);
				rc = 1;
			} else if( non_zero_int_in_range(speed,2,10) ) {
				video_set_ldr_frame(bgn);
				sprintf(lbuf,"FF%3d:%6d:",speed,end);
            video_send_ldr_command(lbuf);
				rc = 1;
			}
		} else if( speed < 0 ) {
			speed = -speed;
			if( non_zero_int_in_range(speed,2,256) ) {
				video_set_ldr_frame(bgn);
				sprintf(lbuf,"LF%3d:%6d:",speed,end);
				video_send_ldr_command(lbuf);
				rc = 1;
			}
		} 
	} else {
		if( ltrace ) {
			fprintf(stderr,"bgn: %d  and  end: %d  define null sequence.\n",
                 bgn,end);
		}
	}
	return( rc );
}

int video_sequence_playback_reverse(int bgn,int end,int speed) {
	int    rc;
	int    ltrace = Video_trace;
	char   lbuf[256];

	if( bgn  && end  && (bgn < end) ) {
		if(        speed > 0 ) {
			if( speed == 1 ) {
				video_set_ldr_frame(end);
				sprintf(lbuf,"PR%6d:",bgn);
				video_send_ldr_command(lbuf);
				rc = 1;
			} else if( non_zero_int_in_range(speed,2,10) ) {
				video_set_ldr_frame(end);
				sprintf(lbuf,"FR%3d:%6d:",speed,bgn);
            video_send_ldr_command(lbuf);
				rc = 1;
			}
		} else if( speed < 0 ) {
			speed = -speed;
			if( non_zero_int_in_range(speed,2,256) ) {
				video_set_ldr_frame(end);
				sprintf(lbuf,"LR%3d:%6d:",speed,bgn);
				video_send_ldr_command(lbuf);
				rc = 1;
			}
		} 
	} else {
		if( ltrace ) {
			fprintf(stderr,"bgn: %d  and  end: %d  define null sequence.\n",
                 bgn,end);
		}
	}
	return( rc );
}

int video_sequence_playback(Video_Control_Type op, double speed,
                            int bgn, int end) {
	int     rc = 0;
	int     ltrace = OFF;

	if( ltrace ) {
		video_trace_on();
	} else {
		video_trace_off();
	}

	speed = fabs(speed);

	if(  Video_ldr_state == OFF ) {
		fprintf(stderr,"Laser disk recorder appears to be off.\n");
		return( rc );
	}

	if( !video_init_ldr("video_sequence_playback",VIDEO_PLAY) ) return 0;

	switch( op ) {

	case Video_Search:
	case Video_Stop:
		if( Video_ldr_mode == VIDEO_RECORD ) {
			fprintf(stderr,"Video_ldr_mode == VIDEO_RECORD Unexpected!\n");
			rc = video_clear_record_mode();
		}
		rc = video_set_ldr_frame(bgn);
		break;
	case Video_Forward:
		rc = video_sequence_playback_forward(bgn,end,1);
		break;
	case Video_Backward:
		rc = video_sequence_playback_reverse(bgn,end,1);
		break;
	case Video_Fast_Forward:
		if( speed != 0.0 ) { 
			if( speed < 1.0 ) {
				speed = (int) (1.0 / speed + 0.5);
				rc = video_sequence_playback_forward(bgn,end,-(int) speed);
			} else {
				rc = video_sequence_playback_forward(bgn,end,(int) speed);
			}
		}
		break;
	case Video_Fast_Backward:
		if( speed != 0.0 ) { 
			if( speed < 1.0 ) {
				speed = (int) (1.0 / speed + 0.5);
				rc = video_sequence_playback_reverse(bgn,end,-(int) speed);
			} else {
				rc = video_sequence_playback_reverse(bgn,end,(int) speed);
			}
		}
		break;
	case Video_Single_Forward:
		if( ltrace ) {
			fprintf(stderr,"Video_Single_Forward op disabled.\n");
		}
		break;
	case Video_Single_Backward:
		if( ltrace ) {
			fprintf(stderr,"Video_Single_Backward op disabled.\n");
		}
		break;
	case Video_Record_Mode:
		if( ltrace ) {
			fprintf(stderr,"Video_Record_Mode op disabled.\n");
		}
		break;
	case Video_Online:
      if( ltrace ) {
         fprintf(stderr,"Video_Online op disabled.\n");
      }
		break;
	case Video_Offline:
      if( ltrace ) {
         fprintf(stderr,"Video_Offline op disabled.\n");
      }
		break;
	case Video_Eject:
      if( ltrace ) {
         fprintf(stderr,"Video_Eject op disabled.\n");
      }
		break;
	default:
		fprintf(stderr,"video_standard_controls: Unrecognized op: %d\n",op);
		rc = 0;
		break;

	}

	return( video_get_ldr_frame() );
}

VIDEO_SEQ_LIST null_video_seq_list(void) {
	VIDEO_SEQ_LIST   null = {NULL, 0, NULL, NULL, NULL, NULL};

	return( null );

}

VIDEO_SEQ_LIST free_video_seq_list(VIDEO_SEQ_LIST the) {
	int    i;

	if( the.name ) free( the.name );
	if( the.nseq > 0 ) {
		for( i = 0; i < the.nseq; i++ ) {
			if( the.labels[i] ) free( the.labels[i] );
		}
		if( the.labels) free( the.labels );
	}
	if( the.bgn )   free( the.bgn );
	if( the.end )   free( the.end );
	if( the.speed ) free( the.speed );
	
	return( null_video_seq_list() );

}

void dump_video_seq_list(VIDEO_SEQ_LIST the) {
	int    i;
	if( the.name ) {
		fprintf(stderr,"dump_video_seq_list: name: <%s>   nseq: %d\n",
				  the.name,the.nseq);
	} else {
		fprintf(stderr,"dump_video_seq_list: No name.  nseq: %d\n",the.nseq);
	}
	if( the.nseq > 0 ) {
		if( the.labels ) {
			fprintf(stderr,"labels:\n");
			for( i = 0; i < the.nseq; i++ ) {
				fprintf(stderr,"<%s>\n",the.labels[i]);
			}
		}
		loc_ivfdump(stderr,the.bgn,the.nseq,"Bgn");
		loc_ivfdump(stderr,the.end,the.nseq,"End");
		loc_ivfdump(stderr,the.speed,the.nseq,"Speed");
	}
}

void loc_ivfdump(FILE *fp,IVEC v,int n,char *s)
{
   int     i, per_line = 6;

   if( v != NULL ) {
      fprintf(fp,"<<< %s >>>\n\n",s);
      for( i = 0; i < n; i++ ) {
         fprintf(fp, ((i + 1) % per_line ? "%12i" : "%12i\n"),v[i]);
      }
      fprintf(fp, (n % per_line) ? "\n\n" : "\n" );
   }
}

VIDEO_SEQ_LIST get_video_seq_list_from_file(char *fname) {
	FILE             *fp;
	VIDEO_SEQ_LIST    the = null_video_seq_list();

	int               ltrace = OFF;

	int               nseq,    bgn,    end,   speed,   i;
	char              label[256];
	
	if( (fp = fopen(fname,"r")) ) {
		nseq = 0;
      while( fscanf(fp,"%d %d %d%[^\n]",&bgn,&end,&speed,label) >= 3 ) {
			++nseq;
		}
		fclose( fp );
		if( ltrace ) {
			fprintf(stderr,"%d sequences in file.\n",nseq);
		}
		if( nseq ) {
			if( (fp = fopen(fname,"r")) ) {
				the.name = strdup(fname);
				the.nseq = nseq;
				the.bgn = (IVEC) malloc(nseq * sizeof(int));
				the.end = (IVEC) malloc(nseq * sizeof(int));
				the.speed = (IVEC) malloc(nseq * sizeof(int));
				the.labels = (char **) malloc(nseq * sizeof(char *));
				nseq = 0;
				while( fscanf(fp,"%d %d %d%[^\n]",the.bgn + nseq,the.end + nseq,
                              the.speed + nseq,label) >= 3 ) {
					for( i = 0; i < strlen(label); i++ ) {
						if( label[i] != ' ' ) {
							the.labels[nseq] = strdup(label+i);
							goto Done;
						}
					}
					the.labels[nseq] = strdup(" ");
Done:
					++nseq;
				}
				fclose( fp );
		   } else {
				fprintf(stderr,
                    "get_video_seq_list_from_file: Error opening <%s>\n",
                    fname);
			}
		}
	} else {
		fprintf(stderr,"get_video_seq_list_from_file: Error opening <%s>\n",
			     fname);
	}

	return( the );
}

int video_set_auto_sequence(char *filename) {
	int    ltrace = Video_auto_sequence_trace;
	char  *lfilename;
	FILE  *fp;

	lfilename = filename ? strdup(filename) : 
                          strdup(DEFAULT_VIDEO_AUTO_SEQUENCE_FILE);
	if( lfilename ) {
		if( (fp = fopen(lfilename,"a")) ) {
			fprintf(stderr,"video_set_auto_sequence: Sequence tracing on <%s>.\n",
                 lfilename);
			fclose(fp);
			if( Video_auto_sequence_filename ) free(Video_auto_sequence_filename);
			Video_auto_sequence_filename = strdup(lfilename);
			Video_auto_sequence_state = VIDEO_AUTO_SEQUENCE_ON;
		} else {
			fprintf(stderr,"video_set_auto_sequence: Could not open ");
			fprintf(stderr,"%s for append.\n",lfilename);
			Video_auto_sequence_state = VIDEO_AUTO_SEQUENCE_OFF;
		}
	} else {
		fprintf(stderr,"video_set_auto_sequence: Null filename.\n");
		Video_auto_sequence_state = VIDEO_AUTO_SEQUENCE_OFF;
	}
	if( lfilename ) free(lfilename);
	return( Video_auto_sequence_state );
}

VIDEO_SEQ_LIST append_seq_list(VIDEO_SEQ_LIST in) {
	VIDEO_SEQ_LIST  out = null_video_seq_list();
	int             i,   nseq;

	nseq = in.nseq + 1;
	if( nseq > 0 ) {
		out.name = in.name ? strdup(in.name) : NULL;
		out.nseq = nseq; 
		out.bgn  = (IVEC) malloc(nseq * sizeof(int));
		out.end  = (IVEC) malloc(nseq * sizeof(int));
		out.speed  = (IVEC) malloc(nseq * sizeof(int));
		out.labels  = (char **) malloc(nseq * sizeof(char *));
		for( i = 0; i < nseq - 1; i++ ) {
			out.bgn[i] = in.bgn[i];
			out.end[i] = in.end[i];
			out.speed[i] = in.speed[i];
			out.labels[i] = in.labels[i] ? strdup(in.labels[i]) : NULL;
		}
		free_video_seq_list(in);
	} else {
		fprintf(stderr,"append_seq_list: nseq <= 0 after increment.\n");
	}
	
	return( out );
}

int put_video_seq_list_to_file(VIDEO_SEQ_LIST the, char *fname) {
	FILE     *fp;
	int       rc = 0, i;

	if( (fp = fopen(fname,"w")) ) {
		for( i = 0; i < the.nseq; i++ ) {
			fprintf(fp,"%6d %6d %6d  %s\n",the.bgn[i],the.end[i],the.speed[i],
                 the.labels[i]);
		}
		fclose(fp);
		rc = the.nseq;
	} else {
		fprintf(stderr,"put_video_seq_list_to_file: Unable to open ");
		fprintf(stderr,"<%s> for write.\n",fname);
	}
	return( rc );
}

int video_set_auto_sequence_label(char *label) {
	int  rc = 0;
	if( Video_auto_sequence_state == VIDEO_AUTO_SEQUENCE_ON ) {
		if( label ) {
			if( Video_auto_sequence_label ) free(Video_auto_sequence_label);
			Video_auto_sequence_label = strdup(label);
			Video_auto_sequence_bgn = VIDEO_NO_START_FRAME;
			Video_auto_sequence_end = VIDEO_NO_START_FRAME;
			fprintf(stderr,"video_set_auto_sequence_label: Sequence label: ");
			fprintf(stderr,"<%s>\n",Video_auto_sequence_label);
			rc = 1;
		} else {
			fprintf(stderr,"video_set_auto_sequence_label: Null label\n");
		}
	} else {
		fprintf(stderr,"video_set_auto_sequence_label: Auto sequencing ");
		fprintf(stderr,"is not enabled.\n");
	}
	return( rc );
}
