/* W I N D O W    T A G    P R O C E S S I N G .   */

#include "project.h"

WT        hiwt        = BASE_WT;     /* High water mark for tag. */
WTBIND   *wtal        = NULL;        /* Window tag active list. */
WTBIND   *save_wtal   = NULL;        /* For SINGLE_MODE.  */
WTND     *wtfl        = NULL;        /* Window tag free list. */

/* Allocates a window tag node ... */

WTND  *make_WTND(wt,next)
WT       wt;
WTND    *next;
{
   WTND     *pwtnd;

   pwtnd = (WTND *) malloc(sizeof(WTND));
   pwtnd->wt = wt; pwtnd->next = next; return pwtnd;
}

/* Frees a window tag node ... */

void free_WTND(pwtnd)
WTND     *pwtnd;
{
   free(pwtnd);
}

/* Allocates a window tag binode ... */

WTBIND  *make_WTBIND(wt,next,prev)
WT         wt;
WTBIND    *prev, *next;
{
   WTBIND     *pwtbind;

   pwtbind = (WTBIND *) malloc((unsigned) sizeof(WTBIND));
   pwtbind->wt = wt; pwtbind->prev = prev; pwtbind->next = next;
   return pwtbind;
}

/* Frees a window tag binode ... */

void free_WTBIND(pwtbind)
WTBIND     *pwtbind;
{
   free(pwtbind);
}

/* Returns new window tag, attaches tag node to active list as side effect. */

WT new_wt()
{
   WT      wt;

   if( Ltrace ) fprintf(stderr,"new_wt: In routine\n");
   if( (wt = new_wt_1()) > 0 ) {
      wtal = make_WTBIND(wt,wtal,NULL);
      if( (wtal->next) != NULL ) (wtal->next)->prev = wtal;
   }
   return(wt);
}

/* Helper function for allocation routines ... */

WT new_wt_1()
{
   WT      wt;

   if( wtfl == NULL ) {  
      if( hiwt <= MAXWT ) {
         wt = hiwt; ++hiwt;
      } else {
         return(-1);
      }
   } else {
      wt = wtfl->wt; wtfl = wtfl->next;
   }
   return wt;
}

/* Tests whether a new window can be created ... */
int wt_available(void) {
   return( (wtfl != NULL) || (hiwt <= MAXWT) );
}

/* Returns new window tag, inserts node *before* supplied wt (if extant)
   so that new window appears *after* supplled wt. Side--effects include
   updating of lal (list of active lists). */

WT new_wt_insert_before_wt(owt)
   WT        owt;
{
   WT        wt;
   PWTBIND   new_node,     p;
   int       ltrace = OFF;

   if( is_active_wt(owt) ) {
      for( p = wtal; p != NULL; p = p->next ) {
         if( p->wt == owt ) {
            if( (wt = new_wt_1()) > 0 ) {
               new_node = make_WTBIND(wt,p,p->prev);
               if( p->prev ) {
                  (p->prev)->next = new_node;
               } else {
                  lal_replace(wtal,new_node);
                  wtal = new_node;
               }
               p->prev = new_node;
            } else {
               return wt;
            }
         }
      }
      return wt;
   } else {
      if( ltrace ) {
         printf("new_wt_insert_before_wt:: Warning: window %d not active.\n",
                owt);
      }
      return new_wt();
   }
}

/* Frees window tag, updating free list as side effect. */

int free_wt(wt)
WT    wt;
{
   WTBIND *p;

   for( p = wtal; p != NULL; p = p->next) {
      if( p->wt == wt ) {
         if( (p->prev) != NULL ) (p->prev)->next = p->next;
         if( (p->next) != NULL ) (p->next)->prev = p->prev;
         wtfl = make_WTND(wt,wtfl);
         if( p == wtal ) wtal = p->next;
         (void) free_WTBIND(p);

         return(1);
      }
   }
   return(0); 
}

#define      LALR(_p1,_p2) if( _p1 == wtal ) { lal_replace(wtal,_p2); wtal = _p2; }

int move_wt_after_wt(wt1,wt2)
   WT        wt1,    wt2;
{
   PWTBIND   p1 = NULL,     p2 = NULL,     p;

   if( is_active_wt(wt1) && is_active_wt(wt2) ) {
      for( p = wtal; !p1 || !p2; p = p->next ) {
         if( p->wt == wt1 ) p1 = p;
         if( p->wt == wt2 ) p2 = p;
      }
      LALR(p1,p1->next)
      LALR(p2,p1)

/*    Delete p1 ... */

      if( p1->prev ) (p1->prev)->next = p1->next;
      if( p1->next ) (p1->next)->prev = p1->prev;

/*    And insert *before* p2 ... */

      p1->prev = p2->prev;
      if( p1->prev ) (p1->prev)->next = p1;
      p1->next = p2;
      p2->prev = p1;

      repanel(n_windows());
      return wt1;
   } else {
      return 0;
   }
}

int move_wt_before_wt(wt1,wt2)
   WT        wt1,    wt2;
{
   PWTBIND   p1 = NULL,     p2 = NULL,     p;

   if( is_active_wt(wt1) && is_active_wt(wt2) ) {
      for( p = wtal; !p1 || !p2; p = p->next ) {
         if( p->wt == wt1 ) p1 = p;
         if( p->wt == wt2 ) p2 = p;
      }
      LALR(p1,p1->next)

/*    Delete p1 ... */

      if( p1->prev ) (p1->prev)->next = p1->next;
      if( p1->next ) (p1->next)->prev = p1->prev;

/*    And insert *after* p2 ... */

      p1->next = p2->next;
      if( p1->next ) (p1->next)->prev = p1;
      p1->prev = p2;
      p2->next = p1;
      repanel(n_windows());
      return wt1;
   } else {
      return 0;
   }
}

#undef       LALR

/* Returns 1 if tag is on active list, 0 otherwise. */

int is_active_wt(wt)
   WT        wt;
{
   WTBIND   *p;

   for( p = wtal; p != NULL; p = p->next) {
      if( p->wt == wt ) 
         return(1);
   }
   return(0);
}

/* Returns window tag from PXYDVBIND head pointer ... */

WT wt_from_xyhead(PXYDVBIND xyhead)
{
   WTBIND   *p;

   int      ltrace = OFF;

   for( p = wtal; p != NULL; p = p->next) {
      if( wds[p->wt].pf.pxyf->xyhead == xyhead ) {
         if( ltrace ) fprintf(stderr,"wt_from_xyhead: Returning %d.\n",p->wt);
         return(p->wt);
      }
   }
   if( ltrace ) fprintf(stderr,"wt_from_xyhead: No match for pointer found.\n");
   return(-1);
}

/* Returns tail of active tag list. */

WTBIND *tail_wtal(wtal)
   WTBIND   *wtal;
{
   WTBIND   *p;
      
   if( wtal != NULL ) {
      for( p = wtal; p->next != NULL; p = p->next );;
      return(p);
   } else {
      return(NULL);
   }
}

/* Returns index (display) of <wt> in active list, -1 if not found. */

int index_from_wt(wt)
   WT       wt;
{
   WTBIND  *p;

   int      i;
   int      ltrace = OFF;

   for( p = tail_wtal(wtal), i = 1; p != NULL; p =p->prev, i++ ) {
      if( p->wt == wt ) {
         if( ltrace ) {
            printf("index_from_wt: Window %d: Index %d.\n",wt,i);
         }
         return i;
      }
   }
   if( ltrace ) {
      printf("index_from_wt: Window %d: No index found, returning -1.\n",i);
   }
   return -1;
}

/* "Inverse" of previous routine. */

WT wt_from_index(i)
   int      i;
{
   WTBIND  *p;

   int      ii;
   int      ltrace = OFF;

   for( p = tail_wtal(wtal), ii = 1; p != NULL; p =p->prev, ii++ ) {
      if( ii == i ) {
         if( ltrace ) {
            printf("wt_from_index: Index %d: Window %i.\n",ii,p->wt);
         }
         return p->wt;
      }
   }
   if( ltrace ) {
      printf("wt_from_index: Index %d: No window found, returning -1.\n",i);
   }
   return (WT) -1;
}

/* Dumps active tag list. */

void dump_wtal()
{
   WTBIND  *p;

   printf("dump_wtal:: Active tag list dump.\n");
   for( p = wtal; p != NULL; p = p->next) {
      printf("%d\n",p->wt);
   }
   printf("dump_wtal:: End dump.\n");
}

/* Dumps active tag list in reverse (allocated) order. */

void rdump_wtal()
{
   WTBIND  *p;

   printf("rdump_wtal:: Active tag list reverse dump.\n");
   for( p = tail_wtal(wtal); p != NULL; p = p->prev) {
      printf("%d\n",p->wt);
   }
   printf("rdump_wtal:: End dump.\n");

}
/* Dumps free tag list and high water mark. */

void dump_wtfl()
{
   WTND    *p;

   printf("<<< dump_wtfl:: Free tag list dump ... >>>\n");
   for( p = wtfl; p != NULL; p = p->next) {
      printf("%d\n",p->wt);
   }
   printf("Next new tag: %d\n",hiwt);
   printf("<<< dump_wtfl:: End dump ... >>>\n");
}

/* Counts number of active windows. */

int n_windows()
{
   WTBIND  *p;
   int      i;

   for( i = 0, p = wtal; p != NULL; i++, p =p->next);
   return(i);
}

int n_windows_1()
{
extern 
   PLALBIND    lalhead;
   PWTBIND     p;
   int         i;

   for( i = 0, p = lalhead->wtl; p != NULL; i++, p = p->next);
   return(i);
}

WTPAIR make_WTPAIR(wt1,wt2)
   WT          wt1,     wt2;
{
   WTPAIR      pair;

   pair.wt1 = wt1;
   pair.wt2 = wt2; 
   return pair;
}

void load_WTPAIR(WTPAIR *pthe,WT wt1,WT wt2) {
   pthe->wt1 = wt1;
   pthe->wt2 = wt2;
}

int is_null_WTPAIR(pair)
   WTPAIR      pair;
{
   return (pair.wt1 < 0) || (pair.wt2 < 0);
}

WTTRIPLE make_WTTRIPLE(wt1,wt2,wt3)
   WT          wt1,     wt2,     wt3;
{
   WTTRIPLE      triple;

   triple.wt1 = wt1;
   triple.wt2 = wt2; 
   triple.wt3 = wt3; 
   return triple;
}

void load_WTTRIPLE(WTTRIPLE *pthe,WT wt1,WT wt2,WT wt3) {
   pthe->wt1 = wt1;
   pthe->wt2 = wt2; 
   pthe->wt3 = wt3; 
}

int is_null_WTTRIPLE(triple)
   WTTRIPLE      triple;
{
   return (triple.wt1 < 0) || (triple.wt2 < 0) || (triple.wt3 < 0);
}

WTQUAD make_WTQUAD(wt1,wt2,wt3,wt4)
   WT          wt1,     wt2,     wt3,     wt4;
{
   WTQUAD      quad;

   quad.wt1 = wt1;
   quad.wt2 = wt2; 
   quad.wt3 = wt3; 
   quad.wt4 = wt4; 
   return quad;
}

void load_WTQUAD(WTQUAD *pthe,WT wt1,WT wt2,WT wt3,WT wt4) {
   pthe->wt1 = wt1;
   pthe->wt2 = wt2;
   pthe->wt3 = wt3;
   pthe->wt4 = wt4;
}

int is_null_WTQUAD(WTQUAD quad) {
   return (quad.wt1 < 0) || (quad.wt2 < 0) || 
          (quad.wt3 < 0) || (quad.wt4 < 0);
}


WTSEX make_WTSEX(wt1,wt2,wt3,wt4,wt5,wt6)
   WT          wt1,     wt2,     wt3,     wt4,    wt5,    wt6;
{
   WTSEX       sex;

   sex.wt1 = wt1;
   sex.wt2 = wt2; 
   sex.wt3 = wt3; 
   sex.wt4 = wt4; 
   sex.wt5 = wt5; 
   sex.wt6 = wt6; 
   return sex;
}

void load_WTSEX(WTSEX *pthe,WT wt1,WT wt2,WT wt3,WT wt4,WT wt5,WT wt6){
   pthe->wt1 = wt1;
   pthe->wt2 = wt2;
   pthe->wt3 = wt3;
   pthe->wt4 = wt4;
   pthe->wt5 = wt5;
   pthe->wt6 = wt6;
}

int is_null_WTSEX(WTSEX sex) {
   return (sex.wt1 < 0) || (sex.wt2 < 0) || 
          (sex.wt3 < 0) || (sex.wt4 < 0) ||
          (sex.wt5 < 0) || (sex.wt6 < 0);
}

/* Routines for applying functions with various number of */
/* auxiliary args to all windows.                         */

#define  FOR   PWTBIND p; for( p = tail_wtal(wtal); p != NULL; p = p->prev)

void apply_all_wt_0(PFV_I f0) {
   FOR {
      if( wds[p->wt].status != WINDOW_NOT_ACTIVE ) {
         (f0)(p->wt);
      }
   }
}

void apply_all_wt_1(PFV_II f1,int arg1) {
   FOR {
      if( wds[p->wt].status != WINDOW_NOT_ACTIVE ) {
         (f1)(p->wt,arg1);
      }
   }
}

void apply_all_wt_d(PFV_ID f,double arg1) {
   FOR {
      if( wds[p->wt].status != WINDOW_NOT_ACTIVE ) {
         (f)(p->wt,arg1);
      }
   }
}

void apply_all_wt_d_d(PFV_IDD f,double arg1,double arg2) {
   FOR {
      if( wds[p->wt].status != WINDOW_NOT_ACTIVE ) {
         (f)(p->wt,arg1,arg2);
      }
   }
}

#undef   FOR

void printf_wt_name(WT wt) {
   printf("Window %d: %s.\n",wt,wds[wt].name);
}

IVECN wt_list_from_wtal(void) {
   IVECN    wt_list = {(IVEC) NULL,0};
   PWTBIND  p = tail_wtal(wtal);
   int      ltrace = OFF;

   while( p ) {
      wt_list = ivecn_append(wt_list,p->wt);
      p = p->prev;
   }

   if( ltrace ) { 
      ivecn_dump(wt_list,"wt_list_from_wtal");
   } 

   return wt_list;
}
