#include "project.h"

/*=========================================================================
Server hostname.
=========================================================================*/
char           *P;

/*=========================================================================
Global window list, panel pointers.

For consistency with 'vs', WINDOWS are indexed from 1.
=========================================================================*/
WINDOW          wds[MAX_WINDOWS + 1];
int             max_windows = MAX_WINDOWS;
int             Max_panel   = MAX_PANEL;
int             Pnl,  Pnl_prev;

/*=========================================================================
Server hostname.
=========================================================================*/
char            shostname[BUFLEN];
int             shnlen = BUFLEN;

/*=========================================================================
Current working directory
=========================================================================*/
char           *Cwd = (char *) NULL;

/*=========================================================================
F  O  R  M  S  
=========================================================================*/
FD_main        *Xvsf           = (FD_main *) NULL;
/* Initial Xvsf (for use with subsequent resizing operations) ... */
FD_main        *Ixvsf          = (FD_main *) NULL;
FD_info        *Xvsf_info      = (FD_info *) NULL;
FD_markers     *Xvsf_markers   = (FD_markers *) NULL;
FD_view        *Xvsf_view      = (FD_view *) NULL;
FD_file        *Xvsf_file      = (FD_file *) NULL;
FD_ops         *Xvsf_ops       = (FD_ops *) NULL;
FD_control     *Xvsf_control   = (FD_control *) NULL;
FD_mpeg        *Xvsf_mpeg      = (FD_mpeg *) NULL;
FD_sm          *Xvsf_sm        = (FD_sm *) NULL;
FD_browse      *Xvsf_browse    = (FD_browse *) NULL;
FD_label       *Xvsf_label     = (FD_label *) NULL;

int             Main_form_border = 5;
int             Control_form_posn_pref = XVS_LEFT;

/*=========================================================================
Last window tag to have data inserted.
=========================================================================*/
int             Last_data_wt = -1;
/*=========================================================================
Deferred working directory for last data window.
=========================================================================*/
char           *Last_data_dir = (char *) NULL;

/*=========================================================================
WWW help info
=========================================================================*/
char            Default_browser[] = "netscape";
char            Help_URL[] = "http://laplace.physics.ubc.ca/~matt/Doc/xvs/";
char           *Browser = (char *) NULL;
/*=========================================================================
Main tracing control
=========================================================================*/
int             Ltrace =  OFF;
int             Ltrace_cb_glcanvas_main = OFF;

/*=========================================================================
Spawns Web browser job with help URL
=========================================================================*/
void xvs_help_www(void) {
   Browser = strdup(getenv("BROWSER") ? getenv("BROWSER") : Default_browser);
   if( Browser ) {
      sprintf(B,"%s %s &",Browser,Help_URL);
      system(B);
   }
}

/*=========================================================================
Does main window have focus?
=========================================================================*/
int xvs_has_focus(void) {
   Window  focus_window;
   int     revert_to;
   if( Xvsf ) {
      XGetInputFocus(fl_display,&focus_window,&revert_to);
      return focus_window == Xvsf->main->window;
   } else {
      return 0;
   }
}

/*=========================================================================
Pnl number -> # of rows & columns.
=========================================================================*/
int pnl_to_ncr(int pnl) {
   if( pnl < 0 ) return 0;
   if( pnl < 1 ) return 1;
   if( pnl < 4 ) return 2;
   if( pnl < 9 ) return 3;
   if( pnl < 16 ) return 4;
   if( pnl < 25 ) return 5;
   return 0;
}

/*=========================================================================
Sets form state and saves prev state.
=========================================================================*/
int xvs_form_state_push(int state) {
   Form_state_prev = Form_state;
   Form_state = state;
   return 0;
}

/*=========================================================================
Restores form state.
=========================================================================*/
int xvs_form_state_pop(void) {
   Form_state = Form_state_prev;
   return 0;
}

/*=========================================================================
Screen width and height from wt.
=========================================================================*/
int xvs_get_wt_w(WT wt) {
   FL_OBJECT  *gl;
   int         rc = 0;
   if( is_active_wt(wt) ) {
      if( (gl = gl_from_wt(wt)) ) {
         rc = gl->w;
      }
   }
   return rc;
}

int xvs_get_wt_h(WT wt) {
   FL_OBJECT  *gl;
   int         rc = 0;
   if( is_active_wt(wt) ) {
      if( (gl = gl_from_wt(wt)) ) {
         rc = gl->h;
      }
   }
   return rc;
}

/*=========================================================================
Attempt to minimize form when no data displayed
=========================================================================*/
void xvs_set_form_geometry(int code) {
   return;
/*
   if( code == 0 ) {
      fl_set_form_geometry(Xvsf->main,Xvsf->main->x,Xvsf->main->y,
         Xvsf->main->w,Ixvsf->main->h);
   } else {
      fl_set_form_geometry(Xvsf->main,Xvsf->main->x,Xvsf->main->y,
         Xvsf->main->w,Xvsf->main->h+100);
      xvs_resize_main_form();
   }
   fl_redraw_form(Xvsf->main);
*/
}

/*=========================================================================
Dumps info about a main form ...
=========================================================================*/
void dump_Xvsf(FD_main *the,char *s) {
   static char R[] = "dump_Xvsf";
   FL_FORM *form = (FL_FORM *) NULL;
   if( the ) {
      mfl_dump_main_form(the,s);
   } else {
      fprintf(stderr,"%s: FATAL 'the' is NULL\n",R);
      exit(1);
   }
}

/*=========================================================================
Initialize all major data structures.
=========================================================================*/
void dump_Pnl(char *s) {
   fprintf(stderr,"****** dump_Pnl: %s  Pnl=%d  Pnl_prev=%d\n",s,Pnl,Pnl_prev);
}

/*=========================================================================
Initialize all major data structures.
=========================================================================*/
void all_ds_init(void) {
   assert(sizeof(APPLY1_TYPE0) == sizeof(APPLY1_TYPE1));
   wds_init();
   lal_init();
   Pnl = -1;
   Pnl_prev = -1;
}

/*=========================================================================
Initialize window list.

Window 0 now contains global default information currently associated 
principally with the control panel (e.g. may need to retrieve wds[0] for 
file operations if no data windows are open.
=========================================================================*/
void wds_init(void) {
   int       iw;

   wds[0].status = WINDOW_DEFAULT;
   if( Cwd ) {
      wds[0].wd     = strdup(Cwd);
      wds[0].wdlen  = strlen(wds[0].wd);
   } else {
      wds[0].wdlen = 0;
   }
   wds[WT_FORM].status = WINDOW_NOT_ACTIVE;
   for( iw = 1; iw < max_windows; iw++ ) {
         wds[iw].status = WINDOW_NOT_ACTIVE;
   }
}

/*=========================================================================
Hides panel, deallocates font structures, and marks WT's as undefined.
=========================================================================*/
void xvs_hide_panel(int pnl) {
   int  ltrace = 0;
   int  i;
   if( pnl < 0 ) return;
   for( i = 0; i <= pnl; i++ ) {
      if( ltrace ) fprintf(stderr,"xvs_hide_panel: Nullling [pnl=%d][i=%d]\n",pnl,i);
      (Xvsf->w)[pnl][i].wt = UNDEF;
      (Xvsf->w)[pnl][i].font_init = 0;
   }
   fl_hide_object(Xvsf->glp[pnl]);
   if( pnl == 0 ) {
      xvs_set_form_geometry(0);
   }
   if( ltrace ) { fprintf(stderr,"\n\n"); sleep(1); }
}

/*=========================================================================
Hides panel, deallocates font structures, and marks WT's as undefined.
=========================================================================*/
void xvs_show_panel(int pnl) {
   int  ltrace = 0;
   if( pnl == 0 ) {
      xvs_set_form_geometry(1);
   }
   fl_show_object(Xvsf->glp[pnl]);
}

/*=========================================================================
Append WT to panel list.  Not currently needed.
=========================================================================*/
int xvs_panel_append(WT wt) {
   int   ltrace = 1;
   int   i, rc;

   if( Pnl < -1 ) Pnl = -1;
   Pnl_prev = Pnl;

   if( ltrace ) {
      fprintf(stderr,"xvs_panel_append: wt=%d  Pnl=%d  Pnl_prev=%d  Max_panel=%d\n",
         wt,Pnl,Pnl_prev,Max_panel);
   }

   if( (Pnl + 1)  < Max_panel ) {
      for( i = 0; i < Pnl + 1; i++ ) {
         (Xvsf->w)[Pnl+1][i].wt = (Xvsf->w)[Pnl][i].wt;
         set_must_redraw((Xvsf->w)[Pnl+1][i].wt);
      }
      (Xvsf->w)[Pnl+1][Pnl+1].wt = wt;
      Pnl = Pnl + 1;
      if( Pnl_prev  >= 0 ) xvs_hide_panel(Pnl_prev);
      if( ltrace ) {
         fprintf(stderr,"xvs_panel_append:  Displaying panel %d\n",Pnl);
      }
      xvs_show_panel(Pnl);
      rc = Pnl;
   } else {
      fprintf(stderr,"xvs_panel_append: Cannot open new panel\n");
      rc = -1;
   }
   return rc;
#undef W
}

/*=========================================================================
Repanels in conjunction with new window insertion.

Pnl must be -1 if there are no windows currently open.
=========================================================================*/
int xvs_open_and_repanel(WT wt_to_open) {
   static char R[] = "xvs_open_and_repanel";
   int         ltrace = Ltrace;
   int         ltrace_sleep = 0;

   PWTBIND     wtp;
   WT          wt;
   int         nw,  iw,    ix,   iy;
   int         rc = -1;
   int         pnl_new;

   if( ltrace ) {
      fprintf(stderr,"%s: Pnl=%d  Pnl_prev=%d  n_windows()=%d\n",
         R,Pnl,Pnl_prev,n_windows());
   }

   if( scope_mode == SINGLE_MODE ) {
      if( ltrace ) { fprintf(stderr,"xvs_open_and_repanel: SINGLE_MODE -> PANEL_MODE\n"); fflush(stderr); }
      (void) scope_to_PANEL_MODE_no_repanel();
      pnl_new = n_windows() - 1;
      if( ltrace ) { fprintf(stderr,"xvs_open_and_repanel: pnl_new=%d\n",pnl_new); fflush(stderr); }
      Pnl_prev = 0;
   } else {
      pnl_new  = Pnl + 1;
      Pnl_prev = Pnl;
   }

   if( pnl_new < Max_panel ) {
      Pnl = pnl_new;
      nw = n_windows();
      wtp = tail_wtal(lal_wtal());    
      if( wtp != NULL ) {
         iw = 0;
         while( wtp != NULL ) {
            (Xvsf->w)[Pnl][iw].wt = wtp->wt;
            window_sense(wtp->wt);    
            ++iw; wtp = wtp->prev;
         }
         set_First_and_Last_wt();
         apply_all_wt_0((PFV_I) set_must_redraw);
         rc = wt_to_open;
      } 
      if( Pnl_prev >= 0 ) {
         for( iw = 0; iw <= Pnl_prev; iw++ ) {
            (Xvsf->w)[Pnl_prev][iw].wt = UNDEF;
         }
         if( ltrace ) fprintf(stderr,"%s: Hiding panel %d\n",R,Pnl_prev);
         xvs_hide_panel(Pnl_prev);
         if( ltrace ) sleep(ltrace_sleep);
      }
      if( ltrace ) fprintf(stderr,"%s: Displaying panel %d\n",R,Pnl);
      xvs_show_panel(Pnl);
      if( ltrace ) sleep(ltrace_sleep);
   } else {
      fprintf(stderr,"%s: Cannot open new panel\n",R);
   }
   if( ltrace ) fprintf(stderr,"%s: rc=%d\n",R,rc);
   return rc;
}


/*=========================================================================
Repositions first nw windows in panel. 
=========================================================================*/
void xvs_repanel(int nw) {
   PWTBIND          wtp;
   WT               wt;
   int              iw,    ix,   iy;

   wtp = tail_wtal(lal_wtal());    
   if( wtp != NULL ) {
      iw = 0;
      while( wtp != NULL ) {
         (Xvsf->w)[Pnl][iw].wt = wtp->wt;
         window_sense(wtp->wt);    
         ++iw; wtp = wtp->prev;
      }
   }
   apply_all_wt_0((PFV_I) set_must_redraw);
   set_First_and_Last_wt();
}

/*=========================================================================
Re-senses visible windows
=========================================================================*/
void xvs_resense(void) {
   int   i;
   if( Pnl >= 0 ) {
      for( i = 0; i <= Pnl; i++ )  {
         window_sense((Xvsf->w)[Pnl][i].wt);
      }
   }
}

/*=========================================================================
Close single window.
=========================================================================*/
int xvs_close_wt(WT wt, int reposition_mouse) {
   static char R[] = "xvs_close_wt";
   int    ltrace = Ltrace;

   int    rc;
   int    on_wt,   oi,   iw;


/* Compute index of window to be activated after closing this one. */

   oi = index_from_wt(wt);
   on_wt = n_windows();
   if( ltrace ) {
      fprintf(stderr,"%s: Number of windows prior to close=%d\n",R,on_wt);
      fflush(stderr);
      fprintf(stderr,"%s: Index of window to be closed=%d\n",R,oi);
      fflush(stderr);
      fprintf(stderr,"%s: Pnl=%d  Pnl_prev=%d\n",R,Pnl,Pnl_prev);
      fflush(stderr);
   }
   Pnl_prev = Pnl;

   if( free_wt(wt) ) {
      if( ltrace ) {
         fprintf(stderr,"%s: Freed wt=%d\n",R,wt); fflush(stderr);
      }
      if( scope_mode == SINGLE_MODE ) {
         lal_pop();
         scope_mode = PANEL_MODE;
      } 

      lal_set();
      free_XYF_storage(wt);
      if( wds[wt].winf.pwf.pxywinf ) {
         free_pxywinf(wds[wt].winf.pwf.pxywinf);
      }

      wds[wt].status = WINDOW_NOT_ACTIVE;

      if( Pnl_prev >= 0 ) {
         for( iw = 0; iw <= Pnl_prev; iw++ ) {
            (Xvsf->w)[Pnl_prev][iw].wt = UNDEF;
         }
         xvs_hide_panel(Pnl_prev);
      }
      Pnl = n_windows() - 1;
      if( Pnl >= 0 ) {
         xvs_show_panel(Pnl);
      }

      if( last_S_wt == wt ) {
         last_S_wt = UNDEF;
      }
      (void) vs_repanel();

      if( reposition_mouse ) {
         if( on_wt > 1  ) {
            centre_mouse_in_wt(wt_from_index(oi == on_wt ? oi - 1 : oi));
         } else {
            centre_mouse_in_wt(scope_wt);
         }
      }

      rc = 0;
   }
   else {
      rc = -1;
   }
   return rc;
}

/*=========================================================================
Close window list.  Other xvs routines are special cases.
=========================================================================*/
int xvs_close_Picked_wt_list(void) {
   static char R[] = "xvs_close_Picked_wt_list";
   int    ltrace = Ltrace;

   int    rc = 0,  rcfree;
   int    on_wt,   oi,   i,  ntoclose,  nclose;

   int    wt = UNDEF;

   on_wt = n_windows();
   ntoclose = Picked_wt_list.n;

   if( ltrace ) {
      fprintf(stderr,"%s: Number of windows prior to close=%d\n",R,on_wt);
      fprintf(stderr,"%s: Number of windows to be closed=%d\n",R,ntoclose);
      fprintf(stderr,"%s: Pnl=%d  Pnl_prev=%d\n",R,Pnl,Pnl_prev);
   }

   if( !ntoclose ) {
      return 0;
   }

/* Compute index of window to be activated after closing this one. */
   oi = index_from_wt(wt);

   Pnl_prev = Pnl;

   for( i = 0, nclose = 0; i < ntoclose; i++ ) {
      wt = Picked_wt_list.v[i];
      rcfree = free_wt(wt);
      if( ltrace ) fprintf(stderr,"xvs_close_Picked_wt_list: free_wt(%d) returns %d\n",
         wt,rcfree);
      if( rcfree ) {
         ++nclose;
         if( ltrace ) {
            fprintf(stderr,"%s: Freed wt=%d nclose=%d\n",R,wt,nclose); 
         }
         if( scope_mode == SINGLE_MODE ) {
            lal_pop();
            scope_mode = PANEL_MODE;
         } 

         lal_set();
         free_XYF_storage(wt);
         if( wds[wt].winf.pwf.pxywinf ) {
            free_pxywinf(wds[wt].winf.pwf.pxywinf);
         }

         wds[wt].status = WINDOW_NOT_ACTIVE;

         if( last_S_wt == wt ) {
            last_S_wt = UNDEF;
         }
      }
   }
   if( Pnl_prev >= 0 ) {
      if( ltrace ) fprintf(stderr,"xvs_close_Picked_wt_list: Hiding panel %d\n",Pnl_prev);
      for( i = 0; i <= Pnl_prev; i++ ) {
         (Xvsf->w)[Pnl_prev][i].wt = UNDEF;
      }
      xvs_hide_panel(Pnl_prev);
   }
   Pnl = n_windows() - 1;
   if( Pnl >= 0 ) {
      if( ltrace ) fprintf(stderr,"xvs_close_Picked_wt_list: Showing panel %d\n",Pnl);
      xvs_show_panel(Pnl);
   }

   (void) vs_repanel();
   if( n_windows() ) {
      centre_mouse_in_wt(wt_from_index(1));
   }
   if( ltrace ) {
      fprintf(stderr,"%s: Number of windows after close=%d\n",R,n_windows());
      fprintf(stderr,"%s: Number of windows to be closed=%d\n",R,nclose);
      fprintf(stderr,"%s: Pnl=%d  Pnl_prev=%d\n",R,Pnl,Pnl_prev);
   }
   return 0;
}

/*=========================================================================
Close all windows.
=========================================================================*/
int xvs_close_all_wt(int reposition_mouse) {
   static char R[] = "xvs_close_all_wt";
   int        ltrace = Ltrace;

   PWTBIND    p;
   int        iw;

   if( ltrace ) {
      fprintf(stderr,"xvs_close_all_wt: n_windows()=%d  Pnl=%d  Pnl_prev=%d\n",
         n_windows(),Pnl,Pnl_prev);
      fprintf(stderr,"xvs_close_all_wt: lal_dump() on entry ...\n");
      lal_dump();
   }
   if( scope_mode == SINGLE_MODE ) {
      assert(Pnl == 0);
      if( ltrace ) {
         fprintf(stderr,"xvs_close_all_wt: Going to PANEL_MODE.\n");
      }
      lal_pop();
      scope_mode = PANEL_MODE;
   } 
   Pnl_prev = Pnl;

   for( p = tail_wtal(wtal); p != NULL; p = p->prev) {
      if( ltrace ) {
         fprintf(stderr,"xvs_close_all_wt: Closing %d.\n",p->wt);
      }
      free_XYF_storage(p->wt);
      if( wds[p->wt].winf.pwf.pxywinf ) {
         free_pxywinf(wds[p->wt].winf.pwf.pxywinf);
      }
      wds[p->wt].status = WINDOW_NOT_ACTIVE;
   }
   while ( p = tail_wtal(wtal) ) {
      if( ltrace ) {
         fprintf(stderr,"xvs_close_all_wt: Freeing %d.\n",p->wt);
      }
      free_wt(p->wt);
   }
   if( Pnl_prev >= 0 ) {
      for( iw = 0; iw <= Pnl_prev; iw++ ) {
         (Xvsf->w)[Pnl_prev][iw].wt = UNDEF;
      }
      if( ltrace ) fprintf(stderr,"%s: Hiding panel %d\n",R,Pnl_prev);
      xvs_hide_panel(Pnl_prev);
   }
   Pnl = n_windows() - 1;
   assert(Pnl == -1);

   lal_set();
   last_S_wt = scope_wt;
   S_wt = -1;
   if( reposition_mouse ) {
      if( mouse_is_near_wt(scope_wt,0.2,3.0) ) {
         centre_mouse_in_wt(scope_wt);
      }
   }
   if( ltrace ) {
      fprintf(stderr,"%s: lal_dump() on exit ...\n",R);
      lal_dump();
      fprintf(stderr,"%s: S_wt=%d last_S_wt=%d\n",R,S_wt,last_S_wt);
   }
   return 0;
}

/*=========================================================================
Close window list
=========================================================================*/
int xvs_close_wt_list(IVECN wtl) {
   static char R[] = "xvs_close_wt_list";
   int        ltrace = Ltrace;

   PWTBIND    p;
   int        iw,  wt;

   if( ltrace ) {
      fprintf(stderr,"xvs_close_wt_list: n_windows()=%d  Pnl=%d  Pnl_prev=%d\n",
         n_windows(),Pnl,Pnl_prev);
      fprintf(stderr,"xvs_close_wt_list: lal_dump() on entry ...\n");
      lal_dump();
   }
   if( scope_mode == SINGLE_MODE ) {
      assert(Pnl == 0);
      if( ltrace ) {
         fprintf(stderr,"xvs_close_all_wt: Going to PANEL_MODE.\n");
      }
      lal_pop();
      scope_mode = PANEL_MODE;
   } 
   Pnl_prev = Pnl;

   for( iw = 0; iw < wtl.n; iw ++) {
      wt = wtl.v[iw];
      if( ltrace ) {
         fprintf(stderr,"xvs_close_wt_list: Closing %d.\n",wt);
      }
      free_XYF_storage(wt);
      if( wds[wt].winf.pwf.pxywinf ) {
         free_pxywinf(wds[wt].winf.pwf.pxywinf);
      }
      free_wt(wt);
      wds[wt].status = WINDOW_NOT_ACTIVE;
   }

   if( Pnl_prev >= 0 ) {
      for( iw = 0; iw <= Pnl_prev; iw++ ) {
         (Xvsf->w)[Pnl_prev][iw].wt = UNDEF;
      }
      xvs_hide_panel(Pnl_prev);
   }
   Pnl = n_windows() - 1;
   if( Pnl >= 0 ) {
      xvs_show_panel(Pnl);
   }

   lal_set();
   last_S_wt = scope_wt;
   S_wt = -1;
   if( mouse_is_near_wt(scope_wt,0.2,3.0) ) {
      centre_mouse_in_wt(scope_wt);
   }
   if( ltrace ) {
      fprintf(stderr,"%s: lal_dump() on exit ...\n",R);
      lal_dump();
   }
   return 0;
}


/*=========================================================================
Helper functions for window initialization; minimal error checking.
=========================================================================*/
void wds_init_1(WT wt,int winno,char *name,int owner,char *wd) {
   wds[wt].sgi_winno = winno;
   wds[wt].name      = strdup(name);
   wds[wt].namelen   = strlen(name);
   wds[wt].owner     = owner;
   wds[wt].wd        = strdup(wd);
   wds[wt].wdlen     = strlen(wd);
   wds[wt].status    = WINDOW_ACTIVE;
}

void wds_init_2(WT wt,DQUAD dq,int scale_option,int max_lxy,int max_nevents) {
   wds[wt].pf.pxyf = make_pxyf(dq,scale_option,MAX_LXY,max_nevents);
   wds[wt].winf.pwf.pxywinf = make_pxywinf();
   link_xywinf_to_xyf(wt);
}

/*=========================================================================
Dump synopsis of contents of server.
=========================================================================*/
int xvs_dump(char *s) {
   static char R[] = "xvs_dump";
   int      rc, i;

   fprintf(stderr,"++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
   fprintf(stderr,"xvs_dump: Invoked from %s\n",s);
   fprintf(stderr,"xvs_dump: Active panel: %d\n",Pnl);
   fprintf(stderr,"xvs_dump: There are %d windows currently active\n",n_windows());
   fprintf(stderr,"xvs_dump: Total number of events stored: %d\n",n_events(wtal));
   fprintf(stderr,"xvs_dump: scope_mode %d\n",scope_mode);
   fprintf(stderr,"\n");
#define F  (Xvsf)->main
   fprintf(stderr,"xvs_dump: Main form X Window ID: %x\n",Xvsf->main->window);
   fprintf(stderr,"xvs_dump: Main form geometry (%d,%d) %d x %d\n",F->x,F->y,F->w,F->h);
#undef  F
#define IQ wds[WT_FORM].winf.wiq
   fprintf(stderr,"xvs_dump: scope iquad (user coords): [%d %d %d %d]\n",
      IQ.x1,IQ.x2,IQ.y1,IQ.y2);
#undef  IQ
   for( i = 0; i < max_windows; i++ ) {
      if( wds[i].status != WINDOW_NOT_ACTIVE ) {
        window_stats(i);
      }
   }
   (void) rdump_wtal();
   (void) lal_dump();

   return 0;
}

/*=========================================================================
Insert SDF data set into server.  If data name is new, as side effect append 
and open new GL canvas to main panel and begin visualizing data.
=========================================================================*/
int xvs_insert_SDFDS(PSDFDS the) {
   static char R[] = "xvs_insert_SDFDS";
   int         ltrace = Ltrace;
   WT          wt;
   PXYDVEC     data;
   double      time;
   int         tag = 0;
   int         i, rc = 0;

   double   zero_tol = 1.0e-60; 

   if( ltrace ) {
      dump_SDFDS(stderr,the,"xvs_insert_SDFDS",0);
   }

   wt = wt_from_SDFDS(the);
   if( wt < 0 ) {
      fprintf(stderr,"xvs_insert_SDFDS: Cannot insert data\n");
      goto Return;
   }
   data = make_XYDVEC(the->dsize);
   time = the->time;
   if( !data ) {
      fprintf(stderr,"xvs_insert_SDFDS: make_XYDVEC(%d) failed\n",the->dsize);
      goto Return;
   }
   if( the->csize != the->dsize ) {
      fprintf(stderr,"xvs_insert_SDFDS: the->csize (%d) <> the->dsize (%d)\n",
         the->csize,the->dsize);
      goto Return;
   }
   for( i = 0; i < the->csize; i++ ) {
         data->x[i] = the->coords[i];
         data->y[i] = the->data[i];
   }

#define     PP wds[wt].winf.pwf.pxywinf  
#define     PF wds[wt].pf.pxyf

/* Filter the data via current quad if necessary ... */

/* Replace ridiculously small by 0.0 */

   if( fabs(time) <= zero_tol ) time = 0.0;
   (void) dvnull_(data->x,&zero_tol,&(data->n));
   (void) dvnull_(data->y,&zero_tol,&(data->n));

   if( PP->domain_mode == FIXED_DOMAIN ) {
#define QUAD (PF->dqc)->dq 
         if( !(data = clip_x_only_XYDVEC(data,QUAD)) ) goto Return;
#undef  QUAD
   }
   if( PF->scale_option == SCALE_AS_LOADED ) {

      if( ltrace ) {
         fprintf(stderr,"xvs_insert_SDFDS:: Computing new limits.\n");
         fprintf(stderr,"Old limits: "); dump_dquad(PF->limits);
      }

      PF->limits = limits_with_new_data(wt,data);

      if( ltrace ) {
         fprintf(stderr,"New limits: "); dump_dquad(PF->limits);
      }

      PF->limits_computed = LIMITS_COMPUTED;
      if( PF->dqb != (PDQBIND) NULL ) free(PF->dqb);
      PF->dqb = make_pdqbind(PF->limits,NULL,NULL);
      if( PP->domain_mode == FLOATING_DOMAIN ) {
         dqc_from_dqb(wt);
         set_must_remap(wt);
         set_must_redraw(wt);
      }
   }

   rc = insert_with_merge_xydvec(wt,data,time,tag);
   (void) lal_set();

   Last_data_wt = wt;
   if( ltrace ) fprintf(stderr,"%s: Last_data_wt=%d\n",R,Last_data_wt);

Return:
   return rc;
}

/*=========================================================================
wt lookup from SDFDS. New wt created and inserted into active list if 
name not found.
=========================================================================*/
WT wt_from_SDFDS(PSDFDS the) {
   int      ltrace = Ltrace;
   WT       wt = -2; 
   DQUAD    dq;
   static char *cwd = (char *) NULL;

   if( !the ) goto Return;

   wt = wt_from_name(the->pname);
   if( ltrace ) {
      fprintf(stderr,"wt_from_SDFDS: the->pname=%s wt_from_name(%s)=%d\n",
         the->pname,the->pname,wt);
   }
   if( wt < 0 ) {
      wt = new_wt();
      lal_set();
      if( wt < 0 ) {
         fprintf(stderr,"wt_from_SDFDS: Cannot open new window '%s'\n",the->pname);
         goto Return;
      } else {
         if( ltrace ) lal_dump();
      }
      dq = dquad_SDFDS(the);
      if( cwd ) free(cwd);
      cwd = getcwd(NULL,BUFLEN);
      if( ltrace ) {
         fprintf(stderr,"wt_from_SDFDS: cwd=%s\n",cwd);
      }
      wds_init_1(wt,0,the->pname,0,cwd);
      wds_init_2(wt,dq,SCALE_AS_LOADED,0,0);
      if( ltrace ) {
         fprintf(stderr,"wt_from_SDFDS:  Initial window stats for wt=%d\n",wt);
         window_stats(wt);
      }

      if( xvs_open_and_repanel(wt) >= 0 ) {
         window_sense(wt);
         centre_mouse_in_wt(wt);
/*
         lal_set();
    
         (void) window_ortho2(wt);
         (void) redraw_window_in_panel(wt);

         if( Default_range_gauge_on_wt_open ) {
            save_S_wt = S_wt; S_wt = wt;
            (void) handle_range_gauge_menu(Default_range_gauge_type);
            S_wt = save_S_wt;
         }
    
         set_First_and_Last_wt(); 
*/
      }
   } 
   if( ltrace ) {
      fprintf(stderr,"wt_from_SDFDS: the->pname=%s  wt=%d\n",the->pname,wt);
   }
Return:
   return wt;
}

/*=========================================================================
gl lookup from wt.
=========================================================================*/
FL_OBJECT *gl_from_wt(WT wt) {
   FL_OBJECT *gl = (FL_OBJECT *) NULL;
   int        ltrace = 0;
   int        i;
   if( Pnl >= 0 ) {
      for( i = 0; i <= Pnl; i++ ) {
         if( (Xvsf->w)[Pnl][i].wt == wt ) {
            gl = (Xvsf->w)[Pnl][i].gl;
            if( Ltrace ) {
               fprintf(stderr,"gl_from_wt: Located wt=%d in panel=%d frame=%d\n",
                  wt,Pnl,i);
            }
            goto Return;
         }
      }
   } 
Return:
   return gl;
}

/*=========================================================================
wt lookup from gl.
=========================================================================*/
WT wt_from_gl(FL_OBJECT *gl) {
   WT       wt = -1;
   int      ltrace = Ltrace;
   int      i;
   
   if( Pnl >= 0 ) {
      for( i = 0; i <= Pnl; i++ ) {
         if( (Xvsf->w)[Pnl][i].gl == gl ) {
            wt = (Xvsf->w)[Pnl][i].wt;
            if( Ltrace ) {
               fprintf(stderr,"wt_from_gl: Located wt=%d in panel=%d frame=%d\n",
                  wt,Pnl,i);
               goto Return;
            }
         }
      }
   } 
Return:
   return wt;
}

/*=========================================================================
font struct lookup from gl.
=========================================================================*/
/*
xvs_dl_font *f_from_gl(FL_OBJECT *gl) {
   xvs_dl_font *f = (xvs_dl_font *) NULL;
   int          ltrace = 1;
   int          i;
   
   if( Pnl >= 0 ) {
      for( i = 0; i <= Pnl; i++ ) {
         if( (Xvsf->w)[Pnl][i].gl == gl ) {
            f = (Xvsf->w)[Pnl][i].f;
            if( !f ) {
               (Xvsf->w)[Pnl][i].f = xvs_dl_font_make();
            }
         }
      }
   } 
Return:
   return f;
}
*/

/*=========================================================================
font struct lookup/set from wt.
=========================================================================*/
/*
xvs_dl_font *f_from_wt(WT wt) {
   xvs_dl_font *f = (xvs_dl_font *) NULL;
   int          ltrace = 1;
   int          i;
   
   if( ltrace ) {
      fprintf(stderr,"f_from_wt: wt=%d  Pnl=%d\n",wt,Pnl);
   }
   if( Pnl >= 0 ) {
      for( i = 0; i <= Pnl; i++ ) {
         if( (Xvsf->w)[Pnl][i].wt == wt ) {
            if( ltrace ) {
               fprintf(stderr,"f_from_wt: Found wt=%d at [Pnl=%d][i=%d]\n",wt,Pnl,i);
            }
            f = (Xvsf->w)[Pnl][i].f;
            if( !f ) {
               f = xvs_dl_font_make();
               (Xvsf->w)[Pnl][i].f = f;
            }
            goto Return;
         }
      }
   } 
Return:
   return f;
}
*/

int xerror_handler(Display *display,XErrorEvent *event) {
   perror("xerror_handler");
   exit(1);
   return 0;
}

/*=========================================================================
Uses the 'import' program to capture window in one of a variety of 
standard graphics format.  Note: 'import' is very slow
even for small captured screen areas.
=========================================================================*/
int xvs_wt_import(WT wt,char *fmt) {
   static char R[] = "xvs_wt_import";
   int ltrace = Ltrace;
   int rc = 0;
   char *lfmt = (char *) NULL;
   FL_OBJECT *gl;

   if( !is_active_wt(wt) ) return 0;

   if( !(gl = gl_from_wt(wt)) ) {
      fprintf(stderr,"%s: FATAL: Could not find gl corresponding to wt=%d\n",
         R,wt);
      exit(1);
   }

   if( !fmt ) {
      fprintf(stderr,"%s: 'fmt' is NULL.\n",R);
      return rc;
   }

   lfmt = str2lc(fmt);
   if(      !strcmp("gif",fmt) ) {
      fprintf(stderr,"%s: GIF capture not completely implemented yet\n",R);
      sprintf(B,"import -window 0x%x -silent xvs.gif",FL_ObjWin(gl));
      fprintf(stderr,"%s: Executing '%s'\n",R,B);
      system(B);
   } else if( !strcmp("ps",fmt) ) {
      fprintf(stderr,"%s: PS capture not completely implemented yet\n",R);
      sprintf(B,"import -window 0x%x -silent xvs.ps",FL_ObjWin(gl));
      fprintf(stderr,"%s: Executing '%s'\n",R,B);
      system(B);
   } else if( !strcmp("jpeg",fmt) || !strcmp("jpg",fmt) ) {
      fprintf(stderr,"%s: JPEG capture not completely implemented yet\n",R);
      sprintf(B,"import -window 0x%x -silent -border -quality 100 xvs.jpeg",
         FL_ObjWin(gl));
      fprintf(stderr,"%s: Executing '%s'\n",R,B);
      system(B);
/*      xvs_wt_to_jpeg(wt,"xvsgl.jpeg",100); */
   } else {
      if( ltrace ) fprintf(stderr,"%s: FATAL: Unrecognized fmt='%s'\n",R,fmt);
      exit(1);
   }
   if( lfmt ) free(lfmt);
   return rc;
}


/*=========================================================================
Slightly modifed version of 'scr_save' from 'DV' (F. Pretorius)
=========================================================================*/
void xvs_wt_to_jpeg(WT wt,char *dir,char *name) {
   static char R[] = "xvs_wt_to_jpeg";
   int vp[4],image_size,width,height,row;
   char *image;
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   JSAMPROW rp;
   FILE *file=0;
   FL_OBJECT *gl;
   int jpeg_qf = 100;
   int ltrace = 1;

   if( !is_active_wt(wt) ) return;

   if( !(gl = gl_from_wt(wt)) ) {
      fprintf(stderr,"%s: FATAL: Could not find gl corresponding to wt=%d\n",
         R,wt);
      exit(1);
   }  
   fl_activate_glcanvas(gl);
   sprintf(B,"%s/%s_%04d.jpg",dir,name,get_display_index(wt));

   // obtain the image dimension, [vp]=[x,y,width,height]
   glGetIntegerv(GL_VIEWPORT,vp);
   if( ltrace ) {
      fprintf(stderr,"%s: Saving %dx%d to '%s'\n",R,vp[2],vp[3],B);
   }

   width=vp[2];
   height=vp[3];

   image_size=sizeof(char)*3*width*height;
   if (!(image=(char *)malloc(image_size)))
   {
      fprintf(stderr,"%s: Out of memory ... can't save the image\n",R);
      return;
   }

   glPixelStorei(GL_PACK_ALIGNMENT,1);
   glReadPixels(0,0,width,height,GL_RGB,GL_UNSIGNED_BYTE,image);

   // following 'libjpeg.doc' to use the jpeg library:
   cinfo.err=jpeg_std_error(&jerr);
   jpeg_create_compress(&cinfo);
   if (!(file=fopen(B,"wb")))
   {
      fprintf(stderr,"%s: cannot open file %s for writing\n",R,B);
      goto cleanup;
   }
   jpeg_stdio_dest(&cinfo,file); 
   cinfo.image_width=width;
   cinfo.image_height=height;
   cinfo.input_components=3;
   cinfo.in_color_space=JCS_RGB;
   jpeg_set_defaults(&cinfo);
   jpeg_set_quality(&cinfo,jpeg_qf,0);
   jpeg_start_compress(&cinfo, TRUE);
   
   // the jpeg library wants the image from top-to-bottom,
   // glReadPixels() returns bottom-to-top, so invert below:
 
   for(row=height-1; row>=0; row--)
   {
      rp=(JSAMPROW)(&image[row*width*3]);
      jpeg_write_scanlines(&cinfo,&rp,1);
   }
   
   jpeg_finish_compress(&cinfo);
   jpeg_destroy_compress(&cinfo);

cleanup:
   if (file) fclose(file);
   free(image);
   return;
}

/* Feb 11 2011:  Added check of environment vbl XVS_SAVE_JPEGS ... */
void xvs_jpeg_seq_to_mpeg(char *dir,char *name,int nseq) {
   static char R[] = "xvs_jpeg_seq_to_mpeg";
   static char Param[] = ".mpeg_encode.param";
   FILE       *fparam;
   int         i;
   int         ltrace = 0;
   int         remove_param_file = 1;
   int         remove_jpeg_files = 1;

   if( getenv("XVS_SAVE_JPEGS") ) {
      remove_jpeg_files = 0;
      remove_param_file = 0;
   }
   if( ltrace ) {
      fprintf(stderr,"%s: dir='%s'\n",R,dir);
      fprintf(stderr,"%s: name='%s'\n",R,name);
      fprintf(stderr,"%s: nseq=%d\n",R,nseq);
      fprintf(stderr,"%s: remove_jpeg_files=%d\n",R,remove_jpeg_files);
      fprintf(stderr,"%s: remove_param_file=%d\n",R,remove_param_file);
   }
   if( !xvs_push_wd_cd(dir) ) {
      sprintf(B,"Could not cd to '%s'",dir);
      scope_mess_and_bell(B,1);
      return;
   }
   sprintf(B,"%s.mpg",name);
   if( Mpeg_fname ) free(Mpeg_fname);
   Mpeg_fname = strdup(B);
   if( ltrace ) {
      fprintf(stderr,"%s: Output to '%s'\n",R,Mpeg_fname);
   }

   if( !(fparam = fopen(Param,"w")) ) {
      sprintf(B,"Could not open '%s' for write",Param);
      scope_mess_and_bell(B,1);
   }
   fprintf(fparam,"PATTERN     IBBPBBPBBP\n");
   fprintf(fparam,"OUTPUT      %s\n",Mpeg_fname);
   fprintf(fparam,"BASE_FILE_FORMAT  JPEG\n");
   fprintf(fparam,"INPUT_CONVERT  *\n");
   fprintf(fparam,"GOP_SIZE 16\n");
   fprintf(fparam,"SLICES_PER_FRAME  1\n");
   fprintf(fparam,"INPUT_DIR %s\n",dir);
   fprintf(fparam,"INPUT\n");
   fprintf(fparam,"%s_*.jpg  [0001-%04d]\n",name,nseq);
   fprintf(fparam,"END_INPUT\n");
   fprintf(fparam,"PIXEL    HALF\n");
   fprintf(fparam,"RANGE    10\n");
   fprintf(fparam,"PSEARCH_ALG LOGARITHMIC\n");
   fprintf(fparam,"BSEARCH_ALG CROSS2\n");
   fprintf(fparam,"IQSCALE     1\n");
   fprintf(fparam,"PQSCALE     1\n");
   fprintf(fparam,"BQSCALE     1\n");
   fprintf(fparam,"REFERENCE_FRAME   ORIGINAL\n");
   fprintf(fparam,"FRAME_RATE 30\n");
   fprintf(fparam,"FORCE_ENCODE_LAST_FRAME\n");
   
   fclose(fparam);
   if( ltrace ) {
      sprintf(B,"cat %s",Param);
      system(B);
   }
   sprintf(B,"mpeg_encode %s",Param);
   sprintf(B1,"Executing '%s'",B);
   if( ltrace ) {
      fprintf(stderr,"%s: %s\n",R,B1);
   }
   scope_mess_perm(B1);
   system(B);
   scope_mess_perm(" ");
   if( remove_param_file) sprintf(B,"/bin/rm %s\n",Param);
   system(B);
   if( remove_jpeg_files ) {
      for( i = 1; i <= nseq; i++ ) {
         sprintf(B,"/bin/rm %s_%04d.jpg\n",name,i);
         system(B);
      }
   }

   xvs_pop_wd();
}

/*=========================================================================
Driver for xvs/ser.
=========================================================================*/
int main(int argc,char **argv) {
   int       xvsf_ss,  xvsport;     

   int       ltrace = Ltrace;
   char      buffer[BUFLEN];

   int       i,   j,  k,  xuc,  yuc, mfb = Main_form_border;

   P = basename(argv[0]);

   if( argc > 1 ) xvs_control(argc,argv);

   xvsport = int_env("XVSPORT",DEFAULT_XVSPORT);
   if( Ltrace ) fprintf(stderr,"%s: Using port %d\n",P,xvsport);

   if( (xvsf_ss = ser0_start_xvs(xvsport)) > 0 ) {
      fprintf(stderr,"%s: Starting service on port %d\n",P,xvsport);
   } else {
      fprintf(stderr,"\n");
      fprintf(stderr,
         "If you want to run xvs, try setting the environment variable\n");
      fprintf(stderr,"XVSPORT to some value other than %d ",xvsport);
      fprintf(stderr,"BEFORE invoking it.\n\n");
      fprintf(stderr,"For example, in tcsh/csh\n\n");
      fprintf(stderr,"  setenv XVSPORT %d\n  xvs\n\n",xvsport+1);
      fprintf(stderr,"or in bash\n\n");
      fprintf(stderr,"  export XVSPORT=%d\n  xvs\n\n",xvsport+1);
      fprintf(stderr,
         "You will then have to set XVSPORT identically in any shell\n");
      fprintf(stderr,
         "in which you invoke, e.g., sdftoxvs.\n\n");
      fprintf(stderr,"Type 'xvs help' for client-side commands\n");
      exit(1);
   }

/* Set global working directory .. */
   if( !(Cwd=getcwd(NULL,BUFLEN)) ) {
      fprintf(stderr,"%s: Couldn't get current working directory.\n",P);
      Cwd = strdup("/");
   } 
   if( Ltrace ) {
      fprintf(stderr,"%s: Cwd='%s'\n",P,Cwd);
   }


/* Attempt to start service ... */

/* Third argument is application class name, fourth specifies how to parse 
   application-specific resources, fifth specifies number of entries in option list */

   fl_initialize(&argc,argv,"OpenGLWindow",0,0);

/* Form set up ... */
   fl_set_border_width(-2);  
   fl_set_app_nomainform(ON);
/* Create main form and keep a copy for subsequent resizes ... */
   Xvsf = create_form_main();
   Ixvsf = create_form_main();
   Form_w_min = Ixvsf->main->w;
   Form_h_min = Ixvsf->main->w;
   fl_set_form_atclose(Xvsf->main,at_close_main,NULL);
   if( Ltrace ) { 
      fprintf(stderr,"Created Xvsf\n");
      dump_Xvsf(Ixvsf,"xvs: Original");
      dump_Xvsf(Xvsf,"xvs: Current");
   }

   fl_set_form_position(Xvsf->main,600,50);
   fl_set_form_minsize(Xvsf->main,Xvsf->main->w,Xvsf->main->h);

/* Create secondary forms, and register atclose callbacks ... */

   Xvsf_info     = create_form_info();
   Xvsf_markers  = create_form_markers();
   Xvsf_view     = create_form_view();
   Xvsf_file     = create_form_file();
   Xvsf_ops      = create_form_ops();
   Xvsf_mpeg     = create_form_mpeg();
   Xvsf_sm       = create_form_sm();
   Xvsf_browse   = create_form_browse();
   Xvsf_label    = create_form_label();

   fl_set_form_atclose(Xvsf_info->info,at_close_hide,NULL);
   fl_set_form_atclose(Xvsf_markers->markers,at_close_markers,NULL);
   fl_set_form_atclose(Xvsf_view->view,at_close_hide,NULL);
   fl_set_form_atclose(Xvsf_file->file,at_close_hide,NULL);
   fl_set_form_atclose(Xvsf_ops->ops,at_close_hide,NULL);
   fl_set_form_atclose(Xvsf_mpeg->mpeg,at_close_mpeg,NULL);
   fl_set_form_atclose(Xvsf_sm->sm,at_close_sm,NULL);
   fl_set_form_atclose(Xvsf_browse->browse,at_close_browse,NULL);
   fl_set_form_atclose(Xvsf_label->label,at_close_hide,NULL);

/* Create tabbed folder, and add secondary forms to it ...*/

   Xvsf_control = create_form_control();
   fl_set_form_atclose(Xvsf_control->control,at_close_control,NULL);

/* Attempt to capture keystrokes in form; not successful thus far ... */
// fl_register_raw_callback(Xvsf_control->control, KeyPressMask, raw_cb_control);

   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"Ops",Xvsf_ops->ops);
   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"View",Xvsf_view->view);
   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"File",Xvsf_file->file);
   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"Markers",Xvsf_markers->markers);
   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"Browse",Xvsf_browse->browse);
   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"Label",Xvsf_label->label);
   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"Mpeg",Xvsf_mpeg->mpeg);
   fl_addto_tabfolder(Xvsf_control->tabfolder_control,"SM",Xvsf_sm->sm);

/* Register idle callback ... */
   fl_set_idle_delta(1);
   fl_set_idle_callback(cb_idle, 0);   


/* Need more than the default 32 pop-ups ... */
   fl_setpup_maxpup(256);

/* Initialize major data structures ... */
   all_ds_init();

/* Visualization related initialization ... */
   vis_init();

/* Initialize widgets ... */
   Xvsf_widgets_init();

/* Initialize "hidden" gl_main window which defines globlal GLXContext */
   xvs_init_gl_main();

// fl_show_form(Xvsf->main,FL_PLACE_SIZE,FL_FULLBORDER,"xvs");
   xvs_set_form_geometry(0);

/* Position form in upper left corner of screen, throw in an extra 8
 * pixels horizontally to compensate for decorations ... */

   fl_set_form_geometry(Xvsf->main,
      fl_scrw - (Xvsf->main->w + mfb + 8), mfb,
      Xvsf->main->w,Xvsf->main->h);
// fl_show_form(Xvsf->main,FL_PLACE_SIZE,FL_FULLBORDER,"xvs");
   fl_show_form(Xvsf->main,FL_PLACE_POSITION,FL_FULLBORDER,"xvs");

   if( Ltrace ) {
      dump_xvs_panel_array(Xvsf->w);
   }

/* Register SIGALRM handler */
   signal(SIGALRM,handle_alarm);

/* Register I/O callback. */
// fl_add_io_callback(xvsf_ss,FL_READ,(FL_IO_CALLBACK) cb_xvs_socket_1,Xvsf);
   fl_add_io_callback(xvsf_ss,FL_READ,(FL_IO_CALLBACK) cb_xvs_socket,Xvsf);
// fl_add_io_callback(xvsf_ss,FL_READ,(FL_IO_CALLBACK) cb_xvs_socket_new,Xvsf);
  
   if( Ltrace ) {
      fprintf(stderr,"Screen width:  %d\n",fl_scrw);
      fprintf(stderr,"Screen height: %d\n",fl_scrh);
   }

/* Initialize Form_state and hand control over to X-forms ... */
   Curr_p = 0;
   Form_state = FORM_STATE_NORMAL;

   fprintf(stderr,"\n");
   fprintf(stderr,"%s: For help/documentation see\n\n",P);
   fprintf(stderr,"http://laplace.physics.ubc.ca/Doc/xvs/\n");
   fprintf(stderr,"\n");
   fl_do_forms();

   return 0;
}
