#include "project.h"

/*=========================================================================
Implements top-level control functions via command line args.

help
killall
kill <name>
closeall
close <name>
get <name> <index>
setpar <par> <value>
getpar <par> 
go <name> <index>

See cb_xvs_socket_1a in cbio.c for server-side routines.
=========================================================================*/ 
enum tlabel_values {
   T_NONE,
   T_GNUPLOT,
   T_MPUT
};

int xvs_control(int argc, char **argv) {
   static char R[] = "xvs_control";
   FILE       *stream = (FILE *) NULL;
   PSDFDS      mess = (PSDFDS) NULL, sdfds = (PSDFDS) NULL;
   int         ltrace = Ltrace;
   int         i, iarg2, narg2, rc;
   char       *cmd;

   int         n;
   double      time, *x = (double *) NULL, *y = (double *) NULL;

   int         tlabel_option;


/* Hack for fact that vsxynt can close stream under certain circumstances ... */
   int         cmd_closes_stream = OFF;

   for( i = 1; i < argc; i++ ) {
      if( ltrace ) fprintf(stderr,"%s: argv[%d]='%s'\n",R,i,argv[i]);
   }

/* Commands are case-insensitive ... */

   cmd = str2lc(argv[1]);
   if( ltrace ) fprintf(stderr,"%s: cmd='%s'\n",R,cmd);
   if(        !strcmp(cmd,"help")       ||
              !strcmp(cmd,"h")              ) {
      xvs_help();
      exit(1);
   }

   if( !(stream = xvs_stream(XVS_OPEN)) ) exit(-1);

   if(        !strcmp(cmd,"stop")       ||
              !strcmp(cmd,"shutdown")   ||
              !strcmp(cmd,"quit")   ||
              !strcmp(cmd,"exit")           ) {
      (void) xvs_control_send_message(stream,"stop","'stop' failed.",P);

   } else if( !strcmp(cmd,"killall")    ||
              !strcmp(cmd,"ka")         || 
              !strcmp(cmd,"closeall")   ||
              !strcmp(cmd,"ca")              ) {
      (void) xvs_control_send_message(stream,"killall","'killall' failed.",P);

   } else if( !strcmp(cmd,"kill")       ||
              !strcmp(cmd,"close")           ) {
      narg2 = argc - 2;
      argv += 2;
      if( narg2 <= 0 ) {
         fprintf(stderr,"usage: %s kill <name> [<name> ...]\n",P);
         fprintf(stderr,"Current names:\n\n");
         xvs_control_list(stream);
         exit(1);
      }
      for( iarg2 = 0; iarg2 < narg2 ; iarg2++ ) {
         sprintf(B,"kill <%s>",argv[iarg2]);
         if( xvs_control_send_message(stream,B,"'kill' failed.",P) ) {
            if( !xvs_recv_status(stream,B1) ) {
               fprintf(stderr,"%s: %s\n",P,B1);
            }
         }
      }

   } else if( !strcmp(cmd,"raise")      ||
              !strcmp(cmd,"pop")             ) {
      (void) xvs_control_send_message(stream,"raise","'raise' failed.",P);

   } else if( !strcmp(cmd,"list")      ||
              !strcmp(cmd,"ls")              ) {
      (void) xvs_control_list(stream);

   } else if( !strcmp(cmd,"nwin")      ||
              !strcmp(cmd,"nw")              ) {
      (void) xvs_control_send_message(stream,"nwin","'nwin' failed.",P);
      xvs_recv_buf(stream,B);
      fprintf(stderr,"%s\n",B);

   } else if( !strcmp(cmd,"save")       ||
              !strcmp(cmd,"sv")              ) {
      narg2 = argc - 2;
      argv += 2;
      if( narg2 <= 0 ) {
         fprintf(stderr,"usage: %s save <name> [<name> ...]\n\n",P);
         fprintf(stderr,"Current names:\n\n");
         xvs_control_list(stream);
         exit(1);
      }
      for( iarg2 = 0; iarg2 < narg2 ; iarg2++ ) {
         sprintf(B,"save <%s>",argv[iarg2]);
         if( xvs_control_send_message(stream,B,"'save' failed.",P) ) {
            if( xvs_recv_status(stream,B1) ) {
               getcwd(B,BUFLEN);
               if( xvs_send_buf(stream,B) ) {
                  if( xvs_recv_status(stream,B1) ) {
                     fprintf(stderr,"%s\n",B1);
                  } else {
                     fprintf(stderr,"%s: save %s failed.\n",P,argv[iarg2]);
                  }
               }
            } else {
               fprintf(stderr,"%s: %s\n",P,B1);
            }
         }
      }
   } else if( !strcmp(cmd,"saveall")    ||
              !strcmp(cmd,"sa")              ) {
      if( xvs_control_send_message(stream,"saveall","'saveall' failed.",P) ) {
         if( xvs_recv_status(stream,B) ) {
            getcwd(B,BUFLEN);
            xvs_send_buf(stream,B);
            while( xvs_recv_status(stream,B) ) {
               fprintf(stderr,"%s\n",B);
            }
         } else {
            fprintf(stderr,"%s: No windows to save\n",P);
         }
      }
   } else if( !strcmp(cmd,"get")         ||
              !strcmp(cmd,"ge")              ) {
      if( ((argc != 4) && (argc != 5)) ||
          ((argc > 4 ) && strcmp(argv[4],"t_gnuplot")) &&
          ((argc > 4 ) && strcmp(argv[4],"t_mput")) ) {
         fprintf(stderr,"usage: %s get <name> <index> [t_gnuplot | t_mput] \n",P);
         exit(1);
      }

      tlabel_option = T_NONE;
      if( argc > 4 && !strcmp(argv[4],"t_gnuplot") ) {
         tlabel_option = T_GNUPLOT;
      }
      if( argc > 4 && !strcmp(argv[4],"t_mput") ) {
         tlabel_option = T_MPUT;
      }

      sprintf(B,"get <%s> %s",argv[2],argv[3]);
      if( xvs_control_send_message(stream,B,"'get' failed.",P) ) {
         if( xvs_recv_status(stream,B) ) {
            sdfds = new_SDFDS();
            rewind(stream);
            if( read_SDFDS(stream,sdfds) ) {
               switch( tlabel_option) {
               case T_NONE:
                  for( i = 0; i < sdfds->dsize; i++ ) {
                     printf("%24.16e %24.16e\n",sdfds->coords[i],sdfds->data[i]);
                  }
                  break;
               case T_GNUPLOT:
                  for( i = 0; i < sdfds->dsize; i++ ) {
                     printf("%24.16e %24.16e %24.16e\n",
                        sdfds->time,sdfds->coords[i],sdfds->data[i]);
                  }
                  break;
               case T_MPUT:
                  printf("%24.16e %d\n",sdfds->time,sdfds->dsize);
                  for( i = 0; i < sdfds->dsize; i++ ) {
                     printf("%24.16e %24.16e\n",sdfds->coords[i],sdfds->data[i]);
                  }
                  break;
               default:
                  break;
               }
            } else {
               fprintf(stderr,"%s: read_SDFDS() failed!\n",P);
            }
            if( sdfds ) free_SDFDS(sdfds);
         } else {
            fprintf(stderr,"%s: %s\n",P,B);
         }
      }
   } else if( !strcmp(cmd,"put")          ||        
              !strcmp(cmd,"pu")              ) {
      if( (argc != 4) || (sscanf(argv[3],"%lf",&time) != 1) ) {
         fprintf(stderr,"usage: %s put <name> <time>\n",P);
         exit(1);
      }
      l_dvvfscanf(stdin,&x,&y,&n);
      if( n ) {
         vsxynt(argv[2],time,x,y,n);
         cmd_closes_stream = ON;
      } else {
         fprintf(stderr,"%s: No data from standard input.  Operation aborted.\n",
            R);
      }
   } else if( !strcmp(cmd,"go")               ) {
      if( argc != 4 ) {
         fprintf(stderr,"usage: %s go <name> <index>\n",P);
         exit(1);
      }
      sprintf(B,"go <%s> %s",argv[2],argv[3]);
      (void) xvs_control_send_message(stream,B,"'go' failed.",P);
      if( !xvs_recv_status(stream,B) ) fprintf(stderr,"%s: %s\n",P,B);
   } else if( !strcmp(cmd,"getpar")          ) {
      fprintf(stderr,"%s: Command '%s' not implemented yet.\n",P,cmd);
   } else if( !strcmp(cmd,"setpar")          ) {
      fprintf(stderr,"%s: Command '%s' not implemented yet.\n",P,cmd);
   } else {
      fprintf(stderr,"%s: Unrecognized command '%s'.\n",P,cmd);
   }

   if( !cmd_closes_stream) (void) xvs_stream(XVS_CLOSE);

   exit(0);

   return 1;
}

/*=========================================================================
Lists dataset names.
=========================================================================*/ 
int xvs_control_list(FILE *stream) {
   (void) xvs_control_send_message(stream,"list","'list' failed.",P);
   while( xvs_recv_status(stream,B) ) {
      fprintf(stderr,"%s\n",B);
   }
   return 1;
}


int xvs_help(void) {
   fprintf(stderr,"%s: Available commands:\n\n",P);
   fprintf(stderr,"    xvs help | h\n");
   fprintf(stderr,"    xvs killall | ka | closeall | ca\n");
   fprintf(stderr,"    xvs kill | close <name>\n");
   fprintf(stderr,"    xvs exit | shutdown | quit | stop\n");
   fprintf(stderr,"    xvs list | ls\n");
   fprintf(stderr,"    xvs raise | pop\n");
   fprintf(stderr,"    xvs nwin | nw\n");
   fprintf(stderr,"    xvs save | sv <name>\n");
   fprintf(stderr,"    xvs saveall | sa\n");
   fprintf(stderr,"    xvs get <name> <index> [t_gnuplot | t_mput] \n");
   fprintf(stderr,"    xvs put <name> <time>\n");
   fprintf(stderr,"    xvs go <name> <index>\n");
   fprintf(stderr,"\n%s: Note: 'xvs put' reads (x,y) pairs from standard input.\n",P);
   fprintf(stderr,"\n%s: For additional, preliminary help, see\n\n",P);
   fprintf(stderr,"    http://laplace.physics.ubc.ca/Doc/xvs/\n");
   return 0;
}
