#include "project.h"

PXYDVEC make_XYDVEC(int n)
{
#define       TYPE     PXYDVEC
#define       X        p->x
#define       Y        p->y
#define       N        p->n
   TYPE       p;

   if( n > 0 ) {
      if( (p = (TYPE) malloc(sizeof(XYDVEC))) != (TYPE) NULL) {
         N = n;
         if( (X = make_DVEC(n)) == NULL ) {
            printf("make_XYDVEC: x DVEC allocation failure.\n");
            p = NULL;
         }
         if( (Y = make_DVEC(n)) == NULL ) {
            if(X) free(X);
            printf("make_XYDVEC: y DVEC allocation failure.\n");
            p = NULL;
         }
/*       printf("make_XYDVEC: allocated DVEC of length %d.\n",N);  */
      }
   } else {
      p = NULL;
   }
   return p;
}

PXYDVEC free_XYDVEC(p)
   TYPE       p;
{
#define       FREE_IF_NOT_NULL(a) if( a != NULL ) free(a)
   if( p != (TYPE) NULL ) {
      FREE_IF_NOT_NULL(X);
      FREE_IF_NOT_NULL(Y);
      FREE_IF_NOT_NULL(p);
   }
   return (TYPE) NULL;
#undef        FREE_IF_NOT_NULL
}

PXYDVEC dump_XYDVEC(p,s)
   TYPE    p;
   char   *s;
{
   if( p != (TYPE) NULL ) {
      printf("dump_xydvec:: %s.\n",s);
      dvprint(X,N,"x"); dvprint(Y,N,"y");
   }
   return p;
}

PXYDVEC test_XYDVEC(int n)
{
   TYPE    p;

   if( (p = make_XYDVEC(n)) == (TYPE) NULL ) {
      err1("test_XYDVEC","malloc failure.");
   }
   X = DViota(X,n);
   Y = DViota(Y,n);

   return p;
#undef              N
#undef              X
#undef              Y
}


PXYDVEC clip_XYDVEC(p,dq_clip)
   PXYDVEC          p;
   DQUAD            dq_clip;
{
   PXYDVEC          pp;
   IVEC             flag;
   DPAIR            dp;
   int              i,   j,   n_pp;

   int              ltrace  =  OFF;

#define             N  p->n
#define             X  pp->x
#define             Y  pp->y
   if( N > 0 ) {
      if( !(flag = make_IVEC(N)) ) {
         err2("clip_XYDVEC","Can't create flag array.");
         return (PXYDVEC) NULL;
      }

      for( i = 0; i < N; i++ ) {
           dp = make_dpair(p->x[i],p->y[i]);
           flag[i] = is_dp_in_dq(dp,dq_clip);
      }
      n_pp = ivsum(flag,N);
      if( ltrace ) {
         printf("clip_XYDVEC:: %d elements in clipping quad.\n",n_pp);
      }

      if( n_pp > 0 ) {
         if( !(pp = make_XYDVEC(n_pp)) ) {
            err2("clip_XYDVEC","Can't create new XYDVEC.");
            return (PXYDVEC) NULL;
         }
         for( j = 0, i = 0; i < N; i++ ) {
            if( flag[i] ) {
               X[j]   = p->x[i];
               Y[j++] = p->y[i];
            }
         }
      } else {
         if( ltrace ) {
            err2("clip_XYDVEC","No XYDVEC elements in clipping quad.");
            err2("clip_XYDVEC","Deleting entire XYDVEC.");
         }
         pp = (PXYDVEC) NULL;
      }
      (void) free_IVEC(flag);
      (void) free_XYDVEC(p);

      return pp;
   } else {
      if( ltrace ) {
         err2("clip_XYDVEC","0 length XYDVEC supplied.");
      }
      return p;
   }
#undef              Y
#undef              X
#undef              N
}

XY_PDATA_AND_PTAG clip_XYDVEC_AND_TAG(PXYDVEC p,PXYTAGVEC tp,DQUAD dq_clip)
{
   XY_PDATA_AND_PTAG  result;

   PXYDVEC            pp;
   PXYTAGVEC          ptp;

   IVEC               flag;
   DPAIR              dp;
   int                i,   j,   n_pp;

   int                ltrace  =  OFF;

#define               N      p->n
#define               X      pp->x
#define               Y      pp->y
#define               NTAG   ptp->n
#define               TAG    ptp->tag
   result.data  = p;
   result.tdata = tp;

   if( N > 0 ) {
      if( !(flag = make_IVEC(N)) ) {
         err2("clip_XYDVEC_AND_TAG","Can't create flag array.");
         return result;
      }

      for( i = 0; i < N; i++ ) {
           dp = make_dpair(p->x[i],p->y[i]);
           flag[i] = is_dp_in_dq(dp,dq_clip);
      }
      n_pp = ivsum(flag,N);
      if( ltrace ) {
         printf("clip_XYDVEC_AND_TAG:: %d elements in clipping quad.\n",n_pp);
      }

      if( n_pp > 0 ) {
         if( !(pp = make_XYDVEC(n_pp)) ) {
            err2("clip_XYDVEC_AND_TAG","Can't create new XYDVEC.");
         } else {
            if( tp ) {
               if( !(ptp = make_XYTAGVEC(n_pp)) ) {
                  err2("clip_XYDVEC_AND_TAG","Can't create new XYTAGVEC.");
               } else {
                  for( j = 0, i = 0; i < N; i++ ) {
                     if( flag[i] ) {
                        X[j]   = p->x[i];
                        Y[j] = p->y[i];
                        TAG[j] = tp->tag[i];
                        j++;
                     }
                  }
               }
            } else {
               ptp = (PXYTAGVEC) NULL;
               for( j = 0, i = 0; i < N; i++ ) {
                  if( flag[i] ) {
                     X[j]   = p->x[i];
                     Y[j++] = p->y[i];
                  }
               }
            }
         }
      } else {
         if( ltrace ) {
            err2("clip_XYDVEC_AND_TAG","No XYDVEC elements in clipping quad.");
            err2("clip_XYDVEC_AND_TAG","Deleting entire XYDVEC.");
         }
         pp  = (PXYDVEC) NULL;
         ptp = (PXYTAGVEC) NULL;
      }
      (void) free_IVEC(flag);
      (void) free_XYDVEC(p);
      if( tp ) free_XYTAGVEC(tp);

      result.data  = pp;
      result.tdata = ptp;

   } else {
      if( ltrace ) {
         err2("clip_XYDVEC_AND_TAG","0 length XYDVEC supplied.");
      }
   }
   
   return result;
#undef              NTAG
#undef              TAG
#undef              Y
#undef              X
#undef              N
}


PXYDVEC clip_x_only_XYDVEC(p,dq_clip)
   PXYDVEC          p;
   DQUAD            dq_clip;
{
   PXYDVEC          pp;
   IVEC             flag;
   int              i,   j,   n_pp;

   int              ltrace  =  OFF;

#define             N  p->n
#define             X  pp->x
#define             Y  pp->y
   if( N > 0 ) {
      if( !(flag = make_IVEC(N)) ) {
         err2("clip_x_only_XYDVEC","Can't create flag array.");
         return (PXYDVEC) NULL;
      }

      for( i = 0; i < N; i++ ) {
           flag[i] = dq_clip.x1 <= p->x[i] && p->x[i] <= dq_clip.x2;
      }
      n_pp = ivsum(flag,N);
      if( ltrace ) {
         printf("clip_x_only_XYDVEC:: %d elements in clipping quad.\n",n_pp);
      }

      if( n_pp > 0 ) {
         if( !(pp = make_XYDVEC(n_pp)) ) {
            err2("clip_x_only_XYDVEC","Can't create new XYDVEC.");
            return (PXYDVEC) NULL;
         }
         for( j = 0, i = 0; i < N; i++ ) {
            if( flag[i] ) {
               X[j]   = p->x[i];
               Y[j++] = p->y[i];
            }
         }
      } else {
         if( ltrace ) {
            err2("clip_x_only_XYDVEC","No XYDVEC elements in clipping quad.");
            err2("clip_x_only_XYDVEC","Deleting entire XYDVEC.");
         }
         pp = (PXYDVEC) NULL;
      }
      (void) free_IVEC(flag);
      (void) free_XYDVEC(p);

      return pp;
   } else {
      if( ltrace ) {
         err2("clip_x_only_XYDVEC","0 length XYDVEC supplied.");
      }
      return p;
   }
#undef              Y
#undef              X
#undef              N
}


#undef        TYPE

void set_lxy(p)
   PXYF       p;
{
   PXYDVBIND  pp;

   int        lxy;
   int        ltrace = OFF;

   for( lxy = 0, pp = p->xyhead; pp != NULL; pp = pp->next, ++lxy);
   p->lxy = lxy;
   if( ltrace ) {  
      fprintf(stderr,"set_lxy: lxy: %d.\n",lxy);
   }
}

void set_lxy_wt(WT wt) {
   unmake_xy_ring(wt);
   set_lxy(wds[wt].pf.pxyf);
   restore_ring_state(wt);
}

int get_lxy_wt(WT wt) {
   if( is_active_wt(wt) ) {
      set_lxy_wt(wt);
      return(wds[wt].pf.pxyf->lxy);
   } else {
      return(0);
   }
}


int full_xyf(p)
   PXYF      p;
{
   return ((p->lxy) == (p->max_lxy));
}

void reindex_XYDVlist(head,start)
   PXYDVBIND     head;
   int           start;
{
   int           ltrace = OFF;

   for( ; head != NULL; head = head->next ) {
      head->index = start++;
   }
   if( ltrace ) {
      fprintf(stderr,"reindex_XYDVlist: last index: %d.\n",start - 1);
   }
}

PXYDVBIND make_XYDVBIND()
{
   PXYDVBIND   lp;

   lp = (PXYDVBIND) malloc(sizeof(XYDVBIND));
   if( lp ) {
      lp->data  = NULL;
      lp->tdata = NULL;
      lp->prev  = NULL;
      lp->next  = NULL;
   }
   return lp;
}

void free_XYDVBIND(PXYDVBIND p) {
   if( p != NULL ){
      if( p->data ) {
         free(p->data->x);
         free(p->data->y);
      }
      if( p->tdata ) {
         free((p->tdata)->tag);
      }
      free(p);
   }
}

PXYDVBIND link_PXYDVBIND(PXYDVBIND p1,PXYDVBIND p2) {
   if( p1 ) {
      p1->next = p2;
   } 
   if( p2 ) {
      p2->prev = p1;
   }
   return(p2);
}

PXYDVBIND clone_XYDVLIST(PXYDVBIND p) {
   PXYDVBIND    pp = NULL, newp;
   for( ; p ; p = p->next ) {
      if( !pp ) {
         pp = link_PXYDVBIND(pp,copy_PXYDVBIND(p));
         newp = pp;
      } else {
         pp = link_PXYDVBIND(pp,copy_PXYDVBIND(p));
      }
   }
   return newp;
}

PXYDVBIND_VECN  make_PXYDVBIND_VECN(int n) {
   PXYDVBIND_VECN   the = {(PXYDVBIND *) NULL,0};

   if( n > 0 ) {
      the.v = (PXYDVBIND *) malloc(n * sizeof(PXYDVBIND));
      the.n = n;
   }

   return(the);
}

void free_PXYDVBIND_VECN(PXYDVBIND_VECN the) {
   if( the.v ) free(the.v);
}

void free_XYDVLIST(p)
   PXYDVBIND     p;
{
   PXYDVBIND     lp;
   int           ltrace = OFF, n = 0;

   for( ; p != NULL; ) {
      lp = p->next;
      free_XYDVBIND(p);
      ++n;
      p = lp;
   }
   if( ltrace ) {
      printf("free_XYDVLIST:: %d XY vectors freed.\n",n);
   }
}

PXYDVBIND double_PXYDVBIND(p)
   PXYDVBIND     p;
{
   int           n = (p->data)->n;

   if( p ) {
      p->data = free_XYDVEC(p->data);
      p->data = make_XYDVEC(2 * n);
      (p->data)->n = 2 * n;
      if( p->tdata ) {
         free_XYTAGVEC(p->tdata);
         p->tdata = make_XYTAGVEC(2 * n);
      }
   }
   return p;
}

PXYDVBIND expand_PXYDVBIND_nfold(p,n_expand)
   PXYDVBIND     p;
   int           n_expand;
{
   int           n = (p->data)->n;

   if( p && (n_expand >= 0) ) {
      p->data = free_XYDVEC(p->data);
      p->data = make_XYDVEC(n_expand * n);
      (p->data)->n = n_expand * n;
      if( p->tdata ) {
         free_XYTAGVEC(p->tdata);
         p->tdata = make_XYTAGVEC(n_expand * n);
      }
   }
   return p;
}

PXYDVBIND copy_PXYDVBIND(PXYDVBIND p) {
   PXYDVBIND   newp = NULL;
#define      N        (p->data)->n
#define      COPY(a)   newp->a = p->a

   if( p ) {
      newp = make_XYDVBIND();
      COPY(index);
      COPY(time);
      COPY(tag);
      newp->data = make_XYDVEC(N);
      DVcopy(p->data->x,newp->data->x,N);
      DVcopy(p->data->y,newp->data->y,N);
      if( p->tdata ) {
         newp->tdata = make_XYTAGVEC(p->tdata->n);
         IVcopy(p->tdata->tag,newp->tdata->tag,p->tdata->n);
      } else {
         newp->tdata = NULL;
      }
   }
   return(newp);
#undef       N
#undef       COPY
}

PXYDVBIND expand_PXYDVBIND_by_n(p,dn)
   PXYDVBIND     p;
   int           dn;
{
   PXYDVEC       old_data;
   PXYTAGVEC     old_tdata;
   int           n = (p->data)->n;

   if( p && (dn > 0) ) {
      old_data = p->data;
      p->data = make_XYDVEC(n + dn);
      (p->data)->n = n + dn;
      dvcopy_(old_data->x,(p->data)->x,&n);
      dvcopy_(old_data->y,(p->data)->y,&n);
      free_XYDVEC(old_data);
      if( p->tdata ) {
         old_tdata = p->tdata;
         p->tdata = make_XYTAGVEC(n + dn);
         ivcopy_(old_tdata->tag,(p->tdata)->tag,&n);
         free_XYTAGVEC(old_tdata);
      }
   }
   return p;
}

PXYDVBIND resize_PXYDVBIND_to_n(PXYDVBIND p,int newn) {
   PXYDVEC       old_data;
   PXYTAGVEC     old_tdata;
   int           n = (p->data)->n;

   if( p && (newn > 0) ) {
      old_data = p->data;
      p->data = make_XYDVEC(newn);
      (p->data)->n = newn;
      free_XYDVEC(old_data);
      old_tdata = p->tdata ? p->tdata : NULL;
      p->tdata = make_XYTAGVEC(newn);
      (p->tdata)->n = newn;
      if( old_tdata ) free_XYTAGVEC(old_tdata);
   }
   return p;
}


int n_XYDV(wt)
   WT           wt;
{
   PXYDVBIND    p;
   int          i;
#define FP wds[wt].pf.pxyf

   unmake_xy_ring(wt);
   i = n_PXYDVBIND(FP->xyhead);
   restore_ring_state(wt);
   return i;

#undef  FP
}

int n_PXYDVBIND(p)
   PXYDVBIND    p;
{
   int          i;
   int          ltrace = OFF;

   for( i = 0; p != NULL; ++i, p = p->next );
   if( ltrace ) printf("n_PXYDVBIND: returning %d.\n",i);
   return i;
}

PXYDVBIND nth_PXYDVBIND(PXYDVBIND phead,int n) {
   PXYDVBIND     p = phead;

   if( n < 1 ) return NULL;

   while( p != NULL ) {
      if( n == 1 ) return p;
      p = p->next; 
      n--;
   }
   return p;
}

double xmin_PXYDVBIND(PXYDVBIND p) {
   double  xmin = 0;
   if( p ) {
      xmin = dvmin(p->data->x,p->data->n);
      for( p = p->next; p != NULL; p = p->next ) {
         xmin = dmin(xmin,dvmin(p->data->x,p->data->n));
      }
   }
   return(xmin);
}

double xmax_PXYDVBIND(PXYDVBIND p) {
   double  xmax = 0;
   if( p ) {
      xmax = dvmax(p->data->x,p->data->n);
      for( p = p->next; p != NULL; p = p->next ) {
         xmax = dmax(xmax,dvmax(p->data->x,p->data->n));
      }
   }
   fprintf(stderr,"xmax_PXYDVBIND: xmax: %g.\n",xmax);
   return(xmax);
}

double ymin_PXYDVBIND(PXYDVBIND p) {
   double  ymin = 0;
   if( p ) {
      ymin = dvmin(p->data->y,p->data->n);
      for( p = p->next; p != NULL; p = p->next ) {
         ymin = dmin(ymin,dvmin(p->data->y,p->data->n));
      }
   }
   return(ymin);
}

double ymax_PXYDVBIND(PXYDVBIND p) {
   double  ymax = 0;
   if( p ) {
      ymax = dvmax(p->data->y,p->data->n);
      for( p = p->next; p != NULL; p = p->next ) {
         ymax = dmax(ymax,dvmax(p->data->y,p->data->n));
      }
   }
   return(ymax);
}

int n_events_wt(wt)
   WT           wt;
{
#define FP wds[wt].pf.pxyf
   PXYDVBIND    p;
   int          accum;
   
   int          ltrace = OFF;

   unmake_xy_ring(wt);
   accum = 0;
   for( p = FP->xyhead; p != NULL; p = p->next ){
      accum += (p->data)->n;
   }
   if( ltrace ) {
      fprintf(stderr,"n_events_wt: %d events in %d.\n",accum,wt);    
   }
   restore_ring_state(wt);
   return accum;
#undef  FP
}

#define FP wds[wt].pf.pxyf
int is_uniform_dx_wt(WT wt, double *pdx) {
   static char R[] = "is_uniform_dx_wt";
   int         ltrace   = Ltrace;
   int         rval     = 0;
   int         first    = 1;
   PXYDVBIND   p = (PXYDVBIND) NULL;
   double      dx,  ndx;
   int         i;

   *pdx = -1.0;

   unmake_xy_ring(wt);
   if( is_active_wt(wt) ) {
       rval = is_uniform_dx_pxydv(FP->xyhead,pdx);
   }
Return:
   if( ltrace ) {
      fprintf(stderr,"%s: wt=%d return value=%d dx=%g\n",R,wt,rval,*pdx);
   }
   restore_ring_state(wt);
   return rval;
}
#undef  FP

#define X  (p->data)->x
#define Y  (p->data)->y
#define N  (p->data)->n
int is_uniform_dx_pxydv(PXYDVBIND xyhead, double *pdx) {
   static char R[] = "is_uniform_dx_pxydv";
   int         ltrace   = Ltrace;
   int         rval     = 0;
   int         first    = 1;
   PXYDVBIND   p = (PXYDVBIND) NULL;
   double      dx,  ndx;
   int         i;

   *pdx = -1.0;

   for( p = xyhead; p != NULL; p = p->next ) {
      if( N > 1 ) {
         for( i = 0; i < N - 1; i++ ) {
            if( first ) {
               dx = X[i+1] - X[i];
               first = 0;
            }
            ndx = X[i+1] - X[i];
            if( !fuzz_equal(dx,ndx,Compare_fuzz) ) goto Return;
         }
         rval = 1;
         *pdx = dx;
      }
   }
Return:
   if( ltrace ) {
      fprintf(stderr,"%s: Return value=%d dx=%g\n",R,rval,*pdx);
   }
   return rval;
}
#undef  X
#undef  Y
#undef  N


/* prepend / append functionality currently not up to insert level. */

int prepend_xydvec(wt,data,time)
   WT          wt;
   PXYDVEC     data;
   double      time;
{
   PXYDVBIND   p,   pp;

#define FP wds[wt].pf.pxyf

   if( FP->max_lxy == 0 ){
      printf("prepend_xydvec: max_lxy = 0, can't store data.\n");
      return(-1);
   }
   p = (PXYDVBIND) malloc(sizeof(XYDVBIND));
   if( p == NULL ) {
      perror("prepend_xydvec");
      return(-1);
   }
   p->index = 1;
   p->time  = time;
   p->data  = data;
   p->prev  = (PXYDVBIND) NULL;
   if( FP->lxy == 0 ) {
      p->next  = (PXYDVBIND) NULL;
      FP->xyhead = FP->xytail = p;
   } else {
      for( pp = FP->xyhead; pp != NULL; pp = pp->next ) {
         ++(pp->index);
      }
      p->next = FP->xyhead;
      (FP->xyhead)->prev = p;
      FP->xyhead = p;
   }
   (FP->nevents) += data->n;
   return(++(FP->lxy));
#undef FP
}

int  append_xydvec(wt,data,time)
   WT          wt;
   PXYDVEC     data;
   double      time;
{
   PXYDVBIND   p,   pp;

#define FP wds[wt].pf.pxyf

   if( FP->max_lxy == 0 ){
      printf("append_xydvec: max_lxy = 0, can't store data.\n");
      return(-1);
   }
   p = (PXYDVBIND) malloc(sizeof(XYDVBIND));
   if( p == NULL ) {
      perror("append_xydvec");
      return(-1);
   }
   p->time  = time;
   p->data  = data;
   p->next  = (PXYDVBIND) NULL;
   if( FP->lxy == 0 ) {
      p->index = 1;
      p->prev  = (PXYDVBIND) NULL;
      FP->xyhead = FP->xytail = p;
   } else {
      p->index = ((FP->xytail)->index) + 1;
      p->prev  = FP->xytail;
      (FP->xytail)->next = p;
      FP->xytail = p;
   }
   (FP->nevents) += data->n;
   return(++(FP->lxy));
#undef FP
}

int insert_xydvec(wt,data,time)
   WT         wt;
   PXYDVEC    data;
   double     time;
{
   PXYDVBIND  p, pp, reindex;
   int        noins, full;
   int        delete, delete_tail = 0, delete_head = 1;

#define FP wds[wt].pf.pxyf

/* printf("insert_xydvec: pre_insertion scan_all_pxy for tag %d.\n",wt);  */
/* scan_all_pxy(wt);  */

   unmake_xy_ring(wt);

   if( FP->max_lxy == 0 ){
      printf("insert_xydvec: max_lxy = 0, can't store data.\n");
      return(-1);
   }
   p = (PXYDVBIND) malloc(sizeof(XYDVBIND));
   if( p == NULL ) {
      perror("insert_xydvec");
      return(-1);
   }
   full = full_xyf(FP);
/* if( full ) {  */
/*    printf("insert_xydvec: window %d is full.\n",wt);  */
/* }  */
   p->time = time;
   p->data = data;

   if( FP->lxy == 0 ) {
/*    printf("insert_xydvec: Inserting %f as new list.\n",time);  */
      p->index = 1;
      p->next = p->prev = (PXYDVBIND) NULL;
      FP->xyhead = FP->xytail = p;
      reindex = NULL;

/* Identical times are ordered on a first received basis. */

   } else if( time <  FP->xyhead->time ) {
/*    printf("insert_xydvec: Inserting %f at head of list.\n",time);      */
      p->index = 1;
      p->prev = (PXYDVBIND) NULL;
      p->next = reindex = FP->xyhead;
      FP->xyhead = p;
      (p->next)->prev = p;
      if( full )
         delete = delete_tail;
   } else if( time >= FP->xytail->time ) {
/*    printf("insert_xydvec: Inserting %f at tail of list.\n",time);      */
      p->index = (FP->xytail->index)+1;
      p->prev = FP->xytail;
      p->next = reindex = (PXYDVBIND) NULL;
      FP->xytail = p; (p->prev)->next = p;
      if( full )
         delete = delete_head;
   } else {
       for( pp = FP->xyhead, noins = 1; (pp != NULL) && noins; pp = pp->next ) {
         if( ((pp->time) <= time)  &&  (time <= (pp->next->time)) ) {
/*          printf("insert_xydvec: Inserting %f between %f and %f.\n",time,
                   pp->time,pp->next->time);   */
            p->index = (pp->index)+1;
            p->prev  = pp; p->next = reindex = pp->next;
            pp->next = p;  (p->next)->prev = p;
            noins = 0;
         }
      }
      if( full )
         delete = delete_head;
   }

/* p still points to most recent data. */

/* Adding assignment of xycurr ...      */

   FP->xymr = FP->xycurr = p;
   FP->new_data = ON;

   if( full ) {
      if( delete == delete_tail ) {
/*       printf("insert_xydvec: deleting tail.\n");  */
         p = FP->xytail;
         FP->xytail = p->prev;
         (FP->xytail)->next = NULL;
         FP->nevents -= p->data->n;
         free_XYDVBIND(p);
      } else if( delete == delete_head ) {
/*       printf("insert_xydvec: deleting head.\n");  */
         p = FP->xyhead;
         FP->xyhead = p->next;
         (FP->xyhead)->prev = NULL;
         FP->nevents -= p->data->n;
         free_XYDVBIND(p);
      }
   }

   FP->xyfirst = FP->xyhead;
   FP->xylast  = FP->xytail;

/* printf("insert_xydvec: post_insertion scan_all_pxy for tag %d.\n",wt);  */
/* scan_all_pxy(wt);  */

   (void) reindex_XYDVlist(FP->xyhead,1);

/* printf("insert_xydvec: post_reindex scan_all_pxy for tag %d.\n",wt);  */
/* scan_all_pxy(wt);  */

   (FP->nevents) += data->n;
/* printf("insert_xyvec: nevents: %d   n_events_wt(): %d.\n",  */
/*        FP->nevents,n_events_wt(wt));  */

   restore_ring_state(wt);

   if( full ) {
      return(FP->lxy);
   } else {
      return(++(FP->lxy));
   }
#undef FP
}

int insert_with_merge_xydvec(wt,data,time,tag)
   WT         wt;
   PXYDVEC    data;
   double     time;
   int        tag;
{
   PXYDVBIND  p, pp, reindex;
   int        noins, full;
   int        delete, delete_tail = 0, delete_head = 1;

   int        ltrace = OFF;
   int        full_trace = OFF;

#define FP wds[wt].pf.pxyf
#define PP wds[wt].winf.pwf.pxywinf

   unmake_xy_ring(wt);

   if( FP->max_lxy == 0 ){
      printf("insert_with_merge_xydvec: max_lxy = 0, can't store data.\n");
      return(-1);
   }
   p = (PXYDVBIND) malloc(sizeof(XYDVBIND));
   if( p == NULL ) {
      perror("insert_xydvec");
      return(-1);
   }
   full = full_xyf(FP);

   p->time = time;
   p->data = data;
   p->tag  = tag;
   p->tdata = NULL;

/* Thin data if auto thinning is enabled ... */

   if( get_auto_thin(wt) ) {
      p = thin_PXYDVBIND(p,Thinning_angle);
   }


   if( ltrace ) {
      printf("\ninsert_with_merge_xydvec: to insert %d events.\n",p->data->n);
   }

   if( FP->lxy == 0 ) {
      if( ltrace ) {
         printf("insert_with_merge_xydvec: Inserting %f as new list.\n",time);
      }
      p->index = 1;
      p->next = p->prev = (PXYDVBIND) NULL;
      FP->xyhead = FP->xytail = p;
      reindex = NULL;

/* Identical times are merged. */

#define EQTIME(a) insert_fuzz_equal(time,a->time)

   } else if( time < FP->xyhead->time || EQTIME(FP->xyhead) ) {
      if( ltrace ) {
         printf("insert_with_merge_xydvec: Inserting %f at head of list.\n",
                time);
      }
      if( EQTIME(FP->xyhead) ) {
         if( ltrace ) {
            printf("insert_with_merge_xydvec: Merging with head.\n");
         }
         full = 0;
         p = merge_pxydv(FP->xyhead,p);
         FP->xyhead = p;
         if( FP->lxy == 1 ) {
            FP->xytail = FP->xyhead;
         }
         if( p->next != NULL ) {
            (p->next)->prev = p;
         }
      } else {
         if( ltrace ) {
            printf("insert_with_merge_xydvec: New insertion\n");
         }
         p->index = 1;
         p->prev = (PXYDVBIND) NULL;
         p->next = reindex = FP->xyhead;
         FP->xyhead = p;
         (p->next)->prev = p;
      }
      if( full )
         delete = delete_tail;
   } else if( time > FP->xytail->time || EQTIME(FP->xytail) ) {
      if( ltrace ) {
         printf("insert_with_merge_xydvec: Inserting %f at tail of list.\n",
                time);
      }
      if( EQTIME(FP->xytail) ) {
         full = 0;
         p = merge_pxydv(FP->xytail,p);
         if( FP->xytail->prev != NULL ) {
            (FP->xytail->prev)->next = p;
         }
         FP->xytail = p;
         if( FP->lxy == 1 ) {
            FP->xyhead = FP->xytail;
            (FP->xyhead)->next = NULL;
         }
      } else {
         p->index = (FP->xytail->index)+1;
         p->prev = FP->xytail;
         p->next = reindex = (PXYDVBIND) NULL;
         FP->xytail = p;
         (p->prev)->next = p;
      }
      if( full )
         delete = delete_head;
   } else {
       for( pp = FP->xyhead, noins = 1; (pp != NULL) && noins; pp = pp->next ) {
         if( ((pp->time) < time || EQTIME(pp)) && time < (pp->next->time) ) {
            if( !EQTIME(pp->next) ) {
               if( ltrace ) {
                  printf(
                  "insert_with_merge_xydvec: Inserting %f between %f and %f.\n",
                  time,pp->time,pp->next->time);
               }
               if( EQTIME(pp) ) {
                  p = merge_pxydv(pp,p);
                  full = 0;
               } else {
                  p->index = (pp->index)+1;
                  p->prev  = pp; p->next = reindex = pp->next;
                  pp->next = p;
               }
               (p->next)->prev = p;
               (p->prev)->next = p;
               noins = 0;
            }
         }
      }
      if( full )
         delete = delete_head;
   }

#undef   EQTIME

/* p still points to most recent data ... also update xycurr . */

   FP->xymr = FP->xycurr = p;
   FP->new_data = ON;

/* If enabled, sort and remove duplicates from AMRD-generated data ... */
   if( Auto_pamr_sort ) {
      p = pxydv_sort_pamr(p);
   }

/* Automatically change marker type predicated on # of data points ... */

   if( p->data->n >= MARKER_HI_N && is_crowded_marker_type(PP->marker_type) ) {
      PP->marker_type = DEFAULT_MARKER_TYPE_HI_N + get_marker_multi_mode(wt);
   }

/* Tag by discretization scale if enabled ... */
   if( Auto_tag_by_dx ) {
      p = pxydv_tag_by_dx_single(p);
      set_marker_multi_mode(wt);
   }

   if( full ) {
      if( delete == delete_tail ) {
         if( ltrace ) {
            printf("insert_with_merge_xydvec: deleting tail.\n");
         }
         p = FP->xytail;
         FP->xytail = p->prev;
         (FP->xytail)->next = NULL;
         FP->nevents -= p->data->n;
         free_XYDVBIND(p);
      } else if( delete == delete_head ) {
         if( ltrace ) {
            printf("insert_with_merge_xydvec: deleting head.\n");
         }
         p = FP->xyhead;
         FP->xyhead = p->next;
         (FP->xyhead)->prev = NULL;
         FP->nevents -= p->data->n;
         free_XYDVBIND(p);
      }
   }

/* Is this side effect always desirable? */

   FP->xyfirst = FP->xyhead;
   FP->xylast  = FP->xytail;

   if( full_trace  ) {
      printf("insert_xydvec: post_insertion scan_all_pxy for tag %d.\n",wt);
      scan_all_pxy(wt);
   }

   (void) reindex_XYDVlist(FP->xyhead,1);

   if( full_trace ) {
      printf("insert_xydvec: post_reindex scan_all_pxy for tag %d.\n",wt);
      scan_all_pxy(wt);
   }

   if( ltrace ) {
      printf("insert_with_merge_xydvec: nevents: %d   n_events_wt(): %d.\n",
             FP->nevents,n_events_wt(wt));
   }

   set_lxy(FP);

   FP->nevents = n_events_wt(wt);
   restore_ring_state(wt);

   return FP->lxy;

#undef FP
}

/* Following routine assumes ring state has been unset ... */

void replace_PXYDVBIND_wt(WT wt,PXYDVBIND p,PXYDVBIND newp) {
#define    FP          wds[wt].pf.pxyf
#define    NEVENTS     FP->nevents
#define    PUPDATE(a)  if(  p == wds[wt].pf.pxyf->a ) wds[wt].pf.pxyf->a = newp
   PXYDVBIND    lp;
   int          dn;
   if( is_active_wt(wt) ) {
      if( p ) {
         if( newp ) {
            for( lp = FP->xyhead; lp != NULL; lp = lp->next ) {
               if( lp == p ) {
                  dn = newp->data->n - p->data->n;
                  NEVENTS += dn;
                  newp->prev = lp->prev;
                  if( lp->prev ) {
                     lp->prev->next = newp;
                  }
                  newp->next = lp->next;
                  if( lp->next ) {
                     lp->next->prev = newp;
                  }
                  PUPDATE(xyhead);
                  PUPDATE(xytail);
                  PUPDATE(xymr);
                  PUPDATE(xyfirst);
                  PUPDATE(xycurr);
                  PUPDATE(xylast);
                  return;
               }
            }
         } else {
            fprintf(stderr,"replace_PXYDVBIND_wt: Null newp.\n");
         }
      } else {
         fprintf(stderr,"replace_PXYDVBIND_wt: Null p.\n");
      }
   } else {
      fprintf(stderr,"replace_PXYDVBIND_wt: window tag %d not active.\n",wt);
   }
#undef           FP
#undef           NEVENTS
#undef           PUPDATE
}

/* Frees storage associated  with XYF structure currently
   allocated in VS_open (make_pxyf(), make_dquad(), make_pxywinf())
   and VS_data_in. Assumes all storage for data can be reached via
   linking from xyhead. */

/* Calls in this routine represent a current bottleneck */

void free_XYF_storage(wt)
   WT      wt;
{
   PXYDVBIND      p;
   PDQBIND        pp, pp_next;
   int            ixy, ipxy, idqs;
   
   int            ltrace = OFF;
   int            ltrace_full = OFF;
   int            ltrace_free = Ltrace * 1000;
   int            nfree = 0;

#define FP wds[wt].pf.pxyf
#define ASSERT_NULL(a,b) if( a != (b) NULL ) {printf("free_XYF_storage: a expected to be NULL.\n"); exit(1);}

   if( ltrace_full ) {
      scan_all_pxy(wt);    
   }

/* FP->nevents = n_events_wt(wt);  */
   unmake_xy_ring(wt);

   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freeing tag %d's storage.\n",wt); fflush(stderr); }
   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freeing tag %d's data.\n",wt); fflush(stderr); }
   for( ixy = 0, ipxy = 0, p = FP->xyhead; p != NULL; p = p->next, ++ixy ) {
         if( ltrace_free && ipxy && !(ipxy % ltrace_free) ) {
            printf("free_XYF_storage: ipxy=%d Freeing %d(%d).\n",ipxy,ixy,(p->data)->n);    
         }
         free( (p->data)->x ); free( (p->data)->y );
         if( p->tdata ) {
            free( (p->tdata)->tag );
         }
         (FP->nevents) -= (p->data)->n;
         if( (p->prev) != NULL) {
            free( p->prev ); ++ipxy;
         }
   }
   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freed tag %d's data.\n",wt); fflush(stderr); }
   assert((FP->nevents) == 0);
   if( FP->lxy > 0 ) {
      assert(ipxy == (FP->lxy - 1));
      free(FP->xytail);
   }
   assert(ixy == FP->lxy);
   if( ltrace ) {
      printf("free_XYF_storage: Freed storage from %d DVECs.\n",FP->lxy);     
   }

   for( idqs = 0, pp = FP->dqsz; pp != NULL; pp = pp_next ) {
      pp_next = pp->next;
      free(pp);
      ++idqs;
   }
   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freed storage from %d dq stack entries.\n", idqs);  fflush(stderr); }

/* ASSERT_NULL(FP->xyfirst,PXYDVBIND);  */
/* ASSERT_NULL(FP->xycurr,PXYDVBIND);  */
/* ASSERT_NULL(FP->xylast,PXYDVBIND);  */

   free(FP);
   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freed FP\n"); fflush(stderr); }
   free(wds[wt].winf.pwf.pxywinf);
   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freed ...pxywinf\n"); fflush(stderr); }
   free(wds[wt].name);
   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freed name\n"); fflush(stderr); }
   free(wds[wt].wd);
   if( ltrace ) { fprintf(stderr,"free_XYF_storage: Freed wd\n"); fflush(stderr); }
#undef FP
#undef ASSERT_NULL
}

DQUAD limits_with_new_data(wt,data)
   WT           wt;
   PXYDVEC      data;
{
extern
   double       dmin(),      dmax();

   double       lxmin,       lxmax,       lymin,        lymax;

#define L wds[wt].pf.pxyf->limits
   lxmin = L.x1; lxmax = L.x2; lymin = L.y1; lymax = L.y2;
/* printf("limits_with_new_data: Original limits %f %f %f %f.\n",  */
/*        lxmin,lxmax,lymin,lymax);  */

   if( data != (PXYDVEC) NULL ) {
      if( data->n > 0 ) {
         lxmin = dmin(lxmin,dvmin(data->x,data->n));
         lxmax = dmax(lxmax,dvmax(data->x,data->n));
         lymin = dmin(lymin,dvmin(data->y,data->n));
         lymax = dmax(lymax,dvmax(data->y,data->n));
      } else {
         printf("limits_with_new_data: 0 length vector in window (wt) %d.\n",wt);
      }
   }
/* printf("limits_with_new_data: Final limits %f %f %f %f.\n",lxmin,lxmax,  */
/*        lymin,lymax);  */
   return make_dquad(lxmin,lymin,lxmax,lymax);
#undef L
}

/* Returns total number of events stored in WTBIND list headed by p. */

int n_events(p)
   PWTBIND     p;
{
   int         accum;
#define        P        wds[p->wt].pf.pxyf
   for( accum = 0; p != NULL; accum += P->nevents, p = p->next);

   return accum;
#undef P
}

/* "Pings" window (if supplied tag is active), setting must_remap,
   new_domain flags if args non--zero. */

void ping_xywin(wt,must_remap,new_domain)
   WT         wt;
   int        must_remap, new_domain;
{
extern
   WINDOW     wds[];
#define       P            wds[wt].winf.pwf.pxywinf

   if( is_active_wt(wt) ) {
      if( must_remap ) {
         P->must_remap = 1;
      }
      if( new_domain ) {
         P->new_domain = 1;
      }
   }
#undef P
}

#define FP wds[wt].pf.pxyf

void set_xytail_wt(WT wt) {
   FP->xytail = FP->xyhead;
   while( FP->xytail && FP->xytail->next ) {
      FP->xytail = FP->xytail->next;
   }
}

void make_xy_ring(wt)
   WT        wt;
{
   PXYDVBIND p;
   int       i, imax = 50;

/* printf("make_xy_ring: wt: %d.\n",wt);  */
   FP->xyfirst = FP->xyhead;
   FP->xylast  = FP->xytail;

   if( FP->xylast != NULL ) {
      (FP->xylast)->next  = FP->xyfirst;
   }
   if( FP->xyfirst != NULL ) {
      (FP->xyfirst)->prev = FP->xylast;
   }
}

void unmake_xy_ring(wt)
   WT        wt;
{
   if( FP->xylast != NULL )
      (FP->xylast)->next  = NULL;
   if( FP->xyfirst != NULL )
      (FP->xyfirst)->prev = NULL;
}
#undef FP

PXYDVBIND dup_XYDVBIND(p)
   PXYDVBIND   p;
{
   PXYDVBIND   pp;

   pp = (PXYDVBIND) malloc(sizeof(XYDVBIND));
   *pp = *p;
   return pp;
}

void set_ring_state(wt)
   WT          wt;
{
extern
   WINDOW      wds[];
#define P      wds[wt].pf.pxyf
   P->ring_state = ON;
}


int get_ring_state(wt)
   WT          wt;
{
extern
   WINDOW      wds[];

   return P->ring_state;
}

void unset_ring_state(wt)
   WT          wt;
{
extern
   WINDOW      wds[];

   P->ring_state = OFF;
}

void restore_ring_state(wt)
   WT          wt;
{
extern
   WINDOW      wds[];

   if( get_ring_state(wt) ) {
      make_xy_ring(wt);
   }
#undef P
}

void set_auto_thin(wt)
   WT          wt;
{
extern
   WINDOW       wds[];
#define FP      wds[wt].pf.pxyf
   FP->thin_option = AUTO_THIN_ON;
}


int  get_auto_thin(wt)
   WT          wt;
{
extern
   WINDOW      wds[];

   return (FP->thin_option == AUTO_THIN_ON);
}

void unset_auto_thin(wt)
   WT          wt;
{
extern
   WINDOW      wds[];

   FP->thin_option = AUTO_THIN_OFF;

#undef FP
}

int set_max_lxy(wt,max_lxy)
   WT           wt;
   int          max_lxy;
{
   int          lxy;
   int          ltrace = OFF;

#define    P    wds[wt].pf.pxyf

   if( ltrace ) {
      printf("set_max_lxy:  wt: %d  max_lxy: %d.\n",wt,max_lxy);
   }
   unmake_xy_ring(wt);
   set_lxy(P);
   if( ltrace ) {
      printf("set_max_lxy: out of set_lxy.\n");
   }
   lxy = P->lxy;
   if( max_lxy > 0 ) {
      if( max_lxy < lxy ) {
/*         P->xyhead = first_n_pxydv(wt,max_lxy);  */
           P->xyhead = last_n_pxydv(wt,max_lxy);
         set_lxy(P);
         assert(P->lxy == max_lxy);

/*       Should limits be automatically recoumpted? */

         P->limits =  dq_xydv_list(P->xyhead);
         set_display_index_current(wt);
         set_must_remap(wt); set_must_redraw(wt);
      }
   }
   if( ltrace ) {
      printf("set_max_lxy: window %d has %d vectors.\n",wt,P->lxy);
   }

   restore_ring_state(wt);
   return (P->max_lxy = max_lxy);
#undef     P
}

PXYDVBIND first_n_pxydv(wt,n)
   WT        wt;
   int       n;
{
   PXYDVBIND p,   pp,   newtail;
   int            i;
#define      P    wds[wt].pf.pxyf

   for( i = 0, p = P->xyhead; i < n; i++, pp = p, p = p->next );
   newtail = pp; (p->prev) = NULL;
   for( pp = p; pp != NULL; pp = pp->next ) {
      free( (pp->data)->x ); free( (pp->data)->y );
      P->nevents -= (pp->data)->n;
      if( (pp->prev) != NULL ) {
         free( pp->prev );
      }
   }
   free(P->xytail);
   P->xytail = newtail; (P->xytail)->next = NULL;
   return     P->xyhead;
#undef       P
}

PXYDVBIND last_n_pxydv(wt,n)
   WT        wt;
   int       n;
{
   PXYDVBIND p,   pp,   newhead;
   int            i;
#define      P    wds[wt].pf.pxyf

   for( i = 0, p = P->xytail; i < n; i++, pp = p, p = p->prev );
   newhead = pp; (p->next) = NULL;
   for( p = P->xyhead; p != NULL; p = p->next ) {
      free( (p->data)->x ); free( (p->data)->y );
      P->nevents -= (p->data)->n;
      if( (p->prev) != NULL ) {
         free( p->prev );
      }
      pp = p;
   }
   free(pp);
   P->xyhead = newhead; (P->xyhead)->prev = NULL;

   P->xyfirst = P->xyhead;
   P->xylast = P->xytail;
   P->xycurr = P->xyhead;
   (void) reindex_XYDVlist(P->xyhead,1);

   return     P->xyhead;
#undef       P
}

/* Merges p1 and p2, return pointer inherits p1's prev, next pointers. */

PXYDVBIND merge_pxydv(p1,p2)
   PXYDVBIND      p1,   p2;
{
   PXYDVBIND      p;
   int            ip1,  ip2,  ip;

/* printf("merge_pxydv: n1: %d  n2: %d.\n",p1->data->n,p2->data->n);      */
   if( (p1 != NULL) && (p2 != NULL ) ) {
      if( (p1->data->n == 0) || (p2->data->n == 0) ) {
         err1("merge_pxydv","one of merge vectors has n == 0.");
      }
/*    printf("merge_pxydv: In routine, past null--check.\n");      */
      p = (PXYDVBIND) malloc( sizeof(XYDVBIND) );
      p->data = (PXYDVEC) malloc(sizeof(XYDVEC));
      p->data->n = p1->data->n + p2->data->n;
/*    printf("merge_pxydv: n: %d  n1: %d  n2: %d.\n",  */
/*           p->data->n, p1->data->n, p2->data->n);   */
      p->data->x = (DVEC) malloc( p->data->n * sizeof(double) );
      p->data->y = (DVEC) malloc( p->data->n * sizeof(double) );
      for( ip = 0, ip1 = 0; ip1 < p1->data->n; ip++, ip1++ ) {
         p->data->x[ip] = p1->data->x[ip1];
         p->data->y[ip] = p1->data->y[ip1];
      }
      for(         ip2 = 0; ip2 < p2->data->n; ip++, ip2++ ) {
         p->data->x[ip] = p2->data->x[ip2];
         p->data->y[ip] = p2->data->y[ip2];
      }
      assert(p->data->n == ip );
      if( p1->tdata && p2->tdata ) {
         assert(p->data->n == p1->tdata->n + p2->tdata->n);
         p->tdata = make_XYTAGVEC(p->data->n);
         for( ip = 0, ip1 = 0; ip1 < p1->tdata->n; ip++, ip1++ ) {
            p->tdata->tag[ip] = p1->tdata->tag[ip1];
         }
         for(         ip2 = 0; ip2 < p2->tdata->n; ip++, ip2++ ) {
            p->tdata->tag[ip] = p2->tdata->tag[ip2];
         }
      } else {
         p->tdata = NULL;
      }
      p->prev = p1->prev;
      p->next = p1->next;
      p->index = p1->index;
      p->time = p1->time;
      p->tag  = p1->tag;

      if( p1->tdata ) free_XYTAGVEC(p1->tdata);
      if( p2->tdata ) free_XYTAGVEC(p2->tdata);
      free(p1->data->x); free(p1->data->y);
      free(p2->data->x); free(p2->data->y);
      free(p1->data);    free(p2->data);
                         free(p2);

      return p;
   } else {
      err1("merge_pxydv","routine called with at least one NULL ptr.");
      return NULL;
   }
}


int fuzz_equal(d1,d2,d3)
   double      d1,     d2,     d3;
{
   if( d1 == 0.0 ) {
      return d2 < fabs(d3);
   } else {
      return (fabs(d1 - d2) / fabs(d1)) <= fabs(d3);
   }
}

/* Fuzz implemented as controllable parameter. */

#include "plot.h"

int insert_fuzz_equal(d1,d2)
   double      d1,     d2;
{
   if( d1 == 0.0 ) {
      return fabs(d2) <= Insert_xy_fuzz;
   } else {
      return (fabs(d1 - d2) / fabs(d1)) <= fabs(Insert_xy_fuzz);
   }
}

PXYDVEC thin_XYDVEC(p,thetac)
   PXYDVEC        p;
   double         thetac;
{
extern
   void           dvthpl_();
#define           TYPE         PXYDVEC
   TYPE           np,          wp;
#define           N             p->n
#define           X             p->x
#define           Y             p->y
#define           NN           np->n
#define           NX           np->x
#define           NY           np->y
#define           WN           wp->n
#define           WX           wp->x
#define           WY           wp->y

   if( p != (TYPE) NULL ) {
      if( N > 0 ) {
         wp = make_XYDVEC(N);
         (void) dvthpl_(X,Y,&(N),WX,WY,&(WN),&thetac);
/*       printf("thin_XYDVEC: nin: %d    nout: %d    thetac: %g.\n",  */
/*              N,WN,thetac);  */
         if( WN > 0 ) {
            np = make_XYDVEC(WN);
            NN = WN;
            NX = DVcopy(WX,NX,WN);
            NY = DVcopy(WY,NY,WN);
            (void) free_XYDVEC(wp);
            return np;
         } else {
            (void) free_XYDVEC(wp);
            return (TYPE) NULL;
         }
      } else {
         return (TYPE) NULL;
      }
   } else {
      return (TYPE) NULL;
   }

#undef            WY
#undef            WX
#undef            WN
#undef            NY
#undef            NX
#undef            NN
#undef            Y
#undef            X
#undef            N
#undef            TYPE
}

PXYDVBIND thin_PXYDVBIND(p,thetac)
   PXYDVBIND      p;
   double         thetac;
{
   PXYDVEC        ndata;

   if( p != (PXYDVBIND) NULL ) {
      if( (ndata = thin_XYDVEC(p->data,thetac)) != (PXYDVEC) NULL ) {
         free_XYDVEC(p->data);
         p->data = ndata;
      } else {
         printf("thin_PXYDVBIND: Warning: thin_PXYDVEC returns NULL. ");
         printf("No thinning.\n");
      }
   }
   return p;
}

PXYDVBIND mean_PXYDVBIND(p)
   PXYDVBIND      p;
{
#define           Y      (p->data)->y
#define           N      (p->data)->n

   if( p != (PXYDVBIND) NULL ) {
      dvdevm_(Y,Y,&N);
   } else {
      printf("mean_PXYDVBIND:: Warning routine invoked with null p.\n");
   }
   return p;
#undef            Y
#undef            N
}

void set_xycurr_nth(wt,n)
   WT             wt;
   int            n;
{
   PXYDVBIND      p;

   int ltrace =   OFF;

   if( ltrace ) printf("set_xycurr_nth: In routine set %d to %d.\n",wt,n);
   unmake_xy_ring(wt);
   for( p = wds[wt].pf.pxyf->xyhead; p != NULL; p = p->next ) {
      if( n == 1 ) {
         wds[wt].pf.pxyf->xycurr = p;
         wds[wt].pf.pxyf->new_data = OFF;
         goto Return;
      }
      n--;
   }
Return:
   if( ltrace ) printf("set_xycurr_nth: In routine set %d to %d.\n",wt,n);
   set_display_index_current(wt);
   set_must_redraw(wt);
   restore_ring_state(wt);
}

DVEC make_tvec(p) 
   PXYDVBIND     p;
{
   DVEC          tvec;
   int           nvec, ivec;
   int           ltrace = ON;

   if( p ) {
      nvec = n_PXYDVBIND(p);
      if( tvec = make_DVEC(nvec) ) {
         for( ivec = 0; p; tvec[ivec++] = p->time, p = p->next );
         return tvec;
      } else {
         printf("make_tvec: make_DVEC(%d) fails.\n",nvec);
         return NULL;
      }
   } else {
      return NULL;
   }
}

DVEC make_ith_yvec(PXYDVBIND p,int ith) 
{
#define          Y      (p->data)->y
#define          N      (p->data)->n
   DVEC          yvec;
   int           nvec,  ivec;

   if( p ) {
      nvec = n_PXYDVBIND(p);
      if( yvec = make_DVEC(nvec) ) {
         for( ivec = 0; p; p = p->next ) {
            if( N > ith ) {
               yvec[ivec++] = Y[ith];
            } else {
               printf(
                  "make_ith_yvec: Warning: No %dth data point for vector %d.\n",
                  ith, ivec);
               yvec[ivec++] = 0;
            }
         }
         return yvec;
      } else {
         printf("make_yvec: make_DVEC(%d) fails.\n",nvec);
         return NULL;
      }
   } else {
      return NULL;
   }
#undef            N
#undef            Y
}

int compatible_PXYDVBIND_lists_t(p1,p2)
   PXYDVBIND        p1,   p2;
{
   double           fuzz = 1.0e-12;

   while( p1 && p2 ) {
      if( !fuzz_equal(p1->time,p2->time,fuzz) ) return 0;
      p1 = p1->next; p2 = p2->next;
   }
   return !(p1 || p2);
}

int compatible_PXYDVBIND_lists_data_n(p1,p2)
   PXYDVBIND        p1,   p2;
{
   double           fuzz = 1.0e-12;

   while( p1 && p2 ) {
      if( p1->data->n != p2->data->n ) return 0;
      p1 = p1->next; p2 = p2->next;
   }
   return !(p1 || p2);
}

PXYTAGVEC make_XYTAGVEC(n)
   int        n;
{
#define       TYPE     PXYTAGVEC
#define       TAG      p->tag
#define       N        p->n
   TYPE       p;

   if( n > 0 ) {
      if( (p = (TYPE) malloc(sizeof(XYTAGVEC))) != (TYPE) NULL) {
         N = n;
         if( (TAG = make_IVEC(n)) == NULL ) {
            printf("make_XYTAGVEC: tag IVEC allocation failure.\n");
            p = NULL;
         }
      }
   } else {
      p = NULL;
   }
   return p;
}

PXYTAGVEC free_XYTAGVEC(p)
   TYPE       p;
{
#define       FREE_IF_NOT_NULL(a) if( a != NULL ) free(a)
   if( p != (TYPE) NULL ) {
      FREE_IF_NOT_NULL(TAG);
      FREE_IF_NOT_NULL(p);
   }
   return (TYPE) NULL;
#undef        FREE_IF_NOT_NULL
}

/* Splits an XYDVEC into one or more XYDVEC according to changes in
   sense of x values ... */
    
PXYDVEC split_XYDVEC_by_x_sense(PXYDVEC in,int *pnout) {
   PXYDVEC       out = (PXYDVEC) NULL;
   int          *n;
   int           i,             j,           iout,        nout;
   double        old_x,         new_x,       new_dx;

   int           ltrace = OFF;

   if( ltrace ) {
      fprintf(stderr,
              "split_XYDVEC_by_x_sense: In routine with length %d XYVEC\n",
              in->n);
   }
   if( in->n > 0 ) {
      n = make_IVEC(in->n);
      ivls(n,0,in->n);
      iout = 0;
      dx_state_reset();
      old_x = in->x[0];
      for( i = 0; i < in->n; i++ ) {
         new_x = in->x[i];
         new_dx = new_x - old_x;
         if( dx_state_changed(new_dx) ) {
            iout = iout++;
         }
         n[iout]++;
         old_x = new_x;
      }
      nout = iout + 1;
      *pnout = nout;
      if( ltrace ) {
         fprintf(stderr,
                 "split_XYDVEC_by_x_sense: Detected %d distinct vectors\n",
                 *pnout);
         ivdump(n,*pnout,"sub-vector lengths");
      }
      assert(ivsum(n,*pnout)==in->n);
      out = MAKEVEC(XYDVEC,nout);
      for( i = 0; i < nout; i++ ) {
         out[i].n = n[i];
         out[i].x = make_DVEC(n[i]);
         out[i].y = make_DVEC(n[i]);
      }
      iout = 0;
      j    = 0;
      dx_state_reset();
      old_x = in->x[0];
      for( i = 0; i < in->n; i++ ) {
         new_x = in->x[i];
         new_dx = new_x - old_x;
         if( dx_state_changed(new_dx) ) {
            iout = iout++;
            j    = 0;
         }
         out[iout].x[j] = in->x[i];
         out[iout].y[j] = in->y[i];
         j++;
         old_x = new_x;
      }
      free(n);
   } else {
      *pnout = 0;
   }
   return( out );
}

void free_split_XYDVEC(PXYDVEC the,int nvec) {
   int      i;

   if( nvec > 0 ) {
#define       FREE_IF_NOT_NULL(a) if( a != NULL ) free(a)
      for( i = 0; i < nvec; i++ ) {
         FREE_IF_NOT_NULL(the[i].x);
         FREE_IF_NOT_NULL(the[i].y);
      }
      FREE_IF_NOT_NULL(the);
#undef        FREE_IF_NOT_NULL
   }
}

#undef        N
#undef        TAG
#undef        TYPE
